/*
 * bpfset.h
 * Copyright (C) 2008, 2009, 2010, 2011, 2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_BPFSET_H__
#define __LIBx1f4_BPFSET_H__

#include <bpfx-names.h>
#include <errors.h>

/*
 * attribute bit definitions
 */
#define X1f4_BPFSET_TRANS_MASK		(1u << 0x00)

/*
 * error definitions
 */
#define X1f4_BPFSET_FREE_ERROR		LIBx1f4L2_FREE_ERROR
#define X1f4_BPFSET_LINK_ERROR		LIBx1f4L2_LINK_ERROR
#define X1f4_BPFSET_MODE_ERROR		LIBx1f4L2_MODE_ERROR

/*
 * error definitions (continued)
 */
#define X1f4_BPFSET_EVER_MATCH		LIBx1f4L2_EVER_MATCH
#define X1f4_BPFSET_MISS_CLASS		LIBx1f4L2_MISS_CLASS

typedef struct x1f4_bpfset_type {
    void *trans;
} x1f4_bpfset_type;

#if !defined __x1f4_trans_type_defined__
typedef struct x1f4_trans_type {
    int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	(*mode) (void *, void **, unsigned);
    void *context;
} x1f4_trans_type;
#endif				/* !__x1f4_trans_type_defined__ */

#if !defined __x1f4_trans_type_defined__
# define __x1f4_trans_type_defined__
#endif				/* !__x1f4_trans_type_defined__ */

extern int x1f4_call_bpfset(unsigned *);
extern int x1f4_cast_bppset(void *, void *);
extern int x1f4_fast_bpfset(void *, unsigned, struct x1f4_bpfset_type *);
extern int x1f4_find_bppset(void *, void *);
extern int x1f4_fini_bpfset(void **);
extern int x1f4_flat_bpfset(void *);
extern int x1f4_flow_bppset(void *);
extern int x1f4_high_bpfset(void *);
extern int x1f4_init_bpfset(void **, unsigned, struct x1f4_bpfset_type *);
extern int x1f4_lime_bpfset(void *, void *, int (*) (void *, void *));
extern int x1f4_post_bppset(void *, void *);

#endif				/* __LIBx1f4_BPFSET_H__ */
