/*
 * bppx.3.c
 * Copyright (C) 2009-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <bpfx-config.h>

#include <stddef.h>
#include <string.h>

#include <bcount.h>
#include <bpfx-defs.h>
#include <bpfx-inter.h>
#include <bpfx-lines.h>
#include <bpfx-names.h>
#include <bpfx-types.h>
#include <bppx-scope.h>
#include <bppx-types.h>
#include <reverse.h>
#include <xfs.h>
#include <xls.h>

#define __HALF_TRANSFER__		1

#define __FINE_EXCESS__			1
#define __FINE_SECOND__			2
#define __FINE_SELECT__			3
#define __FINE_TARGET__			4

#define __FINE_METHOD__			__FINE_EXCESS__

#define ____FINE_METHOD__		__FINE_METHOD__

#define __LOSE_DOUBLE__			1
#define __LOSE_SELECT__			3
#define __LOSE_SINGLE__			4

#define __LOSE_METHOD__			__LOSE_DOUBLE__

#if __HALF_TRANSFER__
# if __LOSE_METHOD__ == __LOSE_SELECT__
#  undef __LOSE_METHOD__
#  define __LOSE_METHOD__			__LOSE_DOUBLE__
# endif				/* __LOSE_METHOD__ == __LOSE_SELECT__ */
#endif				/* __HALF_TRANSFER__ */

#if __HALF_TRANSFER__
# if __LOSE_METHOD__ == __LOSE_SINGLE__
#  undef __LOSE_METHOD__
#  define __LOSE_METHOD__			__LOSE_DOUBLE__
# endif				/* __LOSE_METHOD__ == __LOSE_SINGLE__ */
#endif				/* __HALF_TRANSFER__ */

#define __UNROLL_TRANSFER__		0

#define __HEAD_LEFT__			1

#define __LOSE_CALL__			__BALANCE_COMMON_INTERNAL__

#define __MISS_CALL__			__BALANCE_SECOND_INTERNAL__

#define false(e)			0

#define here_one_count(___h, ___e) \
    if ((___e)) {							      \
	integral_q f;							      \
									      \
	f = (___e);							      \
									      \
	f &= f - 1;							      \
									      \
	if (f) {							      \
	    (___h) = 5;							      \
	} else {							      \
	    (___h) = 4;							      \
	}								      \
    } else {								      \
	(___h) = 0;							      \
    }

#if __LOSE_METHOD__ == __LOSE_SINGLE__
# define here_two_count(___h, ___e, ___v) \
    l2_q_bitcount((___h), (___e));					      \
    (___h) |= (half_integral_q_bits + 1) < (___h) ? 8192 : 4096;	      \
    if ((___v)) {							      \
    } else {								      \
	(___h) |= 2048;							      \
    }
#endif				/* __LOSE_METHOD__ == __LOSE_SINGLE__ */

#if __LOSE_METHOD__ == __LOSE_SINGLE__
# define here_two_l_count(___h, ___e, ___v) \
    here_two_count((___h), (___e), (___v))
#endif				/* __LOSE_METHOD__ == __LOSE_SINGLE__ */

#if __LOSE_METHOD__ == __LOSE_SINGLE__
# define here_two_n_count(___h, ___e, ___v) \
    here_two_count((___h), (___e), (___v))
#endif				/* __LOSE_METHOD__ == __LOSE_SINGLE__ */

#if __LOSE_METHOD__ == __LOSE_SINGLE__
# define here_two_reset(___h) \
    (___h) = 0
#endif				/* __LOSE_METHOD__ == __LOSE_SINGLE__ */

#if __LOSE_METHOD__ == __LOSE_SINGLE__
# define here_two_l_reset(___h) \
    here_two_reset((___h))
#endif				/* __LOSE_METHOD__ == __LOSE_SINGLE__ */

#if __LOSE_METHOD__ == __LOSE_SINGLE__
# define here_two_n_reset(___h) \
    here_two_reset((___h))
#endif				/* __LOSE_METHOD__ == __LOSE_SINGLE__ */

#if __LOSE_METHOD__ == __LOSE_SINGLE__
# define here_two_declarations
#endif				/* __LOSE_METHOD__ == __LOSE_SINGLE__ */

#if __LOSE_METHOD__ == __LOSE_SINGLE__
# define here_two_setup(___f, ___g)
#endif				/* __LOSE_METHOD__ == __LOSE_SINGLE__ */

#define here_two_test(___h) \
    ((___h) & 8192)

#undef here_two_test

/*
 * is this smarter than it looks?
 * what is does look though is that it favors _not_ fullness, but just transfer
 * over merger and closer kinship.
 * what else it might take into account is the amount of data shifting, amount
 * which is computable, though not cheaply.  for the amount of transfer in the
 * target node may not be relevant (if it would be to transfer in the first or
 * last position) and the amount of transfer in the source node would require
 * some zeroes counting.
 */
#if __LOSE_METHOD__ == __LOSE_SELECT__
# define here_two_count(___h, ___e, ___v) \
    ((___h) =								      \
	 (___e) & ~Lx55555555 & (((___e) & ~Lx55555555) - 1) ? 8192 : 4096);  \
    if ((___v)) {							      \
    } else {								      \
	(___h) |= 2048;							      \
    }
#endif				/* __LOSE_METHOD__ == __LOSE_SELECT__ */

#if __LOSE_METHOD__ == __LOSE_SELECT__
# define here_two_l_count(___h, ___e, ___v) \
    here_two_count((___h), (___e), (___v))
#endif				/* __LOSE_METHOD__ == __LOSE_SELECT__ */

#if __LOSE_METHOD__ == __LOSE_SELECT__
# define here_two_n_count(___h, ___e, ___v) \
    here_two_count((___h), (___e), (___v))
#endif				/* __LOSE_METHOD__ == __LOSE_SELECT__ */

#if __LOSE_METHOD__ == __LOSE_SELECT__
# define here_two_count2(___h, ___e, ___v) \
    {									      \
	integral_q ___c;						      \
									      \
	___c = (___e) & ~Lx55555555;					      \
	(___h) = ___c & (___c - 1) ? 8192 : 4096;			      \
	if ((___v)) {							      \
	} else {							      \
	    (___h) |= 2048;						      \
	}								      \
    }
#endif				/* __LOSE_METHOD__ == __LOSE_SELECT__ */

#if __LOSE_METHOD__ == __LOSE_SELECT__
# define here_two_reset(___h) \
    (___h) = 0
#endif				/* __LOSE_METHOD__ == __LOSE_SELECT__ */

#if __LOSE_METHOD__ == __LOSE_SELECT__
# define here_two_l_reset(___h) \
    here_two_reset((___h))
#endif				/* __LOSE_METHOD__ == __LOSE_SELECT__ */

#if __LOSE_METHOD__ == __LOSE_SELECT__
# define here_two_n_reset(___h) \
    here_two_reset((___h))
#endif				/* __LOSE_METHOD__ == __LOSE_SELECT__ */

#if __LOSE_METHOD__ == __LOSE_SELECT__
# define here_two_declarations
#endif				/* __LOSE_METHOD__ == __LOSE_SELECT__ */

#if __LOSE_METHOD__ == __LOSE_SELECT__
# define here_two_setup(___f, ___g)
#endif				/* __LOSE_METHOD__ == __LOSE_SELECT__ */

#if __LOSE_METHOD__ == __LOSE_DOUBLE__
# define here_two_l_count(___h, ___e, ___v) \
    ___n = (___e) & ~Lx55555555;					      \
    ___p = (___n | ___n >> (half_integral_q_bits + 1))			      \
	& ((Q(1) << half_integral_q_bits) - 1);				      \
    if ((___v)) {							      \
	___t = 1024;							      \
    } else {								      \
	___t = 1024 | 2048;						      \
    }
#endif				/* __LOSE_METHOD__ == __LOSE_DOUBLE__ */

#if __LOSE_METHOD__ == __LOSE_DOUBLE__
# define here_two_n_count(___h, ___e, ___v) \
    ___n = (___e) & ~Lx55555555;					      \
    ___p |= (___n | ___n << (half_integral_q_bits - 1))			      \
	& ~Q(0) << half_integral_q_bits;				      \
    if ((___v)) {							      \
	___t |= Q(1024) << half_integral_q_bits;			      \
    } else {								      \
	___t |= Q(1024 | 2048) << half_integral_q_bits;			      \
    }
#endif				/* __LOSE_METHOD__ == __LOSE_DOUBLE__ */

#if __LOSE_METHOD__ == __LOSE_DOUBLE__
# define here_two_l_reset(___h) \
    ___p = 0;								      \
    ___t = 0
#endif				/* __LOSE_METHOD__ == __LOSE_DOUBLE__ */

#if __LOSE_METHOD__ == __LOSE_DOUBLE__
# define here_two_n_reset(___h)
#endif				/* __LOSE_METHOD__ == __LOSE_DOUBLE__ */

#if __LOSE_METHOD__ == __LOSE_DOUBLE__
# define here_two_declarations \
    integral_q ___n, ___p, ___t
#endif				/* __LOSE_METHOD__ == __LOSE_DOUBLE__ */

#if __LOSE_METHOD__ == __LOSE_DOUBLE__
# define here_two_setup(___f, ___g) \
    l2_q_heap_1up(___p);						      \
									      \
    ___p &= Q(0xff) | Q(0xff) << half_integral_q_bits;			      \
									      \
    ___p |= ___t;							      \
    ___p |=								      \
	((___p - ((___t & (Q(1024) | Q(1024) << half_integral_q_bits)) >> 9)) \
	 & (Q(1024) | Q(1024) << half_integral_q_bits)) << 3;		      \
									      \
    (___f) = ___p & ((Q(1) << half_integral_q_bits) - 1);		      \
    (___g) = ___p >> half_integral_q_bits
#endif				/* __LOSE_METHOD__ == __LOSE_DOUBLE__ */

#define here_two_test(___h) \
    ((___h) & 8192)

#define text_rate(late, bpfset) \
    ((late) + 2 < bpfset((bpfset))->link_a.fpnews.link_v.rate)

#define trap(___n)			((___n) + half_integral_q_bits)

#define true(e)				1

#define Q_21_				Q(21)

#define Q_42_				Q(42)

#define Q_63_				Q(63)

#undef Q
#define Q(e)				((integral_q) (e))

#define q_21_				(Q_21_ << (integral_q_bits - 6))

#define q_42_				(Q_42_ << (integral_q_bits - 6))

#define bpfset(bpfset)			((struct bpfset_type *) (bpfset))

#define TYPE_FINE(___f, ___d) \
    {									      \
	(___f)->fine = 1;						      \
	(___f)->mode = (___d);						      \
    }

typedef struct fine_type {
    int fine;
    integral_q mode;
} fine_type;

static int clip(void *, integral_q);
static int east(void *, unsigned);
static int edge(void *, unsigned);
static int fine(void *, void *, integral_q, struct fine_type *);
static int lose(void *, void *, integral_q *, void *, unsigned, integral_q *,
		void *, unsigned, integral_q);
static int miss(void *, unsigned, void *, unsigned, unsigned, integral_q);
static int shut(void *, void *, unsigned, integral_q);

static int
clip(void *node, integral_q mode)
{
    int status;
    integral_q *call;
    unsigned mind, size;

    call = node;

    size = *call;
    if (size) {
	mind = size;

	call++;

	size--;
	while (size) {
	    unsigned half;

	    half = (size + 1) >> 1;
	    if (mode < call[half]) {
		size = half - 1;
	    } else {
		call += half;
		size -= half;
	    }
	}

	if (*call == mode) {
	    status = 0;

	    memmove(call, call + 1,
		    (mind - (call - (integral_q *) node))
		    * sizeof(integral_q));

	    *((integral_q *) node) = mind - 1;
	} else {
	    status = MISS_CLASS;
	}
    } else {
	status = MISS_CLASS;
    }

    return status;
}


static int
east(void *node, unsigned call)
{
#if __LOSE_CALL__
#else
    int status;
#endif				/* __LOSE_CALL__ */
    integral_q news, slip;

    news = *KD(node);

    slip = news & ~Lx55555555;

#if __LOSE_CALL__
#else
    if (slip) {
	status = 0;
#endif				/* __LOSE_CALL__ */

	if (false(call ^ (integral_q_bits - 1))) {
	    call++;

	    if (call & 1) {
		*KD(node) = news & ~(Q(1) << call);

		KD(node)[call - 1] = KD(node)[call];
	    } else {
		if (news & (Q(1) << (call - 1))) {
		    *KD(node) = news & ~(Q(1) << (call - 1));
		    call--;
		    PD(node)[call] = PD(node)[call - 1];
		} else {
		    if (news & (Q(1) << (call + 1))) {
			*KD(node) = news & ~(Q(1) << (call + 1));
			if (call ^ 2) {
			    KD(node)[call - 2] = KD(node)[call];
			    KD(node)[call] = KD(node)[call + 1];
			    PD(node)[call - 1] = PD(node)[call];
			} else {
			    KD(node)[call] = KD(node)[call + 1];
			    PD(node)[call - 1] = PD(node)[call];
			}
		    } else {
			integral_q back;
			unsigned deck, ever;

			back = slip & (~Q(0) << call);
			if (back) {
			    l2_q_xfs(ever, back);
#if __HEAD_LEFT__
			    slip &= ((Q(1) << ever) - (Q(1) << call))
				>> (ever - call);
			    if (slip) {
				l2_q_xls(deck, slip);
				deck ^= integral_q_last;
				if (1) {
				    integral_q *data, *text;

				    *KD(node) = news & ~(Q(1) << deck);
				    ever = (call - deck - 3) >> 1;
				    call--;
				    text = (integral_q *) PD(node) + call;
				    call--;
				    data = KD(node) + call;
				    for (; ever; ever--) {
					*text = *(text - 2);
					text -= 2;
					*data = *(data - 2);
					data -= 2;
				    }
				    {
					*text = *(text - 2);
					text -= 2;
					*data = *(data - 1);
					*text = *(text - 1);
				    }
				}
			    } else {
				if (call ^ 2) {
				    integral_q *data, *text;

				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call -= 2;
				    data = KD(node) + call;
				    call++;
				    text = (integral_q *) PD(node) + call;
				    for (; deck; deck--) {
					*data = *(data + 2);
					data += 2;
					*text = *(text + 2);
					text += 2;
				    }
				    {
					*data = *(data + 2);
					data += 2;
					*text = *(text + 1);
					*data = *(data + 1);
				    }
				} else {
				    integral_q *data, *text;

				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call--;
				    text = (integral_q *) PD(node) + call;
				    call++;
				    data = KD(node) + call;
				    for (; deck; deck--) {
					*text = *(text + 2);
					text += 2;
					*data = *(data + 2);
					data += 2;
				    }
				    {
					*text = *(text + 1);
					*data = *(data + 1);
				    }
				}
			    }
#else
			    l2_q_xls(deck, slip & ((Q(1) << call) - 1));
			    if (deck ^ integral_q_bits) {
				deck ^= integral_q_last;
				if (call - deck < ever - call) {
				    *KD(node) = news & ~(Q(1) << deck);
				    ever = (call - deck - 3) >> 1;
				    call--;
				    for (; ever; ever--) {
					PD(node)[call] = PD(node)[call - 2];
					call--;
					KD(node)[call] = KD(node)[call - 2];
					call--;
				    }
				    {
					PD(node)[call] = PD(node)[call - 2];
					call--;
					KD(node)[call] = KD(node)[call - 1];
					call--;
					PD(node)[call] = PD(node)[call - 1];
				    }
				} else {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call -= 2;
				    for (; deck; deck--) {
					KD(node)[call] = KD(node)[call + 2];
					call++;
					PD(node)[call] = PD(node)[call + 2];
					call++;
				    }
				    {
					KD(node)[call] = KD(node)[call + 2];
					call++;
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				}
			    } else {
				if (call ^ 2) {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call -= 2;
				    for (; deck; deck--) {
					KD(node)[call] = KD(node)[call + 2];
					call++;
					PD(node)[call] = PD(node)[call + 2];
					call++;
				    }
				    {
					KD(node)[call] = KD(node)[call + 2];
					call++;
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				} else {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call--;
				    for (; deck; deck--) {
					PD(node)[call] = PD(node)[call + 2];
					call++;
					KD(node)[call] = KD(node)[call + 2];
					call++;
				    }
				    {
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				}
			    }
#endif				/* __HEAD_LEFT__ */
			} else {
			    l2_q_xls(deck, slip & ((Q(1) << call) - 1));
			    if (1) {
				deck ^= integral_q_last;
				if (1) {
				    integral_q *data, *text;

				    *KD(node) = news & ~(Q(1) << deck);
				    ever = (call - deck - 3) >> 1;
				    call--;
				    text = (integral_q *) PD(node) + call;
				    call--;
				    data = KD(node) + call;
				    for (; ever; ever--) {
					*text = *(text - 2);
					text -= 2;
					*data = *(data - 2);
					data -= 2;
				    }
				    {
					*text = *(text - 2);
					text -= 2;
					*data = *(data - 1);
					*text = *(text - 1);
				    }
				}
			    }
			}
		    }
		}
	    }
	} else {
	    unsigned deck, ever;

	    l2_q_xls(deck, slip);
	    if (1) {
		deck ^= integral_q_last;
		if (deck ^ (integral_q_bits - 1)) {
		    integral_q *data, *text;

		    *KD(node) = news & ~(Q(1) << deck);

		    PD(node)[integral_q_bits - 1] =
			PD(node)[integral_q_bits - 3];

		    call--;
		    data = KD(node) + call;
		    call--;
		    text = (integral_q *) PD(node) + call;

		    ever = (integral_q_bits - 3 - deck) >> 1;
		    for (; ever; ever--) {
			*data = *(data - 2);
			data -= 2;
			*text = *(text - 2);
			text -= 2;
		    }
		    {
			*data = *(data - 1);
			*text = *(text - 1);
		    }
		} else {
		    *KD(node) = news & ~(Q(1) << deck);

		    PD(node)[integral_q_bits - 1] =
			PD(node)[integral_q_bits - 2];
		}
	    }
	}
#if __LOSE_CALL__
#else
    } else {
	if (false(call ^ (integral_q_bits - 1))) {
	    *KD(node) = news & ~(Q(1) << (call - 1));
	} else {
	    *KD(node) = news & ~(Q(1) << (integral_q_bits - 2));

	    PD(node)[integral_q_bits - 1] = PD(node)[integral_q_bits - 3];
	}

	if (1) {
	    status = LAST_ERROR + call;
	}
    }
#endif				/* __LOSE_CALL__ */

#if __LOSE_CALL__
    return 0;
#else
    return status;
#endif				/* __LOSE_CALL__ */
}


static int
edge(void *node, unsigned call)
{
#if __LOSE_CALL__
#else
    int status;
#endif				/* __LOSE_CALL__ */
    integral_q news, slip;

    news = *KD(node);

    slip  = news & ~Lx55555555;

#if __LOSE_CALL__
#else
    if (slip) {
	status = 0;
#endif				/* __LOSE_CALL__ */

	if (1) {
	    call++;

	    if (call & 1) {
		*KD(node) = news & ~(Q(1) << call);
	    } else {
		if (news & (Q(1) << (call - 1))) {
		    *KD(node) = news & ~(Q(1) << (call - 1));
		    KD(node)[call] = KD(node)[call - 1];
		    call--;
		    PD(node)[call] = PD(node)[call - 1];
		} else {
		    if (news & (Q(1) << (call + 1))) {
			*KD(node) = news & ~(Q(1) << (call + 1));
			if (2) {
			    KD(node)[call] = KD(node)[call + 1];
			    PD(node)[call - 1] = PD(node)[call];
			}
		    } else {
			integral_q back;
			unsigned deck, ever;

			back = slip & (~Q(0) << call);
			if (back) {
			    l2_q_xfs(ever, back);
#if __HEAD_LEFT__
			    slip &= ((Q(1) << ever) - (Q(1) << call))
				>> (ever - call);
			    if (slip) {
				l2_q_xls(deck, slip);
				deck ^= integral_q_last;
				if (1) {
				    integral_q *data, *text;

				    *KD(node) = news & ~(Q(1) << deck);
				    ever = (call - deck) >> 1;
				    data = KD(node) + call;
				    call--;
				    text = (integral_q *) PD(node) + call;
				    for (; ever; ever--) {
					*data = *(data - 2);
					data -= 2;
					*text = *(text - 2);
					text -= 2;
				    }
				    {
					*data = *(data - 1);
					*text = *(text - 1);
				    }
				}
			    } else {
				if (2) {
				    integral_q *data, *text;

				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call--;
				    text = (integral_q *) PD(node) + call;
				    call++;
				    data = KD(node) + call;
				    for (; deck; deck--) {
					*text = *(text + 2);
					text += 2;
					*data = *(data + 2);
					data += 2;
				    }
				    {
					*text = *(text + 1);
					*data = *(data + 1);
				    }
				}
			    }
#else
			    l2_q_xls(deck, slip & ((Q(1) << call) - 1));
			    if (deck ^ integral_q_bits) {
				deck ^= integral_q_last;
				if (call - deck < ever - call) {
				    *KD(node) = news & ~(Q(1) << deck);
				    ever = (call - deck) >> 1;
				    for (; ever; ever--) {
					KD(node)[call] = KD(node)[call - 2];
					call--;
					PD(node)[call] = PD(node)[call - 2];
					call--;
				    }
				    {
					KD(node)[call] = KD(node)[call - 1];
					call--;
					PD(node)[call] = PD(node)[call - 1];
				    }
				} else {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call--;
				    for (; deck; deck--) {
					PD(node)[call] = PD(node)[call + 2];
					call++;
					KD(node)[call] = KD(node)[call + 2];
					call++;
				    }
				    {
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				}
			    } else {
				if (2) {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call--;
				    for (; deck; deck--) {
					PD(node)[call] = PD(node)[call + 2];
					call++;
					KD(node)[call] = KD(node)[call + 2];
					call++;
				    }
				    {
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				}
			    }
#endif				/* __HEAD_LEFT__ */
			} else {
			    l2_q_xls(deck, slip & ((Q(1) << call) - 1));
			    if (1) {
				deck ^= integral_q_last;
				if (1) {
				    integral_q *data, *text;

				    *KD(node) = news & ~(Q(1) << deck);
				    ever = (call - deck) >> 1;
				    data = KD(node) + call;
				    call--;
				    text = (integral_q *) PD(node) + call;
				    for (; ever; ever--) {
					*data = *(data - 2);
					data -= 2;
					*text = *(text - 2);
					text -= 2;
				    }
				    {
					*data = *(data - 1);
					*text = *(text - 1);
				    }
				}
			    }
			}
		    }
		}
	    }
	}
#if __LOSE_CALL__
#else
    } else {
	if (1) {
	    if (call ^ 1) {
		*KD(node) = news & ~(Q(1) << (call - 1));

		KD(node)[call + 1] = KD(node)[call - 1];
	    } else {
		*KD(node) = news & ~(Q(1) << (call + 1));
	    }
	}

	if (1) {
	    status = LAST_ERROR + call;
	}
    }
#endif				/* __LOSE_CALL__ */

#if __LOSE_CALL__
    return 0;
#else
    return status;
#endif				/* __LOSE_CALL__ */
}


static int
fine(void *bpfset, void *node, integral_q mode, struct fine_type *fine_data)
{
    int status;

    do {
	integral_q *fast, news, slip;
	unsigned copy;

	news = *KD(node);

	if (1) {
#if SIZEOF_integral_q == 8
	    if (mode < KD(node)[040]) {
#endif				/* SIZEOF_integral_q == 8 */
		if (mode < KD(node)[020]) {
		    if (mode < KD(node)[010]) {
			if (mode < KD(node)[004]) {
			    if (mode < KD(node)[002]) {
				if (news & 2) {
				    if (mode ^ KD(node)[1]) {
					status = MISS_CLASS;
				    } else {
					slip = news & ~Lx55555555;

					*KD(node) = news & ~Q(2);

					TYPE_FINE(fine_data, KD(node)[2]);

					if (slip & (slip - 1)) {
					    status = 0;
					} else {
					    status = LAST_ERROR + 1;
					}
				    }
				} else {
				    status = MISS_CLASS;
				}

				break;
			    } else {
				copy = 002;

				if (mode ^ KD(node)[2]) {
				} else {
				    if (news & 2) {
				    } else {
					if (news & 8) {
					    TYPE_FINE(fine_data, KD(node)[3]);
					} else {
					    TYPE_FINE(fine_data, KD(node)[4]);
					}
				    }
				}
			    }
			} else {
			    if (mode < KD(node)[006]) {
				copy = 004;
			    } else {
				copy = 006;
			    }
			}
		    } else {
			if (mode < KD(node)[014]) {
			    if (mode < KD(node)[012]) {
				copy = 010;
			    } else {
				copy = 012;
			    }
			} else {
			    if (mode < KD(node)[016]) {
				copy = 014;
			    } else {
				copy = 016;
			    }
			}
		    }
		} else {
		    if (mode < KD(node)[030]) {
			if (mode < KD(node)[024]) {
			    if (mode < KD(node)[022]) {
				copy = 020;
			    } else {
				copy = 022;
			    }
			} else {
			    if (mode < KD(node)[026]) {
				copy = 024;
			    } else {
				copy = 026;
			    }
			}
		    } else {
			if (mode < KD(node)[034]) {
			    if (mode < KD(node)[032]) {
				copy = 030;
			    } else {
				copy = 032;
			    }
			} else {
			    if (mode < KD(node)[036]) {
				copy = 034;
			    } else {
				copy = 036;
			    }
			}
		    }
		}
#if SIZEOF_integral_q == 8
	    } else {
		if (mode < KD(node)[060]) {
		    if (mode < KD(node)[050]) {
			if (mode < KD(node)[044]) {
			    if (mode < KD(node)[042]) {
				copy = 040;
			    } else {
				copy = 042;
			    }
			} else {
			    if (mode < KD(node)[046]) {
				copy = 044;
			    } else {
				copy = 046;
			    }
			}
		    } else {
			if (mode < KD(node)[054]) {
			    if (mode < KD(node)[052]) {
				copy = 050;
			    } else {
				copy = 052;
			    }
			} else {
			    if (mode < KD(node)[056]) {
				copy = 054;
			    } else {
				copy = 056;
			    }
			}
		    }
		} else {
		    if (mode < KD(node)[070]) {
			if (mode < KD(node)[064]) {
			    if (mode < KD(node)[062]) {
				copy = 060;
			    } else {
				copy = 062;
			    }
			} else {
			    if (mode < KD(node)[066]) {
				copy = 064;
			    } else {
				copy = 066;
			    }
			}
		    } else {
			if (mode < KD(node)[074]) {
			    if (mode < KD(node)[072]) {
				copy = 070;
			    } else {
				copy = 072;
			    }
			} else {
			    if (mode < KD(node)[076]) {
				copy = 074;
			    } else {
				copy = 076;
			    }
			}
		    }
		}
	    }
#endif				/* SIZEOF_integral_q == 8 */
	}

	slip = news & ~Lx55555555;

	fast = KD(node) + copy;

	if (mode ^ *fast) {
	    copy++;
	    if (news & Q(1) << copy) {
		fast++;
		if (mode ^ *fast) {
		    status = MISS_CLASS;
		    if (1) {
			break;
		    }
		} else {
		    *KD(node) = news & ~(Q(1) << copy);

		    if (slip & (slip - 1)) {
			status = 0;
		    } else {
			status = LAST_ERROR + copy;
		    }

		    if (1) {
			break;
		    }
		}
	    } else {
		status = MISS_CLASS;
		if (1) {
		    break;
		}
	    }
	}

	if (slip & (slip - 1)) {
	    status = 0;

	    if (0) {
	    } else {
		mode = Q(1) << (copy - 1);
		if (news & mode) {
		    *KD(node) = news & ~mode;
		    *fast = *(fast - 1);
		} else {
		    mode <<= 2;
		    if (news & mode) {
			*KD(node) = news & ~mode;
			*fast = *(fast + 1);
		    } else {
#if ____FINE_METHOD__ == __FINE_SELECT__
			integral_q last, next;

			next = slip >> (copy + 1);
			last = slip << (integral_q_bits - copy);
			l2_q_reverse_1up(next);
			next <<= 1;
			if (last < next) {
			    for (; !(next & 15 << (integral_q_bits - 4));
				 next <<= 4) {
				integral_q *side;

				side = fast + 2;
				*fast = *side;
				fast = side + 2;
				*side = *fast;
			    }
			    if (!(next & 3 << (integral_q_bits - 2))) {
				integral_q *side;

				side = fast + 2;
				*fast = *side;
				fast = side;
			    }
			    {
				*fast = *(fast + 1);
			    }

			    *KD(node) =
				news & ~(Q(1) << (fast - KD(node) + 1));
			} else {
			    for (; !(last & 15 << (integral_q_bits - 4));
				 last <<= 4) {
				integral_q *side;

				side = fast - 2;
				*fast = *side;
				fast = side - 2;
				*side = *fast;
			    }
			    if (!(last & 3 << (integral_q_bits - 2))) {
				integral_q *side;

				side = fast - 2;
				*fast = *side;
				fast = side;
			    }
			    {
				*fast = *(fast - 1);
			    }

			    *KD(node) =
				news & ~(Q(1) << (fast - KD(node) - 1));
			}
#elif __FINE_METHOD__ == __FINE_EXCESS__
			integral_q last, next;

			next = slip >> (copy + 1);
			last = slip << (integral_q_bits - copy);
			l2_q_reverse_1up(last);
			last >>= 1;
			slip = last | next;
			slip &= (slip - 1) << 1;
			if (last & slip) {
			    last = slip >> 4;
			    for (; last; last >>= 4) {
				integral_q *side;

				side = fast - 2;
				*fast = *side;
				fast = side - 2;
				*side = *fast;
			    }
			    if (slip & Lx44444444) {
				integral_q *side;

				side = fast - 2;
				*fast = *side;
				fast = side;
			    }
			    {
				*fast = *(fast - 1);
			    }

			    *KD(node) =
				news & ~(Q(1) << (fast - KD(node) - 1));
			} else {
			    next = slip >> 4;
			    for (; next; next >>= 4) {
				integral_q *side;

				side = fast + 2;
				*fast = *side;
				fast = side + 2;
				*side = *fast;
			    }
			    if (slip & Lx44444444) {
				integral_q *side;

				side = fast + 2;
				*fast = *side;
				fast = side;
			    }
			    {
				*fast = *(fast + 1);
			    }

			    *KD(node) =
				news & ~(Q(1) << (fast - KD(node) + 1));
			}
#elif __FINE_METHOD__ == __FINE_SECOND__
			if (copy < half_integral_q_bits) {
			    if (slip & ((Q(1) << copy) - 1)) {
				if (1) {
				    integral_q *fast;

				    slip <<= integral_q_bits - copy;

				    fast = KD(node) + copy;

				    for (; !(slip
					     & 15 << (integral_q_bits - 4));
					 slip <<= 4) {
					integral_q *side;

					side = fast - 2;
					*fast = *side;
					fast = side - 2;
					*side = *fast;
				    }
				    if (!(slip & 3 << (integral_q_bits - 2))) {
					integral_q *side;

					side = fast - 2;
					*fast = *side;
					fast = side;
				    }
				    {
					*fast = *(fast - 1);
				    }

				    *KD(node) = news
					& ~(Q(1) << (fast - KD(node) - 1));
				}
			    } else {
				if (1) {
				    integral_q *fast;

				    slip >>= copy;

				    fast = KD(node) + copy;

				    for (; !(slip & 15); slip >>= 4) {
					integral_q *side;

					side = fast + 2;
					*fast = *side;
					fast = side + 2;
					*side = *fast;
				    }
				    if (!(slip & 3)) {
					integral_q *side;

					side = fast + 2;
					*fast = *side;
					fast = side;
				    }
				    {
					*fast = *(fast + 1);
				    }

				    *KD(node) = news
					& ~(Q(1) << (fast - KD(node) + 1));
				}
			    }
			} else {
			    if (slip & ~Q(0) << copy) {
				if (1) {
				    integral_q *fast;

				    slip >>= copy;

				    fast = KD(node) + copy;

				    for (; !(slip & 15); slip >>= 4) {
					integral_q *side;

					side = fast + 2;
					*fast = *side;
					fast = side + 2;
					*side = *fast;
				    }
				    if (!(slip & 3)) {
					integral_q *side;

					side = fast + 2;
					*fast = *side;
					fast = side;
				    }
				    {
					*fast = *(fast + 1);
				    }

				    *KD(node) = news
					& ~(Q(1) << (fast - KD(node) + 1));
				}
			    } else {
				if (1) {
				    integral_q *fast;

				    slip <<= integral_q_bits - copy;

				    fast = KD(node) + copy;

				    for (; !(slip
					     & 15 << (integral_q_bits - 4));
					 slip <<= 4) {
					integral_q *side;

					side = fast - 2;
					*fast = *side;
					fast = side - 2;
					*side = *fast;
				    }
				    if (!(slip & 3 << (integral_q_bits - 2))) {
					integral_q *side;

					side = fast - 2;
					*fast = *side;
					fast = side;
				    }
				    {
					*fast = *(fast - 1);
				    }

				    *KD(node) = news
					& ~(Q(1) << (fast - KD(node) - 1));
				}
			    }
			}
#elif __FINE_METHOD__ == __FINE_TARGET__
			integral_q back;
			unsigned deck, ever;

			back = slip & (~Q(0) << copy);
			if (back) {
			    l2_q_xfs(ever, back);
# if __HEAD_LEFT__
			    slip &= ((Q(1) << ever) - (Q(1) << copy))
				>> (ever - copy);
			    if (slip) {
				l2_q_xls(deck, slip);
				deck ^= integral_q_last;
				if (1) {
				    *KD(node) = news & ~(Q(1) << deck);

				    ever = (copy - deck) >> 1;
				    deck = ever >> 1;
				    for (; deck; deck--) {
					*fast = *(fast - 2);
					fast -= 2;
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    if (ever & 1) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    {
					*fast = *(fast - 1);
				    }
				}
			    } else {
				if (1) {
				    *KD(node) = news & ~(Q(1) << ever);

				    deck = (ever - copy) >> 1;
				    ever = deck >> 1;
				    for (; ever; ever--) {
					*fast = *(fast + 2);
					fast += 2;
					*fast = *(fast + 2);
					fast += 2;
				    }
				    if (deck & 1) {
					*fast = *(fast + 2);
					fast += 2;
				    }
				    {
					*fast = *(fast + 1);
				    }
				}
			    }
# else
			    back = slip & ((Q(1) << copy) - 1);
			    if (back) {
				l2_q_xls(deck, back);
				deck ^= integral_q_last;
				if (copy - deck < ever - copy) {
				    *KD(node) = news & ~(Q(1) << deck);

				    ever = (copy - deck) >> 1;
				    for (; ever; ever--) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    {
					*fast = *(fast - 1);
				    }
				} else {
				    *KD(node) = news & ~(Q(1) << ever);

				    deck = (ever - copy) >> 1;
				    for (; deck; deck--) {
					*fast = *(fast + 2);
					fast += 2;
				    }
				    {
					*fast = *(fast + 1);
				    }
				}
			    } else {
				if (1) {
				    *KD(node) = news & ~(Q(1) << ever);

				    deck = (ever - copy) >> 1;
				    for (; deck; deck--) {
					*fast = *(fast + 2);
					fast += 2;
				    }
				    {
					*fast = *(fast + 1);
				    }
				}
			    }
# endif				/* __HEAD_LEFT__ */
			} else {
			    l2_q_xls(deck, slip & ((Q(1) << copy) - 1));
			    if (1) {
				deck ^= integral_q_last;
				if (1) {
				    *KD(node) = news & ~(Q(1) << deck);

				    ever = (copy - deck) >> 1;
				    deck = ever >> 1;
				    for (; deck; deck--) {
					*fast = *(fast - 2);
					fast -= 2;
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    if (ever & 1) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    {
					*fast = *(fast - 1);
				    }
				}
			    }
			}
#endif				/* ____FINE_METHOD__ == __FINE_SELECT__ */
		    }
		}
	    }
	} else {
	    *KD(node) = news & ~(Q(1) << copy);
	    if (1) {
		status = LAST_ERROR + copy;
	    }
	}
    } while (0);

    return status;
}


static int
lose(void *bpfset, void *date, integral_q *path, void *high, unsigned pass,
     integral_q *turn, void *near, unsigned rate, integral_q mode)
{
    int status;
    integral_q *file, pipe;
    unsigned call;
    void *node;

    node = PD(high)[pass];

    pipe = *KD(node);

    while (1) {
#if 1
	seek_rate(node, pipe, mode, call);
#else
	seek_copy(node, mode, call);
#endif				/* 0 */

	file = PD(node)[call];

	if (rate) {
	    integral_q lane;

	    lane = *file;

	    if (lane & ~Lx55555555) {
		if (call ^ (integral_q_bits - 1)) {
		    if (pipe & ((Q(1) << call) - 1) << 1) {
			date = NULL;
			near = NULL;
		    } else {
			near = NULL;
			if (date) {
			    date = PD(date)[integral_q_last];
			} else {
			    if (text_rate(rate, bpfset)) {
				integral_q fold, late;

				late = Q(1) << pass;

				fold = *KD(high);
				if (fold & (late - 1) << 1) {
				    if (fold & late) {
					date = PD(high)[pass - 1];
					path = KD(high) + pass;
				    } else {
					date = PD(high)[pass - 2];
					path = KD(high) + pass - 1;
				    }

				    date = PD(date)[integral_q_last];
				}
			    }
			}
		    }
		} else {
		    date = NULL;
		    if (near) {
			if (*KD(near) & 2) {
			    near = PD(near)[0];
			} else {
			    near = PD(near)[1];
			}
		    } else {
			if (text_rate(rate, bpfset)) {
			    if (pass ^ (integral_q_bits - 1)) {
				turn = KD(high) + pass + 1;
				if (pass & 1) {
				    if (*KD(high) & (Q(1) << (pass + 2))) {
					near = PD(high)[pass + 1];
				    } else {
					near = PD(high)[pass + 2];
				    }
				} else {
				    near = PD(high)[pass + 1];
				}

				if (*KD(near) & 2) {
				    near = PD(near)[0];
				} else {
				    near = PD(near)[1];
				}
			    }
			}
		    }
		}

		high = node;

		pass = call;

		node = file;

		pipe = lane;

		rate = rate - 1;
	    } else {
		break;
	    }
	} else {
	    break;
	}
    }

    if (rate) {
	integral_q *club, *head, *past, *tier;

	if (call & 1) {
	    if (pipe & (Q(1) << call)) {
		club = PD(node)[call - 1];
		head = KD(node) + call;
	    } else {
		if (call ^ 1) {
		    club = PD(node)[call - 2];
		    head = KD(node) + call - 1;
		} else {
		    head = path;
		    if (date) {
			club = PD(date)[integral_q_last];
		    } else {
			club = NULL;
			if (text_rate(rate, bpfset)) {
			    integral_q fold;

			    fold = *KD(high);
			    if (fold & ((Q(1) << pass) - 1) << 1) {
				if (fold & (Q(1) << pass)) {
				    date = PD(high)[pass - 1];
				    path = KD(high) + pass;
				} else {
				    date = PD(high)[pass - 2];
				    path = KD(high) + pass - 1;
				}

				head = path;
				club = PD(date)[integral_q_last];
			    }
			}
		    }
		}
	    }
	    if (call ^ (integral_q_bits - 1)) {
		past = KD(node) + call + 1;
		if (pipe & (Q(1) << (call + 2))) {
		    tier = PD(node)[call + 1];
		} else {
		    tier = PD(node)[call + 2];
		}
	    } else {
		past = turn;
		if (near) {
		    if (*KD(near) & 2) {
			tier = PD(near)[0];
		    } else {
			tier = PD(near)[1];
		    }
		} else {
		    tier = NULL;
		    if (text_rate(rate, bpfset)) {
			if (pass ^ (integral_q_bits - 1)) {
			    turn = KD(high) + pass + 1;
			    if (pass & 1) {
				if (*KD(high) & (Q(1) << (pass + 2))) {
				    near = PD(high)[pass + 1];
				} else {
				    near = PD(high)[pass + 2];
				}
			    } else {
				near = PD(high)[pass + 1];
			    }

			    past = turn;
			    if (*KD(near) & 2) {
				tier = PD(near)[0];
			    } else {
				tier = PD(near)[1];
			    }
			}
		    }
		}
	    }
	} else {
	    past = KD(node) + call + 1;
	    tier = PD(node)[call + 1];
	    if (call) {
		club = PD(node)[call - 1];
		head = KD(node) + call;
	    } else {
		head = path;
		if (date) {
		    club = PD(date)[integral_q_last];
		} else {
		    club = NULL;
		    if (text_rate(rate, bpfset)) {
			integral_q fold;

			fold = *KD(high);
			if (fold & ((Q(1) << pass) - 1) << 1) {
			    if (fold & (Q(1) << pass)) {
				date = PD(high)[pass - 1];
				path = KD(high) + pass;
			    } else {
				date = PD(high)[pass - 2];
				path = KD(high) + pass - 1;
			    }

			    head = path;
			    club = PD(date)[integral_q_last];
			}
		    }
		}
		{
		}
		{
		}
	    }
	}

	do {
#if __LOSE_CALL__
#else
	    int excess;
#endif				/* __LOSE_CALL__ */

	    if (false(*file & ~Lx55555555)) {
		status = 0;
	    } else {
		status = _libx1f4l2_lift_bppset
		    (bpfset, date, path, &head, &club, pipe, &node, &call,
		     &file, &tier, &past, turn, near, mode);
		if (status == FREE_ERROR) {
		    break;
		} else {
		    pipe = *KD(node);
		}
	    }

#if __LOSE_CALL__
	    status = lose
		(bpfset, club, head, node, call, past, tier, rate - 1, mode);
#else
	    excess = lose
		(bpfset, club, head, node, call, past, tier, rate - 1, mode);
	    if (excess) {
		if (excess < LAST_ERROR) {
		    status = excess;
		} else {
		    integral_q last, next;
		    unsigned copy, deck, ever;

		    copy = excess - LAST_ERROR;

		    if (club) {
			last = *club & ~Lx55555555;
			here_one_count(deck, last);
			if (pipe & ((Q(1) << call) - 1) << 1) {
			    deck += 2;
			}
		    } else {
			deck = 0;

			/*
			 * rubbish warnings...
			 */
			last = 0;
		    }

		    if (tier) {
			next = *tier & ~Lx55555555;
			here_one_count(ever, next);
			if (call ^ (integral_q_bits - 1)) {
			    ever += 2;
			}
		    } else {
			ever = 0;
		    }

		    if (deck < ever) {
			if (1) {
			    integral_q data, *fast, *side, text;

			    fast = tier;

			    text = next;
			    text &= (text - 1) << 1;

			    *KD(fast) = Lx55555555 | (next & ~text);

			    text >>= 2;
			    if (text) {
				fast += 2;

				side = fast + integral_q_bits - 1;

				last = *fast;

				data = *side;

				text >>= 2;

				for (; text; text >>= 2) {
				    *fast = *(fast + 2);
				    fast += 2;
				    *side = *(side + 2);
				    side += 2;
				}
				{
				    *fast = *(fast + 1);
				}
				{
				    *side = *(side + 1);
				}
			    } else {
				last = fast[1];

				data = fast[integral_q_bits];
			    }

			    fast = file;

			    *KD(fast) = Lx55555555;

			    if (copy ^ integral_q_last) {
				if (copy ^ 1) {
				    fast += copy - 1;

				    side = fast + integral_q_last + 2;

				    ever = (integral_q_last - copy) >> 1;
				    for (; ever; ever--) {
					*fast = *(fast + 2);
					fast += 2;
					*side = *(side + 2);
					side += 2;
				    }
				    {
					*fast = *past;
				    }
				    {
					*side = data;
				    }
				} else {
				    fast += 2;

				    side = fast + integral_q_bits - 1;

				    ever = half_integral_q_bits - 2;
				    for (; ever; ever--) {
					*fast = *(fast + 2);
					fast += 2;
					*side = *(side + 2);
					side += 2;
				    }
				    {
					*fast = *past;
				    }
				    {
					*side = *(side + 2);
					side += 2;
				    }
				    {
					*side = data;
				    }
				}
			    } else {
				fast += integral_q_bits - 2;

				*fast = *past;

				fast += integral_q_bits + 1;

				{
				    *(fast - 2) = *fast;
				}

				{
				    *fast = data;
				}
			    }
			}

			*past = last;
		    } else {
			if (1) {
			    integral_q data, *fast, *side, text;

			    fast = club;

			    text = last;

			    if (text & Q(1) << integral_q_last) {
				next = fast[integral_q_last];

			        fast += (integral_q_bits << 1) - 1;

				data = *fast;
				*fast = *(fast - 1);

				*KD(club) = Lx55555555
				    | (last ^ Q(1) << integral_q_last);
			    } else {
				fast += integral_q_bits - 2;

				side = fast + integral_q_bits + 1;

				next = *fast;

				data = *side;

				text <<= 2;

				for (; !(text & Q(1) << integral_q_last);
				     text <<= 2) {
				    *fast = *(fast - 2);
				    fast -= 2;
				    *side = *(side - 2);
				    side -= 2;
				}
				{
				    *fast = *(fast - 1);
				}
				{
				    *side = *(side - 2);
				    side -= 2;
				}
				{
				    *side = *(side - 1);
				}

				*KD(club) = Lx55555555
				    | (last ^ Q(1) << (fast - club - 1));
			    }

			    fast = file;

			    *KD(fast) = Lx55555555;

			    if (copy ^ integral_q_last) {
				if (copy ^ 1) {
				    fast += copy - 1;

				    side = fast + 1 + integral_q_bits;

				    ever = (copy - 3) >> 1;
				    for (; ever; ever--) {
					*fast = *(fast - 2);
					fast -= 2;
					*side = *(side - 2);
					side -= 2;
				    }
				    {
					*fast = *head;
				    }
				    {
					*side = *(side - 2);
					side -= 2;
				    }
				    {
					*side = data;
				    }
				} else {
				    fast += 2;

				    *fast = *head;

				    fast += integral_q_last;

				    {
					*fast = data;
				    }
				}
			    } else {
				fast += integral_q_bits - 2;

				side = fast + integral_q_bits - 1;

				ever = half_integral_q_bits - 2;
				for (; ever; ever--) {
				    *fast = *(fast - 2);
				    fast -= 2;
				    *side = *(side - 2);
				    side -= 2;
				}
				{
				    *fast = *head;
				}
				{
				    *side = data;
				}
			    }
			}

			*head = next;
		    }
		}
	    }
#endif				/* __LOSE_CALL__ */
	} while (0);
    } else {
	struct fine_type lead;

	lead.fine = 0;

	status = fine(bpfset, file, mode, &lead);
	if (status) {
	    if (status < LAST_ERROR) {
	    } else {
		integral_q *club, *head, last, next, *tier;
		unsigned copy, deck, ever;

		copy = status - LAST_ERROR;

		status = 0;

		if (call & 1) {
		    if (pipe & (Q(1) << call)) {
			club = PD(node)[call - 1];
			head = KD(node) + call;
			date = NULL;
		    } else {
			if (call ^ 1) {
			    club = PD(node)[call - 2];
			    head = KD(node) + call - 1;
			    date = NULL;
			} else {
			    head = path;
			    if (date) {
				club = PD(date)[integral_q_last];
			    } else {
				club = NULL;
				if (text_rate(0, bpfset)) {
				    integral_q fold;

				    fold = *KD(high);
				    if (fold & ((Q(1) << pass) - 1) << 1) {
					if (fold & (Q(1) << pass)) {
					    date = PD(high)[pass - 1];
					    path = KD(high) + pass;
					} else {
					    date = PD(high)[pass - 2];
					    path = KD(high) + pass - 1;
					}

					head = path;
					club = PD(date)[integral_q_last];
				    } else {
					lead.fine = 0;
				    }
				} else {
				    lead.fine = 0;
				}
			    }
			}
		    }
		    if (call ^ (integral_q_bits - 1)) {
			near = NULL;
			if (pipe & (Q(1) << (call + 2))) {
			    tier = PD(node)[call + 1];
			} else {
			    tier = PD(node)[call + 2];
			}
		    } else {
			if (near) {
			    if (*KD(near) & 2) {
				tier = PD(near)[0];
			    } else {
				tier = PD(near)[1];
			    }
			} else {
			    tier = NULL;
			    if (text_rate(0, bpfset)) {
				if (pass ^ (integral_q_bits - 1)) {
				    turn = KD(high) + pass + 1;
				    if (pass & 1) {
					if (*KD(high) & (Q(1) << (pass + 2))) {
					    near = PD(high)[pass + 1];
					} else {
					    near = PD(high)[pass + 2];
					}
				    } else {
					near = PD(high)[pass + 1];
				    }

				    if (*KD(near) & 2) {
					tier = PD(near)[0];
				    } else {
					tier = PD(near)[1];
				    }
				}
			    }
			}
		    }
		} else {
		    near = NULL;
		    if (call) {
			club = PD(node)[call - 1];
			tier = PD(node)[call + 1];
			head = KD(node) + call;
			date = NULL;
		    } else {
			head = path;
			if (date) {
			    club = PD(date)[integral_q_last];
			} else {
			    club = NULL;
			    if (text_rate(0, bpfset)) {
				integral_q fold;

				fold = *KD(high);
				if (fold & ((Q(1) << pass) - 1) << 1) {
				    if (fold & (Q(1) << pass)) {
					date = PD(high)[pass - 1];
					path = KD(high) + pass;
				    } else {
					date = PD(high)[pass - 2];
					path = KD(high) + pass - 1;
				    }

				    head = path;
				    club = PD(date)[integral_q_last];
				} else {
				    lead.fine = 0;
				}
			    } else {
				lead.fine = 0;
			    }
			}
			{
			    tier = PD(node)[call + 1];
			}
		    }
		}

		{
		    here_two_declarations;

		    if (club) {
			last = *club;
			here_two_l_count(deck, last, date);
		    } else {
			here_two_l_reset(deck);

			/*
			 * rubbish warnings...
			 */
			last = 0;
		    }

		    if (tier) {
			next = *tier;
			here_two_n_count(ever, next, near);
		    } else {
			here_two_n_reset(ever);

			/*
			 * rubbish warnings...
			 */
			next = 0;
		    }

		    here_two_setup(deck, ever);
		}

		if (here_two_test(deck | ever)) {
		    if (deck < ever) {
			if (lead.fine) {
			    *head = lead.mode;
			}

#if __HALF_TRANSFER__
			if (1) {
			    integral_q *fast, *side, text;

			    fast = file;

			    if (copy & 1) {
			    } else {
				text = *fast & ~Lx55555555;

				if (text >> copy) {
				    fast += copy;

				    next = text >> copy >> 2;

# if __UNROLL_TRANSFER__
				    mode = next >> 2;
				    for (; mode; mode >>= 4) {
					*fast = *(fast + 2);
					fast += 2;
					*fast = *(fast + 2);
					fast += 2;
				    }
				    if (next & Lx22222222) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# else
				    for (; next; next >>= 2) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast + 1);
				    }
				} else {
				    fast += copy;

				    next = text
					<< (integral_q_bits - copy + 1);

# if __UNROLL_TRANSFER__
				    mode = next << 2;
				    for (; mode; mode <<= 4) {
					*fast = *(fast - 2);
					fast -= 2;
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    if (next & Lx44444444) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
# else
				    for (; next; next <<= 2) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast - 1);
				    }
				}
			    }

			    ever &= 254;

			    fast = file;

			    *KD(fast) =
				Lx55555555 | ~Q(0) << (integral_q_bits - ever);

			    fast += integral_q_bits - ever + 1;

			    side = fast + 1;

			    ever >>= 1;

			    deck = ever - 1;
			    for (; deck; deck--) {
				*fast = *side;
				side += 2;
				fast++;
			    }

			    side = tier;

			    text = *side;

			    while (ever & ~1) {
				if (text & Q(2)) {
				    side++;
				    *fast = *side;
				    fast++;
				    side++;
				    *fast = *side;
				    fast++;

				    ever -= 2;
				} else {
				    ever--;

				    side += 2;
				    *fast = *side;
				    fast++;
				}

				text >>= 2;
			    }
			    if (ever) {
				if (text & Q(2)) {
				    side++;
				    *fast = *side;
				    fast++;

				    side++;

				    fast = tier + 2;

				    text >>= 1;
				} else {
				    side += 2;
				    *fast = *side;
				    fast++;

				    side++;

				    fast = tier + 2;

				    text >>= 2;

				    if (fast < side) {
					text >>= 1;
					if (text & Q(1)) {
					    *fast = *side;
					    fast += 2;
					} else {
					}

					side++;
				    }
				}
			    } else {
				side++;

				fast = tier + 2;

				if (fast < side) {
				    text >>= 1;
				    if (text & Q(1)) {
					*fast = *side;
					fast += 2;
				    } else {
				    }

				    side++;
				}
			    }
			    while (fast < side) {
				if (text & Q(4)) {
				    *fast = *side;
				    fast += 2;
				    side++;
				    *fast = *side;
				    fast += 2;
				    side++;
				} else {
				    *fast = *side;
				    fast += 2;
				    side += 2;
				}

				text >>= 2;
			    }

			    *KD(tier) &= Lx55555555 | ~Q(0) << (fast - tier);
			}

			if (call ^ (integral_q_bits - 1)) {
			    KD(node)[call + 1] = *(tier + 2);
			} else {
			    if (true(near)) {
				*turn = *(tier + 2);
			    } else {
			    }
			}
#else
			integral_q port;

			if (1) {
			    integral_q *fast, text;

			    fast = tier;

			    text = next & ~Lx55555555;
			    text &= (text - 1) << 1;

			    text >>= 2;
			    if (text) {
				fast += 2;

				last = *fast;

				text >>= 2;
				if (text) {
				    port = *(fast + 2);
				    *fast = port;
				    fast += 2;

				    text >>= 2;

# if __UNROLL_TRANSFER__
				    mode = text >> 2;
				    for (; mode; mode >>= 4) {
					*fast = *(fast + 2);
					fast += 2;
					*fast = *(fast + 2);
					fast += 2;
				    }
				    if (text & Lx22222222) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# else
				    for (; text; text >>= 2) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast + 1);
				    }
				} else {
				    port = fast[1];
				    *fast = port;
				}
			    } else {
				last = fast[1];
				port = fast[2];
			    }

			    *KD(tier) = next ^ Q(1) << (fast - tier + 1);

			    fast = file;

			    if (copy & 1) {
				*KD(fast) =
				    Lx55555555 | (Q(1) << integral_q_last);

				*(fast + integral_q_last) = last;
			    } else {
				text = *fast & ~Lx55555555;

				if (text >> copy) {
				    *KD(fast) =
					Lx55555555 | (Q(1) << integral_q_last);

				    fast += copy;

				    next = text >> copy >> 2;

# if __UNROLL_TRANSFER__
				    mode = next >> 2;
				    for (; mode; mode >>= 4) {
					*fast = *(fast + 2);
					fast += 2;
					*fast = *(fast + 2);
					fast += 2;
				    }
				    if (next & Lx22222222) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# else
				    for (; next; next >>= 2) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast + 1);
				    }

				    {
					*(file + integral_q_last) = last;
				    }
				} else {
				    if (Q(1) << copy
					< text << (integral_q_bits - copy)) {
					*KD(fast) = Lx55555555
					    | (Q(1) << integral_q_last);

					fast += copy;

					next = text
					    << (integral_q_bits - copy + 1);

# if __UNROLL_TRANSFER__
					mode = next << 2;
					for (; mode; mode <<= 4) {
					    *fast = *(fast - 2);
					    fast -= 2;
					    *fast = *(fast - 2);
					    fast -= 2;
					}
					if (next & Lx44444444) {
					    *fast = *(fast - 2);
					    fast -= 2;
					}
# else
					for (; next; next <<= 2) {
					    *fast = *(fast - 2);
					    fast -= 2;
					}
# endif				/* __UNROLL_TRANSFER__ */
					{
					    *fast = *(fast - 1);
					}

					{
					    *(file + integral_q_last) = last;
					}
				    } else {
					*KD(fast) = text | Lx55555555;

					fast += copy;

					deck =
					    (integral_q_bits - 2 - copy) >> 1;
# if __UNROLL_TRANSFER__
					ever = deck >> 1;
					for (; ever; ever--) {
					    *fast = *(fast + 2);
					    fast += 2;
					    *fast = *(fast + 2);
					    fast += 2;
					}
					if (deck & 1) {
					    *fast = *(fast + 2);
					    fast += 2;
					}
# else
					for (; deck; deck--) {
					    *fast = *(fast + 2);
					    fast += 2;
					}
# endif				/* __UNROLL_TRANSFER__ */
					{
					    *fast = last;
					}
				    }
				}
			    }
			}

			if (call ^ (integral_q_bits - 1)) {
			    KD(node)[call + 1] = port;
			} else {
			    if (true(near)) {
				*turn = port;
			    } else {
			    }
			}
#endif				/* __HALF_TRANSFER__ */
		    } else {
#if __HALF_TRANSFER__
			if (1) {
			    integral_q *fast, *side, text;

			    fast = file;

			    if (copy & 1) {
			    } else {
				text = *fast & ~Lx55555555;

				if (text >> copy) {
				    fast += copy;

				    last = text >> copy >> 2;

# if __UNROLL_TRANSFER__
				    mode = last >> 2;
				    for (; mode; mode >>= 4) {
					*fast = *(fast + 2);
					fast += 2;
					*fast = *(fast + 2);
					fast += 2;
				    }
				    if (last & Lx22222222) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# else
				    for (; last; last >>= 2) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast + 1);
				    }
				} else {
				    fast += copy;

				    last = text
					<< (integral_q_bits - copy + 1);

# if __UNROLL_TRANSFER__
				    mode = last << 2;
				    for (; mode; mode <<= 4) {
					*fast = *(fast - 2);
					fast -= 2;
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    if (last & Lx44444444) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
# else
				    for (; last; last <<= 2) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast - 1);
				    }
				}
			    }

			    deck &= 254;

			    fast = file;

			    *KD(fast) = Lx55555555 | ((Q(1) << deck) - 1);

			    fast += deck - 1;

			    side = fast - 1;

			    deck >>= 1;

			    ever = deck - 1;
			    for (; ever; ever--) {
				*fast = *side;
				side -= 2;
				fast--;
			    }

			    side = club;

			    text = *side;

			    side += integral_q_bits;

			    while (deck & ~1) {
				if (text & Q(1) << integral_q_last) {
				    side--;
				    *fast = *side;
				    fast--;
				    side--;
				    *fast = *side;
				    fast--;

				    deck -= 2;
				} else {
				    deck--;

				    side -= 2;
				    *fast = *side;
				    fast--;
				}

				text <<= 2;
			    }
			    if (deck) {
				if (text & Q(1) << integral_q_last) {
				    side--;
				    *fast = *side;

				    side--;

				    *head = *fast;

				    fast = club + integral_q_bits - 2;

				    text <<= 1;
				} else {
				    side -= 2;
				    *fast = *side;

				    side--;

				    *head = *fast;

				    fast = club + integral_q_bits - 2;

				    text <<= 2;

				    if (side < fast) {
					if (text & Q(1) << integral_q_last) {
					    *fast = *side;
					    fast -= 2;
					}

					side--;

					text <<= 1;
				    }
				}
			    } else {
				side--;

				*head = *(fast + 1);

				fast = club + integral_q_bits - 2;

				if (side < fast) {
				    if (text & Q(1) << integral_q_last) {
					*fast = *side;
					fast -= 2;
				    }

				    side--;

				    text <<= 1;
				}
			    }
			    while (side < fast) {
				if (text & Q(1) << (integral_q_bits - 2)) {
				    *fast = *side;
				    fast -= 2;
				    side--;
				    *fast = *side;
				    fast -= 2;
				    side--;
				} else {
				    *fast = *side;
				    fast -= 2;
				    side -= 2;
				}

				text <<= 2;
			    }

			    *KD(club) &=
				Lx55555555 | ((Q(1) << (fast - club)) - 1);
			}
#else
			if (1) {
			    integral_q *fast, text;

			    fast = club;

			    text = *fast;
			    if (text & Q(1) << integral_q_last) {
				fast += integral_q_bits;
				next = *(fast - 1);
			    } else {
				fast += integral_q_bits - 2;

				next = *fast;

				text <<= 2;

# if __UNROLL_TRANSFER__
				for (;
				     !(text & Q(10) << (integral_q_bits - 4));
				     text <<= 4) {
				    *fast = *(fast - 2);
				    fast -= 2;
				    *fast = *(fast - 2);
				    fast -= 2;
				}
				if (!(text & Q(2) << (integral_q_bits - 2))) {
				    *fast = *(fast - 2);
				    fast -= 2;
				}
# else
				for (; !(text & Q(1) << integral_q_last);
				     text <<= 2) {
				    *fast = *(fast - 2);
				    fast -= 2;
				}
# endif				/* __UNROLL_TRANSFER__ */
				{
				    *fast = *(fast - 1);
				}
			    }

			    *KD(club) = last ^ Q(1) << (fast - club - 1);

			    fast = file;

			    if (copy & 1) {
				*KD(fast) = Lx55555555 | 2;

				*(fast + 1) = next;
			    } else {
				text = *fast & ~Lx55555555;

				if (text >> copy) {
				    if (text >> copy < Q(1) << copy) {
					*KD(fast) = Lx55555555 | 2;

					fast += copy;

					last = text >> copy >> 2;

# if __UNROLL_TRANSFER__
					mode = last >> 2;
					for (; mode; mode >>= 4) {
					    *fast = *(fast + 2);
					    fast += 2;
					    *fast = *(fast + 2);
					    fast += 2;
					}
					if (last & Lx22222222) {
					    *fast = *(fast + 2);
					    fast += 2;
					}
# else
					for (; last; last >>= 2) {
					    *fast = *(fast + 2);
					    fast += 2;
					}
# endif				/* __UNROLL_TRANSFER__ */
					{
					    *fast = *(fast + 1);
					}

					{
					    *(file + 1) = next;
					}
				    } else {
					*KD(fast) = text | Lx55555555;

					fast += copy;

					ever = (copy - 2) >> 1;
# if __UNROLL_TRANSFER__
					deck = ever >> 1;
					for (; deck; deck--) {
					    *fast = *(fast - 2);
					    fast -= 2;
					    *fast = *(fast - 2);
					    fast -= 2;
					}
					if (ever & 1) {
					    *fast = *(fast - 2);
					    fast -= 2;
					}
# else
					for (; ever; ever--) {
					    *fast = *(fast - 2);
					    fast -= 2;
					}
# endif				/* __UNROLL_TRANSFER__ */
					{
					    *fast = next;
					}
				    }
				} else {
				    *KD(fast) = Lx55555555 | 2;

				    fast += copy;

				    last = text
					<< (integral_q_bits - copy + 1);

# if __UNROLL_TRANSFER__
				    mode = last << 2;
				    for (; mode; mode <<= 4) {
					*fast = *(fast - 2);
					fast -= 2;
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    if (last & Lx44444444) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
# else
				    for (; last; last <<= 2) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast - 1);
				    }

				    {
					*(file + 1) = next;
				    }
				}
			    }
			}

			*head = next;
#endif				/* __HALF_TRANSFER__ */
		    }
		} else {
		    do {
			integral_q *fast;

			if (tier) {
			    if (club) {
			    } else {
				_libx1f4l2_four_bppset
				    (bpfset, last, &club, pipe, &node, &call,
				     &head, &file, &tier, next, &copy);
				if (club) {
				} else {
				    break;
				}
			    }
			} else {
			    if (1) {
				_libx1f4l2_nine_bppset
				    (bpfset, last, &club, pipe, &node, &call,
				     &head, &file, &tier, next, &copy);
				if (tier) {
				} else {
				    break;
				}
			    }
			}

			if (1) {
			    integral_q text;

			    fast = file;

			    if (copy & 1) {
			    } else {
				text = *fast & ~Lx55555555;

				if (text >> copy) {
				    fast += copy;

				    next = text >> copy >> 2;

#if __UNROLL_TRANSFER__
				    mode = next >> 2;
				    for (; mode; mode >>= 4) {
					*fast = *(fast + 2);
					fast += 2;
					*fast = *(fast + 2);
					fast += 2;
				    }
				    if (next & Lx22222222) {
					*fast = *(fast + 2);
					fast += 2;
				    }
#else
				    for (; next; next >>= 2) {
					*fast = *(fast + 2);
					fast += 2;
				    }
#endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast + 1);
				    }
				} else {
				    fast += copy;

				    next = text
					<< (integral_q_bits - copy + 1);

#if __UNROLL_TRANSFER__
				    mode = next << 2;
				    for (; mode; mode <<= 4) {
					*fast = *(fast - 2);
					fast -= 2;
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    if (next & Lx44444444) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
#else
				    for (; next; next <<= 2) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
#endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast - 1);
				    }
				}
			    }
			}

			{
			    integral_q news, *side;

			    fast = club;

			    news = *fast & ~Lx55555555;

			    if (news & Q_42_) {
				*fast = (Lxdddddddd | Q(2))
				    ^ Q(1) << integral_q_last;

				if (news & 2) {
				    fast += 4;

				    side = fast - 1;

				    *side = *fast;
				    fast += 2;
				    side++;
				} else {
				    if (news & 8) {
					fast += 2;

					side = fast - 1;

					*side = *fast;
					fast++;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
					fast += 2;
					side++;
				    } else {
					fast += 2;

					side = fast - 1;

					*side = *fast;
					fast += 2;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
					fast++;
					side++;
				    }
				}
			    } else {
				fast += 4;

				side = fast - 1;

				*side = *fast;
				fast += 2;
				side++;
			    }

			    news |= Q(1) << (integral_q_bits - 2);
			    news >>= 6;

			    for (; !(news & Q_63_); news >>= 6) {
				*side = *fast;
				fast += 2;
				side += 2;
				*side = *fast;
				fast += 2;
				side++;
				*side = *fast;
				fast += 2;
				side++;
			    }

			    if (news & ~((Q(1) << 6) - 1)) {
				*club = (Lxdddddddd | Q(2) << (side - club))
				    ^ Q(1) << integral_q_last;

				if (news & 2) {
				    *side = *fast;
				    fast++;
				    side++;
				    *side = *fast;
				    fast++;
				    side++;
				    *side = *fast;
				    fast += 2;
				    side++;
				    *side = *fast;
				    fast += 2;
				    side++;
				} else {
				    if (news & 8) {
					*side = *fast;
					fast += 2;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
					fast += 2;
					side++;
				    } else {
					*side = *fast;
					fast += 2;
					side++;
					*side = *fast;
					fast += 2;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
					fast++;
					side++;
				    }
				}

				news >>= 6;
			    }

			    for (; !(news & Q_21_); news >>= 6) {
				*side = *fast;
				fast += 2;
				side += 2;
				*side = *fast;
				fast += 2;
				side++;
				*side = *fast;
				fast += 2;
				side++;
			    }

			    if (news & Q_42_) {
				*club = (Lxdddddddd | Q(2) << (side - club))
				    ^ Q(1) << integral_q_last;

				if (half_integral_q_bits % 3 ^ 2) {
				    *side = *fast;
				    fast++;
				    side++;
				    *side = *fast;

				    fast = file;

				    fast += 2;
				    side++;
				    *side = *fast;
				    fast += 2;
				    side++;
				    *side = *fast;
				    side++;
				} else {
				    *side = *fast;
				    if (news & 2) {
					fast++;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
				    } else {
					fast += 2;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
				    }

				    fast = file;

				    fast += 2;
				    side++;
				    *side = *fast;
				    side++;
				}
			    } else {
				if (half_integral_q_bits % 3 ^ 2) {
				    *side = *fast;

				    fast = file;

				    fast += 2;
				    side += 2;
				    *side = *fast;
				    fast += 2;
				    side++;
				    *side = *fast;
				    side++;
				} else {
				    *side = *fast;
				    fast += 2;
				    side += 2;
				    *side = *fast;

				    fast = file;

				    fast += 2;
				    side++;
				    *side = *fast;
				    side++;
				}
			    }

			    deck =
				integral_q_bits / 4 - integral_q_bits / 6 - 2;
			    for (; deck; deck--) {
				fast += 2;
				*side = *fast;
				side += 2;
				fast += 2;
				*side = *fast;
				side++;
				fast += 2;
				*side = *fast;
				side++;
			    }

			    {
				fast += 2;
				*side = *fast;
				side += 2;
				fast += 2;
				*side = *fast;
			    }

			    fast = tier;

			    news = *tier & ~Lx55555555;

			    fast += integral_q_bits;

			    side = fast;

			    news |= Q(1);

			    for (; !(news & q_42_); news <<= 6) {
				side--;
				fast -= 2;
				*side = *fast;
				side--;
				fast -= 2;
				*side = *fast;
				side -= 2;
				fast -= 2;
				*side = *fast;
			    }

			    if (news & q_21_) {
			    } else {
				*tier = Lxdddddddd | Q(2) << (side - tier - 4);

				if (news & Q(1) << (integral_q_bits - 1)) {
				    side--;
				    fast--;
				    *side = *fast;
				    side--;
				    fast--;
				    *side = *fast;
				    side--;
				    fast -= 2;
				    *side = *fast;
				    side--;
				    fast -= 2;
				    *side = *fast;
				} else {
				    if (news & Q(1) << (integral_q_bits - 3)) {
					side--;
					fast -= 2;
					*side = *fast;
					side--;
					fast--;
					*side = *fast;
					side--;
					fast--;
					*side = *fast;
					side--;
					fast -= 2;
					*side = *fast;
				    } else {
					side--;
					fast -= 2;
					*side = *fast;
					side--;
					fast -= 2;
					*side = *fast;
					side--;
					fast--;
					*side = *fast;
					side--;
					fast--;
					*side = *fast;
				    }
				}

				news <<= 6;
			    }

			    for (; !(news & q_21_); news <<= 6) {
				side--;
				fast -= 2;
				*side = *fast;
				side--;
				fast -= 2;
				*side = *fast;
				side -= 2;
				fast -= 2;
				*side = *fast;
			    }

			    if (news & q_42_) {
				*tier = Lxdddddddd | Q(2) << (side - tier - 4);

				if (half_integral_q_bits % 3 ^ 2) {
				    side--;
				    fast--;
				    *side = *fast;

				    fast = file + integral_q_bits - 2;

				    side--;
				    *side = *fast;
				    side--;
				    fast -= 2;
				    *side = *fast;
				    side--;
				    fast -= 2;
				    *side = *fast;
				} else {
				    side--;
				    if (news & Q(1) << (integral_q_bits - 1)) {
					fast--;
					*side = *fast;
					side--;
					fast--;
					*side = *fast;
				    } else {
					fast -= 2;
					*side = *fast;
					side--;
					fast--;
					*side = *fast;
				    }

				    fast = file + integral_q_bits - 2;

				    side--;
				    *side = *fast;
				    side--;
				    fast -= 2;
				    *side = *fast;
				}
			    } else {
				if (half_integral_q_bits % 3 ^ 2) {
				    fast = file + integral_q_bits - 2;

				    side--;
				    *side = *fast;
				    side--;
				    fast -= 2;
				    *side = *fast;
				    side -= 2;
				    fast -= 2;
				    *side = *fast;
				} else {
				    side--;
				    fast -= 2;
				    *side = *fast;

				    fast = file + integral_q_bits - 2;

				    side--;
				    *side = *fast;
				    side -= 2;
				    fast -= 2;
				    *side = *fast;
				}
			    }

			    deck =
				integral_q_bits / 4 - integral_q_bits / 6 - 2;
			    for (; deck; deck--) {
				side--;
				fast -= 2;
				*side = *fast;
				side--;
				fast -= 2;
				*side = *fast;
				side -= 2;
				fast -= 2;
				*side = *fast;
			    }

			    {
				side--;
				fast -= 2;
				*side = *fast;
				side--;
				fast -= 2;
				*side = *fast;
			    }
			}

			if (call ^ integral_q_last) {
			    int excess;

			    *head = *fast;

			    status = edge(node, call);

			    excess = bpfset(bpfset)->link_m.free
				(bpfset(bpfset)->link_m.data, file);
			    if (excess) {
				status = FREE_ERROR;
			    }
			} else {
			    int excess;

			    *turn = *fast;

			    status = east(node, call);

			    excess = bpfset(bpfset)->link_m.free
				(bpfset(bpfset)->link_m.data, file);
			    if (excess) {
				status = FREE_ERROR;
			    }
			}
		    } while (0);
		}
	    }
	} else {
	    if (lead.fine) {
		if (call < 2) {
		    if (pipe & ((Q(1) << call) - 1) << 1) {
			KD(node)[1] = lead.mode;
		    } else {
			if (date) {
			    *path = lead.mode;
			} else {
			    if (text_rate(0, bpfset)) {
				integral_q fold;

				fold = *KD(high);
				if (fold & ((Q(1) << pass) - 1) << 1) {
				    if (fold & (Q(1) << pass)) {
					KD(high)[pass] = lead.mode;
				    } else {
					KD(high)[pass - 1] = lead.mode;
				    }
				} else {
				    *path = lead.mode;
				}
			    } else {
				*path = lead.mode;
			    }
			}
		    }
		} else {
		    if (call & 1) {
			if (pipe & (Q(1) << call)) {
			    KD(node)[call] = lead.mode;
			} else {
			    KD(node)[call - 1] = lead.mode;
			}
		    } else {
			KD(node)[call] = lead.mode;
		    }
		}
	    }
	}
    }

    return status;
}


static int
miss(void *bpfset, unsigned mind, void *node, unsigned pick, unsigned rate,
     integral_q mode)
{
    int status;

    if (1) {
	integral_q *club, *head, *past, *tier, zero;

	if (pick) {
	    club = ND(node)[pick - 1];
	    head = KD(node) + pick;
	    if (pick ^ mind) {
		past = KD(node) + pick + 1;
		tier = ND(node)[pick + 1];
	    } else {
		past = &zero;
		tier = NULL;
	    }
	} else {
	    club = NULL;
	    head = &zero;
	    past = KD(node) + pick + 1;
	    tier = ND(node)[pick + 1];
	}

	do {
	    integral_q *file;

	    file = ND(node)[pick];

	    if (*file & ~Lx55555555) {
	    } else {
		status = _libx1f4l2_sand_bppset
		    (bpfset, mind, node, &club, &head, &pick, file, &past,
		     &tier, rate, mode, &zero);
		if (status) {
		    if (status < LAST_ERROR) {
			break;
		    } else {
			if (status ^ (LAST_ERROR + integral_q_bits)) {
			    status = shut
				(bpfset, bpfset(bpfset)->link_a.fpnews.node,
				 rate, mode);

			    break;
			} else {
			    mind--;

			    file = ND(node)[pick];
			}
		    }
		}
	    }

	    status = lose
		(bpfset, club, head, node, trap(pick), past, tier, rate - 1,
		 mode);
#if __MISS_CALL__
#else
	    if (status) {
		if (status < LAST_ERROR) {
		} else {
		    integral_q last, next;
		    unsigned copy, deck, ever;

		    copy = status - LAST_ERROR;

		    status = 0;

		    if (hold) {
			if (hold ^ 2) {
			    deck = 1;
			    ever = 2;

			    club = tier;

			    tier = ND(node)[pick + 2];

			    next = *tier & ~Lx55555555;
			} else {
			    deck = 2;
			    ever = 1;

			    tier = club;

			    club = ND(node)[pick - 2];

			    last = *club & ~Lx55555555;
			}
		    } else {
			if (club) {
			    last = *club & ~Lx55555555;
			    l2_q_bitcount(deck, last);
			} else {
			    deck = 0;

			    /*
			     * rubbish warnings...
			     */
			    last = 0;
			}

			if (tier) {
			    next = *tier & ~Lx55555555;
			    l2_q_bitcount(ever, next);
			} else {
			    ever = 0;
			}
		    }

		    if (deck < ever) {
			if (1) {
			    integral_q data, *fast, *side, text;

			    fast = tier;

			    text = next;
			    text &= (text - 1) << 1;

			    *KD(fast) = Lx55555555 | (next & ~text);

			    text >>= 2;
			    if (text) {
				fast += 2;

				side = fast + integral_q_bits - 1;

				last = *fast;

				data = *side;

				text >>= 2;

				for (; text; text >>= 2) {
				    *fast = *(fast + 2);
				    fast += 2;
				    *side = *(side + 2);
				    side += 2;
				}
				{
				    *fast = *(fast + 1);
				}
				{
				    *side = *(side + 1);
				}
			    } else {
				last = fast[1];

				data = fast[integral_q_bits];
			    }

			    if (hold) {
				fast = club + 2;

				side = fast + integral_q_bits - 1;

				text = KD(node)[pick + 2];

				KD(node)[pick + 2] = last;

				last = *fast;

				next = data;

				data = *side;

				ever = half_integral_q_bits - 2;
				for (; ever; ever--) {
				    *fast = *(fast + 2);
				    fast += 2;
				    *side = *(side + 2);
				    side += 2;
				}
				{
				    *side = *(side + 2);
				    side += 2;
				}
				{
				    *fast = text;
				}
				{
				    *side = next;
				}
			    }

			    fast = file;

			    *KD(fast) = Lx55555555;

			    if (copy ^ integral_q_last) {
				if (copy ^ 1) {
				    fast += copy - 1;

				    side = fast + integral_q_last + 2;

				    ever = (integral_q_last - copy) >> 1;
				    for (; ever; ever--) {
					*fast = *(fast + 2);
					fast += 2;
					*side = *(side + 2);
					side += 2;
				    }
				    {
					*fast = *past;
				    }
				    {
					*side = data;
				    }
				} else {
				    fast += 2;

				    side = fast + integral_q_bits - 1;

				    ever = half_integral_q_bits - 2;
				    for (; ever; ever--) {
					*fast = *(fast + 2);
					fast += 2;
					*side = *(side + 2);
					side += 2;
				    }
				    {
					*fast = *past;
				    }
				    {
					*side = *(side + 2);
					side += 2;
				    }
				    {
					*side = data;
				    }
				}
			    } else {
				fast += integral_q_bits - 2;

				*fast = *past;

				fast += integral_q_bits + 1;

				{
				    *(fast - 2) = *fast;
				}

				{
				    *fast = data;
				}
			    }
			}

			KD(node)[pick + 1] = last;
		    } else {
			if (1) {
			    integral_q data, *fast, *side, text;

			    fast = club;

			    text = last;

			    if (text & Q(1) << integral_q_last) {
				next = fast[integral_q_last];

			        fast += (integral_q_bits << 1) - 1;

				data = *fast;
				*fast = *(fast - 1);

				*KD(club) = Lx55555555
				    | (last ^ Q(1) << integral_q_last);
			    } else {
				fast += integral_q_bits - 2;

				side = fast + integral_q_bits + 1;

				next = *fast;

				data = *side;

				text <<= 2;

				for (; !(text & Q(1) << integral_q_last);
				     text <<= 2) {
				    *fast = *(fast - 2);
				    fast -= 2;
				    *side = *(side - 2);
				    side -= 2;
				}
				{
				    *fast = *(fast - 1);
				}
				{
				    *side = *(side - 2);
				    side -= 2;
				}
				{
				    *side = *(side - 1);
				}

				*KD(club) = Lx55555555
				    | (last ^ Q(1) << (fast - club - 1));
			    }

			    if (hold) {
				fast = tier + integral_q_bits - 2;

				side = fast + integral_q_bits + 1;

				text = KD(node)[pick - 1];

				KD(node)[pick - 1] = next;

				next = *fast;

				last = data;

				data = *side;

				ever = half_integral_q_bits - 2;
				for (; ever; ever--) {
				    *fast = *(fast - 2);
				    fast -= 2;
				    *side = *(side - 2);
				    side -= 2;
				}
				{
				    *side = *(side - 2);
				    side -= 2;
				}
				{
				    *fast = text;
				}
				{
				    *side = last;
				}
			    }

			    fast = file;

			    *KD(fast) = Lx55555555;

			    if (copy ^ integral_q_last) {
				if (copy ^ 1) {
				    fast += copy - 1;

				    side = fast + 1 + integral_q_bits;

				    ever = (copy - 3) >> 1;
				    for (; ever; ever--) {
					*fast = *(fast - 2);
					fast -= 2;
					*side = *(side - 2);
					side -= 2;
				    }
				    {
					*fast = *head;
				    }
				    {
					*side = *(side - 2);
					side -= 2;
				    }
				    {
					*side = data;
				    }
				} else {
				    fast += 2;

				    *fast = *head;

				    fast += integral_q_last;

				    {
					*fast = data;
				    }
				}
			    } else {
				fast += integral_q_bits - 2;

				side = fast + integral_q_bits - 1;

				ever = half_integral_q_bits - 2;
				for (; ever; ever--) {
				    *fast = *(fast - 2);
				    fast -= 2;
				    *side = *(side - 2);
				    side -= 2;
				}
				{
				    *fast = *head;
				}
				{
				    *side = data;
				}
			    }
			}

			KD(node)[pick] = next;
		    }
		}
	    }
#endif				/* __MISS_CALL__ */
	} while (0);
    }

    return status;
}


static int
shut(void *bpfset, void *node, unsigned rate, integral_q mode)
{
    int status;
    integral_q *call;
    unsigned mind, pick, size;

    call = node;

    size = *call;

    mind = size;

    call++;

    size--;
    while (size) {
	unsigned half;

	half = (size + 1) >> 1;
	if (mode < call[half]) {
	    size = half - 1;
	} else {
	    call += half;
	    size -= half;
	}
    }

    if (mode < *call) {
    } else {
	call++;
    }

    pick = call - 1 - (integral_q *) node;

    if (rate ^ 1) {
	status = miss(bpfset, mind, node, pick, rate - 1, mode);
    } else {
	integral_q *file;
	struct fine_type lead;

	lead.fine = 0;

	/*
	 * rubbish warnings...
	 */
	lead.mode = 0;

	file = (integral_q *) ND(node)[pick];

	status = fine(bpfset, file, mode, &lead);
	if (status) {
	    if (status < LAST_ERROR) {
	    } else {
		integral_q *club, last, next, *tier;
		unsigned copy, deck, ever;

		copy = status - LAST_ERROR;

		status = 0;

		if (pick) {
		} else {
		    lead.fine = 0;
		}

		if (pick) {
		    club = (integral_q *) ND(node)[pick - 1];
		    last = *club;
		    l2_q_bitcount(deck, last);
		} else {
		    deck = 0;

		    /*
		     * rubbish warnings...
		     */
		    last = 0;
		    club = NULL;
		}

		if (pick ^ mind) {
		    tier = (integral_q *) ND(node)[pick + 1];
		    next = *tier;
		    l2_q_bitcount(ever, next);
		} else {
		    ever = 0;

		    /*
		     * rubbish warnings...
		     */
		    next = 0;
		    tier = NULL;
		}

		if (half_integral_q_bits + 1 < deck
		    || half_integral_q_bits + 1 < ever) {
		    if (deck < ever) {
			if (lead.fine) {
			    KD(node)[pick] = lead.mode;
			}

#if __HALF_TRANSFER__
			if (1) {
			    integral_q *fast, *side, text;

			    fast = file;

			    if (copy & 1) {
			    } else {
				text = *fast & ~Lx55555555;

				if (text >> copy) {
				    fast += copy;

				    next = text >> copy >> 2;

# if __UNROLL_TRANSFER__
				    mode = next >> 2;
				    for (; mode; mode >>= 4) {
					*fast = *(fast + 2);
					fast += 2;
					*fast = *(fast + 2);
					fast += 2;
				    }
				    if (next & Lx22222222) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# else
				    for (; next; next >>= 2) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast + 1);
				    }
				} else {
				    fast += copy;

				    next = text
					<< (integral_q_bits - copy + 1);

# if __UNROLL_TRANSFER__
				    mode = next << 2;
				    for (; mode; mode <<= 4) {
					*fast = *(fast - 2);
					fast -= 2;
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    if (next & Lx44444444) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
# else
				    for (; next; next <<= 2) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast - 1);
				    }
				}
			    }

			    ever -= half_integral_q_bits;
			    ever &= ~1;

			    fast = file;

			    *KD(fast) =
				Lx55555555 | ~Q(0) << (integral_q_bits - ever);

			    fast += integral_q_bits - ever + 1;

			    side = fast + 1;

			    ever >>= 1;

			    deck = ever - 1;
			    for (; deck; deck--) {
				*fast = *side;
				side += 2;
				fast++;
			    }

			    side = tier;

			    text = *side;

			    while (ever & ~1) {
				if (text & Q(2)) {
				    side++;
				    *fast = *side;
				    fast++;
				    side++;
				    *fast = *side;
				    fast++;

				    ever -= 2;
				} else {
				    ever--;

				    side += 2;
				    *fast = *side;
				    fast++;
				}

				text >>= 2;
			    }
			    if (ever) {
				if (text & Q(2)) {
				    side++;
				    *fast = *side;
				    fast++;

				    side++;

				    fast = tier + 2;

				    text >>= 1;
				} else {
				    side += 2;
				    *fast = *side;
				    fast++;

				    side++;

				    fast = tier + 2;

				    text >>= 2;

				    if (fast < side) {
					text >>= 1;
					if (text & Q(1)) {
					    *fast = *side;
					    fast += 2;
					} else {
					}

					side++;
				    }
				}
			    } else {
				side++;

				fast = tier + 2;

				if (fast < side) {
				    text >>= 1;
				    if (text & Q(1)) {
					*fast = *side;
					fast += 2;
				    } else {
				    }

				    side++;
				}
			    }
			    while (fast < side) {
				if (text & Q(4)) {
				    *fast = *side;
				    fast += 2;
				    side++;
				    *fast = *side;
				    fast += 2;
				    side++;
				} else {
				    *fast = *side;
				    fast += 2;
				    side += 2;
				}

				text >>= 2;
			    }

			    *KD(tier) &= Lx55555555 | ~Q(0) << (fast - tier);
			}

			KD(node)[pick + 1] = *(tier + 2);
#else
			integral_q port;

			if (1) {
			    integral_q *fast, text;

			    fast = tier;

			    text = next & ~Lx55555555;
			    text &= (text - 1) << 1;

			    text >>= 2;
			    if (text) {
				fast += 2;

				last = *fast;

				text >>= 2;
				if (text) {
				    port = *(fast + 2);
				    *fast = port;
				    fast += 2;

				    text >>= 2;

# if __UNROLL_TRANSFER__
				    mode = text >> 2;
				    for (; mode; mode >>= 4) {
					*fast = *(fast + 2);
					fast += 2;
					*fast = *(fast + 2);
					fast += 2;
				    }
				    if (text & Lx22222222) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# else
				    for (; text; text >>= 2) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast + 1);
				    }
				} else {
				    port = fast[1];
				    *fast = port;
				}
			    } else {
				last = fast[1];
				port = fast[2];
			    }

			    *KD(tier) = next ^ Q(1) << (fast - tier + 1);

			    fast = file;

			    if (copy & 1) {
				*KD(fast) =
				    Lx55555555 | (Q(1) << integral_q_last);

				*(fast + integral_q_last) = last;
			    } else {
				text = *fast & ~Lx55555555;

				if (text >> copy) {
				    *KD(fast) =
					Lx55555555 | (Q(1) << integral_q_last);

				    fast += copy;

				    next = text >> copy >> 2;

# if __UNROLL_TRANSFER__
				    mode = next >> 2;
				    for (; mode; mode >>= 4) {
					*fast = *(fast + 2);
					fast += 2;
					*fast = *(fast + 2);
					fast += 2;
				    }
				    if (next & Lx22222222) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# else
				    for (; next; next >>= 2) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast + 1);
				    }

				    {
					*(file + integral_q_last) = last;
				    }
				} else {
				    if (Q(1) << copy
					< text << (integral_q_bits - copy)) {
					*KD(fast) = Lx55555555
					    | (Q(1) << integral_q_last);

					fast += copy;

					next = text
					    << (integral_q_bits - copy + 1);

# if __UNROLL_TRANSFER__
					mode = next << 2;
					for (; mode; mode <<= 4) {
					    *fast = *(fast - 2);
					    fast -= 2;
					    *fast = *(fast - 2);
					    fast -= 2;
					}
					if (next & Lx44444444) {
					    *fast = *(fast - 2);
					    fast -= 2;
					}
# else
					for (; next; next <<= 2) {
					    *fast = *(fast - 2);
					    fast -= 2;
					}
# endif				/* __UNROLL_TRANSFER__ */
					{
					    *fast = *(fast - 1);
					}

					{
					    *(file + integral_q_last) = last;
					}
				    } else {
					*KD(fast) = text | Lx55555555;

					fast += copy;

					deck =
					    (integral_q_bits - 2 - copy) >> 1;
# if __UNROLL_TRANSFER__
					ever = deck >> 1;
					for (; ever; ever--) {
					    *fast = *(fast + 2);
					    fast += 2;
					    *fast = *(fast + 2);
					    fast += 2;
					}
					if (deck & 1) {
					    *fast = *(fast + 2);
					    fast += 2;
					}
# else
					for (; deck; deck--) {
					    *fast = *(fast + 2);
					    fast += 2;
					}
# endif				/* __UNROLL_TRANSFER__ */
					{
					    *fast = last;
					}
				    }
				}
			    }
			}

			KD(node)[pick + 1] = port;
#endif				/* __HALF_TRANSFER__ */
		    } else {
#if __HALF_TRANSFER__
			if (1) {
			    integral_q *fast, *side, text;

			    fast = file;

			    if (copy & 1) {
			    } else {
				text = *fast & ~Lx55555555;

				if (text >> copy) {
				    fast += copy;

				    last = text >> copy >> 2;

# if __UNROLL_TRANSFER__
				    mode = last >> 2;
				    for (; mode; mode >>= 4) {
					*fast = *(fast + 2);
					fast += 2;
					*fast = *(fast + 2);
					fast += 2;
				    }
				    if (last & Lx22222222) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# else
				    for (; last; last >>= 2) {
					*fast = *(fast + 2);
					fast += 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast + 1);
				    }
				} else {
				    fast += copy;

				    last = text
					<< (integral_q_bits - copy + 1);

# if __UNROLL_TRANSFER__
				    mode = last << 2;
				    for (; mode; mode <<= 4) {
					*fast = *(fast - 2);
					fast -= 2;
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    if (last & Lx44444444) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
# else
				    for (; last; last <<= 2) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast - 1);
				    }
				}
			    }

			    deck -= half_integral_q_bits;
			    deck &= ~1;

			    fast = file;

			    *KD(fast) = Lx55555555 | ((Q(1) << deck) - 1);

			    fast += deck - 1;

			    side = fast - 1;

			    deck >>= 1;

			    ever = deck - 1;
			    for (; ever; ever--) {
				*fast = *side;
				side -= 2;
				fast--;
			    }

			    side = club;

			    text = *side;

			    side += integral_q_bits;

			    while (deck & ~1) {
				if (text & Q(1) << integral_q_last) {
				    side--;
				    *fast = *side;
				    fast--;
				    side--;
				    *fast = *side;
				    fast--;

				    deck -= 2;
				} else {
				    deck--;

				    side -= 2;
				    *fast = *side;
				    fast--;
				}

				text <<= 2;
			    }
			    if (deck) {
				if (text & Q(1) << integral_q_last) {
				    side--;
				    *fast = *side;

				    side--;

				    KD(node)[pick] = *fast;

				    fast = club + integral_q_bits - 2;

				    text <<= 1;
				} else {
				    side -= 2;
				    *fast = *side;

				    side--;

				    KD(node)[pick] = *fast;

				    fast = club + integral_q_bits - 2;

				    text <<= 2;

				    if (side < fast) {
					if (text & Q(1) << integral_q_last) {
					    *fast = *side;
					    fast -= 2;
					}

					side--;

					text <<= 1;
				    }
				}
			    } else {
				side--;

				KD(node)[pick] = *(fast + 1);

				fast = club + integral_q_bits - 2;

				if (side < fast) {
				    if (text & Q(1) << integral_q_last) {
					*fast = *side;
					fast -= 2;
				    }

				    side--;

				    text <<= 1;
				}
			    }
			    while (side < fast) {
				if (text & Q(1) << (integral_q_bits - 2)) {
				    *fast = *side;
				    fast -= 2;
				    side--;
				    *fast = *side;
				    fast -= 2;
				    side--;
				} else {
				    *fast = *side;
				    fast -= 2;
				    side -= 2;
				}

				text <<= 2;
			    }

			    *KD(club) &=
				Lx55555555 | ((Q(1) << (fast - club)) - 1);
			}
#else
			if (1) {
			    integral_q *fast, text;

			    fast = club;

			    text = *fast;
			    if (text & Q(1) << integral_q_last) {
				fast += integral_q_bits;
				next = *(fast - 1);
			    } else {
				fast += integral_q_bits - 2;

				next = *fast;

				text <<= 2;

# if __UNROLL_TRANSFER__
				for (;
				     !(text & Q(10) << (integral_q_bits - 4));
				     text <<= 4) {
				    *fast = *(fast - 2);
				    fast -= 2;
				    *fast = *(fast - 2);
				    fast -= 2;
				}
				if (!(text & Q(2) << (integral_q_bits - 2))) {
				    *fast = *(fast - 2);
				    fast -= 2;
				}
# else
				for (; !(text & Q(1) << integral_q_last);
				     text <<= 2) {
				    *fast = *(fast - 2);
				    fast -= 2;
				}
# endif				/* __UNROLL_TRANSFER__ */
				{
				    *fast = *(fast - 1);
				}
			    }

			    *KD(club) = last ^ Q(1) << (fast - club - 1);

			    fast = file;

			    if (copy & 1) {
				*KD(fast) = Lx55555555 | 2;

				*(fast + 1) = next;
			    } else {
				text = *fast & ~Lx55555555;

				if (text >> copy) {
				    if (text >> copy < Q(1) << copy) {
					*KD(fast) = Lx55555555 | 2;

					fast += copy;

					last = text >> copy >> 2;

# if __UNROLL_TRANSFER__
					mode = last >> 2;
					for (; mode; mode >>= 4) {
					    *fast = *(fast + 2);
					    fast += 2;
					    *fast = *(fast + 2);
					    fast += 2;
					}
					if (last & Lx22222222) {
					    *fast = *(fast + 2);
					    fast += 2;
					}
# else
					for (; last; last >>= 2) {
					    *fast = *(fast + 2);
					    fast += 2;
					}
# endif				/* __UNROLL_TRANSFER__ */
					{
					    *fast = *(fast + 1);
					}

					{
					    *(file + 1) = next;
					}
				    } else {
					*KD(fast) = text | Lx55555555;

					fast += copy;

					ever = (copy - 2) >> 1;
# if __UNROLL_TRANSFER__
					deck = ever >> 1;
					for (; deck; deck--) {
					    *fast = *(fast - 2);
					    fast -= 2;
					    *fast = *(fast - 2);
					    fast -= 2;
					}
					if (ever & 1) {
					    *fast = *(fast - 2);
					    fast -= 2;
					}
# else
					for (; ever; ever--) {
					    *fast = *(fast - 2);
					    fast -= 2;
					}
# endif				/* __UNROLL_TRANSFER__ */
					{
					    *fast = next;
					}
				    }
				} else {
				    *KD(fast) = Lx55555555 | 2;

				    fast += copy;

				    last = text
					<< (integral_q_bits - copy + 1);

# if __UNROLL_TRANSFER__
				    mode = last << 2;
				    for (; mode; mode <<= 4) {
					*fast = *(fast - 2);
					fast -= 2;
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    if (last & Lx44444444) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
# else
				    for (; last; last <<= 2) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
# endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast - 1);
				    }

				    {
					*(file + 1) = next;
				    }
				}
			    }
			}

			KD(node)[pick] = next;
#endif				/* __HALF_TRANSFER__ */
		    }
		} else {
		    do {
			integral_q *fast;

			if (mind ^ 1) {
			} else {
			    status = _libx1f4l2_fold_bppset
				(bpfset, club, file, tier, pick, copy, node);
			    if (1) {
				break;
			    }
			}

			if (pick) {
			    if (pick ^ mind) {
			    } else {
				_libx1f4l2_gray_bppset
				    (bpfset, last, &club, &node, &pick, &file,
				     &tier, next, &copy);
				if (tier) {
				} else {
				    break;
				}
			    }
			} else {
			    if (1) {
				_libx1f4l2_blue_bppset
				    (bpfset, last, &club, &node, &pick, &file,
				     &tier, next, &copy);
				if (club) {
				} else {
				    break;
				}
			    }
			}

			if (1) {
			    integral_q text;

			    fast = file;

			    if (copy & 1) {
			    } else {
				text = *fast & ~Lx55555555;

				if (text >> copy) {
				    fast += copy;

				    next = text >> copy >> 2;

#if __UNROLL_TRANSFER__
				    mode = next >> 2;
				    for (; mode; mode >>= 4) {
					*fast = *(fast + 2);
					fast += 2;
					*fast = *(fast + 2);
					fast += 2;
				    }
				    if (next & Lx22222222) {
					*fast = *(fast + 2);
					fast += 2;
				    }
#else
				    for (; next; next >>= 2) {
					*fast = *(fast + 2);
					fast += 2;
				    }
#endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast + 1);
				    }
				} else {
				    fast += copy;

				    next = text
					<< (integral_q_bits - copy + 1);

#if __UNROLL_TRANSFER__
				    mode = next << 2;
				    for (; mode; mode <<= 4) {
					*fast = *(fast - 2);
					fast -= 2;
					*fast = *(fast - 2);
					fast -= 2;
				    }
				    if (next & Lx44444444) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
#else
				    for (; next; next <<= 2) {
					*fast = *(fast - 2);
					fast -= 2;
				    }
#endif				/* __UNROLL_TRANSFER__ */
				    {
					*fast = *(fast - 1);
				    }
				}
			    }
			}

			{
			    integral_q news, *side;

			    fast = club;

			    news = *fast & ~Lx55555555;

			    if (news & Q_42_) {
				*fast = (Lxdddddddd | Q(2))
				    ^ Q(1) << integral_q_last;

				if (news & 2) {
				    fast += 4;

				    side = fast - 1;

				    *side = *fast;
				    fast += 2;
				    side++;
				} else {
				    if (news & 8) {
					fast += 2;

					side = fast - 1;

					*side = *fast;
					fast++;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
					fast += 2;
					side++;
				    } else {
					fast += 2;

					side = fast - 1;

					*side = *fast;
					fast += 2;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
					fast++;
					side++;
				    }
				}
			    } else {
				fast += 4;

				side = fast - 1;

				*side = *fast;
				fast += 2;
				side++;
			    }

			    news |= Q(1) << (integral_q_bits - 2);
			    news >>= 6;

			    for (; !(news & Q_63_); news >>= 6) {
				*side = *fast;
				fast += 2;
				side += 2;
				*side = *fast;
				fast += 2;
				side++;
				*side = *fast;
				fast += 2;
				side++;
			    }

			    if (news & ~((Q(1) << 6) - 1)) {
				*club = (Lxdddddddd | Q(2) << (side - club))
				    ^ Q(1) << integral_q_last;

				if (news & 2) {
				    *side = *fast;
				    fast++;
				    side++;
				    *side = *fast;
				    fast++;
				    side++;
				    *side = *fast;
				    fast += 2;
				    side++;
				    *side = *fast;
				    fast += 2;
				    side++;
				} else {
				    if (news & 8) {
					*side = *fast;
					fast += 2;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
					fast += 2;
					side++;
				    } else {
					*side = *fast;
					fast += 2;
					side++;
					*side = *fast;
					fast += 2;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
					fast++;
					side++;
				    }
				}

				news >>= 6;
			    }

			    for (; !(news & Q_21_); news >>= 6) {
				*side = *fast;
				fast += 2;
				side += 2;
				*side = *fast;
				fast += 2;
				side++;
				*side = *fast;
				fast += 2;
				side++;
			    }

			    if (news & Q_42_) {
				*club = (Lxdddddddd | Q(2) << (side - club))
				    ^ Q(1) << integral_q_last;

				if (half_integral_q_bits % 3 ^ 2) {
				    *side = *fast;
				    fast++;
				    side++;
				    *side = *fast;

				    fast = file;

				    fast += 2;
				    side++;
				    *side = *fast;
				    fast += 2;
				    side++;
				    *side = *fast;
				    side++;
				} else {
				    *side = *fast;
				    if (news & 2) {
					fast++;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
				    } else {
					fast += 2;
					side++;
					*side = *fast;
					fast++;
					side++;
					*side = *fast;
				    }

				    fast = file;

				    fast += 2;
				    side++;
				    *side = *fast;
				    side++;
				}
			    } else {
				if (half_integral_q_bits % 3 ^ 2) {
				    *side = *fast;

				    fast = file;

				    fast += 2;
				    side += 2;
				    *side = *fast;
				    fast += 2;
				    side++;
				    *side = *fast;
				    side++;
				} else {
				    *side = *fast;
				    fast += 2;
				    side += 2;
				    *side = *fast;

				    fast = file;

				    fast += 2;
				    side++;
				    *side = *fast;
				    side++;
				}
			    }

			    deck =
				integral_q_bits / 4 - integral_q_bits / 6 - 2;
			    for (; deck; deck--) {
				fast += 2;
				*side = *fast;
				side += 2;
				fast += 2;
				*side = *fast;
				side++;
				fast += 2;
				*side = *fast;
				side++;
			    }

			    {
				fast += 2;
				*side = *fast;
				side += 2;
				fast += 2;
				*side = *fast;
			    }

			    fast = tier;

			    news = *tier & ~Lx55555555;

			    fast += integral_q_bits;

			    side = fast;

			    news |= Q(1);

			    for (; !(news & q_42_); news <<= 6) {
				side--;
				fast -= 2;
				*side = *fast;
				side--;
				fast -= 2;
				*side = *fast;
				side -= 2;
				fast -= 2;
				*side = *fast;
			    }

			    if (news & q_21_) {
			    } else {
				*tier = Lxdddddddd | Q(2) << (side - tier - 4);

				if (news & Q(1) << (integral_q_bits - 1)) {
				    side--;
				    fast--;
				    *side = *fast;
				    side--;
				    fast--;
				    *side = *fast;
				    side--;
				    fast -= 2;
				    *side = *fast;
				    side--;
				    fast -= 2;
				    *side = *fast;
				} else {
				    if (news & Q(1) << (integral_q_bits - 3)) {
					side--;
					fast -= 2;
					*side = *fast;
					side--;
					fast--;
					*side = *fast;
					side--;
					fast--;
					*side = *fast;
					side--;
					fast -= 2;
					*side = *fast;
				    } else {
					side--;
					fast -= 2;
					*side = *fast;
					side--;
					fast -= 2;
					*side = *fast;
					side--;
					fast--;
					*side = *fast;
					side--;
					fast--;
					*side = *fast;
				    }
				}

				news <<= 6;
			    }

			    for (; !(news & q_21_); news <<= 6) {
				side--;
				fast -= 2;
				*side = *fast;
				side--;
				fast -= 2;
				*side = *fast;
				side -= 2;
				fast -= 2;
				*side = *fast;
			    }

			    if (news & q_42_) {
				*tier = Lxdddddddd | Q(2) << (side - tier - 4);

				if (half_integral_q_bits % 3 ^ 2) {
				    side--;
				    fast--;
				    *side = *fast;

				    fast = file + integral_q_bits - 2;

				    side--;
				    *side = *fast;
				    side--;
				    fast -= 2;
				    *side = *fast;
				    side--;
				    fast -= 2;
				    *side = *fast;
				} else {
				    side--;
				    if (news & Q(1) << (integral_q_bits - 1)) {
					fast--;
					*side = *fast;
					side--;
					fast--;
					*side = *fast;
				    } else {
					fast -= 2;
					*side = *fast;
					side--;
					fast--;
					*side = *fast;
				    }

				    fast = file + integral_q_bits - 2;

				    side--;
				    *side = *fast;
				    side--;
				    fast -= 2;
				    *side = *fast;
				}
			    } else {
				if (half_integral_q_bits % 3 ^ 2) {
				    fast = file + integral_q_bits - 2;

				    side--;
				    *side = *fast;
				    side--;
				    fast -= 2;
				    *side = *fast;
				    side -= 2;
				    fast -= 2;
				    *side = *fast;
				} else {
				    side--;
				    fast -= 2;
				    *side = *fast;

				    fast = file + integral_q_bits - 2;

				    side--;
				    *side = *fast;
				    side -= 2;
				    fast -= 2;
				    *side = *fast;
				}
			    }

			    deck =
				integral_q_bits / 4 - integral_q_bits / 6 - 2;
			    for (; deck; deck--) {
				side--;
				fast -= 2;
				*side = *fast;
				side--;
				fast -= 2;
				*side = *fast;
				side -= 2;
				fast -= 2;
				*side = *fast;
			    }

			    {
				side--;
				fast -= 2;
				*side = *fast;
				side--;
				fast -= 2;
				*side = *fast;
			    }
			}

			if (1) {
			    integral_q *side;
			    void **trap;

			    trap = ND(node) + pick;

			    memmove(trap, trap + 1,
				    (mind - pick) * sizeof(void *));

			    side = KD(node) + pick;

			    *side = *fast;

			    side++;

			    memmove(side, side + 1,
				    (mind - pick) * sizeof(integral_q));

			    *KD(node) = mind - 1;

			    status = bpfset(bpfset)->link_m.free
				(bpfset(bpfset)->link_m.data, file);
			    if (status) {
				status = FREE_ERROR;
			    }
			}
		    } while (0);
		}
	    }
	} else {
	    if (lead.fine) {
		if (pick) {
		    KD(node)[pick] = lead.mode;
		} else {
		}
	    }
	}
    }

    return status;
}


int
x1f4_cast_bppset(void *bpfset, integral_q mode)
{
    int status;
    unsigned rate;

    rate = bpfset(bpfset)->link_a.fpnews.link_v.rate;
    if (rate) {
	status = shut(bpfset, bpfset(bpfset)->link_a.fpnews.node, rate, mode);
    } else {
	status = clip(bpfset(bpfset)->link_a.fpnews.node, mode);
    }

    return status;
}
