/*
 * l4xx.4.c
 * Copyright (C) 2008-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <stddef.h>

#include <l4xx-defs.h>
#include <l4xx-inter.h>
#include <l4xx-types.h>

#define l4list(l4list)			((struct l4list_type *) (l4list))

static int link_list(struct l4list_type *, struct list64_type *);
static int push_list(struct l4list_type *, struct list64_type *, unsigned);

static int
link_list(struct l4list_type *l4list_data, struct list64_type *list64_data)
{
    l4list_data->link_a.list88.count = 1;
    l4list_data->link_a.list88.head64_data = list64_data;
    l4list_data->link_a.list88.stride = 1;
    l4list_data->link_a.list88.tail64_data = list64_data;

    l4list_data->link_c.class = 0;
    l4list_data->link_c.list64_data = list64_data;

    list64_data->last_0 = NULL;
    list64_data->last_1 = NULL;
    list64_data->next_1 = NULL;
    list64_data->next_7 = NULL;

    return 0;
}


static int
push_list(struct l4list_type *l4list_data, struct list64_type *list64_data,
	  unsigned index)
{
    int status;
    unsigned Count, count, stride;

    count = l4list_data->link_a.list88.count;

    Count = count + 1;

    l4list_data->link_a.list88.count = Count;

    stride = l4list_data->link_a.list88.stride;

    if (!index) {
	_libx1f4l2_hfix_l4list(l4list_data, stride == count, list64_data);
    } else {
	if (index == count) {
	    _libx1f4l2_tfix_l4list(l4list_data, stride == count, list64_data);
	} else {
	    _libx1f4l2_mfix_l4list
		(l4list_data, count, stride, index, list64_data);
	}
    }

    l4list_data->link_c.class = index;
    l4list_data->link_c.list64_data = list64_data;

    status = _libx1f4l2_tack_l4list(l4list_data);

    return status;
}


int
x1f4_push_l4list(void *l4list, unsigned index, unsigned size, void **push)
{
    int status;
    unsigned count;

    count = l4list(l4list)->link_a.list88.count;

    if (count < index) {
	status = DECK_ERROR;
    } else {
	void *list64;

	status = l4list(l4list)->link_m.link
	    (l4list(l4list)->link_m.data, &list64,
	     sizeof(struct list64_type) + size);
	if (status) {
	    status = LINK_ERROR;
	} else {
	    if (!count) {
		status = link_list(l4list, list64);
	    } else {
		status = push_list(l4list, list64, index);
	    }
	    if (status) {
	    } else {
		*push = ((struct list64_type *) list64) + 1;
	    }
	}
    }

    return status;
}
