/*
 * l4xx.6.c
 * Copyright (C) 2008-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <math.h>
#include <stddef.h>
#include <string.h>

#include <l4xx-defs.h>
#include <l4xx-inter.h>
#include <l4xx-types.h>

static int copy_list(struct l4list_type *, struct l4list_type *, unsigned,
		     int (*) (unsigned *, const void *));

static int
copy_list(struct l4list_type *l4list_data, struct l4list_type *l4list_text,
	  unsigned count, int (*size) (unsigned *, const void *))
{
    int (*link) (void *, void **, unsigned), status;
    struct list64_type *list64_data, *list64_miss = NULL;
    void *data;

    list64_data = l4list_text->link_a.list88.head64_data;

    l4list_data->link_a.list88.count = count;

    data = l4list_data->link_m.data;
    link = l4list_data->link_m.link;

    while (1) {
	unsigned copy;
	void *list64;

	size(&copy, list64_data + 1);

	status = link(data, &list64, sizeof(struct list64_type) + copy);
	if (status) {
	    status = LINK_ERROR;

	    _libx1f4l2_rack_l4list(l4list_data, list64_miss);

	    break;
	} else {
	    struct list64_type *list64_copy;

	    list64_copy = list64;

	    if (list64_miss) {
		list64_copy->last_1 = list64_miss;
		list64_miss->next_1 = list64_copy;
	    } else {
		list64_copy->last_1 = NULL;

		l4list_data->link_a.list88.head64_data = list64_copy;
	    }

	    list64_miss = list64_copy;

	    memcpy(list64_copy + 1, list64_data + 1, copy);
	}

	count--;
	if (!count) {
	    unsigned count, stride;

	    count = l4list_data->link_a.list88.count;

	    stride = sqrt(count);
	    if (stride < 1) {
		stride = 1;
	    } else {
		if (stride < count) {
		} else {
		    if (count == 1) {
			stride = 1;
		    } else {
			stride = count - 1;
		    }
		}
	    }

	    l4list_data->link_a.list88.stride = stride;

	    l4list_data->link_a.list88.tail64_data = list64;

	    _libx1f4l2_link_l4list
		(l4list_data->link_a.list88.head64_data, count, stride,
		 list64);

	    break;
	} else {
	    list64_data = list64_data->next_1;
	}
    }

    return status;
}


int
x1f4_copy_l4list(void **l4list, struct l4list_type *l4list_text,
		 int (*size) (unsigned *, const void *))
{
    int status;
    void *text;

    status = l4list_text->link_m.link
	(l4list_text->link_m.data, &text, sizeof(struct l4list_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	struct l4list_type *l4list_data;
	unsigned count;

	l4list_data = text;

	l4list_data->link_m = l4list_text->link_m;

	count = l4list_text->link_a.list88.count;
	if (!count) {
	    l4list_data->link_a.list88.count = 0;

	    *l4list = text;
	} else {
	    status = copy_list(l4list_data, l4list_text, count, size);
	    if (status) {
		l4list_text->link_m.free(l4list_text->link_m.data, text);
	    } else {
		*l4list = text;

		l4list_data->link_c.class = 0;
		l4list_data->link_c.list64_data =
		    l4list_data->link_a.list88.head64_data;
	    }
	}
    }

    return status;
}
