/*
 * l4xx.8.c
 * Copyright (C) 2008-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <l4xx-defs.h>
#include <l4xx-inter.h>
#include <l4xx-types.h>

#define l4list(l4list)			((struct l4list_type *) (l4list))

static int fast_list(struct l4list_type *, unsigned, unsigned);

static int
fast_list(struct l4list_type *l4list_data, unsigned index, unsigned class)
{
    unsigned count, stride;
    void *list64;

    count = l4list_data->link_a.list88.count;

    stride = l4list_data->link_a.list88.stride;

    _libx1f4l2_seek_l4list(l4list_data, index, &list64);

    _libx1f4l2_fast_l4list(l4list_data, index, list64);
    if (index < class) {
	class--;
    }
    if (!class) {
	_libx1f4l2_hfix_l4list(l4list_data, stride == count, list64);
    } else {
	if (class == count) {
	    _libx1f4l2_tfix_l4list(l4list_data, stride == count, list64);
	} else {
	    _libx1f4l2_mfix_l4list(l4list_data, count, stride, class, list64);
	}
    }

    l4list_data->link_c.class = class;

    return 0;
}


int
x1f4_move_l4list(void *l4list, unsigned index, unsigned class)
{
    int status;
    unsigned count;

    count = l4list(l4list)->link_a.list88.count;

    if (!(index < count)) {
	status = DECK_ERROR;
    } else {
	if (count < class) {
	    status = DECK_ERROR;
	} else {
	    if (index == class) {
		status = 0;
	    } else {
		if (index + 1 == class) {
		    status = 0;
		} else {
		    status = fast_list(l4list, index, class);
		}
	    }
	}
    }

    return status;
}
