/*
 * lsas-types.h
 * Copyright (C) 2008-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_lsas_types_H__
#define __LIBx1f4_lsas_types_H__

#include <config.h>

#include <lsas-config.h>

#include <trans.h>

#if SIZEOF_VOID_P == SIZEOF_LONG
# define integral_q			unsigned long
#else
# define integral_q			unsigned
#endif				/* SIZEOF_VOID_P == SIZEOF_LONG */

#if SIZEOF_LONG == 8
# define scale(___c) \
    (___c) >>= 3
#elif SIZEOF_LONG == 4
# define scale(___c) \
    (___c) >>= 2
#else
# define scale(___c) \
    (___c) /= SIZEOF_LONG
#endif				/* SIZEOF_LONG == 8 */

#define F1ST(near) \
    L1ST(*((near) + 1))

#define F2ND(near) \
    L2ND(*((near) + 1))

#define L1ST(near) \
    ((near) & 65535)

#define L2ND(near) \
    ((near) >> 16)

#define M1ST(near) \
    (near)

#define M2ND(near) \
    ((near) << 16)

#define P1ST(near) \
    ((near) & 65535)

#define P2ND(near) \
    ((near) & ~65535)

#define S1ST(near) \
    ((near) &= 65535)

#define S2ND(near) \
    ((near) >>= 16)

#define vert_noir(vert) \
    ((((vert) + 1) + 7) & ~7)

#define bleu_noir(bleu, noir) \
    (vert_noir(bleu) + (noir))

#define frame_rate(c) \
    ((c) | POINT_CAST)

#define seek_bit(c) \
    ((c) >> SHIFT_SEEK & 1)

#define point_rate(c) \
    ((c) & ~(POINT_CAST | POINT_SEEK))

#define lead_shrews			3

#define lead_switch \
    (sizeof(unsigned) * lead_shrews)

#define call_sizeof \
    (lead_switch + sizeof(struct fpcase_type))

#define dana_sizeof \
    (lead_switch + sizeof(struct fpfast_type))

#define X1ST(node, near) \
    ((void *) ((char *) (node) + F1ST((near))))

#define X2ND(node, near) \
    ((void *) ((char *) (node) + F2ND((near))))

typedef unsigned char byte;

typedef struct fpcase_type {
    unsigned miss[2];
    void *data[2];
} fpcase_type;

typedef struct fpfast_type {
    unsigned miss[2];
    void *data[1];
} fpfast_type;

typedef struct fplist_type {
    unsigned call, dana, rate;
    void *fpnode, *(*fvnode)[1];
} fplist_type;

typedef struct screen_type {
    unsigned inner, outer;
    struct trans_type *trans_data;
} screen_type;

typedef struct lsnote_type {
    struct {
	struct fplist_type fplist;
    } link_a;
#if __POST_SPLIT_CURE__ == __TREE_CURE__
    struct {
	unsigned lane;
	void *mail;
    } link_e;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
    struct {
	int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	    (*mode) (void *, void **, unsigned);
	void *data;
    } link_m;
} lsnote_type;

typedef struct quarter_type {
    const byte *name;
    unsigned aile, call, dana, deck, fast, four, half, lead;
#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
    unsigned mind;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
    unsigned near, pick;
#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
    unsigned post;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
    unsigned shut, side, slip, tear, tier, zero;
} quarter_type;

#endif				/* __LIBx1f4_lsas_types_H__ */
