/*
 * lsas.0.c
 * Copyright (C) 2008-2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * TODO
 *
 *.record unlinked nodes for insertion failures in lsaf.m.join and lsaf.w.long
 *     - add test for insertion failure
 *.fix the messed/mixed node accessing (fpcase/fpfast, unsigned, etc)
 *.reduce the size of the internal critical bit tree nodes to 8 bytes?
 *.B+trees?  prefix B+trees?
 *.a B+tree idea: since the short keys have the potential of lengthening the
 * internal CBTs paths, replace them with longer ones (e.g. instead 0xab use
 * 0xaaff)
 *.a B+tree idea: for the inner nodes put the CBT nodes in the slots
 * corresponding the index composed of their most significant bits.  this can
 * be done only on special operations - like node splitting.  when searching
 * the nodes, the slots corresponding the current search will be examined - for
 * success a number of path vertexes can be skipped.  for failure the number of
 * examining bits can be reduced until a fit is found.  the normal CBT search
 * procedure is applied for the remainder of the path.  the number of index
 * bits is determined by the number of slots (2 ^ number of bits).  this will
 * work well if the inner node CBTs are dense (the distance between vertexes is
 * small, ideally zero).
 *.use 64bit words (where available and fast) for the bit cache to reduce
 * number of tests (and maybe bit loads)
 */

#include <stddef.h>

#include <lsas-defs.h>
#include <lsas-inter.h>
#include <lsas-names.h>
#include <lsas-types.h>
#include <mask.0.h>
#include <trans.h>

int
x1f4_init_lsnote(void **lsnote, unsigned flags,
		 struct screen_type *screen_data)
{
    int (*link) (void *, void **, unsigned), status;
    void *data, *text;

    if (flags & TRANS_MASK) {
	struct trans_type *trans_data;

	trans_data = screen_data->trans_data;

	link = trans_data->link;
	data = trans_data->data;
    } else {
	link = _libx1f4l2_link_data;
	data = (void *) 0;
    }

    status = link(data, &text, sizeof(struct lsnote_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	struct lsnote_type *lsnote_data;

	lsnote_data = text;

	status = _libx1f4l2_fast_lsnote(lsnote_data, flags, screen_data);

	if (status) {
	    int (*free) (void *, void *);

	    if (flags & TRANS_MASK) {
		free = screen_data->trans_data->free;
	    } else {
		free = _libx1f4l2_free_data;
	    }

	    free(data, text);
	} else {
	    *lsnote = text;
	}
    }

    return status;
}
