/*
 * lsas.1.c
 * Copyright (C) 2008-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lsas-config.h>

#include <stddef.h>

#include <lsas-defs.h>
#include <lsas-names.h>
#include <lsas-types.h>
#include <mask.0.h>
#include <trans.h>

#define lsnote(lsnote)			((struct lsnote_type *) (lsnote))

int
x1f4_fast_lsnote(void *lsnote, unsigned flags,
		 struct screen_type *screen_data)
{
#if __POST_SPLIT_CURE__ == __TREE_CURE__
    int status;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
    unsigned call, dana;
#if __POST_SPLIT_CURE__ == __TREE_CURE__
    unsigned lane;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

    if (flags & INNER_MASK) {
	call = screen_data->inner / call_sizeof;
	if (call < 8) {
	    call = 8;
	} else {
	    if (8192 < call) {
		call = 8192;
	    }
	}
    } else {
	call = CALL_INNER / call_sizeof;
    }

    while (65536 < call_sizeof * call) {
	call >>= 1;
    }

    call &= ~1;

    if (flags & OUTER_MASK) {
	dana = screen_data->outer / dana_sizeof;
	if (dana < 8) {
	    dana = 8;
	} else {
	    if (8192 < dana) {
		dana = 8192;
	    }
	}
    } else {
	dana = DANA_OUTER / dana_sizeof;
    }

    while (65536 < dana_sizeof * dana) {
	dana >>= 1;
    }

    dana &= ~1;

#if __POST_SPLIT_CURE__ == __TREE_CURE__
    lane = (call < dana ? dana : call) >> 1;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

    if (flags & TRANS_MASK) {
	struct trans_type *trans_data;

	trans_data = screen_data->trans_data;

	lsnote(lsnote)->link_m.data = trans_data->data;
	lsnote(lsnote)->link_m.free = trans_data->free;
	lsnote(lsnote)->link_m.link = trans_data->link;
	lsnote(lsnote)->link_m.mode = trans_data->mode;
    } else {
	lsnote(lsnote)->link_m.free = _libx1f4l2_free_data;
	lsnote(lsnote)->link_m.link = _libx1f4l2_link_data;
	lsnote(lsnote)->link_m.mode = _libx1f4l2_mode_data;
    }

#if __POST_SPLIT_CURE__ == __TREE_CURE__
    status = lsnote(lsnote)->link_m.link
	(lsnote(lsnote)->link_m.data, &lsnote(lsnote)->link_e.mail,
	 lane * (sizeof(unsigned *) + sizeof(unsigned) * (2 + lead_shrews)));
    if (status) {
	status = LINK_ERROR;
#elif 0
    }
#else
    if (0) {
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
    } else {
	lsnote(lsnote)->link_a.fplist.fpnode = NULL;
	lsnote(lsnote)->link_a.fplist.fvnode = NULL;

	lsnote(lsnote)->link_a.fplist.call = call;
	lsnote(lsnote)->link_a.fplist.dana = dana;
	lsnote(lsnote)->link_a.fplist.rate = 0;

#if __POST_SPLIT_CURE__ == __TREE_CURE__
	lsnote(lsnote)->link_e.lane = lane;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
    }

#if __POST_SPLIT_CURE__ == __TREE_CURE__
    return status;
#else
    return 0;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
}
