/*
 * lsas.6.c
 * Copyright (C) 2008-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lsas-defs.h>
#include <lsas-names.h>
#include <lsas-types.h>

#define lsnote(lsnote)			((struct lsnote_type *) (lsnote))

static int back_node(void *, unsigned, void *,
		     int (*) (void *, void *, const char *, unsigned));
static int back_note(void *, unsigned, unsigned, void *,
		     int (*) (void *, void *, const char *, unsigned));
static int back_slip(void *, void *,
		     int (*) (void *, void *, const char *, unsigned));
static int back_tear(void *, unsigned, unsigned, unsigned, void *,
		     int (*) (void *, void *, const char *, unsigned));

static int
back_node(void *node, unsigned slip, void *back,
	  int (*call) (void *, void *, const char *, unsigned))
{
    int delete;
    unsigned *near, note;

    near = (void *) ((char *) node + slip);
    note = *near;
    if (note & POINT_CAST) {
	note = near[1];
	delete = back_node(node, L1ST(note), back, call);
	if (delete) {
	} else {
	    delete = back_node(node, L2ND(note), back, call);
	}
    } else {
	void *data;

	data = ((struct fpfast_type *) near)->data[0];
	delete = call(back, (char *) data + vert_noir(note), data, note);
    }

    return delete;
}


static int
back_note(void *node, unsigned miss, unsigned rate, void *back,
	  int (*call) (void *, void *, const char *, unsigned))
{
    int delete;
    struct fpcase_type *fpcase_data;
    unsigned *near, note;

    rate--;

    fpcase_data = (void *) ((char *) node + miss);
    if (rate) {
	delete = back_note(fpcase_data->data[1], miss, rate, back, call);
    } else {
	delete = back_slip(fpcase_data->data[1], back, call);
    }
    if (delete) {
    } else {
	near = node;
	near = (void *) ((char *) node + *near);
	note = *near;
	if (note & POINT_CAST) {
	    note = near[1];
	    delete = back_tear(node, L1ST(note), miss, rate, back, call);
	    if (delete) {
	    } else {
		delete = back_tear(node, L2ND(note), miss, rate, back, call);
	    }
	} else {
	    void *data;

	    data = ((struct fpcase_type *) near)->data[0];
	    delete = call(back, (char *) data + vert_noir(note), data, note);
	    if (delete) {
	    } else {
		if (rate) {
		    delete = back_note
			(((struct fpcase_type *) near)->data[1], miss, rate,
			 back, call);
		} else {
		    delete = back_slip
			(((struct fpcase_type *) near)->data[1], back, call);
		}
	    }
	}
    }

    return delete;
}


static int
back_slip(void *node, void *back,
	  int (*call) (void *, void *, const char *, unsigned))
{
    int delete;
    unsigned *near, note;

    near = node;
    near = (void *) ((char *) node + *near);
    note = *near;
    if (note & POINT_CAST) {
	note = near[1];
	delete = back_node(node, L1ST(note), back, call);
	if (delete) {
	} else {
	    delete = back_node(node, L2ND(note), back, call);
	}
    } else {
	void *data;

	data = ((struct fpfast_type *) near)->data[0];
	delete = call(back, (char *) data + vert_noir(note), data, note);
    }

    return delete;
}


static int
back_tear(void *node, unsigned slip, unsigned miss, unsigned rate, void *back,
	  int (*call) (void *, void *, const char *, unsigned))
{
    int delete;
    unsigned *near, note;

    near = (void *) ((char *) node + slip);
    note = *near;
    if (note & POINT_CAST) {
	note = near[1];
	delete = back_tear(node, L1ST(note), miss, rate, back, call);
	if (delete) {
	} else {
	    delete = back_tear(node, L2ND(note), miss, rate, back, call);
	}
    } else {
	void *data;

	data = ((struct fpcase_type *) near)->data[0];
	delete = call(back, (char *) data + vert_noir(note), data, note);
	if (delete) {
	} else {
	    if (rate) {
		delete = back_note
		    (((struct fpcase_type *) near)->data[1], miss, rate, back,
		     call);
	    } else {
		delete = back_slip
		    (((struct fpcase_type *) near)->data[1], back, call);
	    }
	}
    }

    return delete;
}


int
x1f4_span_lsnote(void *lsnote, void *back,
		 int (*call) (void *, void *, const char *, unsigned))
{
    int delete;
    unsigned rate;

    rate = lsnote(lsnote)->link_a.fplist.rate;
    if (rate) {
	delete = back_note
	    (lsnote(lsnote)->link_a.fplist.fpnode,
	     lead_switch * lsnote(lsnote)->link_a.fplist.call, rate, back,
	     call);
    } else {
	if (lsnote(lsnote)->link_a.fplist.fpnode) {
	    delete = back_slip
		(lsnote(lsnote)->link_a.fplist.fpnode, back, call);
	} else {
	    delete = 0;
	}
    }

    return delete;
}
