/*
 * lsas.l.c
 * Copyright (C) 2008-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lsas-defs.h>
#include <lsas-names.h>
#include <lsas-types.h>

#define lsnote(lsnote)			((struct lsnote_type *) (lsnote))

static int back_node(void *, unsigned, void *, int (*) (void *, void *));
static int back_note(void *, unsigned, unsigned, void *,
		     int (*) (void *, void *));
static int back_slip(void *, void *, int (*) (void *, void *));
static int back_tear(void *, unsigned, unsigned, unsigned, void *,
		     int (*) (void *, void *));

static int
back_node(void *node, unsigned slip, void *data, int (*free) (void *, void *))
{
    int status;
    unsigned *near, note;

    near = (void *) ((char *) node + slip);
    note = *near;
    if (note & POINT_CAST) {
	int excess;

	note = near[1];
	status = back_node(node, L1ST(note), data, free);
	excess = back_node(node, L2ND(note), data, free);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    } else {
	status = free(data, ((struct fpfast_type *) near)->data[0]);
	if (status) {
	    status = FREE_ERROR;
	}
    }

    return status;
}


static int
back_note(void *node, unsigned miss, unsigned rate, void *data,
	  int (*free) (void *, void *))
{
    int excess, status;
    struct fpcase_type *fpcase_data;
    unsigned *near, note;

    rate--;

    fpcase_data = (void *) ((char *) node + miss);
    if (rate) {
	status = back_note(fpcase_data->data[1], miss, rate, data, free);
    } else {
	status = back_slip(fpcase_data->data[1], data, free);
    }

    near = node;
    near = (void *) ((char *) node + *near);
    note = *near;
    if (note & POINT_CAST) {
	note = near[1];
	excess = back_tear(node, L1ST(note), miss, rate, data, free);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
	excess = back_tear(node, L2ND(note), miss, rate, data, free);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    } else {
	excess = free(data, ((struct fpcase_type *) near)->data[0]);
	if (excess) {
	    if (status) {
	    } else {
		status = FREE_ERROR;
	    }
	}
	if (rate) {
	    excess = back_note
		(((struct fpcase_type *) near)->data[1], miss, rate, data,
		 free);
	} else {
	    excess = back_slip
		(((struct fpcase_type *) near)->data[1], data, free);
	}
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    excess = free(data, node);
    if (excess) {
	if (status) {
	} else {
	    status = FREE_ERROR;
	}
    }

    return status;
}


static int
back_slip(void *node, void *data, int (*free) (void *, void *))
{
    int excess, status;
    unsigned *near, note;

    near = node;
    near = (void *) ((char *) node + *near);
    note = *near;
    if (note & POINT_CAST) {
	note = near[1];
	status = back_node(node, L1ST(note), data, free);
	excess = back_node(node, L2ND(note), data, free);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    } else {
	status = free(data, ((struct fpfast_type *) near)->data[0]);
	if (status) {
	    status = FREE_ERROR;
	}
    }

    excess = free(data, node);
    if (excess) {
	if (status) {
	} else {
	    status = FREE_ERROR;
	}
    }

    return status;
}


static int
back_tear(void *node, unsigned slip, unsigned miss, unsigned rate, void *data,
	  int (*free) (void *, void *))
{
    int excess, status;
    unsigned *near, note;

    near = (void *) ((char *) node + slip);
    note = *near;
    if (note & POINT_CAST) {
	note = near[1];
	status = back_tear(node, L1ST(note), miss, rate, data, free);
	excess = back_tear(node, L2ND(note), miss, rate, data, free);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    } else {
	status = free(data, ((struct fpcase_type *) near)->data[0]);
	if (status) {
	    status = FREE_ERROR;
	}

	if (rate) {
	    excess = back_note
		(((struct fpcase_type *) near)->data[1], miss, rate, data,
		 free);
	} else {
	    excess = back_slip
		(((struct fpcase_type *) near)->data[1], data, free);
	}
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


int
_libx1f4l2_ceil_lsnote(void *lsnote)
{
    int excess, (*free) (void *, void *), status;
    unsigned rate;
    void *data;

    data = lsnote(lsnote)->link_m.data;
    free = lsnote(lsnote)->link_m.free;

    rate = lsnote(lsnote)->link_a.fplist.rate;
    if (rate) {
	status = back_note
	    (lsnote(lsnote)->link_a.fplist.fpnode,
	     lead_switch * lsnote(lsnote)->link_a.fplist.call, rate, data,
	     free);
    } else {
	if (lsnote(lsnote)->link_a.fplist.fpnode) {
	    status = back_slip
		(lsnote(lsnote)->link_a.fplist.fpnode, data, free);
	} else {
	    status = 0;
	}
    }

    excess = free(data, lsnote(lsnote)->link_a.fplist.fvnode);
    if (excess) {
	if (status) {
	} else {
	    status = FREE_ERROR;
	}
    }

    return status;
}
