/*
 * lsas.o.c
 * Copyright (C) 2011-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lsas-config.h>

#include <lsas-defs.h>
#include <lsas-names.h>
#include <lsas-types.h>

#if __POST_SPLIT_CURE__ == __TREE_CURE__
# define fix3()
# define fix4()
# define set3()
# define set4()
#else
# define fix3() \
    Gxx3 = 0;
# define fix4() \
    Gxx4 = 0;
# define set3() \
    Gxx3 = 1;
# define set4() \
    Gxx4 = 1;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

static int case_date(unsigned, unsigned, void *, unsigned *);
static int case_late(unsigned, unsigned, void *, unsigned *);
static int case_node(void *, void *, unsigned, void *);

static int
case_date(unsigned seek, unsigned mind, void *node, unsigned *tear)
{
    unsigned call, diff, last, *near, note, side;

    side = point_rate(seek);

    call = *tear;

    last = point_rate(call);
    if (call & POINT_SEEK) {
    } else {
	last++;
    }

    diff = side - last;

    if (seek & POINT_SEEK) {
    } else {
	side++;
    }

    note = tear[1];

    near = (void *) ((char *) node + L1ST(note));
    if (*near & POINT_CAST) {
    } else {
	if (last ^ side) {
	    if (32 < diff) {
		near[1] = 0;
	    } else {
		unsigned fine, miss, tile;

		tile = side - last;

		fine = near[1];
		miss = fine >> 24;
		fine &= (1 << 24) - 1;
		if (seek & POINT_SEEK) {
		} else {
		    fine >>= 1;
		    fine |= 1 << 23;
		    miss++;
		    tile--;
		}

		if (23 < tile) {
		    fine = mind >> (tile - 24);
		} else {
		    fine = mind << (24 - tile) | fine >> tile;
		}

		if (24 < miss) {
		    miss = 24;
		}

		fine &= (1 << 24) - 1;

		near[1] = miss << 24 | fine;
	    }
	}
    }

    return 0;
}


static int
case_late(unsigned seek, unsigned mind, void *node, unsigned *tear)
{
    unsigned call, diff, last, *near, note, side;

    side = point_rate(seek);

    call = *tear;

    last = point_rate(call);
    if (call & POINT_SEEK) {
    } else {
	last++;
    }

    diff = side - last;

    if (seek & POINT_SEEK) {
    } else {
	side++;
    }

    note = tear[1];

    near = (void *) ((char *) node + L2ND(note));
    if (*near & POINT_CAST) {
    } else {
	if (last ^ side) {
	    if (32 < diff) {
		near[1] = 0;
	    } else {
		unsigned fine, miss, tile;

		tile = side - last;

		fine = near[1];
		miss = fine >> 24;
		fine &= (1 << 24) - 1;
		if (seek & POINT_SEEK) {
		} else {
		    fine >>= 1;
		    miss++;
		    tile--;
		}

		if (23 < tile) {
		    fine = mind >> (tile - 24);
		} else {
		    fine = mind << (24 - tile) | fine >> tile;
		}

		if (24 < miss) {
		    miss = 24;
		}

		fine &= (1 << 24) - 1;

		near[1] = miss << 24 | fine;
	    }
	}
    }

    return 0;
}


static int
case_node(void *node, void *side, unsigned slip, void *back)
{
    int delete = 0;
    struct quarter_type *quarter_data;
    unsigned *near, note;

    near = (void *) ((char *) node + slip);
    note = *near;

    quarter_data = back;

    if (note & POINT_CAST) {
#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
	int Gxx3 = 0, Gxx4 = 0;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

	note = near[1];

	if (case_node(node, side, L1ST(note), back)) {
	    delete = 1;
	    if (quarter_data->shut) {
		if (~quarter_data->shut) {
		    if (quarter_data->lead) {
			if (~quarter_data->lead) {
			} else {
			    quarter_data->lead = 0;
			}
		    } else {
			quarter_data->dana = *near;
			quarter_data->four = near[2];
			quarter_data->lead = L1ST(note);
		    }
		    {
			unsigned *wear;

			wear = (void *) ((char *) side + slip);
			wear[1] = quarter_data->shut | L2ND(note) << 16;
			case_date
			    (quarter_data->aile, quarter_data->zero, side,
			     wear);
		    }

		    set4();

		    quarter_data->shut = 0;
		} else {
		    set3();
		    set4();

		    quarter_data->aile = *near;
		    quarter_data->zero = near[2];
		    quarter_data->shut = L2ND(note);
		}
	    } else {
		if (quarter_data->slip == quarter_data->half) {
		    set3();
		    set4();

		    if (quarter_data->lead) {
			if (~quarter_data->lead) {
			    quarter_data->aile = *near;
			    quarter_data->zero = near[2];
			    quarter_data->shut = L2ND(note);
			}
		    } else {
			quarter_data->dana = *near;
			quarter_data->four = near[2];
			quarter_data->lead = L1ST(note);
			quarter_data->aile = *near;
			quarter_data->zero = near[2];
			quarter_data->shut = L2ND(note);
		    }
		} else {
		    if (quarter_data->slip < quarter_data->half) {
			set3();
		    } else {
			set4();
		    }
		}
	    }
	} else {
	    if (quarter_data->slip == quarter_data->half) {
		delete = 1;
		quarter_data->lead = ~0;
		quarter_data->aile = *near;
		quarter_data->zero = near[2];
		quarter_data->shut = L2ND(note);
		quarter_data->call = slip;
		quarter_data->tear = L1ST(note);
	    } else {
		if (quarter_data->slip < quarter_data->half) {
		    set3();
		} else {
		    set4();
		}
	    }
	}

	if (case_node(node, side, L2ND(note), back)) {
	    delete = 1;
	    if (quarter_data->lead) {
		if (~quarter_data->lead) {
		    if (quarter_data->shut) {
		    } else {
			if (quarter_data->slip ^ quarter_data->half) {
			    quarter_data->aile = *near;
			    quarter_data->zero = near[2];
			    quarter_data->shut = L2ND(note);
			}
		    }
		    {
			near[1] = L1ST(note) | quarter_data->lead << 16;
			case_late
			    (quarter_data->dana, quarter_data->four, node,
			     near);
		    }

		    set3();

		    quarter_data->lead = 0;
		} else {
		    set3();
		    set4();

		    quarter_data->dana = *near;
		    quarter_data->four = near[2];
		    quarter_data->lead = L1ST(note);
		}
	    } else {
		if (quarter_data->slip ^ quarter_data->half) {
		    if (quarter_data->shut) {
		    } else {
			quarter_data->dana = *near;
			quarter_data->four = near[2];
			quarter_data->lead = L1ST(note);
			quarter_data->aile = *near;
			quarter_data->zero = near[2];
			quarter_data->shut = L2ND(note);

			set3();
			set4();
		    }
		}
	    }
	} else {
	    if (quarter_data->slip == quarter_data->half) {
		delete = 1;
		quarter_data->dana = *near;
		quarter_data->four = near[2];
		quarter_data->lead = L1ST(note);
		quarter_data->call = slip;
		quarter_data->tear = L2ND(note);

		fix3();
		fix4();
	    }
	}

#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
	if (Gxx3) {
	    unsigned *sail;

	    sail = (void *) ((char *) side + slip);
	    *sail = quarter_data->mind;
	    quarter_data->mind = slip;
	}
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
	if (Gxx4) {
	    *near = quarter_data->post;
	    quarter_data->post = slip;
	}
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
    } else {
	quarter_data->slip++;

	if (quarter_data->slip < quarter_data->half) {
	    unsigned *sail;

	    sail = (void *) ((char *) side + slip);
	    *sail = quarter_data->fast;
	    quarter_data->fast = slip;
	} else {
	    if (quarter_data->half < quarter_data->slip) {
		*near = quarter_data->near;
		quarter_data->near = slip;
	    }
	}
    }

    return delete;
}


int
_libx1f4l2_moon_lsnote(void *node, void *side, unsigned slip, void *back)
{
    int delete = 0;
    struct quarter_type *quarter_data;
    unsigned *near, note, rate;

    near = (void *) ((char *) node + slip);
    note = *near;

    quarter_data = back;

    if (note & POINT_CAST) {
	int Gyy3 = 0, Gyy4 = 0, break_record, d2, d3;
	unsigned lead, pick;

	pick = point_rate(quarter_data->pick);

	rate = point_rate(note);

	if (pick < rate) {
	    break_record = 1;

	    lead = 0;
	} else {
	    break_record = 0;

#if 0
	    if (pick == rate) {
		if (quarter_data->pick & POINT_SEEK) {
		    /*
		     * looks like -break_record- has a "not care" value here,
		     * it works fine either way.  So I'll spare it.
		     */
		    break_record = 1;
		}
	    }
#endif				/* 0 */

	    if (note & POINT_SEEK) {
		lead = 2;
	    } else {
		lead = (quarter_data->name[rate >> 3] >> (7 - (rate & 7)) & 1)
		    + 1;
	    }
	}

	if (break_record) {
	    if (quarter_data->tier) {
		unsigned deck, half;

		deck = quarter_data->slip;
		half = quarter_data->half - 1;
		quarter_data->deck = deck;
		if (deck < half) {
		    quarter_data->slip = deck + 1;
		} else {
		    if (deck == half) {
			quarter_data->slip = deck + 1;
			delete = 1;
		    } else {
		    }
		}
	    }
	}

	note = near[1];

	if (lead & 1) {
	    d2 = _libx1f4l2_moon_lsnote(node, side, L1ST(note), back);
	} else {
	    d2 = case_node(node, side, L1ST(note), back);
	}

	if (d2) {
	    delete = 1;
	    if (quarter_data->shut) {
		if (~quarter_data->shut) {
		    if (quarter_data->lead) {
			if (~quarter_data->lead) {
			} else {
			    quarter_data->lead = 0;
			}
		    } else {
			quarter_data->dana = *near;
			quarter_data->four = near[2];
			quarter_data->lead = L1ST(note);
		    }
		    {
			unsigned *wear;

			wear = (void *) ((char *) side + slip);
			wear[1] = quarter_data->shut | L2ND(note) << 16;
			case_date
			    (quarter_data->aile, quarter_data->zero, side,
			     wear);
		    }

		    Gyy4 = 1;

		    quarter_data->shut = 0;
		} else {
		    Gyy3 = 1;
		    Gyy4 = 1;

		    quarter_data->aile = *near;
		    quarter_data->zero = near[2];
		    quarter_data->shut = L2ND(note);
		}
	    } else {
		if (quarter_data->slip == quarter_data->half) {
		    Gyy3 = 1;
		    Gyy4 = 1;

		    if (quarter_data->lead) {
			if (~quarter_data->lead) {
			    quarter_data->aile = *near;
			    quarter_data->zero = near[2];
			    quarter_data->shut = L2ND(note);
			}
		    } else {
			quarter_data->dana = *near;
			quarter_data->four = near[2];
			quarter_data->lead = L1ST(note);
			quarter_data->aile = *near;
			quarter_data->zero = near[2];
			quarter_data->shut = L2ND(note);
		    }
		} else {
		    if (quarter_data->slip < quarter_data->half) {
			Gyy3 = 1;
		    } else {
			Gyy4 = 1;
		    }
		}
	    }
	} else {
	    if (quarter_data->slip == quarter_data->half) {
		delete = 1;
		quarter_data->lead = ~0;
		quarter_data->aile = *near;
		quarter_data->zero = near[2];
		quarter_data->shut = L2ND(note);
		quarter_data->call = slip;
		quarter_data->tear = L1ST(note);
	    } else {
		if (quarter_data->slip < quarter_data->half) {
		    Gyy3 = 1;
		} else {
		    Gyy4 = 1;
		}
	    }
	}

	if (lead & 2) {
	    d3 = _libx1f4l2_moon_lsnote(node, side, L2ND(note), back);
	} else {
	    d3 = case_node(node, side, L2ND(note), back);
	}

	if (d3) {
	    delete = 1;
	    if (quarter_data->lead) {
		if (~quarter_data->lead) {
		    if (quarter_data->shut) {
		    } else {
			if (quarter_data->slip ^ quarter_data->half) {
			    quarter_data->aile = *near;
			    quarter_data->zero = near[2];
			    quarter_data->shut = L2ND(note);
			}
		    }
		    {
			near[1] = L1ST(note) | quarter_data->lead << 16;
			case_late
			    (quarter_data->dana, quarter_data->four, node,
			     near);
		    }

		    Gyy3 = 1;

		    quarter_data->lead = 0;
		} else {
		    Gyy3 = 1;
		    Gyy4 = 1;

		    quarter_data->dana = *near;
		    quarter_data->four = near[2];
		    quarter_data->lead = L1ST(note);
		}
	    } else {
		if (quarter_data->slip ^ quarter_data->half) {
		    if (quarter_data->shut) {
		    } else {
			quarter_data->dana = *near;
			quarter_data->four = near[2];
			quarter_data->lead = L1ST(note);
			quarter_data->aile = *near;
			quarter_data->zero = near[2];
			quarter_data->shut = L2ND(note);

			Gyy3 = 1;
			Gyy4 = 1;
		    }
		}
	    }
	} else {
	    if (quarter_data->slip == quarter_data->half) {
		delete = 1;
		quarter_data->dana = *near;
		quarter_data->four = near[2];
		quarter_data->lead = L1ST(note);
		quarter_data->call = slip;
		quarter_data->tear = L2ND(note);

		Gyy3 = 0;
		Gyy4 = 0;
	    }
	}

	if (break_record) {
	    if (quarter_data->tier) {
	    } else {
		unsigned deck, half;

		deck = quarter_data->slip;
		half = quarter_data->half - 1;
		quarter_data->deck = deck;
		if (deck < half) {
		    quarter_data->slip = deck + 1;
		} else {
		    if (deck == half) {
			quarter_data->slip = deck + 1;
			delete = 1;
		    } else {
		    }
		}
	    }
	}

#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
	if (Gyy3) {
	    unsigned *sail;

	    sail = (void *) ((char *) side + slip);
	    *sail = quarter_data->mind;
	    quarter_data->mind = slip;
	}
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
	if (Gyy4) {
	    *near = quarter_data->post;
	    quarter_data->post = slip;
	}
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
    } else {
	if (1) {
	    if (quarter_data->tier) {
		unsigned deck, half;

		deck = quarter_data->slip;
		half = quarter_data->half - 1;
		quarter_data->deck = deck;
		if (deck < half) {
		    quarter_data->slip = deck + 1;
		} else {
		    if (deck == half) {
			quarter_data->slip = deck + 1;
			delete = 1;
		    } else {
		    }
		}
	    }
	}

	quarter_data->slip++;

	if (quarter_data->slip < quarter_data->half) {
	    unsigned *sail;

	    sail = (void *) ((char *) side + slip);
	    *sail = quarter_data->fast;
	    quarter_data->fast = slip;
	} else {
	    if (quarter_data->half < quarter_data->slip) {
		*near = quarter_data->near;
		quarter_data->near = slip;
	    }
	}

	if (1) {
	    if (quarter_data->tier) {
	    } else {
		unsigned deck, half;

		deck = quarter_data->slip;
		half = quarter_data->half - 1;
		quarter_data->deck = deck;
		if (deck < half) {
		    quarter_data->slip = deck + 1;
		} else {
		    if (deck == half) {
			quarter_data->slip = deck + 1;
			delete = 1;
		    } else {
		    }
		}
	    }
	}
    }

    return delete;
}
