/*
 * lvlist.h
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_LVLIST_H__
#define __LIBx1f4_LVLIST_H__

#include <errors.h>

/*
 * attribute bit definitions
 */
#define X1f4_LVLIST_TRANS_MASK		(1u << 0x00)

/*
 * error definitions
 */
#define X1f4_LVLIST_CALL_ERROR		LIBx1f4L2_CALL_ERROR
#define X1f4_LVLIST_DECK_ERROR		LIBx1f4L2_DECK_ERROR
#define X1f4_LVLIST_FREE_ERROR		LIBx1f4L2_FREE_ERROR
#define X1f4_LVLIST_LINK_ERROR		LIBx1f4L2_LINK_ERROR
#define X1f4_LVLIST_MODE_ERROR		LIBx1f4L2_MODE_ERROR
#define X1f4_LVLIST_SIZE_ERROR		LIBx1f4L2_SIZE_ERROR

typedef struct x1f4_lvlist_type {
    void *trans;
} x1f4_lvlist_type;

#if !defined __x1f4_trans_type_defined__
typedef struct x1f4_trans_type {
    int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	(*mode) (void *, void **, unsigned);
    void *context;
} x1f4_trans_type;
#endif				/* !__x1f4_trans_type_defined__ */

#if !defined __x1f4_trans_type_defined__
# define __x1f4_trans_type_defined__
#endif				/* !__x1f4_trans_type_defined__ */

extern int x1f4_call_lvlist(unsigned *);
extern int x1f4_clip_lvlist
    (void *, unsigned, void *, int (*) (void *, void *));
extern int x1f4_deck_lvlist(void *, int);
extern int x1f4_deux_lvlist(void *, unsigned, void *, unsigned);
extern int x1f4_ever_lvlist(void *, void **);
extern int x1f4_fast_lvlist(void *, unsigned, struct x1f4_lvlist_type *);
extern int x1f4_fgxp_lvlist
    (void *, void *, unsigned, int (*) (void *, void *), unsigned *);
extern int x1f4_find_lvlist
    (void *, void *, unsigned, int (*) (void *, void *), void **);
extern int x1f4_fini_lvlist(void **);
extern int x1f4_flat_lvlist(void *);
extern int x1f4_flip_lvlist(void *);
extern int x1f4_flow_lvlist(void *);
extern int x1f4_head_lvlist(void *, unsigned, void **);
extern int x1f4_high_lvlist(void *);
extern int x1f4_init_lvlist(void **, unsigned, struct x1f4_lvlist_type *);
extern int x1f4_land_lvlist
    (void *, void *, int (*) (void *, void *), unsigned, unsigned, void **);
extern int x1f4_last_lvlist(void **, void *);
extern int x1f4_lead_lvlist(void *, void **);
extern int x1f4_lime_lvlist(void *, void *, int (*) (void *, void *));
extern int x1f4_line_lvlist
    (void *, void *, int (*) (void *, unsigned *, const void *),
     int (*) (void *, void *, const void *), int (*) (void *, void *), void *);
extern int x1f4_mind_lvlist(void *);
extern int x1f4_miss_lvlist(void *, unsigned);
extern int x1f4_mode_lvlist(void *, unsigned, unsigned, void **);
extern int x1f4_move_lvlist(void *, unsigned, unsigned);
extern int x1f4_next_lvlist(void **, void *);
extern int x1f4_peek_lvlist(void *, unsigned, void **);
extern int x1f4_push_lvlist(void *, unsigned, unsigned, void **);
extern int x1f4_rand_lvlist(void *);
extern int x1f4_seek_lvlist
    (void *, void *, unsigned, int (*) (void *, void *), unsigned *);
extern int x1f4_side_lvlist(void *, void *);
extern int x1f4_size_lvlist(void *, unsigned *);
extern int x1f4_skew_lvlist(void *, void *);
extern int x1f4_slip_lvlist(void *);
extern int x1f4_sort_lvlist
    (void *, void *, unsigned, int (*) (void *, void *, void *));
extern int x1f4_spin_lvlist(void *, unsigned, unsigned);
extern int x1f4_tail_lvlist(void *, unsigned, void **);
extern int x1f4_text_lvlist
    (void **, void *, int (*) (void *, unsigned *, const void *),
     int (*) (void *, void *, const void *), int (*) (void *, void *), void *);
extern int x1f4_tick_lvlist
    (void *, void *, unsigned, int (*) (void *, void *), unsigned *, void **);
extern int x1f4_tile_lvlist(unsigned *, void *);
extern int x1f4_turn_lvlist(void *);
extern int x1f4_wind_lvlist(void *, unsigned, void *, unsigned);

#endif				/* __LIBx1f4_LVLIST_H__ */
