/*
 * nclist.h
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_NCLIST_H__
#define __LIBx1f4_NCLIST_H__

#include <errors.h>
#include <ncxx-names.h>

/*
 * attribute bit definitions
 */
#define X1f4_NCLIST_TRANS_MASK		(1u << 0x00)

/*
 * error definitions
 */
#define X1f4_NCLIST_CALL_ERROR		LIBx1f4L2_CALL_ERROR
#define X1f4_NCLIST_DECK_ERROR		LIBx1f4L2_DECK_ERROR
#define X1f4_NCLIST_FREE_ERROR		LIBx1f4L2_FREE_ERROR
#define X1f4_NCLIST_LINK_ERROR		LIBx1f4L2_LINK_ERROR
#define X1f4_NCLIST_MODE_ERROR		LIBx1f4L2_MODE_ERROR
#define X1f4_NCLIST_SIZE_ERROR		LIBx1f4L2_SIZE_ERROR

typedef struct x1f4_nclist_type {
    void *trans;
} x1f4_nclist_type;

#if !defined __x1f4_trans_type_defined__
typedef struct x1f4_trans_type {
    int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	(*mode) (void *, void **, unsigned);
    void *context;
} x1f4_trans_type;
#endif				/* !__x1f4_trans_type_defined__ */

#if !defined __x1f4_trans_type_defined__
# define __x1f4_trans_type_defined__
#endif				/* !__x1f4_trans_type_defined__ */

extern int x1f4_call_nclist(unsigned *);
extern int x1f4_clip_nclist
    (void *, unsigned, void *, int (*) (void *, void *));
extern int x1f4_deck_nclist(void *, int);
extern int x1f4_deux_nclist(void *, unsigned, void *, unsigned);
extern int x1f4_ever_nclist(void *, void **);
extern int x1f4_fast_nclist(void *, unsigned, struct x1f4_nclist_type *);
extern int x1f4_fgxp_nclist
    (void *, void *, unsigned, int (*) (void *, void *), unsigned *);
extern int x1f4_find_nclist
    (void *, void *, unsigned, int (*) (void *, void *), void **);
extern int x1f4_fine_nclist(void *, void *, int (*) (void *, void *));
extern int x1f4_fini_nclist(void **);
extern int x1f4_flat_nclist(void *);
extern int x1f4_flip_nclist(void *);
extern int x1f4_flow_nclist(void *);
extern int x1f4_head_nclist(void *, unsigned, void **);
extern int x1f4_high_nclist(void *);
extern int x1f4_init_nclist(void **, unsigned, struct x1f4_nclist_type *);
extern int x1f4_land_nclist
    (void *, void *, int (*) (void *, void *), unsigned, unsigned, void **);
extern int x1f4_lead_nclist(void *, void **);
extern int x1f4_lime_nclist(void *, void *, int (*) (void *, void *));
extern int x1f4_line_nclist
    (void *, void *, int (*) (void *, unsigned *, const void *),
     int (*) (void *, void *, const void *), int (*) (void *, void *), void *);
extern int x1f4_mind_nclist(void *);
extern int x1f4_miss_nclist(void *, unsigned);
extern int x1f4_mode_nclist(void *, unsigned, unsigned, void **);
extern int x1f4_move_nclist(void *, unsigned, unsigned);
extern int x1f4_nice_nclist(void *, void *, int (*) (void *, void *));
extern int x1f4_peek_nclist(void *, unsigned, void **);
extern int x1f4_pull_nclist
    (void *, unsigned, void *, int (*) (void *, void *));
extern int x1f4_push_nclist(void *, unsigned, unsigned, void **);
extern int x1f4_rand_nclist(void *);
extern int x1f4_seek_nclist
    (void *, void *, unsigned, int (*) (void *, void *), unsigned *);
extern int x1f4_size_nclist(void *, unsigned *);
extern int x1f4_skew_nclist(void *, void *);
extern int x1f4_slip_nclist(void *);
extern int x1f4_sort_nclist
    (void *, void *, unsigned, int (*) (void *, void *, void *));
extern int x1f4_spin_nclist(void *, unsigned, unsigned);
extern int x1f4_tail_nclist(void *, unsigned, void **);
extern int x1f4_text_nclist
    (void **, void *, int (*) (void *, unsigned *, const void *),
     int (*) (void *, void *, const void *), int (*) (void *, void *), void *);
extern int x1f4_tick_nclist
    (void *, void *, unsigned, int (*) (void *, void *), unsigned *, void **);
extern int x1f4_turn_nclist(void *);
extern int x1f4_wind_nclist(void *, unsigned, void *, unsigned);

#endif				/* __LIBx1f4_NCLIST_H__ */
