/*
 * ncxx.c.c
 * Copyright (C) 2008-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <ncxx-defs.h>
#include <ncxx-inter.h>
#include <ncxx-names.h>
#include <ncxx-types.h>
#include <trans.h>

static int fini_list(void **);
static int init_list(void **, struct nclist_type *);
static int text_list(struct nclist_type *, struct nclist_type *,
		     int (*) (void *, unsigned *, const void *),
		     int (*) (void *, void *, const void *),
		     int (*) (void *, void *), void *);

static int
fini_list(void **text)
{
    return _libx1f4l2_fini_nclist(text);
}


static int
init_list(void **text, struct nclist_type *nclist_data)
{
    int status;
    struct screen_type screen;
    struct trans_type trans;

    screen.trans_data = &trans;

    trans.data = nclist_data->link_m.data;
    trans.free = nclist_data->link_m.free;
    trans.link = nclist_data->link_m.link;
    trans.mode = nclist_data->link_m.mode;

    status = _libx1f4l2_init_nclist(text, TRANS_MASK, &screen);

    return status;
}


static int
text_list(struct nclist_type *nclist_data, struct nclist_type *nclist_text,
	  int (*size) (void *, unsigned *, const void *),
	  int (*copy) (void *, void *, const void *),
	  int (*free) (void *, void *), void *context)
{
    int status;

    if (nclist_text->link_a.fplist.fplink.fqnode.size) {
	status = _libx1f4l2_rail_nclist
	    (nclist_data, nclist_text, size, copy, free, context);
    } else {
	status = 0;
    }

    return status;
}


int
x1f4_text_nclist(void **nclist, void *tctext,
		 int (*size) (void *, unsigned *, const void *),
		 int (*copy) (void *, void *, const void *),
		 int (*free) (void *, void *), void *context)
{
    int status;
    void *text;

    status = init_list(&text, tctext);
    if (status) {
    } else {
	status = text_list(text, tctext, size, copy, free, context);
	if (status) {
	    fini_list(&text);
	} else {
	    *nclist = text;
	}
    }

    return status;
}
