/*
 * qccc.f.c
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <mask.0.h>
#include <qccc-bits.h>
#include <qccc-defs.h>
#include <qccc-inter.h>
#include <qccc-names.h>
#include <qccc-types.h>
#include <trans.h>

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

#define qcrate(qcrate)			((struct qcrate_type *) (qcrate))

static int init_list(struct qcrate_type *, unsigned, unsigned);

static int
init_list(struct qcrate_type *qcrate_data, unsigned mall, unsigned lock)
{
    int status;
    void *data;

    status = qcrate_data->link_m.link
	(qcrate_data->link_m.data, &data,
	 ((sizeof(unsigned) << 1) + mall) << lock);
    if (status) {
	status = LINK_ERROR;
    } else {
	qcrate_data->link_a.fplist.fpnode_data = data;
	_libx1f4l2_type_qcrate(data, 1 << lock, mall);
	S_NODE(M_TEXT(data, sizeof(unsigned)), 0);
    }

    return status;
}


int
x1f4_fast_qcrate(void *qcrate, unsigned flags,
		 struct screen_type *screen_data)
{
    int status;
    unsigned lock, mall;

    mall = screen_data->size;

    lock = TEXT_RATE;
    if (flags & RATE_FRAME) {
	lock = screen_data->rate;
    }

    while (1 << 14 < ((sizeof(unsigned) << 1) + mall) << lock) {
	lock--;
    }

    if ((int) lock < 4) {
	/*
	 * improper error code, should invent another
	 */
	status = DECK_ERROR;
    } else {
	if (flags & TRANS_MASK) {
	    struct trans_type *trans_data;

	    trans_data = screen_data->trans_data;

	    qcrate(qcrate)->link_m.data = trans_data->data;
	    qcrate(qcrate)->link_m.free = trans_data->free;
	    qcrate(qcrate)->link_m.link = trans_data->link;
	    qcrate(qcrate)->link_m.mode = trans_data->mode;
	} else {
	    qcrate(qcrate)->link_m.free = _libx1f4l2_free_data;
	    qcrate(qcrate)->link_m.link = _libx1f4l2_link_data;
	    qcrate(qcrate)->link_m.mode = _libx1f4l2_mode_data;
	}

	status = init_list(qcrate, mall, lock);
	if (status) {
	} else {
	    qcrate(qcrate)->link_a.fplist.lock = lock;
	    qcrate(qcrate)->link_a.fplist.mall = mall;
	    qcrate(qcrate)->link_a.fplist.rate = 0;

	    qcrate(qcrate)->link_e.fplink_data = NULL;
	    qcrate(qcrate)->link_e.fplink_text = NULL;
	}
    }

    return status;
}
