/*
 * qccc.q.c
 * Copyright (C) 2009-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qccc-config.h>

#include <config.h>

#include <string.h>

#include <qccc-bits.h>
#include <qccc-names.h>
#include <qccc-types.h>

#define __UNBALANCED__			1

#define true(e)				1

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

static void copy_data(void *, unsigned, unsigned, unsigned, void **);
#if SIZEOF_UNSIGNED == 4
static void copy_deck(void *, unsigned, unsigned, unsigned, void **);
#endif				/* SIZEOF_UNSIGNED == 4 */
static void type_list(void *, unsigned, unsigned);
static void type_mini(void *, unsigned, unsigned);

static void
copy_data(void *node, unsigned violet, unsigned pick, unsigned call,
	  void **back)
{
    do {
	unsigned slip, text;

	text = Q_NODE(M_TEXT(node, pick));

	slip = L_LAST(text);
	if (slip < violet) {
#if __UNBALANCED__
	    unsigned ever, miss;
#endif				/* __UNBALANCED__ */

#if __UNBALANCED__
	    miss = Q_NODE(M_TEXT(node, slip));
#endif				/* __UNBALANCED__ */

#if __UNBALANCED__
	    ever = L_LAST(miss);
	    if (ever < violet) {
		copy_data(node, violet, ever, call, back);
	    } else {
		char *u, *v;
		unsigned i;

		v = M_TEXT(node, ever);
		u = *back;

		i = call;
		for (; i; i--) {
		    *u++ = *v++;
		}
		{
		    *u++ = *v;
		}

		*back = u;
	    }
#endif				/* __UNBALANCED__ */

#if __UNBALANCED__
#else
	    copy_data(node, violet, slip, call, back);
#endif				/* __UNBALANCED__ */

#if __UNBALANCED__
	    ever = L_NEXT(miss);
	    if (ever < violet) {
		copy_data(node, violet, ever, call, back);
	    } else {
		char *u, *v;
		unsigned i;

		v = M_TEXT(node, ever);
		u = *back;

		i = call;
		for (; i; i--) {
		    *u++ = *v++;
		}
		{
		    *u++ = *v;
		}

		*back = u;
	    }
#endif				/* __UNBALANCED__ */
	} else {
	    char *u, *v;
	    unsigned i;

	    v = M_TEXT(node, slip);
	    u = *back;

	    i = call;
	    for (; i; i--) {
		*u++ = *v++;
	    }
	    {
		*u++ = *v;
	    }

	    *back = u;
	}

	pick = L_NEXT(text);
    } while (pick < violet);
    if (0) {
    } else {
	if (1) {
	    char *u, *v;
	    unsigned i;

	    v = M_TEXT(node, pick);
	    u = *back;

	    i = call;
	    for (; i; i--) {
		*u++ = *v++;
	    }
	    {
		*u++ = *v;
	    }

	    *back = u;
	}
    }
}


#if SIZEOF_UNSIGNED == 4
static void
copy_deck(void *node, unsigned violet, unsigned pick, unsigned call,
	  void **back)
{
    do {
	unsigned slip, text;

	text = Q_NODE(M_TEXT(node, pick));

	slip = L_LAST(text);
	if (slip < violet) {
# if __UNBALANCED__
	    unsigned ever, miss;
# endif				/* __UNBALANCED__ */

# if __UNBALANCED__
	    miss = Q_NODE(M_TEXT(node, slip));
# endif				/* __UNBALANCED__ */

# if __UNBALANCED__
	    ever = L_LAST(miss);
	    if (ever < violet) {
		copy_deck(node, violet, ever, call, back);
	    } else {
		unsigned i, *u, *v;

		v = M_TEXT(node, ever);
		u = *back;

		i = call;
		for (; i; i--) {
		    *u++ = *v++;
		}
		{
		    *u++ = *v;
		}

		*back = u;
	    }
# endif				/* __UNBALANCED__ */

# if __UNBALANCED__
# else
	    copy_deck(node, violet, slip, call, back);
# endif				/* __UNBALANCED__ */

# if __UNBALANCED__
	    ever = L_NEXT(miss);
	    if (ever < violet) {
		copy_deck(node, violet, ever, call, back);
	    } else {
		unsigned i, *u, *v;

		v = M_TEXT(node, ever);
		u = *back;

		i = call;
		for (; i; i--) {
		    *u++ = *v++;
		}
		{
		    *u++ = *v;
		}

		*back = u;
	    }
# endif				/* __UNBALANCED__ */
	} else {
	    unsigned i, *u, *v;

	    v = M_TEXT(node, slip);
	    u = *back;

	    i = call;
	    for (; i; i--) {
		*u++ = *v++;
	    }
	    {
		*u++ = *v;
	    }

	    *back = u;
	}

	pick = L_NEXT(text);
    } while (pick < violet);
    if (0) {
    } else {
	if (1) {
	    unsigned i, *u, *v;

	    v = M_TEXT(node, pick);
	    u = *back;

	    i = call;
	    for (; i; i--) {
		*u++ = *v++;
	    }
	    {
		*u++ = *v;
	    }

	    *back = u;
	}
    }
}
#endif				/* SIZEOF_UNSIGNED == 4 */


static void
type_list(void *node, unsigned rate, unsigned size)
{
    unsigned copy, list, text;
    void *side;

    text = (sizeof(unsigned) << 1 << rate) + (size << rate >> 1);

    side = M_TEXT(node, text);

    list = sizeof(unsigned) << rate;

    S_NODE(node, Q_NODE(node) | B_LIST(list));

    node = M_TEXT(node, list);

    copy = 1 << rate;
    copy >>= 1;
    copy--;
    for (; copy; copy--) {
	S_NODE(node, text);
	text += size;
	list += sizeof(unsigned) << 1;
	S_NODE(side, list);
	side = M_TEXT(side, size);
	node = M_TEXT(node, sizeof(unsigned) << 1);
    }
    {
	S_NODE(node, text);
	S_NODE(side, 0);
    }
}


static void
type_mini(void *node, unsigned rate, unsigned size)
{
#if ____REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST__
    unsigned *line, pick, type = 0;

    pick = sizeof(unsigned) << 2;

    line = node;

    rate -= 2;

    line += 2;

    for (; type < rate; type++) {
	unsigned call, i;

	call = 1 << (rate - type);

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + (sizeof(unsigned) << 1))
		   | B_TILT(0));
	    pick += sizeof(unsigned) << 2;
	    line++;
	    S_NODE(line, call);
	    line++;
	}
    }

    pick = sizeof(unsigned) << 3 << rate;

    {
	unsigned i;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line, B_LAST(pick) | B_NEXT(pick + size) | B_TILT(0));
	    pick += size << 1;
	    line++;
	    S_NODE(line, 1);
	    line++;
	}
    }
#elif __REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST_DOUBLE_LAYERED__
    unsigned beta, *line, pick, slip, type = 0;

    pick = sizeof(unsigned) << 3;
    slip = sizeof(unsigned) << 2;

    line = node;

    line += 2;

    beta = (rate - 3) >> 1;
    for (; beta; beta--) {
	unsigned call, i;

	call = 1 << (rate - 3 - type);

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line++;
	    S_NODE(line, call << 1);
	    line++;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 6)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 014;
	    line++;
	    S_NODE(line, call << 0);
	    line++;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 6)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 014;
	    line++;
	    S_NODE(line, call << 0);
	    line++;
	}

	type += 2;
    }
    if (rate & 1) {
	unsigned i;

	pick = sizeof(unsigned) << 1 << rate;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line++;
	    S_NODE(line, 2);
	    line++;
	    S_NODE(line, B_LAST(pick) | B_NEXT(pick + size) | B_TILT(0));
	    pick += size << 1;
	    line++;
	    S_NODE(line, 1);
	    line++;
	    S_NODE(line, B_LAST(pick) | B_NEXT(pick + size) | B_TILT(0));
	    pick += size << 1;
	    line++;
	    S_NODE(line, 1);
	    line++;
	}
    } else {
	unsigned i;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line++;
	    S_NODE(line, 4);
	    line++;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 004;
	    line++;
	    S_NODE(line, 2);
	    line++;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 004;
	    line++;
	    S_NODE(line, 2);
	    line++;
	}

	pick = sizeof(unsigned) << 1 << rate;

	i = 4 << type;
	for (; i; i--) {
	    S_NODE(line, B_LAST(pick) | B_NEXT(pick + size) | B_TILT(0));
	    pick += size << 1;
	    line++;
	    S_NODE(line, 1);
	    line++;
	}
    }
#elif __REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST_TRIPLE_LAYERED__
    unsigned beta, *line, mass, pick, slip, type = 0;

    pick = sizeof(unsigned) << 4;
    slip = sizeof(unsigned) << 2;

    line = node;

    line += 2;

    beta = (rate - 4) / 3;

    mass = rate - beta * 3;

    for (; beta; beta--) {
	unsigned call, i;

	call = 1 << (rate - 4 - type);

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 004;
	    line++;
	    S_NODE(line, call << 2);
	    line++;
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 004;
	    line++;
	    S_NODE(line, call << 1);
	    line++;
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line++;
	    S_NODE(line, call << 1);
	    line++;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 016)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 034;
	    line++;
	    S_NODE(line, call << 0);
	    line++;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 016)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 034;
	    line++;
	    S_NODE(line, call << 0);
	    line++;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 016)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 034;
	    line++;
	    S_NODE(line, call << 0);
	    line++;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 016)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 034;
	    line++;
	    S_NODE(line, call << 0);
	    line++;
	}

	type += 3;
    }
    if (mass == 4) {
	unsigned i;

	pick = sizeof(unsigned) << 1 << rate;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 004;
	    line++;
	    S_NODE(line, 4);
	    line++;
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 004;
	    line++;
	    S_NODE(line, 2);
	    line++;
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line++;
	    S_NODE(line, 2);
	    line++;
	    S_NODE(line, B_LAST(pick) | B_NEXT(pick + size) | B_TILT(0));
	    pick += size << 1;
	    line++;
	    S_NODE(line, 1);
	    line++;
	    S_NODE(line, B_LAST(pick) | B_NEXT(pick + size) | B_TILT(0));
	    pick += size << 1;
	    line++;
	    S_NODE(line, 1);
	    line++;
	    S_NODE(line, B_LAST(pick) | B_NEXT(pick + size) | B_TILT(0));
	    pick += size << 1;
	    line++;
	    S_NODE(line, 1);
	    line++;
	    S_NODE(line, B_LAST(pick) | B_NEXT(pick + size) | B_TILT(0));
	    pick += size << 1;
	    line++;
	    S_NODE(line, 1);
	    line++;
	}
    } else {
	if (mass == 5) {
	    unsigned i;

	    i = 1 << type;
	    for (; i; i--) {
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 004;
		line++;
		S_NODE(line, 010);
		line++;
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 004;
		line++;
		S_NODE(line, 004);
		line++;
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 006;
		line++;
		S_NODE(line, 004);
		line++;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 002)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 004;
		line++;
		S_NODE(line, 002);
		line++;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 002)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 004;
		line++;
		S_NODE(line, 002);
		line++;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 002)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 004;
		line++;
		S_NODE(line, 002);
		line++;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 002)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 004;
		line++;
		S_NODE(line, 002);
		line++;
	    }

	    pick = sizeof(unsigned) << 1 << rate;

	    i = 8 << type;
	    for (; i; i--) {
		S_NODE(line, B_LAST(pick) | B_NEXT(pick + size) | B_TILT(0));
		pick += size << 1;
		line++;
		S_NODE(line, 001);
		line++;
	    }
	} else {
	    unsigned i;

	    i = 1 << type;
	    for (; i; i--) {
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 004;
		line++;
		S_NODE(line, 020);
		line++;
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 004;
		line++;
		S_NODE(line, 010);
		line++;
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 006;
		line++;
		S_NODE(line, 010);
		line++;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 006)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 014;
		line++;
		S_NODE(line, 004);
		line++;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 006)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 014;
		line++;
		S_NODE(line, 004);
		line++;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 006)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 014;
		line++;
		S_NODE(line, 004);
		line++;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 006)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 014;
		line++;
		S_NODE(line, 004);
		line++;
	    }

	    pick = sizeof(unsigned) << 1 << rate;

	    i = 8 << type;
	    for (; i; i--) {
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 006;
		line++;
		S_NODE(line, 002);
		line++;
		S_NODE(line, B_LAST(pick) | B_NEXT(pick + size) | B_TILT(0));
		pick += size << 1;
		line++;
		S_NODE(line, 001);
		line++;
		S_NODE(line, B_LAST(pick) | B_NEXT(pick + size) | B_TILT(0));
		pick += size << 1;
		line++;
		S_NODE(line, 001);
		line++;
	    }
	}
    }
#endif				/* ____REFORM_MEMORY_LAYOUT__ == __BREADT... */
}


int
_libx1f4l2_part_qcrate(void *node, void *side, unsigned rate, unsigned size)
{
    unsigned violet;
    void *slip;

    violet = sizeof(unsigned) << 1 << rate;

    slip = M_TEXT(side, violet);

#if SIZEOF_UNSIGNED == 4
    if (size & 3) {
#endif				/* SIZEOF_UNSIGNED == 4 */
	copy_data(node, violet, P_ROOT(Q_NODE(node)), size - 1, &slip);
#if SIZEOF_UNSIGNED == 4
    } else {
	copy_deck(node, violet, P_ROOT(Q_NODE(node)), (size >> 2) - 1,
		  &slip);
    }
#endif				/* SIZEOF_UNSIGNED == 4 */

    memcpy((char *) node + violet,
	   (char *) side + violet + (size << rate >> 1), size << rate >> 1);

    S_NODE(node, B_ROOT(sizeof(unsigned) << 1));

    type_mini(node, rate, size);

    type_list(node, rate, size);

    S_NODE(side, B_ROOT(sizeof(unsigned) << 1));

    type_mini(side, rate, size);

    type_list(side, rate, size);

    return 1;
}
