/*
 * qlcc-types.h
 * Copyright (C) 2008-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#ifndef __LIBx1f4_qlcc_types_H__
#define __LIBx1f4_qlcc_types_H__

#include <qlcc-config.h>

#include <trans.h>

#define __long_unsigned_qlnode__	1
#define __unsigned_qlnode__		2

#if SIZEOF_UNSIGNED_LONG == 8
# define __qlnode__			__long_unsigned_qlnode__
#else
# define __qlnode__			__unsigned_qlnode__
#endif				/* SIZEOF_UNSIGNED_LONG == 8 */

#define ____qlnode__			__qlnode__

#if ____qlnode__ == __long_unsigned_qlnode__
# define QLSIZE				SIZEOF_UNSIGNED_LONG
#elif __qlnode__ == __unsigned_qlnode__
# define QLSIZE				SIZEOF_UNSIGNED
#endif				/* ____qlnode__ == __unsigned_qlnode__ */

#if 1
# define QLBITS				(QLSIZE << 3)
#endif				/* 1 */

#if 1
# define HALF_QLBITS			(QLBITS >> 1)
#endif				/* 1 */

#if ____qlnode__ == __long_unsigned_qlnode__
# define QLNODE				unsigned long
#elif __qlnode__ == __unsigned_qlnode__
# define QLNODE				unsigned
#endif				/* ____qlnode__ == __unsigned_qlnode__ */

#if __DELAY_MERGE__
typedef struct dhmiss_type {
    QLNODE *meta;
# if __STAMP_MERGE__
# else
    unsigned black;
# endif				/* __STAMP_MERGE__ */
    unsigned late;
    void *hook, *side;
} dhmiss_type;
#endif				/* __DELAY_MERGE__ */

typedef struct fplink_type {
    struct fplink_type *fplink_data;
} fplink_type;

typedef struct fpnode_type {
    QLNODE node;
    unsigned call;
    void *star;
} fpnode_type;

typedef struct fplist_type {
    QLNODE node;
    unsigned lock, mall, rate, size;
    void *fpnode;
} fplist_type;

typedef struct screen_type {
    unsigned rate, size;
    struct trans_type *trans_data;
} screen_type;

typedef struct qlrate_type {
    struct {
	struct fplist_type fplist;
    } link_a;
    struct {
	struct fplink_type *fplink_data, *fplink_text;
    } link_e;
    struct {
	int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	    (*mode) (void *, void **, unsigned);
	void *data;
    } link_m;
    struct {
	int (*ever) (void *, unsigned, unsigned, struct fpnode_type *,
		     struct fpnode_type *, unsigned *),
	    (*last) (void *, unsigned, void *, void *, QLNODE *, void **),
	    (*lead) (void *, unsigned, unsigned, struct fpnode_type *,
		     struct fpnode_type *, unsigned *),
	    (*mind) (void *, unsigned, struct fpnode_type *,
		     struct fpnode_type *),
	    (*near) (void *, unsigned, struct fpnode_type *, QLNODE,
		     unsigned *),
	    (*post) (void *, unsigned, void *, void *, QLNODE *, void **),
	    (*plus) (void *, unsigned, struct fpnode_type *, QLNODE),
	    (*slip) (void *, unsigned, struct fpnode_type *,
		     struct fpnode_type *);
    } link_v;
} qlrate_type;

#endif				/* __LIBx1f4_qlcc_types_H__ */
