/*
 * qlcc.0.c
 * Copyright (C) 2008-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * improvement research directions:
 *
 * replace the merging of two half full nodes into one full node with merging
 *     of three half full nodes into two two thirds full nodes
 * keep node population is a tighter range, something along 5/8 to 7/8 line.
 *     for 32 slots per node some nice figures would be 21 and 28.  the
 *     benefit would be avoid the worst insert and delete performers.  it is
 *     also reasonable to expect the average node population would improve.
 *     to preserve insert/delete asymmetry, it should go more like:
 *	   2 x 30 -> 3 x 20 and 4 x 20 -> 3 x 26.(6)
 *     or:
 *	   2 x 27 -> 3 x 18 and 4 x 18 -> 3 x 24
 *     or in between.
 */

/*
 * TODO
 *
 * MICRO_RATES - see qsrate
 */

#include <stddef.h>

#include <mask.0.h>
#include <qlcc-defs.h>
#include <qlcc-inter.h>
#include <qlcc-names.h>
#include <qlcc-types.h>
#include <trans.h>

int
x1f4_init_qlrate(void **qlrate, unsigned flags,
		 struct screen_type *screen_data)
{
    int (*link) (void *, void **, unsigned), status;
    void *text;

    if (flags & TRANS_MASK) {
	struct trans_type *trans_data;

	trans_data = screen_data->trans_data;

	link = trans_data->link;
	text = trans_data->data;
    } else {
	link = _libx1f4l2_link_data;
	text = (void *) 0;
    }

    status = link(text, &text, sizeof(struct qlrate_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	struct qlrate_type *qlrate_data;

	qlrate_data = text;

	status = _libx1f4l2_fast_qlrate(qlrate_data, flags, screen_data);

	if (status) {
	    int (*free) (void *, void *);

	    if (flags & TRANS_MASK) {
		struct trans_type *trans_data;

		trans_data = screen_data->trans_data;

		free = trans_data->free;
		text = trans_data->data;
	    } else {
		free = _libx1f4l2_free_data;
	    }

	    free(text, qlrate_data);
	} else {
	    *qlrate = text;
	}
    }

    return status;
}
