/*
 * qlcc.2.c
 * Copyright (C) 2010-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qlcc-config.h>

#include <stddef.h>
#include <string.h>

#include <qlcc-bits.h>
#include <qlcc-defs.h>
#include <qlcc-inter.h>
#include <qlcc-names.h>
#include <qlcc-types.h>

#define qlrate(qlrate)			((struct qlrate_type *) (qlrate))

#if __DELAY_SPLIT__
static int part_late(struct qlrate_type *, unsigned, void *, void *, unsigned,
		     struct fpnode_type **, struct fpnode_type **);
#endif				/* __DELAY_SPLIT__ */

#if __DELAY_SPLIT__
static int
part_late(struct qlrate_type *qlrate_data, unsigned rate, void *miss,
	  void *node, unsigned index, struct fpnode_type **link,
	  struct fpnode_type **push)
{
    int status = 0;
    struct fpnode_type *fpnode_call, *fpnode_data;
    unsigned collect, lock;
    void *fpstar;

    lock = qlrate_data->link_a.fplist.lock;

    fpnode_call = *link;
    fpnode_data = *push;

    fpstar = fpnode_data->star;

    collect = fpnode_data - (struct fpnode_type *) fpnode_call->star;

    fpnode_call = miss;
    fpnode_data = node;

    rate--;
    while (rate) {
	int excess;
	unsigned text;

	text = ((struct fpnode_type *) fpnode_data->star)->call;

	excess = _libx1f4l2_sand_qlrate
	    (qlrate_data, fpnode_call, &fpnode_data, &index, 1 + rate, fpstar);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}

	fpnode_call = fpnode_data;

	fpnode_data = fpnode_data->star;

	rate--;

	index = text >> lock;
	fpnode_data += (text & ((1 << lock >> 1) - 1)) << 1;
    }
    {
	int excess;

	excess = _libx1f4l2_sand_qlrate
	    (qlrate_data, fpnode_call, &fpnode_data, &index, 1, fpstar);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}

	fpnode_call = fpnode_data;

	fpnode_data = fpnode_data->star;
    }

    fpnode_data += (collect & ((1 << lock >> 1) - 1)) << 1;

    *link = fpnode_call;
    *push = fpnode_data;

    return status;
}
#endif				/* __DELAY_SPLIT__ */


int
x1f4_push_qlrate(void *qlrate, unsigned index, void **push)
{
    int status;
    unsigned count;

    count = qlrate(qlrate)->link_a.fplist.size;

    if (count < index) {
	status = DECK_ERROR;
    } else {
	unsigned rate;

	status = 0;

	rate = qlrate(qlrate)->link_a.fplist.rate;

	if (rate) {
	    do {
		struct fpnode_type *fpnode_data;
#if __DELAY_SPLIT__
		unsigned black = 0, late = 0;
#endif				/* __DELAY_SPLIT__ */
		unsigned lock;
#if __DELAY_SPLIT__
		void *hook = NULL;
#endif				/* __DELAY_SPLIT__ */
		void *node;
#if __DELAY_SPLIT__
		void *sail = NULL;
#endif				/* __DELAY_SPLIT__ */

		lock = qlrate(qlrate)->link_a.fplist.lock;

		if (qlrate(qlrate)->link_a.fplist.node >> lock) {
		    status = _libx1f4l2_fall_qlrate(qlrate);
		    if (status) {
			break;
		    } else {
			rate++;
		    }
		}

		qlrate(qlrate)->link_a.fplist.size = count + 1;

		fpnode_data = qlrate(qlrate)->link_a.fplist.fpnode;

		if (1) {
		    struct fpnode_type *fpnode_call = NULL;

		    if (1) {
			unsigned half;

			half = 1 << lock >> 1;

			while (half) {
			    struct fpnode_type *fpnode_slip;
			    unsigned pick;

			    fpnode_slip = fpnode_data + half;
			    pick = fpnode_slip->call;
			    if (pick < index) {
				index -= pick;
				fpnode_data = fpnode_slip;
			    } else {
				if (pick) {
				    fpnode_slip->call = pick + 1;
				} else {
				    fpnode_data = fpnode_slip;
				}
			    }

			    half >>= 1;
			}

			rate--;
		    }

		    while (rate) {
			unsigned half;

			if (fpnode_data->node >> lock) {
#if __DELAY_SPLIT__
			    if (hook) {
				struct fpnode_type *fpnode_slip;

				fpnode_slip = fpnode_call->star;
				fpnode_slip->call = index << lock
				    | (fpnode_data - fpnode_slip);
			    } else {
				hook = fpnode_data;
				sail = fpnode_call;
				if (1) {
				    late = rate;
				    if (1) {
					black = index;
				    }
				}
			    }
#else
			    int excess;

			    excess = _libx1f4l2_sand_qlrate
				(qlrate, fpnode_call, &fpnode_data, &index,
				 rate);
			    if (excess) {
				if (status) {
				} else {
				    status = excess;
				}
			    }
#endif				/* __DELAY_SPLIT__ */
#if __DELAY_SPLIT__
			} else {
			    hook = NULL;
#endif				/* __DELAY_SPLIT__ */
			}

			fpnode_call = fpnode_data;

			fpnode_data = fpnode_data->star;

			half = 1 << lock >> 1;

			while (4 < half) {
			    struct fpnode_type *fpnode_slip;
			    unsigned pick;

			    fpnode_slip = fpnode_data + half;
			    pick = fpnode_slip->call;
			    if (pick < index) {
				index -= pick;
				fpnode_data = fpnode_slip;
			    } else {
				fpnode_slip->call = pick + 1;
			    }

			    half >>= 1;

			    fpnode_slip = fpnode_data + half;
			    pick = fpnode_slip->call;
			    if (pick < index) {
				index -= pick;
				fpnode_data = fpnode_slip;
			    } else {
				fpnode_slip->call = pick + 1;
			    }

			    half >>= 1;

			    fpnode_slip = fpnode_data + half;
			    pick = fpnode_slip->call;
			    if (pick < index) {
				index -= pick;
				fpnode_data = fpnode_slip;
			    } else {
				fpnode_slip->call = pick + 1;
			    }

			    half >>= 1;

			    fpnode_slip = fpnode_data + half;
			    pick = fpnode_slip->call;
			    if (pick < index) {
				index -= pick;
				fpnode_data = fpnode_slip;
			    } else {
				fpnode_slip->call = pick + 1;
			    }

			    half >>= 1;
			}
			while (half) {
			    struct fpnode_type *fpnode_slip;
			    unsigned pick;

			    fpnode_slip = fpnode_data + half;
			    pick = fpnode_slip->call;
			    if (pick < index) {
				index -= pick;
				fpnode_data = fpnode_slip;
			    } else {
				fpnode_slip->call = pick + 1;
			    }

			    half >>= 1;
			}

			rate--;
		    }

		    if (fpnode_data->node) {
		    } else {
			int excess;

#if __DELAY_SPLIT__
			if (hook) {
			    excess = part_late
				(qlrate, late, sail, hook, black, &fpnode_call,
				 &fpnode_data);
			    if (excess) {
				if (status) {
				} else {
				    status = excess;
				}
			    }
			}
#endif				/* __DELAY_SPLIT__ */

			excess = _libx1f4l2_mold_qlrate
			    (qlrate, fpnode_call, fpnode_data, index, push);
			if (excess) {
			    if (status) {
			    } else {
				status = excess;
			    }
			}

			break;
		    }

		    count = fpnode_data->node;
		}

		if (index) {
		    QLNODE dana, news, slip;

		    news = fpnode_data->node;

		    dana = news;

		    index--;

		    slip = (QLNODE) 1 << index;

		    if (news & slip) {
			unsigned trans;

			news ^= slip;

			slip--;

			slip &= news;

			ql_bitcount(trans, slip);

			fpnode_data->node = slip | (news ^ slip) << 1;

			index += trans;
			index++;

			*push = (char *) fpnode_data->star
			    + index * qlrate(qlrate)->link_a.fplist.mall;
		    } else {
			QLNODE call;
			unsigned mall, trans;

			mall = qlrate(qlrate)->link_a.fplist.mall;

			node = fpnode_data->star;

			slip--;

			slip &= news;

			ql_bitcount(trans, slip);

			trans += index;

			/*
			 * meant to be shifted by QLBITS - index.
			 * will shift less for zero index and adapt
			 * accordingly.
			 */
			slip = news << (QLBITS - 1 - index);
			ql_reverse(slip);

			call = news >> index;
			if (call) {
			    call ^= call - 1;

			    if (slip & call >> 1) {
				unsigned flat, miss;

				flat = index;

				ql_xfs(miss, slip);

				flat -= miss;

				miss &= ~1;
				miss += ~trans & 1;

				node = (char *) node + (trans - miss) * mall;

				memmove(node, (char *) node + mall,
					miss * mall);

				call = 1;
				call <<= flat;
				call--;
				slip = ~call;
				slip <<= 1;
				fpnode_data->node =
				    (dana & call) | (dana & slip) << 1;

				*push = (char *) node + miss * mall;
			    } else {
				unsigned flat, miss;

				ql_bitcount(miss, call);

				flat = miss - 1;

				miss &= ~1;
				miss -= trans & 1;

				node = (char *) node + (trans + 1) * mall;

				memmove((char *) node + mall, node,
					miss * mall);

				flat += index;
				call = 1;
				call <<= flat;
				call--;
				slip = ~call;
				slip <<= 1;
				fpnode_data->node =
				    (dana & call) | (dana & slip) << 1;

				*push = node;
			    }
			} else {
			    unsigned flat, miss;

			    flat = index;

			    ql_xfs(miss, slip);

			    flat -= miss;

			    miss &= ~1;
			    miss += ~trans & 1;

			    node = (char *) node + (trans - miss) * mall;

			    memmove(node, (char *) node + mall, miss * mall);

			    call = 1;
			    call <<= flat;
			    call--;
			    slip = ~call;
			    slip <<= 1;
			    fpnode_data->node =
				(dana & call) | (dana & slip) << 1;

			    *push = (char *) node + miss * mall;
			}
		    }
		} else {
		    QLNODE news, slip;
		    unsigned mall, miss;

		    news = fpnode_data->node;

		    node = fpnode_data->star;

		    ql_xfs(miss, news);

		    miss++;

		    mall = qlrate(qlrate)->link_a.fplist.mall;

		    memmove((char *) node + mall, node, miss * mall);

		    slip = ~(QLNODE) 0 << miss;
		    fpnode_data->node = (news & slip) << 1;

		    *push = node;
		}
	    } while (0);
	} else {
	    if (count ^ 3 * QLBITS >> 1) {
		unsigned mall;
		void *node;

		qlrate(qlrate)->link_a.fplist.size = count + 1;

		node = qlrate(qlrate)->link_a.fplist.fpnode;

		mall = qlrate(qlrate)->link_a.fplist.mall;

		node = (char *) node + index * mall;

		count -= index;
		if (count) {
		    memmove((char *) node + mall, node, count * mall);
		} else {
		}

		*push = node;
	    } else {
		status = _libx1f4l2_rail_qlrate(qlrate, index, push);
	    }
	}
    }

    return status;
}
