/*
 * qlcc.f.c
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <stddef.h>

#include <mask.0.h>
#include <qlcc-defs.h>
#include <qlcc-inter.h>
#include <qlcc-names.h>
#include <qlcc-types.h>
#include <trans.h>

#if SIZEOF_UNSIGNED == 4
# define RANGE_SHIFT			2
#endif				/* SIZEOF_UNSIGNED == 4 */

#if SIZEOF_UNSIGNED == 8
# define RANGE_SHIFT			3
#endif				/* SIZEOF_UNSIGNED == 8 */

#define qlrate(qlrate)			((struct qlrate_type *) (qlrate))

static int init_list(struct qlrate_type *, unsigned);

static int
init_list(struct qlrate_type *qlrate_data, unsigned mall)
{
    int status;
    void *data;

    status = qlrate_data->link_m.link
	(qlrate_data->link_m.data, &data, mall * 3 * QLBITS >> 1);
    if (status) {
	status = LINK_ERROR;
    } else {
	qlrate_data->link_a.fplist.fpnode = data;
    }

    return status;
}


int
x1f4_fast_qlrate(void *qlrate, unsigned flags,
		 struct screen_type *screen_data)
{
    int status;
    unsigned lock, mall;

    mall = screen_data->size;

    lock = TEXT_RATE;
    if (flags & RATE_FRAME) {
	lock = screen_data->rate;
    }

    if ((int) lock < 4) {
	/*
	 * improper error code, should invent another
	 */
	status = DECK_ERROR;
    } else {
	if (flags & TRANS_MASK) {
	    struct trans_type *trans_data;

	    trans_data = screen_data->trans_data;

	    qlrate(qlrate)->link_m.data = trans_data->data;
	    qlrate(qlrate)->link_m.free = trans_data->free;
	    qlrate(qlrate)->link_m.link = trans_data->link;
	    qlrate(qlrate)->link_m.mode = trans_data->mode;
	} else {
	    qlrate(qlrate)->link_m.free = _libx1f4l2_free_data;
	    qlrate(qlrate)->link_m.link = _libx1f4l2_link_data;
	    qlrate(qlrate)->link_m.mode = _libx1f4l2_mode_data;
	}

	status = init_list(qlrate, mall);
	if (status) {
	} else {
	    qlrate(qlrate)->link_a.fplist.lock = lock;
	    qlrate(qlrate)->link_a.fplist.mall = mall;

	    qlrate(qlrate)->link_a.fplist.node = 0;
	    qlrate(qlrate)->link_a.fplist.rate = 0;
	    qlrate(qlrate)->link_a.fplist.size = 0;

	    qlrate(qlrate)->link_e.fplink_data = NULL;
	    qlrate(qlrate)->link_e.fplink_text = NULL;

#if SIZEOF_UNSIGNED & (SIZEOF_UNSIGNED - 1)
#else
	    if (mall & (SIZEOF_UNSIGNED - 1)) {
#endif				/* SIZEOF_UNSIGNED & (SIZEOF_UNSIGNED - 1) */
		qlrate(qlrate)->link_v.ever = _libx1f4l2_ms00_qlrate;
		qlrate(qlrate)->link_v.last = _libx1f4l2_cs00_qlrate;
		qlrate(qlrate)->link_v.lead = _libx1f4l2_mv00_qlrate;
		qlrate(qlrate)->link_v.mind = _libx1f4l2_rv00_qlrate;
		qlrate(qlrate)->link_v.near = _libx1f4l2_hs00_qlrate;
		qlrate(qlrate)->link_v.post = _libx1f4l2_cv00_qlrate;
		qlrate(qlrate)->link_v.plus = _libx1f4l2_hv00_qlrate;
		qlrate(qlrate)->link_v.slip = _libx1f4l2_rs00_qlrate;
#if SIZEOF_UNSIGNED & (SIZEOF_UNSIGNED - 1)
#else
	    } else {
		mall >>= RANGE_SHIFT;
		if (mall >> 2) {
		    switch (mall & 3) {
		    case 0:
			qlrate(qlrate)->link_v.ever = _libx1f4l2_mx10_qlrate;
			qlrate(qlrate)->link_v.last = _libx1f4l2_cx10_qlrate;
			qlrate(qlrate)->link_v.lead = _libx1f4l2_mz10_qlrate;
			qlrate(qlrate)->link_v.mind = _libx1f4l2_rz10_qlrate;
			qlrate(qlrate)->link_v.near = _libx1f4l2_hx10_qlrate;
			qlrate(qlrate)->link_v.post = _libx1f4l2_cz10_qlrate;
			qlrate(qlrate)->link_v.plus = _libx1f4l2_hz10_qlrate;
			qlrate(qlrate)->link_v.slip = _libx1f4l2_rx10_qlrate;
			break;
		    case 1:
			qlrate(qlrate)->link_v.ever = _libx1f4l2_mx11_qlrate;
			qlrate(qlrate)->link_v.last = _libx1f4l2_cx11_qlrate;
			qlrate(qlrate)->link_v.lead = _libx1f4l2_mz11_qlrate;
			qlrate(qlrate)->link_v.mind = _libx1f4l2_rz11_qlrate;
			qlrate(qlrate)->link_v.near = _libx1f4l2_hx11_qlrate;
			qlrate(qlrate)->link_v.post = _libx1f4l2_cz11_qlrate;
			qlrate(qlrate)->link_v.plus = _libx1f4l2_hz11_qlrate;
			qlrate(qlrate)->link_v.slip = _libx1f4l2_rx11_qlrate;
			break;
		    case 2:
			qlrate(qlrate)->link_v.ever = _libx1f4l2_mx12_qlrate;
			qlrate(qlrate)->link_v.last = _libx1f4l2_cx12_qlrate;
			qlrate(qlrate)->link_v.lead = _libx1f4l2_mz12_qlrate;
			qlrate(qlrate)->link_v.mind = _libx1f4l2_rz12_qlrate;
			qlrate(qlrate)->link_v.near = _libx1f4l2_hx12_qlrate;
			qlrate(qlrate)->link_v.post = _libx1f4l2_cz12_qlrate;
			qlrate(qlrate)->link_v.plus = _libx1f4l2_hz12_qlrate;
			qlrate(qlrate)->link_v.slip = _libx1f4l2_rx12_qlrate;
			break;
		    default:
			qlrate(qlrate)->link_v.ever = _libx1f4l2_mx13_qlrate;
			qlrate(qlrate)->link_v.last = _libx1f4l2_cx13_qlrate;
			qlrate(qlrate)->link_v.lead = _libx1f4l2_mz13_qlrate;
			qlrate(qlrate)->link_v.mind = _libx1f4l2_rz13_qlrate;
			qlrate(qlrate)->link_v.near = _libx1f4l2_hx13_qlrate;
			qlrate(qlrate)->link_v.post = _libx1f4l2_cz13_qlrate;
			qlrate(qlrate)->link_v.plus = _libx1f4l2_hz13_qlrate;
			qlrate(qlrate)->link_v.slip = _libx1f4l2_rx13_qlrate;
		    }
		} else {
		    switch (mall) {
		    case 1:
			qlrate(qlrate)->link_v.ever = _libx1f4l2_ms11_qlrate;
			qlrate(qlrate)->link_v.last = _libx1f4l2_cs11_qlrate;
			qlrate(qlrate)->link_v.lead = _libx1f4l2_mv11_qlrate;
			qlrate(qlrate)->link_v.mind = _libx1f4l2_rv11_qlrate;
			qlrate(qlrate)->link_v.near = _libx1f4l2_hs11_qlrate;
			qlrate(qlrate)->link_v.post = _libx1f4l2_cv11_qlrate;
			qlrate(qlrate)->link_v.plus = _libx1f4l2_hv11_qlrate;
			qlrate(qlrate)->link_v.slip = _libx1f4l2_rs11_qlrate;
			break;
		    case 2:
			qlrate(qlrate)->link_v.ever = _libx1f4l2_ms12_qlrate;
			qlrate(qlrate)->link_v.last = _libx1f4l2_cs12_qlrate;
			qlrate(qlrate)->link_v.lead = _libx1f4l2_mv12_qlrate;
			qlrate(qlrate)->link_v.mind = _libx1f4l2_rv12_qlrate;
			qlrate(qlrate)->link_v.near = _libx1f4l2_hs12_qlrate;
			qlrate(qlrate)->link_v.post = _libx1f4l2_cv12_qlrate;
			qlrate(qlrate)->link_v.plus = _libx1f4l2_hv12_qlrate;
			qlrate(qlrate)->link_v.slip = _libx1f4l2_rs12_qlrate;
			break;
		    default:
			qlrate(qlrate)->link_v.ever = _libx1f4l2_ms13_qlrate;
			qlrate(qlrate)->link_v.last = _libx1f4l2_cs13_qlrate;
			qlrate(qlrate)->link_v.lead = _libx1f4l2_mv13_qlrate;
			qlrate(qlrate)->link_v.mind = _libx1f4l2_rv13_qlrate;
			qlrate(qlrate)->link_v.near = _libx1f4l2_hs13_qlrate;
			qlrate(qlrate)->link_v.post = _libx1f4l2_cv13_qlrate;
			qlrate(qlrate)->link_v.plus = _libx1f4l2_hv13_qlrate;
			qlrate(qlrate)->link_v.slip = _libx1f4l2_rs13_qlrate;
		    }
		}
	    }
#endif				/* SIZEOF_UNSIGNED & (SIZEOF_UNSIGNED - 1) */
	}
    }

    return status;
}
