/*
 * qlcc.o.c
 * Copyright (C) 2010-2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qlcc-config.h>

#include <qlcc-bits.h>
#include <qlcc-inter.h>
#include <qlcc-lines.h>
#include <qlcc-names.h>
#include <qlcc-types.h>

#define qlrate(qlrate)			((struct qlrate_type *) (qlrate))

#if __DELAY_MERGE__
int
_libx1f4l2_trap_qlrate(void *qlrate, QLNODE *near,
		       struct fpnode_type *fpnode_line,
		       struct fpnode_type *fpnode_data, unsigned index,
		       void *dhmiss)
#else
int
_libx1f4l2_trap_qlrate(void *qlrate, QLNODE *near,
		       struct fpnode_type *fpnode_line,
		       struct fpnode_type *fpnode_data, unsigned index)
#endif				/* __DELAY_MERGE__ */
{
    QLNODE ever, lead;
    int status = 0;
    struct fpnode_type *fpnode_last, *fpnode_post;
    unsigned call, last, line, post;

    line = 1 << qlrate(qlrate)->link_a.fplist.lock;

    call = fpnode_data - fpnode_line;

    fpnode_last = fpnode_data;

    fpnode_post = fpnode_data;

    if (near == &qlrate(qlrate)->link_a.fplist.node) {
	last = call;
	while (last) {
	    last--;
	    fpnode_last--;
	    if (~fpnode_last->node) {
		break;
	    }
	}

	post = line - 1 - call;
	while (post) {
	    post--;
	    fpnode_post++;
	    if (~fpnode_post->node) {
		break;
	    }
	}

	fpnode_data->node = ~(QLNODE) 0;

	lead = fpnode_last->node;

	ql_bitcount(last, lead);

	ever = fpnode_post->node;

	ql_bitcount(post, ever);
    } else {
	if (call & 1) {
	    fpnode_last--;
	    lead = fpnode_last->node;
	    if (lead ^ Lx55555555) {
		ql_bitcount(last, lead);
		if ((call + 1) ^ line) {
		    fpnode_post++;
		    ever = fpnode_post->node;
		    ql_bitcount(post, ever);
		    if (post < last) {
		    } else {
			post = QLBITS;
		    }
		} else {
		    ever = ~(QLNODE) 0;
		    post = QLBITS;
		}
	    } else {
		ever = ~(QLNODE) 0;
		last = HALF_QLBITS;
		post = QLBITS;
	    }
	} else {
	    if (call) {
		fpnode_post++;
		ever = fpnode_post->node;
		if (~ever) {
		    ql_bitcount(post, ever);
		    if (ever ^ Lx55555555) {
			last = QLBITS;
			lead = ~(QLNODE) 0;
		    } else {
			fpnode_last--;
			lead = fpnode_last->node;

			if (~lead) {
			    ql_bitcount(last, lead);
			} else {
			    fpnode_last--;
			    lead = fpnode_last->node;
			    ql_bitcount(last, lead);

			    if (last < post) {
			    } else {
				last = QLBITS;
			    }
			}
		    }
		} else {
		    fpnode_last--;
		    lead = fpnode_last->node;

		    if (~lead) {
			ql_bitcount(last, lead);
			if ((call + 2) ^ line) {
			    fpnode_post++;
			    ever = fpnode_post->node;
			    ql_bitcount(post, ever);
			    if (post < last) {
			    } else {
				post = QLBITS;
			    }
			} else {
			    post = QLBITS;
			}
		    } else {
			fpnode_last--;
			lead = fpnode_last->node;
			ql_bitcount(last, lead);

			if ((call + 2) ^ line) {
			    fpnode_post++;
			    ever = fpnode_post->node;
			    ql_bitcount(post, ever);
			} else {
			    post = QLBITS;
			}
		    }
		}
	    } else {
		last = QLBITS;
		lead = ~(QLNODE) 0;

		fpnode_post++;
		ever = fpnode_post->node;
		if (~ever) {
		} else {
		    fpnode_post++;
		    ever = fpnode_post->node;
		}

		ql_bitcount(post, ever);
	    }
	}
    }

    line >>= 1;

    if (last < post) {
	if (lead ^ Lx55555555) {
	    unsigned seek;

	    qlrate(qlrate)->link_v.lead
		(qlrate, index, last, fpnode_data, fpnode_last, &seek);

	    if (near == &qlrate(qlrate)->link_a.fplist.node) {
		last = fpnode_last - fpnode_line;

		last++;
		do {
		    (fpnode_line + last)->call -= seek;
		    last += last & ~(last - 1);
		} while (last < line << 1);

		call++;
		while (call < line << 1) {
		    (fpnode_line + call)->call += seek;
		    call += call & ~(call - 1);
		}
	    } else {
		fpnode_data->call -= seek;

		if (call & 1) {
		} else {
		    unsigned slip;

		    last = fpnode_last - fpnode_line;

		    if (last & 1) {
		    } else {
			(fpnode_last + 1)->call -= seek;
		    }

		    slip = 1;
		    do {
			(fpnode_data + slip)->call += seek;
			slip <<= 1;
		    } while (slip & ~call);
		}
	    }
	} else {
#if __DELAY_MERGE__
	    if (((struct dhmiss_type *) dhmiss)->hook) {
# if __STAMP_MERGE__
# else
		unsigned mind = 0;
# endif				/* __STAMP_MERGE__ */

# if __STAMP_MERGE__
# else
		if (index == fpnode_data->node - 1) {
		    mind = 1;
		}
# endif				/* __STAMP_MERGE__ */

# if __STAMP_MERGE__
		status = _libx1f4l2_type_qlrate
		    (qlrate, dhmiss, &near, &fpnode_line, &fpnode_data);
# else
		status = _libx1f4l2_type_qlrate
		    (qlrate, dhmiss, mind, &near, &fpnode_line, &fpnode_data);
# endif				/* __STAMP_MERGE__ */

		call = fpnode_data - fpnode_line;

		fpnode_last = fpnode_data - 1;
		if (~fpnode_last->node) {
		} else {
		    fpnode_last--;
		}
	    }
#endif				/* __DELAY_MERGE__ */

	    qlrate(qlrate)->link_v.mind
		(qlrate, index, fpnode_data, fpnode_last);

	    last = fpnode_last - fpnode_line;

	    if (near == &qlrate(qlrate)->link_a.fplist.node) {
#if __DELAY_MERGE__
		int excess;
#else
# define excess				status
#endif				/* __DELAY_MERGE__ */

		excess = _libx1f4l2_bill_qlrate
		    (qlrate, near, fpnode_line, fpnode_last, last, fpnode_data,
		     call);
#if __DELAY_MERGE__
		if (excess) {
		    if (status) {
		    } else {
			status = excess;
		    }
		}
#endif				/* __DELAY_MERGE__ */
	    } else {
		unsigned text;

		text = HALF_QLBITS - 1;

		if (call & 1) {
#if __DELAY_MERGE__
		    int excess;
#else
# define excess				status
#endif				/* __DELAY_MERGE__ */

		    fpnode_data->call += text;

		    excess = _libx1f4l2_beta_qlrate
			(qlrate, near, fpnode_line, call);
#if __DELAY_MERGE__
		    if (excess) {
			if (status) {
			} else {
			    status = excess;
			}
		    }
#endif				/* __DELAY_MERGE__ */

#if __DELAY_MERGE__
#else
# undef excess
#endif				/* __DELAY_MERGE__ */
		} else {
		    unsigned slip;

		    if (last & 1) {
#if __DELAY_MERGE__
			int excess;
#else
# define excess				status
#endif				/* __DELAY_MERGE__ */
			void *star;

			star = fpnode_data->star;

			fpnode_data->node = fpnode_last->node;
			fpnode_data->star = fpnode_last->star;

			fpnode_last->node = ~(QLNODE) 0;

			fpnode_last->star = star;

			text++;

			fpnode_data->call -= text;

			call++;

			slip = 1;
			do {
			    (fpnode_data + slip)->call += text;
			    slip <<= 1;
			} while (slip & ~call);

			call--;

			excess = _libx1f4l2_beta_qlrate
			    (qlrate, near, fpnode_line, last);
#if __DELAY_MERGE__
			if (excess) {
			    if (status) {
			    } else {
				status = excess;
			    }
			}
#endif				/* __DELAY_MERGE__ */

#if __DELAY_MERGE__
#else
# undef excess
#endif				/* __DELAY_MERGE__ */
		    } else {
#if __DELAY_MERGE__
			int excess;
#else
# define excess				status
#endif				/* __DELAY_MERGE__ */

			slip = (fpnode_data + 1)->call;

			(fpnode_data + 1)->call = 0;

			fpnode_data->call += slip;

			(fpnode_data - 1)->call += slip;

			excess = _libx1f4l2_beta_qlrate
			    (qlrate, near, fpnode_line, call);
#if __DELAY_MERGE__
			if (excess) {
			    if (status) {
			    } else {
				status = excess;
			    }
			}
#endif				/* __DELAY_MERGE__ */

#if __DELAY_MERGE__
#else
# undef excess
#endif				/* __DELAY_MERGE__ */
		    }
		}
	    }
	}
    } else {
	if (ever ^ Lx55555555) {
	    unsigned seek;

	    qlrate(qlrate)->link_v.ever
		(qlrate, index, post, fpnode_data, fpnode_post, &seek);

	    post = fpnode_post - fpnode_line;

	    if (near == &qlrate(qlrate)->link_a.fplist.node) {
		post++;
		while (post < line << 1) {
		    (fpnode_line + post)->call -= seek;
		    post += post & ~(post - 1);
		}

		call++;
		do {
		    (fpnode_line + call)->call += seek;
		    call += call & ~(call - 1);
		} while (call < line << 1);
	    } else {
		fpnode_post->call += seek;

		if (post & 1) {
		} else {
		    unsigned slip;

		    if (call & 1) {
		    } else {
			(fpnode_post - 1)->call += seek;
		    }

		    slip = 1;
		    do {
			(fpnode_post + slip)->call -= seek;
			slip <<= 1;
		    } while (slip & ~post);
		}
	    }
	} else {
#if __DELAY_MERGE__
	    if (((struct dhmiss_type *) dhmiss)->hook) {
# if __STAMP_MERGE__
# else
		unsigned mind = 0;
# endif				/* __STAMP_MERGE__ */

# if __STAMP_MERGE__
# else
		if (index == fpnode_data->node - 1) {
		    mind = 1;
		}
# endif				/* __STAMP_MERGE__ */

# if __STAMP_MERGE__
		status = _libx1f4l2_type_qlrate
		    (qlrate, dhmiss, &near, &fpnode_line, &fpnode_data);
# else
		status = _libx1f4l2_type_qlrate
		    (qlrate, dhmiss, mind, &near, &fpnode_line, &fpnode_data);
# endif				/* __STAMP_MERGE__ */

		call = fpnode_data - fpnode_line;

		fpnode_post = fpnode_data + 1;
		if (~fpnode_post->node) {
		} else {
		    fpnode_post++;
		}
	    }
#endif				/* __DELAY_MERGE__ */

	    qlrate(qlrate)->link_v.slip
		(qlrate, index, fpnode_data, fpnode_post);

	    if (near == &qlrate(qlrate)->link_a.fplist.node) {
#if __DELAY_MERGE__
		int excess;
#else
# define excess				status
#endif				/* __DELAY_MERGE__ */

		excess = _libx1f4l2_mail_qlrate
		    (qlrate, near, fpnode_line, fpnode_data, call,
		     fpnode_post);
#if __DELAY_MERGE__
		if (excess) {
		    if (status) {
		    } else {
			status = excess;
		    }
		}
#endif				/* __DELAY_MERGE__ */
	    } else {
		unsigned slip;
		void *fpnode, *fqnode;

		fpnode = fpnode_data->star;
		fqnode = fpnode_post->star;

		fpnode_data->star = fqnode;
		fpnode_post->star = fpnode;

		fpnode_data->node = ~(QLNODE) 0;
index &= ~1;
		fpnode_post->node = (QLNODE) 1 << index;

		if (call & 1) {
#if __DELAY_MERGE__
		    int excess;
#else
# define excess				status
#endif				/* __DELAY_MERGE__ */

		    fpnode_post->call -= HALF_QLBITS - 1;

		    call++;

		    slip = 1;
		    do {
			(fpnode_post + slip)->call += HALF_QLBITS - 1;
			slip <<= 1;
		    } while (slip & ~call);

		    call--;

		    excess = _libx1f4l2_beta_qlrate
			(qlrate, near, fpnode_line, call);
#if __DELAY_MERGE__
		    if (excess) {
			if (status) {
			} else {
			    status = excess;
			}
		    }
#endif				/* __DELAY_MERGE__ */

#if __DELAY_MERGE__
#else
# undef excess
#endif				/* __DELAY_MERGE__ */
		} else {
		    post = fpnode_post - fpnode_line;
		    if (post & 1) {
#if __DELAY_MERGE__
			int excess;
#else
# define excess				status
#endif				/* __DELAY_MERGE__ */
			void *star;

			star = fpnode_data->star;

			fpnode_data->node = fpnode_post->node;
			fpnode_data->star = fpnode_post->star;

			fpnode_post->star = star;

			fpnode_post->call = QLBITS - 1;

			excess = _libx1f4l2_beta_qlrate
			    (qlrate, near, fpnode_line, post);
#if __DELAY_MERGE__
			if (excess) {
			    if (status) {
			    } else {
				status = excess;
			    }
			}
#endif				/* __DELAY_MERGE__ */

#if __DELAY_MERGE__
#else
# undef excess
#endif				/* __DELAY_MERGE__ */
		    } else {
#if __DELAY_MERGE__
			int excess;
#else
# define excess				status
#endif				/* __DELAY_MERGE__ */

			(fpnode_post - 1)->call = 0;

			fpnode_post->call -= HALF_QLBITS - 1;

			call += 2;

			slip = 1;
			do {
			    (fpnode_post + slip)->call += HALF_QLBITS - 1;
			    slip <<= 1;
			} while (slip & ~call);

			call -= 2;

			excess = _libx1f4l2_beta_qlrate
			    (qlrate, near, fpnode_line, call);
#if __DELAY_MERGE__
			if (excess) {
			    if (status) {
			    } else {
				status = excess;
			    }
			}
#endif				/* __DELAY_MERGE__ */

#if __DELAY_MERGE__
#else
# undef excess
#endif				/* __DELAY_MERGE__ */
		    }
		}
	    }
	}
    }

    return status;
}
