/*
 * qlcc.y.c
 * Copyright (C) 2010-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qlcc-defs.h>
#include <qlcc-names.h>
#include <qlcc-types.h>

#define qlrate(qlrate)			((struct qlrate_type *) (qlrate))

static int back_rate
    (struct qlrate_type *, struct fpnode_type *, unsigned, unsigned);
static int flat_rate(struct qlrate_type *, void *, unsigned, unsigned);

static int
back_rate(struct qlrate_type *qlrate_data, struct fpnode_type *fpnode_data,
	  unsigned rate, unsigned line)
{
    int excess, status = 0;
    unsigned i;

    i = line;
    if (rate) {
	for (; i; i--) {
	    if (fpnode_data->node) {
		excess = flat_rate(qlrate_data, fpnode_data->star, rate, line);
		if (excess) {
		    if (status) {
		    } else {
			status = FREE_ERROR;
		    }
		}
	    }

	    fpnode_data++;
	}
    } else {
	for (; i; i--) {
	    if (~fpnode_data->node) {
		excess = flat_rate(qlrate_data, fpnode_data->star, rate, line);
		if (excess) {
		    if (status) {
		    } else {
			status = FREE_ERROR;
		    }
		}
	    }

	    fpnode_data++;
	}
    }

    return status;
}


static int
flat_rate(struct qlrate_type *qlrate_data, void *node, unsigned rate,
	  unsigned line)
{
    int excess, status;

    if (rate) {
	status = back_rate(qlrate_data, node, rate - 1, line);
    } else {
	status = 0;
    }

    excess = qlrate_data->link_m.free(qlrate_data->link_m.data, node);
    if (excess) {
	if (status) {
	} else {
	    status = FREE_ERROR;
	}
    }

    return status;
}


int
_libx1f4l2_ceil_qlrate(void *qlrate)
{
    int excess, status;
    unsigned rate;
    void *land;

    rate = qlrate(qlrate)->link_a.fplist.rate;
    if (rate) {
	void *node;

	node = qlrate(qlrate)->link_a.fplist.fpnode;
	status = back_rate
	    (qlrate, node, rate - 1, 1 << qlrate(qlrate)->link_a.fplist.lock);
    } else {
	status = 0;
    }

    land = qlrate(qlrate)->link_e.fplink_text;
    if (land) {
	excess = qlrate(qlrate)->link_m.free
	    (qlrate(qlrate)->link_m.data, land);
	if (excess) {
	    if (status) {
	    } else {
		status = FREE_ERROR;
	    }
	}
    }

    land = qlrate(qlrate)->link_e.fplink_data;
    while (land) {
	struct fplink_type *fplink_data;

	fplink_data = ((struct fplink_type *) land)->fplink_data;

	excess = qlrate(qlrate)->link_m.free
	    (qlrate(qlrate)->link_m.data, land);
	if (excess) {
	    if (status) {
	    } else {
		status = FREE_ERROR;
	    }
	}

	land = fplink_data;
    }

    return status;
}
