/*
 * qlrate.h
 * Copyright (C) 2010, 2011, 2012, 2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_QLRATE_H__
#define __LIBx1f4_QLRATE_H__

#include <errors.h>
#include <qlcc-names.h>

/*
 * attribute bit definitions
 */
#define X1f4_QLRATE_RATE_FRAME		(1u << 0x01)
#define X1f4_QLRATE_TRANS_MASK		(1u << 0x02)

/*
 * error definitions
 */
#define X1f4_QLRATE_CALL_ERROR		LIBx1f4L2_CALL_ERROR
#define X1f4_QLRATE_DECK_ERROR		LIBx1f4L2_DECK_ERROR
#define X1f4_QLRATE_FREE_ERROR		LIBx1f4L2_FREE_ERROR
#define X1f4_QLRATE_LINK_ERROR		LIBx1f4L2_LINK_ERROR
#define X1f4_QLRATE_MODE_ERROR		LIBx1f4L2_MODE_ERROR
#define X1f4_QLRATE_SIZE_ERROR		LIBx1f4L2_SIZE_ERROR

typedef struct x1f4_qlrate_type {
    unsigned rate, size;
    void *trans;
} x1f4_qlrate_type;

#if !defined __x1f4_trans_type_defined__
typedef struct x1f4_trans_type {
    int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	(*mode) (void *, void **, unsigned);
    void *context;
} x1f4_trans_type;
#endif				/* !__x1f4_trans_type_defined__ */

#if !defined __x1f4_trans_type_defined__
# define __x1f4_trans_type_defined__
#endif				/* !__x1f4_trans_type_defined__ */

extern int x1f4_call_qlrate(unsigned *);
extern int x1f4_case_qlrate(void *);
extern int x1f4_ever_qlrate(void *, void **);
extern int x1f4_fast_qlrate(void *, unsigned, struct x1f4_qlrate_type *);
extern int x1f4_fini_qlrate(void **);
extern int x1f4_flat_qlrate(void *);
extern int x1f4_high_qlrate(void *);
extern int x1f4_init_qlrate(void **, unsigned, struct x1f4_qlrate_type *);
extern int x1f4_lead_qlrate(void *, void **);
extern int x1f4_lime_qlrate(void *, void *, int (*) (void *, void *));
extern int x1f4_miss_qlrate(void *, unsigned);
extern int x1f4_peek_qlrate(void *, unsigned, void **);
extern int x1f4_push_qlrate(void *, unsigned, void **);
extern int x1f4_size_qlrate(void *, unsigned *);
extern int x1f4_swap_qlrate(void *, void *);

#endif				/* __LIBx1f4_QLRATE_H__ */
