/*
 * qxcc.3.c
 * Copyright (C) 2010-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qxcc-config.h>

#include <stddef.h>
#include <string.h>

#include <qxcc-bits.h>
#include <qxcc-defs.h>
#include <qxcc-inter.h>
#include <qxcc-types.h>

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

#define qxrate(qxrate)			((struct qxrate_type *) (qxrate))

#if __DELAY_MERGE__
static int join_late(struct qxrate_type *, unsigned, void *, unsigned,
		     unsigned, unsigned *, void **);
#endif				/* __DELAY_MERGE__ */

#if __DELAY_MERGE__
static int
join_late(struct qxrate_type *qxrate_data, unsigned rate, void *node,
	  unsigned index, unsigned mind, unsigned *text, void **miss)
{
    int status = 0;
    unsigned epslip, level, lock, trans, violet;
    void *pick;

    lock = qxrate_data->link_a.fplist.lock;

    violet = sizeof(unsigned) << 1 << lock;

    index -= mind;

    while (1) {
	unsigned class, epnode, *lpnode;

	trans = index;

	lpnode = node;

	epnode = P_ROOT(Q_NODE(node));

	do {
	    lpnode = M_TEXT(node, epnode);

	    epslip = *lpnode;

	    level = index;

	    lpnode++;

	    class = *lpnode;
	    if (index < class) {
		epnode = L_LAST(epslip);
		class++;
	    } else {
		epnode = L_NEXT(epslip);
		index -= class;
	    }
	} while (epnode < violet);

	if (1) {
	    pick = *(void **) M_TEXT(node, epnode);
	}

	if (1) {
	    int excess;

	    trans += mind;
	    level += mind;
	    index += mind;

	    excess = _libx1f4l2_away_qxrate
		(qxrate_data, node, trans, level, class, &index, epslip,
		 lpnode, &pick);

	    index -= mind;

	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	{
	    node = pick;
	}

	if (rate) {
	    rate--;
	} else {
	    break;
	}
    }

    index += mind;

    *miss = node;
    *text = index;

    return status;
}
#endif				/* __DELAY_MERGE__ */


int
x1f4_miss_qxrate(void *qxrate, unsigned index)
{
    int status;
    unsigned count;

    count = qxrate(qxrate)->link_a.fplist.size;

    if (!(index < count)) {
	status = DECK_ERROR;
    } else {
	unsigned rate;
	void *node, *pick;

	node = qxrate(qxrate)->link_a.fplist.fpnode_data;

	qxrate(qxrate)->link_a.fplist.size = count - 1;

	status = 0;

	rate = qxrate(qxrate)->link_a.fplist.rate;

	do {
	    if (rate) {
#if __DELAY_MERGE__
		unsigned black = 0;
#endif				/* __DELAY_MERGE__ */
		unsigned epslip;
#if __DELAY_MERGE__
		unsigned late = 0;
#endif				/* __DELAY_MERGE__ */
		unsigned level, lock, *lpnode, trans, upper, violet;
#if __DELAY_MERGE__
		void *hook;
#endif				/* __DELAY_MERGE__ */

#if __DELAY_MERGE__
		hook = NULL;
#endif				/* __DELAY_MERGE__ */

		lock = qxrate(qxrate)->link_a.fplist.lock;

		violet = sizeof(unsigned) << 1 << lock;

		rate--;

		while (rate) {
		    unsigned class, epnode, *lpnode;

		    trans = index;

		    lpnode = node;

		    epnode = P_ROOT(Q_NODE(node));

		    lpnode[1]--;

		    do {
			lpnode = M_TEXT(node, epnode);

			epslip = *lpnode;

			level = index;

			lpnode++;

			class = *lpnode;
			if (index < class) {
			    epnode = L_LAST(epslip);
			    *lpnode = class - 1;
			} else {
			    epnode = L_NEXT(epslip);
			    index -= class;
			}
		    } while (epnode < violet);

		    if (1) {
			pick = *(void **) M_TEXT(node, epnode);
			if (1) {
			    rate--;
			}
		    }

		    if (1) {
			unsigned list;

			list = P_LIST(Q_NODE(pick));
			if (Q_NODE(M_TEXT(pick, list + sizeof(unsigned)))
			    ^ ((1 << lock >> 1) - 1)) {
#if __DELAY_MERGE__
			    hook = NULL;
#endif				/* __DELAY_MERGE__ */
			} else {
#if __DELAY_MERGE__
			    if (hook) {
			    } else {
				hook = node;
				late = rate;
				if (1) {
				    black = trans;
				}
			    }
#else
			    int excess;

			    excess = _libx1f4l2_away_qxrate
				(qxrate, node, trans, level, class, &index,
				 epslip, lpnode, &pick);
			    if (excess) {
				if (status) {
				} else {
				    status = excess;
				}
			    }
#endif				/* __DELAY_MERGE__ */
			}
		    }

		    {
			node = pick;
		    }
		}
		{
		    unsigned epnode;

		    trans = index;

		    lpnode = node;

		    epnode = P_ROOT(Q_NODE(node));

		    lpnode++;

		    count = *lpnode;

		    *lpnode = count - 1;

		    do {
			unsigned class;

			lpnode = M_TEXT(node, epnode);

			epslip = *lpnode;

			upper = count;

			level = index;

			lpnode++;

			class = *lpnode;
			if (index < class) {
			    epnode = L_LAST(epslip);
			    *lpnode = class - 1;
			    count = class;
			} else {
			    epnode = L_NEXT(epslip);
			    index -= class;
			    count -= class;
			}
		    } while (epnode < violet);

		    if (1) {
			pick = *(void **) M_TEXT(node, epnode);
		    }
		}

		if (count << 1 ^ qxrate(qxrate)->link_a.fplist.line) {
		} else {
#if __DELAY_MERGE__
		    if (hook) {
			int excess;
			unsigned mind = 0;

			if (index == count - 1) {
			    mind = 1;
			}

			excess = join_late
			    (qxrate, late, hook, black, mind, &index, &node);
			if (excess) {
			    if (status) {
			    } else {
				status = excess;
			    }
			}

			{
			    unsigned epnode;

			    index -= mind;

			    trans = index;

			    lpnode = node;

			    epnode = P_ROOT(Q_NODE(node));

			    lpnode++;

			    count = *lpnode + 1;

			    do {
				unsigned class;

				lpnode = M_TEXT(node, epnode);

				epslip = *lpnode;

				upper = count;

				level = index;

				lpnode++;

				class = *lpnode;
				if (index < class) {
				    epnode = L_LAST(epslip);
				    class++;
				    count = class;
				} else {
				    epnode = L_NEXT(epslip);
				    index -= class;
				    count -= class;
				}
			    } while (epnode < violet);

			    index += mind;
			    level += mind;
			    trans += mind;
			}
		    }
#endif				/* __DELAY_MERGE__ */

		    status = _libx1f4l2_long_qxrate
			(qxrate, violet, trans, level, index, count, upper,
			 epslip, lpnode, node);
		    if (1) {
			break;
		    }
		}
	    } else {
		pick = node;
	    }

	    count -= index;
	    count--;
 	    if (count) {
		unsigned mall;

		mall = qxrate(qxrate)->link_a.fplist.mall;

		pick = (char *) pick + index * mall;

		memmove(pick, (char *) pick + mall, count * mall);
	    } else {
	    }
	} while (0);
    }

    return status;
}
