/*
 * qxcc.f.c
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <mask.0.h>
#include <qxcc-bits.h>
#include <qxcc-defs.h>
#include <qxcc-inter.h>
#include <qxcc-names.h>
#include <qxcc-types.h>
#include <trans.h>

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

#define qxrate(qxrate)			((struct qxrate_type *) (qxrate))

static int init_list(struct qxrate_type *, unsigned, unsigned);

static int
init_list(struct qxrate_type *qxrate_data, unsigned mall, unsigned lock)
{
    int status;
    void *data;

    status = qxrate_data->link_m.link
	(qxrate_data->link_m.data, &data, mall * 3 * lock >> 1);
    if (status) {
	status = LINK_ERROR;
    } else {
	qxrate_data->link_a.fplist.fpnode_data = data;
    }

    return status;
}


int
x1f4_fast_qxrate(void *qxrate, unsigned flags,
		 struct screen_type *screen_data)
{
    int status;
    unsigned line, lock, mall;

    mall = screen_data->size;

    line = TEXT_LINE;
    if (flags & HALF_FRAME) {
	line = screen_data->half << 2;
    }

    lock = TEXT_RATE;
    if (flags & RATE_FRAME) {
	lock = screen_data->rate;
    }

    while (1 << 14 < ((sizeof(unsigned) << 1) + mall) << lock) {
	lock--;
    }

    if ((int) lock < 4) {
	/*
	 * improper error code, should invent another
	 */
	status = DECK_ERROR;
    } else {
	if (flags & TRANS_MASK) {
	    struct trans_type *trans_data;

	    trans_data = screen_data->trans_data;

	    qxrate(qxrate)->link_m.data = trans_data->data;
	    qxrate(qxrate)->link_m.free = trans_data->free;
	    qxrate(qxrate)->link_m.link = trans_data->link;
	    qxrate(qxrate)->link_m.mode = trans_data->mode;
	} else {
	    qxrate(qxrate)->link_m.free = _libx1f4l2_free_data;
	    qxrate(qxrate)->link_m.link = _libx1f4l2_link_data;
	    qxrate(qxrate)->link_m.mode = _libx1f4l2_mode_data;
	}

	status = init_list(qxrate, mall, line);
	if (status) {
	} else {
	    qxrate(qxrate)->link_a.fplist.line = line;
	    qxrate(qxrate)->link_a.fplist.lock = lock;
	    qxrate(qxrate)->link_a.fplist.mall = mall;
	    qxrate(qxrate)->link_a.fplist.rate = 0;
	    qxrate(qxrate)->link_a.fplist.size = 0;

	    qxrate(qxrate)->link_e.fplink_data = NULL;
	    qxrate(qxrate)->link_e.fplink_text = NULL;
	}
    }

    return status;
}
