/*
 * qxcc.u.c
 * Copyright (C) 2010-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qxcc-config.h>

#include <qxcc-bits.h>
#include <qxcc-defs.h>
#include <qxcc-inter.h>
#include <qxcc-types.h>

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

#define qxrate(qxrate)			((struct qxrate_type *) (qxrate))

#if __DELAY_MERGE__
# define last_index(index, pick) \
    ((index) == Q_NODE(M_TEXT((pick), sizeof(unsigned))) ? 1 : 0)
#else
# define last_index(index, pick) \
    ((index) == Q_NODE(M_TEXT((pick), sizeof(unsigned))) - 1 ? 1 : 0)
#endif				/* __DELAY_MERGE__ */

#define run_list(list, lock, bare) \
    {									      \
	(list) = P_LIST(Q_NODE((bare))) + sizeof(unsigned);		      \
	if ((list) ^ sizeof(unsigned)) {				      \
	    (list) =							      \
		Q_NODE(M_TEXT((bare), (list))) ^ ((1 << (lock) >> 1) - 1);    \
	}								      \
    }

#define set_list(list, lock, bare) \
    {									      \
	(list) = P_LIST(Q_NODE((bare))) + sizeof(unsigned);		      \
	if ((list) ^ sizeof(unsigned)) {				      \
	    (list) =							      \
		Q_NODE(M_TEXT((bare), (list))) ^ ((1 << lock) - 3);	      \
	}								      \
    }

int
_libx1f4l2_away_qxrate(void *qxrate, void *node, unsigned trans,
		       unsigned level, unsigned class, unsigned *shift,
		       unsigned epslip, unsigned *lpnode, void **away)
{
    int status = 0;
    unsigned index, list, lock, violet;
    void *pick;

    pick = *away;

    index = *shift;

    lock = qxrate(qxrate)->link_a.fplist.lock;

    violet = sizeof(unsigned) << 1 << lock;

    if (level < class) {
	unsigned esnode;

	esnode = L_NEXT(epslip);
	if (esnode < violet) {
	    unsigned esslip;
	    void *delete, *mare;

	    delete = M_TEXT(node, esnode);

	    esslip = Q_NODE(delete);

	    mare = *(void **) M_TEXT(node, L_LAST(esslip));

	    run_list(list, lock, mare);
	    if (list) {
		unsigned news;
		void *mind;

		_libx1f4l2_fair_qxrate(qxrate, mare, &news, &mind);
		_libx1f4l2_bill_qxrate(qxrate, pick, news, mind);

		*lpnode += news;

		delete = M_TEXT(delete, sizeof(unsigned));
		S_NODE(delete, Q_NODE(delete) - news);
	    } else {
		void *ware;

		ware = *(void **) M_TEXT(node, L_NEXT(esslip));

		run_list(list, lock, ware);
		if (list) {
		    unsigned clip, news;
		    void *mind;

		    _libx1f4l2_fair_qxrate(qxrate, mare, &news, &mind);
		    _libx1f4l2_bill_qxrate(qxrate, pick, news, mind);

		    *lpnode += news;

		    _libx1f4l2_fair_qxrate(qxrate, ware, &clip, &mind);
		    _libx1f4l2_bill_qxrate(qxrate, mare, clip, mind);

		    delete = M_TEXT(delete, sizeof(unsigned));
		    S_NODE(delete, Q_NODE(delete) + clip - news);
		} else {
		    status = _libx1f4l2_beta_qxrate
			(qxrate, 1, node, last_index(index, pick), trans,
			 shift, away);
		}
	    }
	} else {
	    void *mare;

	    mare = *(void **) M_TEXT(node, esnode);

	    run_list(list, lock, mare);
	    if (list) {
		unsigned news;
		void *mind;

		_libx1f4l2_fair_qxrate(qxrate, mare, &news, &mind);
		_libx1f4l2_bill_qxrate(qxrate, pick, news, mind);

		*lpnode += news;
	    } else {
		set_list(list, lock, node);
		if (list) {
		    status = _libx1f4l2_beta_qxrate
			(qxrate, 1, node, last_index(index, pick), trans,
			 shift, away);
		} else {
		    status = _libx1f4l2_moon_qxrate(qxrate, away);
		    if (status) {
		    } else {
			*shift = trans;
		    }
		}
	    }
	}
    } else {
	unsigned esnode;

	esnode = L_LAST(epslip);
	if (esnode < violet) {
	    unsigned esslip;
	    void *delete, *mare;

	    delete = M_TEXT(node, esnode);

	    esslip = Q_NODE(delete);

	    mare = *(void **) M_TEXT(node, L_NEXT(esslip));

	    run_list(list, lock, mare);
	    if (list) {
		unsigned news;
		void *slip;

		_libx1f4l2_else_qxrate(qxrate, mare, &news, &slip);
		_libx1f4l2_mail_qxrate(qxrate, pick, news, slip);

		*lpnode -= news;

		*shift = index + news;
	    } else {
		void *ware;

		ware = *(void **) M_TEXT(node, L_LAST(esslip));

		run_list(list, lock, ware);
		if (list) {
		    unsigned news;
		    void *slip;

		    _libx1f4l2_else_qxrate(qxrate, mare, &news, &slip);
		    _libx1f4l2_mail_qxrate(qxrate, pick, news, slip);

		    *lpnode -= news;

		    *shift = index + news;

		    _libx1f4l2_else_qxrate(qxrate, ware, &news, &slip);
		    _libx1f4l2_mail_qxrate(qxrate, mare, news, slip);

		    delete = M_TEXT(delete, sizeof(unsigned));
		    S_NODE(delete, Q_NODE(delete) - news);
		} else {
		    status = _libx1f4l2_beta_qxrate
			(qxrate, 1, node, last_index(index, pick), trans,
			 shift, away);
		}
	    }
	} else {
	    void *mare;

	    mare = *(void **) M_TEXT(node, esnode);

	    run_list(list, lock, mare);
	    if (list) {
		unsigned news;
		void *slip;

		_libx1f4l2_else_qxrate(qxrate, mare, &news, &slip);
		_libx1f4l2_mail_qxrate(qxrate, pick, news, slip);

		*lpnode -= news;

		*shift = index + news;
	    } else {
		set_list(list, lock, node);
		if (list) {
		    status = _libx1f4l2_beta_qxrate
			(qxrate, 1, node, last_index(index, pick), trans,
			 shift, away);
		} else {
		    status = _libx1f4l2_moon_qxrate(qxrate, away);
		    if (status) {
		    } else {
			*shift = trans;
		    }
		}
	    }
	}
    }

    return status;
}
