/*
 * s8xv.6.c
 * Copyright (C) 2009-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <s8xx-defs.h>
#include <s8xx-names.h>
#include <s8xx-types.h>

#define fast(miss)			((const byte *) (miss))

#define s8deck(s8deck)			((struct s8deck_type *) (s8deck))

#define fpnode_data(fpnode) \
    ((struct fpnode_type **) ((struct fpnode_type *) (fpnode) + 1))

static int pass_name(void *, const byte *, unsigned, struct fpnode_type *);

static int
pass_name(void *s8deck, const byte *name, unsigned size,
	  struct fpnode_type *fpnode_data)
{
    int status;
    struct fpnode_type **fpnode, **fqnode = NULL, *fqnode_data = NULL;
    unsigned call, last = 0, rate, tier = 0;

    fpnode = &s8deck(s8deck)->link_a.fpdeck.fpnode_data;

    size <<= 3;

    call = fpnode_data->data.call;
    if (call & POINT_CAST) {
	while (1) {
	    unsigned seek;

	    rate = point_rate(call);

	    fqnode_data = fpnode_data;

	    fqnode = fpnode;

	    seek = call & POINT_SEEK;

	    if (rate < size) {
		if (seek) {
		    tier = 1;
		} else {
		    tier = name[rate >> 3] >> (7 - (rate & 7)) & 1;
		}
	    } else {
		tier = 0;
	    }

	    fpnode = fpnode_data(fpnode_data) + tier;

	    fpnode_data = *fpnode;

	    call = fpnode_data->data.call;
	    if (call & POINT_CAST) {
		last = seek ? rate : rate + 1;
	    } else {
		break;
	    }
	}
    }

    size >>= 3;

    if (call ^ size) {
	status = 0;
    } else {
	if (memcmp((void **) (fpnode_data + 1) + 1, name, size)) {
	    status = 0;
	} else {
	    status = s8deck(s8deck)->link_m.free
		(s8deck(s8deck)->link_m.data, fpnode_data);
	    if (status) {
		status = FREE_ERROR;
	    } else {
		if (fqnode) {
		    unsigned fail;

		    tier ^= 1;

		    fpnode_data = fpnode_data(fqnode_data)[tier];

		    call = fqnode_data->data.call;
		    fail = fqnode_data->fine;

		    rate = point_rate(call);
		    if (call & POINT_SEEK) {
		    } else {
			rate++;
		    }

		    status = s8deck(s8deck)->link_m.free
			(s8deck(s8deck)->link_m.data, fqnode_data);
		    if (status) {
			status = FREE_ERROR;
		    } else {
		    }

		    /*
		     * edit tree even for free failure
		     */

		    *fqnode = fpnode_data;

		    if (fpnode_data->data.call & POINT_CAST) {
		    } else {
			if (last ^ rate) {
			    unsigned slip;

			    if (call & POINT_SEEK) {
			    } else {
				rate--;
			    }

			    slip = rate - last;
			    if (slip < 25) {
				unsigned bits, fine;

				fine = fpnode_data->fine;
				bits = fine >> 24;
				fine &= (1 << 24) - 1;
				fine >>= slip;
				if (call & POINT_SEEK) {
				    fine |= fail << (24 - slip);
				} else {
				    fine |= fail << (25 - slip);
				    fine |= tier << (24 - slip);
				    fine >>= 1;
				    bits++;
				}

				fine &= (1 << 24) - 1;
				bits += slip;
				if (24 < bits) {
				    bits = 24;
				}

				fpnode_data->fine = fine | bits << 24;
			    } else {
				const byte *word;
				unsigned beta, fail, fine = 0, near, turn;

				size = fpnode_data->data.call;

				turn = (size << 3) - last;
				if (24 < turn) {
				    turn = 24;
				}

				beta = last >> 3;

				word = (byte *)
				    ((void **) (fpnode_data + 1) + 1) + beta;

				beta = size - beta;

				fail = last & 7;

				if (fail) {
				    if (4 < beta) {
					beta = 4;
				    }
				} else {
				    if (3 < beta) {
					beta = 3;
				    }
				}

				near = beta;

				while (beta) {
				    beta--;
				    fine <<= 8;
				    fine |= *word++;
				}

				if (fail) {
				    fine <<= fail;
				    fine <<= (4 - near) << 3;
				    fine >>= 8;
				} else {
				    fine <<= (3 - near) << 3;
				}

				fpnode_data->fine = fine | turn << 24;
			    }
			}
		    }
		} else {
		    s8deck(s8deck)->link_a.fpdeck.fpnode_data = NULL;
		}
	    }
	}
    }

    return status;
}


int
x1f4_pass_s8path(void *s8deck, const char *name, unsigned size)
{
    int status;
    struct fpnode_type *fpnode_data;

    fpnode_data = s8deck(s8deck)->link_a.fpdeck.fpnode_data;
    if (fpnode_data) {
	status = pass_name(s8deck, fast(name), size, fpnode_data);
    } else {
	status = 0;
    }

    return status;
}
