/*
 * scxv.7.c
 * Copyright (C) 2009-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include <scxx-defs.h>
#include <scxx-names.h>
#include <scxx-patch.h>
#include <scxx-types.h>

#define back(miss)			((const char *) (miss))
#define fast(miss)			((const byte *) (miss))

#define null				(&scdeck(scdeck)->link_i.zero)

#define scdeck(scdeck)			((struct scdeck_type *) (scdeck))

static int case_deck(void *, struct fpnode_type *, const byte *, unsigned,
		     const byte *, unsigned, const void *);
static int ever_deck(void *, struct fpnode_type *, const byte *, unsigned,
		     const void *);
static int fast_deck(void *, struct fpnode_type **, unsigned *, const void *);
static int line_deck(void *, struct fpnode_type **, struct fpnode_type *,
		     unsigned, unsigned, unsigned *, const byte *, unsigned,
		     const void *);
static int mind_deck(void *, struct fpnode_type *, const byte *, unsigned,
		     const void *);
static int none_deck(void *, struct fpnode_type *, byte *);
static int post_deck(void *, struct fpnode_type **, unsigned *, const byte *,
		     unsigned, const void *);
static int post_line(void *, const byte *, unsigned, const void *);
static int post_name(void *, const byte *, unsigned, const void *);
static int post_pair(void *, const byte *, unsigned, const void *);
static int post_slip(void *, const byte *, unsigned, const void *);

static int
case_deck(void *scdeck, struct fpnode_type *fpnode_slip, const byte *slip,
	  unsigned length, const byte *name, unsigned Size, const void *data)
{
    int status;
    void *node;

    status = scdeck(scdeck)->link_m.link
	(scdeck(scdeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	int c, e;
	unsigned bits, prefix;

	e = *slip;

	prefix = slip - fpnode_slip->base;

	bits = fpnode_slip->bits;

	if (Size == 1) {
	    unsigned suffix;

	    c = Size ? *name : 0;

	    suffix = length;
	    if (suffix == 1) {
		struct fpnode_type *fpnode_data;

		fpnode_data = node;

		if (c < e + 1) {
		    fpnode_data->base = null;
		    fpnode_data->bits = dset(c, 0);
		    fpnode_data->data = data;

		    fpnode_data++;

		    fpnode_data->base = null;
		    fpnode_data->bits = pset(e, 0, bits);
		    fpnode_data->data = fpnode_slip->data;
		} else {
		    fpnode_data->base = null;
		    fpnode_data->bits = pset(e, 0, bits);
		    fpnode_data->data = fpnode_slip->data;

		    fpnode_data++;

		    fpnode_data->base = null;
		    fpnode_data->bits = dset(c, 0);
		    fpnode_data->data = data;
		}

		fpnode_slip->data = node;
	    } else {
		void *flat;

		status = scdeck(scdeck)->link_m.link
		    (scdeck(scdeck)->link_m.data, &flat, suffix);
		if (status) {
		    status = LINK_ERROR;

		    scdeck(scdeck)->link_m.free
			(scdeck(scdeck)->link_m.data, node);
		} else {
		    struct fpnode_type *fpnode_data;

		    fpnode_data = node;

		    memcpy(flat, slip + 1, suffix);

		    suffix--;

		    if (c < e + 1) {
			fpnode_data->base = null;
			fpnode_data->bits = dset(c, 0);
			fpnode_data->data = data;

			fpnode_data++;

			fpnode_data->base = flat;
			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;
		    } else {
			fpnode_data->base = flat;
			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;

			fpnode_data++;

			fpnode_data->base = null;
			fpnode_data->bits = dset(c, 0);
			fpnode_data->data = data;
		    }

		    fpnode_slip->data = node;
		}
	    }
	} else {
	    void *sand;

	    status = scdeck(scdeck)->link_m.link
		(scdeck(scdeck)->link_m.data, &sand, Size);
	    if (status) {
		status = LINK_ERROR;

		scdeck(scdeck)->link_m.free(scdeck(scdeck)->link_m.data, node);
	    } else {
		unsigned suffix;

		c = *name;

		suffix = length;
		if (suffix == 1) {
		    struct fpnode_type *fpnode_data;

		    fpnode_data = node;

		    memcpy(sand, name + 1, Size);

		    Size--;

		    if (c < e + 1) {
			fpnode_data->base = sand;
			fpnode_data->bits = dset(c, Size);
			fpnode_data->data = data;

			fpnode_data++;

			fpnode_data->base = null;
			fpnode_data->bits = pset(e, 0, bits);
			fpnode_data->data = fpnode_slip->data;
		    } else {
			fpnode_data->base = null;
			fpnode_data->bits = pset(e, 0, bits);
			fpnode_data->data = fpnode_slip->data;

			fpnode_data++;

			fpnode_data->base = sand;
			fpnode_data->bits = dset(c, Size);
			fpnode_data->data = data;
		    }

		    fpnode_slip->data = node;
		} else {
		    void *flat;

		    status = scdeck(scdeck)->link_m.link
			(scdeck(scdeck)->link_m.data, &flat, suffix);
		    if (status) {
			status = LINK_ERROR;

			scdeck(scdeck)->link_m.free
			    (scdeck(scdeck)->link_m.data, sand);
			scdeck(scdeck)->link_m.free
			    (scdeck(scdeck)->link_m.data, node);
		    } else {
			struct fpnode_type *fpnode_data;

			fpnode_data = node;

			memcpy(sand, name + 1, Size);

			Size--;

			memcpy(flat, slip + 1, suffix);

			suffix--;

			if (c < e + 1) {
			    fpnode_data->base = sand;
			    fpnode_data->bits = dset(c, Size);
			    fpnode_data->data = data;

			    fpnode_data++;

			    fpnode_data->base = flat;
			    fpnode_data->bits = pset(e, suffix, bits);
			    fpnode_data->data = fpnode_slip->data;
			} else {
			    fpnode_data->base = flat;
			    fpnode_data->bits = pset(e, suffix, bits);
			    fpnode_data->data = fpnode_slip->data;

			    fpnode_data++;

			    fpnode_data->base = sand;
			    fpnode_data->bits = dset(c, Size);
			    fpnode_data->data = data;
			}

			fpnode_slip->data = node;
		    }
		}
	    }
	}
    }

    return status;
}


static int
ever_deck(void *scdeck, struct fpnode_type *fpnode_slip, const byte *name,
	  unsigned Size, const void *data)
{
    int status;
    void *node;

    status = scdeck(scdeck)->link_m.link
	(scdeck(scdeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	if (Size == 1) {
	    struct fpnode_type *fpnode_data;

	    fpnode_data = node;

	    fpnode_data->base = null;
	    fpnode_data->bits = dset(0, 0);
	    fpnode_data->data = fpnode_slip->data;

	    fpnode_data++;

	    fpnode_data->base = null;
	    fpnode_data->bits = dset(*name, 0);
	    fpnode_data->data = data;

	    fpnode_slip->bits = jset(fpnode_slip->bits, 2);
	    fpnode_slip->data = node;
	} else {
	    void *sand;

	    status = scdeck(scdeck)->link_m.link
		(scdeck(scdeck)->link_m.data, &sand, Size);
	    if (status) {
		status = LINK_ERROR;

		scdeck(scdeck)->link_m.free(scdeck(scdeck)->link_m.data, node);
	    } else {
		struct fpnode_type *fpnode_data;

		fpnode_data = node;

		memcpy(sand, name + 1, Size);

		Size--;

		fpnode_data->base = null;
		fpnode_data->bits = dset(0, 0);
		fpnode_data->data = fpnode_slip->data;

		fpnode_data++;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, Size);
		fpnode_data->data = data;

		fpnode_slip->bits = jset(fpnode_slip->bits, 2);
		fpnode_slip->data = node;
	    }
	}
    }

    return status;
}


static int
fast_deck(void *scdeck, struct fpnode_type **fpnode, unsigned *size,
	  const void *data)
{
    int status;
    struct fpnode_type *fpnode_slip;

    fpnode_slip = *fpnode;

    if (prefix(fpnode_slip)) {
	do {
	    unsigned fast;

	    fast = *size;

	    if (fast & 3) {
		status = 0;
	    } else {
		void *node;

		node = fpnode_slip;

		status = scdeck(scdeck)->link_m.mode
		    (scdeck(scdeck)->link_m.data, &node,
		     (fast + 4) * sizeof(struct fpnode_type));
		if (status) {
		    status = MODE_ERROR;
		    if (1) {
			break;
		    }
		} else {
		    *fpnode = node;

		    fpnode_slip = node;
		}
	    }

	    memmove(fpnode_slip + 1, fpnode_slip,
		    fast * sizeof(struct fpnode_type));

	    fpnode_slip->base = null;
	    fpnode_slip->bits = dset(0, 0);
	    fpnode_slip->data = data;

	    *size = fast + 1;
	} while (0);
    } else {
	status = EVER_MATCH;
    }

    return status;
}


static int
line_deck(void *scdeck, struct fpnode_type **fpnode,
	  struct fpnode_type *fpnode_data, unsigned line, unsigned fast,
	  unsigned *size, const byte *name, unsigned Size, const void *data)
{
    int status;

    do {
	if (fast & 3) {
	} else {
	    void *node;

	    node = fpnode_data;

	    status = scdeck(scdeck)->link_m.mode
		(scdeck(scdeck)->link_m.data, &node,
		 (fast + 4) * sizeof(struct fpnode_type));
	    if (status) {
		status = MODE_ERROR;
		if (1) {
		    break;
		}
	    } else {
		*fpnode = node;

		fpnode_data = node;
	    }
	}

	if (Size == 1) {
	    fpnode_data += line;

	    memmove(fpnode_data + 1, fpnode_data,
		    (fast - line) * sizeof(struct fpnode_type));

	    fpnode_data->base = null;
	    fpnode_data->bits = dset(*name, 0);
	    fpnode_data->data = data;

	    *size = fast + 1;

	    status = 0;
	} else {
	    void *sand;

	    status = scdeck(scdeck)->link_m.link
		(scdeck(scdeck)->link_m.data, &sand, Size);
	    if (status) {
		status = LINK_ERROR;
	    } else {
		fpnode_data += line;

		memmove(fpnode_data + 1, fpnode_data,
			(fast - line) * sizeof(struct fpnode_type));

		memcpy(sand, name + 1, Size);

		Size--;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, Size);
		fpnode_data->data = data;

		*size = fast + 1;
	    }
	}
    } while (0);

    return status;
}


static int
mind_deck(void *scdeck, struct fpnode_type *fpnode_data, const byte *name,
	  unsigned Size, const void *data)
{
    int status;
    struct fpnode_type *fpnode_slip;
    unsigned bits, size;

    bits = fpnode_data->bits;
    fpnode_slip = (struct fpnode_type *) fpnode_data->data;

    size = Spread(bits);

    status = post_deck(scdeck, &fpnode_slip, &size, name, Size, data);
    if (status) {
    } else {
	fpnode_data->bits = iset(bits, size);
	fpnode_data->data = fpnode_slip;
    }

    return status;
}


static int
none_deck(void *scdeck, struct fpnode_type *fpnode_slip, byte *slip)
{
    int status;
    unsigned prefix;
    void *miss;

    prefix = slip - fpnode_slip->base;

    fpnode_slip->bits = nset(prefix, fpnode_slip->bits, 2);

    miss = fpnode_slip->base;

    if (prefix) {
	*slip = 0;

	status = scdeck(scdeck)->link_m.mode
	    (scdeck(scdeck)->link_m.data, &miss, prefix + 1);
	if (status) {
	    status = MODE_ERROR;
	} else {
	    fpnode_slip->base = miss;
	}
    } else {
	status = scdeck(scdeck)->link_m.free
	    (scdeck(scdeck)->link_m.data, miss);
	if (status) {
	    status = FREE_ERROR;
	} else {
	    fpnode_slip->base = null;
	}
    }

    return status;
}


static int
post_deck(void *scdeck, struct fpnode_type **fpnode, unsigned *size,
	  const byte *name, unsigned Size, const void *data)
{
    int status;

    if (Size) {
	int c;
	struct fpnode_type *fpnode_data, *fpnode_slip;
	unsigned bits, mind;

	c = *name;

	mind = *size;

	fpnode_slip = *fpnode;

	fpnode_data = fpnode_slip;

	Progress_node(fpnode_data, mind - 1);

	bits = fpnode_data->bits;

	if (Prefix(bits) == c) {
	    byte *slip;
	    int e;
	    unsigned length;

	    Size--;
	    name++;

	    slip = fpnode_data->base;

	    length = length(fpnode_data);
	    if (length < Record) {
	    } else {
		length += strlen(back(slip) + Record);
	    }

	    if (0) {
	    } else {
		while (1) {
		    if (length) {
			e = *slip;
			if (Size) {
			    c = *name;
			    if (c == e) {
				slip++;
				name++;

				Size--;
				length--;
			    } else {
				break;
			    }
			} else {
			    break;
			}
		    } else {
			break;
		    }
		}
	    }
	    if (length) {
		status = case_deck
		    (scdeck, fpnode_data, slip, length, name, Size, data);
		if (status) {
		} else {
		    status = none_deck(scdeck, fpnode_data, slip);
		}
	    } else {
		if (Extant(bits)) {
		    status = mind_deck(scdeck, fpnode_data, name, Size, data);
		} else {
		    if (Size) {
			status = ever_deck
			    (scdeck, fpnode_data, name, Size, data);
		    } else {
			status = EVER_MATCH;
		    }
		}
	    }
	} else {
	    unsigned line;

	    line = fpnode_data - fpnode_slip;

	    if (c < Prefix(bits)) {
	    } else {
		line++;
	    }

	    status = line_deck
		(scdeck, fpnode, fpnode_slip, line, mind, size, name, Size,
		 data);
	}
    } else {
	status = fast_deck(scdeck, fpnode, size, data);
    }

    return status;
}


static int
post_line(void *scdeck, const byte *name, unsigned size, const void *data)
{
    int status;
    void *node;

    status = scdeck(scdeck)->link_m.link
	(scdeck(scdeck)->link_m.data, &node, sizeof(struct fpnode_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	if (size == 1) {
	    struct fpnode_type *fpnode_data;

	    scdeck(scdeck)->link_a.fpdeck.fpnode_data = node;

	    fpnode_data = node;

	    fpnode_data->base = null;
	    fpnode_data->bits = dset(*name, 0);
	    fpnode_data->data = data;
	} else {
	    void *sand;

	    status = scdeck(scdeck)->link_m.link
		(scdeck(scdeck)->link_m.data, &sand, size);
	    if (status) {
		status = LINK_ERROR;

		scdeck(scdeck)->link_m.free(scdeck(scdeck)->link_m.data, node);
	    } else {
		struct fpnode_type *fpnode_data;

		scdeck(scdeck)->link_a.fpdeck.fpnode_data = node;

		fpnode_data = node;

		memcpy(sand, name + 1, size);

		size--;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, size);
		fpnode_data->data = data;
	    }
	}
    }

    return status;
}


static int
post_name(void *scdeck, const byte *name, unsigned Size, const void *data)
{
    int status;
    struct fpnode_type *fpnode_data, *fpnode_slip;
    unsigned size;

    fpnode_data = scdeck(scdeck)->link_a.fpdeck.fpnode_data;

    size = spread(fpnode_data);
    fpnode_slip = (struct fpnode_type *) fpnode_data->data;

    status = post_deck(scdeck, &fpnode_slip, &size, name, Size, data);
    if (status) {
    } else {
	fpnode_data->bits = eset(size);
	fpnode_data->data = fpnode_slip;
    }

    return status;
}


static int
post_pair(void *scdeck, const byte *name, unsigned Size, const void *data)
{
    int status;
    void *node;

    status = scdeck(scdeck)->link_m.link
	(scdeck(scdeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	struct fpnode_type *fpnode_data, *fpnode_slip;
	unsigned size = 1;

	fpnode_data = scdeck(scdeck)->link_a.fpdeck.fpnode_data;

	fpnode_slip = node;

	*fpnode_slip = *fpnode_data;

	fpnode_data->base = null;
	fpnode_data->bits = eset(1);
	fpnode_data->data = fpnode_slip;

	status = post_deck(scdeck, &fpnode_slip, &size, name, Size, data);
	if (status) {
	} else {
	    fpnode_data->bits = eset(size);
	    fpnode_data->data = fpnode_slip;
	}
    }

    return status;
}


static int
post_slip(void *scdeck, const byte *name, unsigned Size, const void *data)
{
    int status;
    struct fpnode_type *fpnode_data;
    unsigned size = 1;

    fpnode_data = scdeck(scdeck)->link_a.fpdeck.fpnode_data;

    status = post_deck(scdeck, &fpnode_data, &size, name, Size, data);
    if (status) {
    } else {
	if (0) {
	    scdeck(scdeck)->link_a.fpdeck.fpnode_data = fpnode_data;
	}
    }

    return status;
}


int
x1f4_post_scpath(void *scdeck, const char *name, unsigned size,
		 const void *data)
{
    int status;
    struct fpnode_type *fpnode_data;

    fpnode_data = scdeck(scdeck)->link_a.fpdeck.fpnode_data;
    if (fpnode_data) {
	int prefix;

	prefix = prefix(fpnode_data);
	if (prefix) {
	    if (prefix == *name) {
		status = post_slip(scdeck, fast(name), size, data);
	    } else {
		status = post_pair(scdeck, fast(name), size, data);
	    }
	} else {
	    status = post_name(scdeck, fast(name), size, data);
	}
    } else {
	status = post_line(scdeck, fast(name), size, data);
    }

    return status;
}
