/*
 * scxx.e.c
 * Copyright (C) 2009-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include <scxx-defs.h>
#include <scxx-names.h>
#include <scxx-patch.h>
#include <scxx-types.h>

#define back(miss)			((const char *) (miss))
#define fast(miss)			((const byte *) (miss))

#define null				(&scdeck(scdeck)->link_i.zero)

#define scdeck(scdeck)			((struct scdeck_type *) (scdeck))

static int case_deck(void *, struct fpnode_type *, const byte *, int, int,
		     const byte *, const void ***);
static int ever_deck(void *, struct fpnode_type *, const byte *,
		     const void ***);
static int fast_deck(void *, struct fpnode_type **, unsigned *,
		     const void ***);
static int line_deck(void *, struct fpnode_type **, struct fpnode_type *,
		     unsigned, unsigned, unsigned *, const byte *,
		     const void ***);
static int mind_deck(void *, struct fpnode_type *, const byte *,
		     const void ***);
static int none_deck(void *, struct fpnode_type *, byte *);
static int post_deck(void *, struct fpnode_type **, unsigned *, const byte *,
		     const void ***);
static int post_line(void *, const byte *, const void ***);
static int post_name(void *, const byte *, const void ***);
static int post_pair(void *, const byte *, const void ***);
static int post_slip(void *, const byte *, const void ***);

static int
case_deck(void *scdeck, struct fpnode_type *fpnode_slip, const byte *slip,
	  int e, int c, const byte *name, const void ***ware)
{
    int status;
    void *node;

    status = scdeck(scdeck)->link_m.link
	(scdeck(scdeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	unsigned bits, prefix, length;

	prefix = slip - fpnode_slip->base;

	bits = fpnode_slip->bits;

	length = strlen(back(name));
	if (length < 2) {
	    unsigned suffix;

	    suffix = strlen(back(slip));
	    if (suffix == 1) {
		struct fpnode_type *fpnode_data;

		fpnode_data = node;

		if (c < e + 1) {
		    fpnode_data->base = null;
		    fpnode_data->bits = dset(c, 0);

		    *ware = &fpnode_data->data;

		    fpnode_data++;

		    fpnode_data->base = null;
		    fpnode_data->bits = pset(e, 0, bits);
		    fpnode_data->data = fpnode_slip->data;
		} else {
		    fpnode_data->base = null;
		    fpnode_data->bits = pset(e, 0, bits);
		    fpnode_data->data = fpnode_slip->data;

		    fpnode_data++;

		    fpnode_data->base = null;
		    fpnode_data->bits = dset(c, 0);

		    *ware = &fpnode_data->data;
		}

		fpnode_slip->data = node;
	    } else {
		void *flat;

		status = scdeck(scdeck)->link_m.link
		    (scdeck(scdeck)->link_m.data, &flat, suffix);
		if (status) {
		    status = LINK_ERROR;

		    scdeck(scdeck)->link_m.free
			(scdeck(scdeck)->link_m.data, node);
		} else {
		    struct fpnode_type *fpnode_data;

		    fpnode_data = node;

		    memcpy(flat, slip + 1, suffix);

		    suffix--;

		    if (c < e + 1) {
			fpnode_data->base = null;
			fpnode_data->bits = dset(c, 0);

			*ware = &fpnode_data->data;

			fpnode_data++;

			fpnode_data->base = flat;
			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;
		    } else {
			fpnode_data->base = flat;
			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;

			fpnode_data++;

			fpnode_data->base = null;
			fpnode_data->bits = dset(c, 0);

			*ware = &fpnode_data->data;
		    }

		    fpnode_slip->data = node;
		}
	    }
	} else {
	    void *sand;

	    status = scdeck(scdeck)->link_m.link
		(scdeck(scdeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;

		scdeck(scdeck)->link_m.free(scdeck(scdeck)->link_m.data, node);
	    } else {
		unsigned suffix;

		suffix = strlen(back(slip));
		if (suffix == 1) {
		    struct fpnode_type *fpnode_data;

		    fpnode_data = node;

		    memcpy(sand, name + 1, length);

		    length--;

		    if (c < e + 1) {
			fpnode_data->base = sand;
			fpnode_data->bits = dset(c, length);

			*ware = &fpnode_data->data;

			fpnode_data++;

			fpnode_data->base = null;
			fpnode_data->bits = pset(e, 0, bits);
			fpnode_data->data = fpnode_slip->data;
		    } else {
			fpnode_data->base = null;
			fpnode_data->bits = pset(e, 0, bits);
			fpnode_data->data = fpnode_slip->data;

			fpnode_data++;

			fpnode_data->base = sand;
			fpnode_data->bits = dset(c, length);

			*ware = &fpnode_data->data;
		    }

		    fpnode_slip->data = node;
		} else {
		    void *flat;

		    status = scdeck(scdeck)->link_m.link
			(scdeck(scdeck)->link_m.data, &flat, suffix);
		    if (status) {
			status = LINK_ERROR;

			scdeck(scdeck)->link_m.free
			    (scdeck(scdeck)->link_m.data, sand);
			scdeck(scdeck)->link_m.free
			    (scdeck(scdeck)->link_m.data, node);
		    } else {
			struct fpnode_type *fpnode_data;

			fpnode_data = node;

			memcpy(sand, name + 1, length);

			length--;

			memcpy(flat, slip + 1, suffix);

			suffix--;

			if (c < e + 1) {
			    fpnode_data->base = sand;
			    fpnode_data->bits = dset(c, length);

			    *ware = &fpnode_data->data;

			    fpnode_data++;

			    fpnode_data->base = flat;
			    fpnode_data->bits = pset(e, suffix, bits);
			    fpnode_data->data = fpnode_slip->data;
			} else {
			    fpnode_data->base = flat;
			    fpnode_data->bits = pset(e, suffix, bits);
			    fpnode_data->data = fpnode_slip->data;

			    fpnode_data++;

			    fpnode_data->base = sand;
			    fpnode_data->bits = dset(c, length);

			    *ware = &fpnode_data->data;
			}

			fpnode_slip->data = node;
		    }
		}
	    }
	}
    }

    return status;
}


static int
ever_deck(void *scdeck, struct fpnode_type *fpnode_slip, const byte *name,
	  const void ***ware)
{
    int status;
    void *node;

    status = scdeck(scdeck)->link_m.link
	(scdeck(scdeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	unsigned length;

	length = strlen(back(name));
	if (length == 1) {
	    struct fpnode_type *fpnode_data;

	    fpnode_data = node;

	    fpnode_data->base = null;
	    fpnode_data->bits = dset(0, 0);
	    fpnode_data->data = fpnode_slip->data;

	    fpnode_data++;

	    fpnode_data->base = null;
	    fpnode_data->bits = dset(*name, 0);

	    *ware = &fpnode_data->data;

	    fpnode_slip->bits = jset(fpnode_slip->bits, 2);
	    fpnode_slip->data = node;
	} else {
	    void *sand;

	    status = scdeck(scdeck)->link_m.link
		(scdeck(scdeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;

		scdeck(scdeck)->link_m.free(scdeck(scdeck)->link_m.data, node);
	    } else {
		struct fpnode_type *fpnode_data;

		fpnode_data = node;

		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->base = null;
		fpnode_data->bits = dset(0, 0);
		fpnode_data->data = fpnode_slip->data;

		fpnode_data++;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, length);

		*ware = &fpnode_data->data;

		fpnode_slip->bits = jset(fpnode_slip->bits, 2);
		fpnode_slip->data = node;
	    }
	}
    }

    return status;
}


static int
fast_deck(void *scdeck, struct fpnode_type **fpnode, unsigned *size,
	  const void ***ware)
{
    int status;
    struct fpnode_type *fpnode_slip;

    fpnode_slip = *fpnode;

    if (prefix(fpnode_slip)) {
	do {
	    unsigned fast;

	    fast = *size;

	    if (fast & 3) {
		status = 0;
	    } else {
		void *node;

		node = fpnode_slip;

		status = scdeck(scdeck)->link_m.mode
		    (scdeck(scdeck)->link_m.data, &node,
		     (fast + 4) * sizeof(struct fpnode_type));
		if (status) {
		    status = MODE_ERROR;
		    if (1) {
			break;
		    }
		} else {
		    *fpnode = node;

		    fpnode_slip = node;
		}
	    }

	    memmove(fpnode_slip + 1, fpnode_slip,
		    fast * sizeof(struct fpnode_type));

	    fpnode_slip->base = null;
	    fpnode_slip->bits = dset(0, 0);

	    *ware = &fpnode_slip->data;

	    *size = fast + 1;
	} while (0);
    } else {
	status = EVER_MATCH;

	*ware = &fpnode_slip->data;
    }

    return status;
}


static int
line_deck(void *scdeck, struct fpnode_type **fpnode,
	  struct fpnode_type *fpnode_data, unsigned line, unsigned fast,
	  unsigned *size, const byte *name, const void ***ware)
{
    int status;

    do {
	unsigned length;

	if (fast & 3) {
	} else {
	    void *node;

	    node = fpnode_data;

	    status = scdeck(scdeck)->link_m.mode
		(scdeck(scdeck)->link_m.data, &node,
		 (fast + 4) * sizeof(struct fpnode_type));
	    if (status) {
		status = MODE_ERROR;
		if (1) {
		    break;
		}
	    } else {
		*fpnode = node;

		fpnode_data = node;
	    }
	}

	length = strlen(back(name));
	if (length == 1) {
	    fpnode_data += line;

	    memmove(fpnode_data + 1, fpnode_data,
		    (fast - line) * sizeof(struct fpnode_type));

	    fpnode_data->base = null;
	    fpnode_data->bits = dset(*name, 0);

	    *ware = &fpnode_data->data;

	    *size = fast + 1;

	    status = 0;
	} else {
	    void *sand;

	    status = scdeck(scdeck)->link_m.link
		(scdeck(scdeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;
	    } else {
		fpnode_data += line;

		memmove(fpnode_data + 1, fpnode_data,
			(fast - line) * sizeof(struct fpnode_type));

		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, length);

		*ware = &fpnode_data->data;

		*size = fast + 1;
	    }
	}
    } while (0);

    return status;
}


static int
mind_deck(void *scdeck, struct fpnode_type *fpnode_data, const byte *name,
	  const void ***ware)
{
    int status;
    struct fpnode_type *fpnode_slip;
    unsigned bits, size;

    bits = fpnode_data->bits;
    fpnode_slip = (struct fpnode_type *) fpnode_data->data;

    size = Spread(bits);

    status = post_deck(scdeck, &fpnode_slip, &size, name, ware);
    if (status) {
    } else {
	fpnode_data->bits = iset(bits, size);
	fpnode_data->data = fpnode_slip;
    }

    return status;
}


static int
none_deck(void *scdeck, struct fpnode_type *fpnode_slip, byte *slip)
{
    int status;
    unsigned prefix;
    void *miss;

    prefix = slip - fpnode_slip->base;

    *slip = 0;

    fpnode_slip->bits = nset(prefix, fpnode_slip->bits, 2);

    miss = fpnode_slip->base;

    if (prefix) {
	status = scdeck(scdeck)->link_m.mode
	    (scdeck(scdeck)->link_m.data, &miss, prefix + 1);
	if (status) {
	    status = MODE_ERROR;
	} else {
	    fpnode_slip->base = miss;
	}
    } else {
	status = scdeck(scdeck)->link_m.free
	    (scdeck(scdeck)->link_m.data, miss);
	if (status) {
	    status = FREE_ERROR;
	} else {
	    fpnode_slip->base = null;
	}
    }

    return status;
}


static int
post_deck(void *scdeck, struct fpnode_type **fpnode, unsigned *size,
	  const byte *name, const void ***ware)
{
    int c, status;

    c = *name;
    if (c) {
	struct fpnode_type *fpnode_data, *fpnode_slip;
	unsigned mind;

	mind = *size;

	fpnode_slip = *fpnode;

	fpnode_data = fpnode_slip;

	Progress_node(fpnode_data, mind - 1);

	if (prefix(fpnode_data) == c) {
	    byte *slip;
	    int e;

	    name++;

	    slip = fpnode_data->base;
	    if (1) {
		while (1) {
		    c = *name;
		    e = *slip;
		    if (c == e) {
			if (c) {
			    slip++;
			    name++;
			} else {
			    break;
			}
		    } else {
			break;
		    }
		}
	    }
	    if (e) {
		status = case_deck
		    (scdeck, fpnode_data, slip, e, c, name, ware);
		if (status) {
		} else {
		    status = none_deck(scdeck, fpnode_data, slip);
		}
	    } else {
		if (extant(fpnode_data)) {
		    status = mind_deck(scdeck, fpnode_data, name, ware);
		} else {
		    if (c) {
			status = ever_deck(scdeck, fpnode_data, name, ware);
		    } else {
			status = EVER_MATCH;

			*ware = &fpnode_data->data;
		    }
		}
	    }
	} else {
	    unsigned line;

	    line = fpnode_data - fpnode_slip;

	    if (c < prefix(fpnode_data)) {
	    } else {
		line++;
	    }

	    status = line_deck
		(scdeck, fpnode, fpnode_slip, line, mind, size, name, ware);
	}
    } else {
	status = fast_deck(scdeck, fpnode, size, ware);
    }

    return status;
}


static int
post_line(void *scdeck, const byte *name, const void ***ware)
{
    int status;
    void *node;

    status = scdeck(scdeck)->link_m.link
	(scdeck(scdeck)->link_m.data, &node, sizeof(struct fpnode_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	unsigned length;

	length = strlen(back(name));
	if (length < 2) {
	    struct fpnode_type *fpnode_data;

	    scdeck(scdeck)->link_a.fpdeck.fpnode_data = node;

	    fpnode_data = node;

	    fpnode_data->base = null;
	    fpnode_data->bits = dset(*name, 0);

	    *ware = &fpnode_data->data;
	} else {
	    void *sand;

	    status = scdeck(scdeck)->link_m.link
		(scdeck(scdeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;

		scdeck(scdeck)->link_m.free(scdeck(scdeck)->link_m.data, node);
	    } else {
		struct fpnode_type *fpnode_data;

		scdeck(scdeck)->link_a.fpdeck.fpnode_data = node;

		fpnode_data = node;

		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, length);

		*ware = &fpnode_data->data;
	    }
	}
    }

    return status;
}


static int
post_name(void *scdeck, const byte *name, const void ***ware)
{
    int status;
    struct fpnode_type *fpnode_data, *fpnode_slip;
    unsigned size;

    fpnode_data = scdeck(scdeck)->link_a.fpdeck.fpnode_data;

    size = spread(fpnode_data);
    fpnode_slip = (struct fpnode_type *) fpnode_data->data;

    status = post_deck(scdeck, &fpnode_slip, &size, name, ware);
    if (status) {
    } else {
	fpnode_data->bits = eset(size);
	fpnode_data->data = fpnode_slip;
    }

    return status;
}


static int
post_pair(void *scdeck, const byte *name, const void ***ware)
{
    int status;
    void *node;

    status = scdeck(scdeck)->link_m.link
	(scdeck(scdeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	struct fpnode_type *fpnode_data, *fpnode_slip;
	unsigned size = 1;

	fpnode_data = scdeck(scdeck)->link_a.fpdeck.fpnode_data;

	fpnode_slip = node;

	*fpnode_slip = *fpnode_data;

	fpnode_data->base = null;
	fpnode_data->bits = eset(1);
	fpnode_data->data = fpnode_slip;

	status = post_deck(scdeck, &fpnode_slip, &size, name, ware);
	if (status) {
	} else {
	    fpnode_data->bits = eset(size);
	    fpnode_data->data = fpnode_slip;
	}
    }

    return status;
}


static int
post_slip(void *scdeck, const byte *name, const void ***ware)
{
    int status;
    struct fpnode_type *fpnode_data;
    unsigned size = 1;

    fpnode_data = scdeck(scdeck)->link_a.fpdeck.fpnode_data;

    status = post_deck(scdeck, &fpnode_data, &size, name, ware);
    if (status) {
    } else {
	if (0) {
	    scdeck(scdeck)->link_a.fpdeck.fpnode_data = fpnode_data;
	}
    }

    return status;
}


int
x1f4_bind_scdeck(void *scdeck, const char *name, const void ***ware)
{
    int status;
    struct fpnode_type *fpnode_data;

    fpnode_data = scdeck(scdeck)->link_a.fpdeck.fpnode_data;
    if (fpnode_data) {
	int prefix;

	prefix = prefix(fpnode_data);
	if (prefix) {
	    if (prefix == *name) {
		status = post_slip(scdeck, fast(name), ware);
	    } else {
		status = post_pair(scdeck, fast(name), ware);
	    }
	} else {
	    if (extant(fpnode_data)) {
		status = post_name(scdeck, fast(name), ware);
	    } else {
		if (*name) {
		    status = post_pair(scdeck, fast(name), ware);
		} else {
		    status = post_slip(scdeck, fast(name), ware);
		}
	    }
	}
    } else {
	status = post_line(scdeck, fast(name), ware);
    }

    return status;
}
