/*
 * slxi.w.c
 * Copyright (C) 2009, 2010, 2011, 2012, 2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include <slxx-defs.h>
#include <slxx-inter.h>
#include <slxx-names.h>
#include <slxx-types.h>

#define false(e)			0
#define true(e)				1

#define back(miss)			((const char *) (miss))

#define sldeck(sldeck)			((struct sldeck_type *) (sldeck))

static int land_deck(struct fpnode_type *, unsigned, unsigned, unsigned *);
static int miss_head(void *, struct fpnode_type *, unsigned);
static int miss_tail(void *, struct fpnode_type *, unsigned);
static int miss_zero(void *, struct fpnode_type *);
static int pass_link(void *, struct fpnode_type *, struct fpnode_type *);
static int pass_name(void *, struct fpnode_type *, struct fpnode_type *);

static int
land_deck(struct fpnode_type *fpnode_data, unsigned extent, unsigned miss,
	  unsigned *f)
{
    unsigned left, p, slip;

    slip = *f;

    left = 0;
    extent >>= 4;
    for (; extent; extent--) {
	p = prefix(fpnode_data);
	slip += p;
	left += p * LINEAR_d_SCALE;
	if (left < miss) {
	} else {
	    break;
	}

	fpnode_data += 16;
    }

    *f = slip;

    return extent;
}


static int
miss_head(void *sldeck, struct fpnode_type *fpnode_call, unsigned slip)
{
    int status;
    struct fpnode_type *fpnode_slip;
    unsigned bits, call, extent, zero;

    bits = fpnode_call->bits;

    extent = Extent(bits);
    call = extent - slip;

    fpnode_slip = (void *) fpnode_call->data;

    if (Target(bits)) {
	fpnode_slip++;
	zero = 1;
    } else {
	zero = 0;
    }

    memmove
	(fpnode_slip, fpnode_slip + slip, call * sizeof(struct fpnode_type));

    fpnode_call->bits = fset(bits, (Offset(bits) + slip) | (extent - 16) >> 4);
    if (zero) {
	fpnode_call->bits = zset(fpnode_call->bits);
    }

    status = miss_tail(sldeck, fpnode_call, slip);

    return status;
}


static int
miss_tail(void *sldeck, struct fpnode_type *fpnode_call, unsigned slip)
{
    int status;
    unsigned bits, extent, zero;
    void *node;

    bits = fpnode_call->bits;

    extent = Extent(bits) - slip;

    if (Target(bits)) {
	zero = 1;
    } else {
	zero = 0;
    }

    node = (void *) fpnode_call->data;
    status = sldeck(sldeck)->link_m.mode
	(sldeck(sldeck)->link_m.data, &node,
	 (extent + zero) * sizeof(struct fpnode_type));
    if (status) {
	status = MODE_ERROR;
    } else {
	fpnode_call->data = node;

	fpnode_call->bits = fset(bits, Offset(bits) | (extent - 16) >> 4);
	if (zero) {
	    fpnode_call->bits = zset(fpnode_call->bits);
	}
    }

    return status;
}


static int
miss_zero(void *sldeck, struct fpnode_type *fpnode_call)
{
    int status;
    struct fpnode_type *fpnode_slip;
    unsigned bits, extent;
    void *node;

    bits = fpnode_call->bits;

    extent = Extent(bits);

    fpnode_slip = (void *) fpnode_call->data;
    memmove(fpnode_slip, fpnode_slip + 1, extent * sizeof(struct fpnode_type));

    fpnode_call->bits = wset(bits);

    node = fpnode_slip;
    status = sldeck(sldeck)->link_m.mode
	(sldeck(sldeck)->link_m.data, &node,
	 extent * sizeof(struct fpnode_type));
    if (status) {
	status = MODE_ERROR;
    } else {
	fpnode_call->data = node;
    }

    return status;
}


static int
pass_link(void *sldeck, struct fpnode_type *fpnode_slip,
	  struct fpnode_type *fpnode_data)
{
    byte *base, *lock;
    int prefix, status;
    unsigned bits, length, miss, offset;

    bits = fpnode_slip->bits;

    base = fpnode_slip->base;

    length = Length(bits);
    if (length < Record) {
    } else {
	length += strlen(back(base) + Record);
    }

    if (fpnode_data == fpnode_slip->data) {
	fpnode_data++;
    } else {
	fpnode_data--;
    }

    prefix = Prefix(bits);
    if (fpnode_slip == &sldeck(sldeck)->link_a.fpdeck.fpnode) {
	prefix = prefix(fpnode_slip);
    } else {
	prefix = 1;
    }

    miss = fpnode_data->bits;

    lock = fpnode_data->base;

    if (prefix) {
	if (Prefix(miss)) {
	    offset = Length(miss);
	    if (offset < Record) {
		if (sizeof(byte *) < offset) {
		} else {
		    lock = (byte *) &fpnode_data->base;
		}
	    } else {
		offset += strlen(back(lock) + Record);
	    }

	    offset++;
	} else {
	    offset = 0;
	}
    } else {
	offset = 0;
    }

    if (offset) {
	do {
	    unsigned mirror;
	    void *copy;

	    mirror = length + offset;

	    if (sizeof(byte *) < mirror) {
		if (sizeof(byte *) < length) {
		    copy = base;
		    status = sldeck(sldeck)->link_m.mode
			(sldeck(sldeck)->link_m.data, &copy, mirror + 1);
		    if (status) {
			status = MODE_ERROR;
			if (1) {
			    break;
			}
		    }
		} else {
		    status = sldeck(sldeck)->link_m.link
			(sldeck(sldeck)->link_m.data, &copy, mirror + 1);
		    if (status) {
			status = LINK_ERROR;
			if (1) {
			    break;
			}
		    }

		    memcpy(copy, &fpnode_slip->base, length);
		}

		fpnode_slip->base = copy;
	    } else {
		copy = &fpnode_slip->base;
	    }
	    if (0) {
	    } else {
		byte *deck;
		void *slip;

		deck = copy;
		deck += length;
		*deck = Prefix(miss);
		deck++;
		offset--;
		if (offset) {
		    memcpy(deck, lock, offset);
		    if (1) {
			deck += offset;
		    }
		}

		if (sizeof(byte *) < mirror) {
		    *deck = 0;
		}

		slip = (void *) fpnode_slip->data;

		fpnode_slip->bits = pset(Prefix(bits), mirror, miss);
		fpnode_slip->data = fpnode_data->data;

		status = sldeck(sldeck)->link_m.free
		    (sldeck(sldeck)->link_m.data, slip);
		if (status) {
		    status = FREE_ERROR;
		} else {
		    if (sizeof(byte *) < offset) {
			status = sldeck(sldeck)->link_m.free
			    (sldeck(sldeck)->link_m.data, lock);
			if (status) {
			    status = FREE_ERROR;
			}
		    }
		}
	    }
	} while (0);
    } else {
	const void *link;

	link = fpnode_data->data;

	status = sldeck(sldeck)->link_m.free
	    (sldeck(sldeck)->link_m.data, (void *) fpnode_slip->data);
	if (status) {
	    status = FREE_ERROR;
	} else {
	    if (prefix) {
		fpnode_slip->bits = pset(Prefix(bits), length, miss);
	    } else {
		fpnode_slip->base = lock;
		if (false(prefix(fpnode_slip))) {
		    if (true(fpnode_slip
			     == sldeck(sldeck)->link_a.fpdeck.fpnode_data)) {
			/*
			 * can't get here, for prefix would be set.
			 */
		    } else {
			/*
			 * can't get here, for that would mean that both parent
			 * and child node prefixes are 0.
			 */
		    }
		} else {
		    fpnode_slip->bits = miss;
		}
	    }

	    fpnode_slip->data = link;
	}
    }

    return status;
}


static int
pass_name(void *sldeck, struct fpnode_type *fpnode_slip,
	  struct fpnode_type *fpnode_data)
{
    int status;
    unsigned bits, size, slip;
    void *list;

    bits = fpnode_slip->bits;

    size = Effect(bits);

    fpnode_slip->bits = iset(bits, size);

    list = (void *) fpnode_slip->data;

    slip = fpnode_data - (struct fpnode_type *) list;

    if (slip ^ size) {
	memmove(fpnode_data, fpnode_data + 1, (size - slip)
		* sizeof(struct fpnode_type));
    } else {
    }

    if (size & 3) {
	status = 0;
    } else {
	status = sldeck(sldeck)->link_m.mode
	    (sldeck(sldeck)->link_m.data, &list,
	     sizeof(struct fpnode_type) * size);
	if (status) {
	    status = MODE_ERROR;
	} else {
	    fpnode_slip->data = list;
	}
    }

    return status;
}


int
_libx1f4l2_blue_sldeck(void *sldeck, struct fpnode_type *fpnode_slip,
		       struct fpnode_type *fpnode_data)
{
    int status;

    do {
	unsigned bits, fast;

	fast = fpnode_data->bits;

	if (sizeof(byte *) < length(fpnode_data)) {
	    status = sldeck(sldeck)->link_m.free
		(sldeck(sldeck)->link_m.data, fpnode_data->base);
	    if (status) {
		status = FREE_ERROR;

		break;
	    } else {
		fpnode_data->bits = vset(0, fast);
	    }
	}

	bits = fpnode_slip->bits;

	if (Linear(bits)) {
	    struct fpnode_type *fpnode_miss, *fpnode_text;
	    unsigned count = 0, extent, land, target;

	    fpnode_text = (void *) fpnode_slip->data;

	    if (Target(bits)) {
		if (fpnode_data == fpnode_text) {
		    status = miss_zero(sldeck, fpnode_slip);
		    if (1) {
			break;
		    }
		} else {
		    fpnode_text++;
		}
	    }

	    fpnode_data->bits = lset();

	    extent = Extent(bits);

	    target = extent;

	    land = fpnode_data - fpnode_text;

	    fpnode_miss = fpnode_text;
	    fpnode_miss += land & ~15;

	    if (land & 15) {
		fast = fpnode_miss->bits;
		fpnode_miss->bits = sset(Prefix(fast) - 1, fast);
	    } else {
		fpnode_miss->bits = rset(Prefix(fast) - 1);
	    }

	    if (extent < 32) {
	    } else {
		unsigned c;

		c = fpnode_data - fpnode_text;
		if (c < 16) {
		    count = prefix(fpnode_miss);
		    if (count) {
			extent -= 16;
			fpnode_text += 16;
		    } else {
			unsigned slip = 16;

			while (!prefix(fpnode_text + slip)) {
			    slip += 16;
			}

			status = miss_head(sldeck, fpnode_slip, slip);

			if (1) {
			    break;
			}
		    }
		} else {
		    if (extent - c < 16) {
			count = prefix(fpnode_miss);
			if (count) {
			    extent -= 16;
			} else {
			    unsigned slip = 16;

			    fpnode_text += extent - 16;
			    while (!prefix(fpnode_text - slip)) {
				slip += 16;
			    }

			    status = miss_tail(sldeck, fpnode_slip, slip);

			    if (1) {
				break;
			    }
			}
		    }
		}
	    }

	    if (land_deck
		(fpnode_text, extent, target * LINEAR_n_SCALE, &count)) {
		status = 0;
	    } else {
		status = _libx1f4l2_else_sldeck(sldeck, fpnode_slip, count);
	    }
	} else {
	    unsigned fast;

	    fast = Effect(bits);
	    if (fast == 1) {
		status = pass_link(sldeck, fpnode_slip, fpnode_data);
	    } else {
		status = pass_name(sldeck, fpnode_slip, fpnode_data);
		if (status) {
		} else {
		    unsigned call, fail, head, tail, zero = 0;

		    fpnode_data = (void *) fpnode_slip->data;

		    head = prefix(fpnode_data);
		    if (head) {
		    } else {
			zero = 1;
			head = prefix(fpnode_data + 1);
		    }

		    tail = prefix(fpnode_data + fast - 1);

		    head &= ~15;

		    fail = fast;

		    if (zero) {
			fail--;
		    }

		    call = ((tail + 16) & ~15) - head;
		    if (fail * LINEAR_d_SCALE < LINEAR_n_SCALE * call) {
		    } else {
			void *node;

			status = sldeck(sldeck)->link_m.link
			    (sldeck(sldeck)->link_m.data, &node,
			     (call + zero) * sizeof(struct fpnode_type));
			if (status) {
			    status = LINK_ERROR;
			} else {
			    struct fpnode_type *fpnode_text;
			    unsigned i, slip;

			    fpnode_text = node;

			    fpnode_slip->data = node;

			    i = call + zero;
			    for (; i; i--) {
				fpnode_text->bits = lset();
				fpnode_text++;
			    }

			    fpnode_text = node;

			    if (zero) {
				*fpnode_text++ = *fpnode_data++;
			    }

			    i = call >> 4;
			    slip = head;
			    for (; i; i--) {
				unsigned e = 0, prefix;

				prefix = prefix(fpnode_data) - slip;
				while (prefix < 16) {
				    fail--;
				    e++;
				    fpnode_text[prefix] = *fpnode_data++;
				    if (fail) {
					prefix = prefix(fpnode_data) - slip;
				    } else {
					break;
				    }
				}

				fpnode_text->bits = sset(e, fpnode_text->bits);

				slip += 16;

				fpnode_text += 16;
			    }

			    status = sldeck(sldeck)->link_m.free
				(sldeck(sldeck)->link_m.data,
				 fpnode_data - fast);
			    if (status) {
				status = FREE_ERROR;
				if (1) {
				    break;
				}
			    }

			    fpnode_slip->bits =
				fset(fpnode_slip->bits,
				     head | ((call >> 4) - 1));
			    if (zero) {
				fpnode_slip->bits = zset(fpnode_slip->bits);
			    }
			}
		    }
		}
	    }
	}
    } while (0);

    return status;
}
