/*
 * slxx-types.h
 * Copyright (C) 2008, 2009, 2010, 2011, 2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_slxx_types_H__
#define __LIBx1f4_slxx_types_H__

#include <config.h>

#include <stddef.h>

#include <trans.h>

#define Extant_bits			3
#define Length_bits			13

#define Extant_offset			16
#define Length_offset			19

#define Record				((1 << Length_bits) - 1)

#define Extant(mode) \
    ((mode) & ((1 << Extant_bits) - 1) << Extant_offset)
#define Length(mode) \
    ((mode) >> Length_offset)
#define Prefix(mode) \
    ((mode) & 255)
#define Shadow(mode) \
    ((mode) >> 8 & 255)
#define Spread(mode) \
    (((mode) >> 8 & 255) + 1)

#define Recall(mode) \
    ((mode) & ((1 << (Extant_bits - 1)) - 1) << Extant_offset)

#define Effect(mode) \
    ((mode) >> 8 & 255)

#define Extent(mode) \
    (((mode) >> 4 & 240) + 16)
#define Offset(mode) \
    ((mode) >> 8 & 240)

#define Linear(mode) \
    (Recall(mode) == (2 << Extant_offset))
#define Target(mode) \
    (Extant(mode) & 4 << Extant_offset)
#define Upward(mode) \
    (Extant(mode) >> (Extant_offset + 2) & 1)
#define Vacant(mode) \
    (Extant(mode) == (3 << Extant_offset))

#define extant(node) \
    Extant(((struct fpnode_type *) (node))->bits)
#define extent(node) \
    Extent(((struct fpnode_type *) (node))->bits)
#define length(node) \
    (((struct fpnode_type *) (node))->bits >> Length_offset)
#define prefix(node) \
    (((struct fpnode_type *) (node))->bits & 255)
#define spread(node) \
    Spread(((struct fpnode_type *) (node))->bits)

#define vacant(node) \
    Vacant(((struct fpnode_type *) (node))->bits)

#define effect(node) \
    Effect(((struct fpnode_type *) (node))->bits)

#if WORDS_BIGENDIAN
# define prefixof() \
    (offsetof(struct fpnode_type, bits) + sizeof(unsigned) - 1)
#else
# define prefixof() \
    offsetof(struct fpnode_type, bits)
#endif				/* WORDS_BIGENDIAN */

#define dset(prefix, length) \
    ((prefix) | sfff(length) << Length_offset)

#define eset(spread) \
    (((spread) - 1) << 8 | 65536)

#define fset(bits, linear) \
    ((linear) << 8 | ((bits) & ~524032) | 131072)

#define iset(bits, spread) \
    (((spread) - 1) << 8 | ((bits) & ~65280))

#define jset(bits, spread) \
    (((spread) - 1) << 8 | (bits) | 65536)

#define kset(bits, spread) \
    (((spread) - 1) << 8 | ((bits) & ~524032) | 65536)

#define lset() \
    196608

#define nset(length, bits, spread) \
    (sfff(length) << Length_offset | ((spread) - 1) << 8 | ((bits) & 255)     \
     | 65536)

#define pset(prefix, length, bits) \
    (dset(prefix, length) | ((bits) & 524032))

#define rset(prefix) \
    ((prefix) | 196608)

#define sfff(length) \
    ((length) < Record ? (length) : Record)

#define sset(prefix, bits) \
    (((bits) & ~255) | (prefix))

#define vset(length, bits) \
    (((bits) & ~524287) | sfff(length) << Length_offset)

#define wset(bits) \
    (~(4 << Extant_offset) & (bits))

#define zset(bits) \
    (4 << Extant_offset | (bits))

typedef unsigned char byte;

typedef struct fpnode_type {
    byte *base;
    unsigned bits;
    const void *data;
} fpnode_type;

typedef struct fpdeck_type {
    struct fpnode_type fpnode;
} fpdeck_type;

typedef struct screen_type {
    struct trans_type *trans_data;
} screen_type;

typedef struct sldeck_type {
    struct {
	struct fpdeck_type fpdeck;
    } link_a;
    struct {
	int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	    (*mode) (void *, void **, unsigned);
	void *data;
    } link_m;
} sldeck_type;

#endif				/* __LIBx1f4_slxx_types_H__ */
