/*
 * slxx.2.c
 * Copyright (C) 2008, 2009, 2010, 2011, 2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <slxx-defs.h>
#include <slxx-names.h>
#include <slxx-types.h>

#define sldeck(sldeck)			((struct sldeck_type *) (sldeck))

static int free_list(struct fpnode_type *, void *, int (*) (void *, void *));
static int free_node(struct fpnode_type *, void *, int (*) (void *, void *));

static int
free_list(struct fpnode_type *fpnode_data, void *data,
	  int (*free) (void *, void *))
{
    int status;
    unsigned bits;

    bits = fpnode_data->bits;

    if (Extant(bits)) {
	struct fpnode_type *fpnode_slip;

	fpnode_slip = (struct fpnode_type *) fpnode_data->data;

	fpnode_data = fpnode_slip;

	if (Linear(bits)) {
	    unsigned size;

	    size = Shadow(bits);

	    status = 0;

	    size &= 15;
	    size++;
	    size <<= 4;

	    if (Target(bits)) {
		size++;
	    }

	    for (; size; size--) {
		if (vacant(fpnode_data)) {
		} else {
		    status = free_node(fpnode_data, data, free);
		    if (status) {
			break;
		    }
		}

		fpnode_data++;
	    }
	} else {
	    unsigned size;

	    size = Spread(bits);

	    while (1) {
		status = free_node(fpnode_data, data, free);
		if (status) {
		    break;
		} else {
		    size--;
		    if (size) {
			fpnode_data++;
		    } else {
			break;
		    }
		}
	    }
	}

	if (status) {
	} else {
	    status = free(data, fpnode_slip);
	    if (status) {
		status = FREE_ERROR;
	    }
	}
    } else {
	status = 0;
    }

    return status;
}


static int
free_node(struct fpnode_type *fpnode_data, void *data,
	  int (*free) (void *, void *))
{
    int status;

    if (1) {
	if (sizeof(byte *) < length(fpnode_data)) {
	    status = free(data, fpnode_data->base);
	    if (status) {
		status = FREE_ERROR;
	    } else {
		status = free_list(fpnode_data, data, free);
	    }
	} else {
	    status = free_list(fpnode_data, data, free);
	}
    }

    return status;
}


int
x1f4_flat_sldeck(void *sldeck)
{
    int status;
    struct fpnode_type *fpnode_data;

    fpnode_data = &sldeck(sldeck)->link_a.fpdeck.fpnode;
    if (vacant(fpnode_data)) {
	status = 0;
    } else {
	status = free_node
	    (fpnode_data, sldeck(sldeck)->link_m.data,
	     sldeck(sldeck)->link_m.free);
    }

    return status;
}
