/*
 * slxx.4.c
 * Copyright (C) 2009, 2010, 2011, 2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include <slxx-names.h>
#include <slxx-types.h>

#define back(miss)			((const char *) (miss))

#define sldeck(sldeck)			((struct sldeck_type *) (sldeck))

static int back_node(const struct fpnode_type *, void *,
		     int (*) (void *, const void *));

static int
back_node(const struct fpnode_type *fpnode_data, void *back,
	  int (*call) (void *, const void *))
{
    int delete;
    unsigned bits;

    bits = fpnode_data->bits;
    if (Extant(bits)) {
	fpnode_data = fpnode_data->data;

	if (Linear(bits)) {
	    unsigned offset, size;

	    size = Shadow(bits);

	    delete = 0;

	    offset = size & ~15;

	    size &= 15;
	    size++;
	    size <<= 4;

	    offset += size;

	    do {
		if (Target(bits)) {
		    delete = back_node(fpnode_data, back, call);
		    if (delete) {
			break;
		    } else {
			fpnode_data++;
		    }
		}

		for (; size; size--) {
		    if (vacant(fpnode_data)) {
		    } else {
			delete = back_node(fpnode_data, back, call);
			if (delete) {
			    break;
			}
		    }

		    fpnode_data++;
		}
	    } while (0);
	} else {
	    unsigned size;

	    size = Spread(bits);

	    while (1) {
		delete = back_node(fpnode_data, back, call);
		if (delete) {
		    break;
		} else {
		    size--;
		    if (size) {
			fpnode_data++;
		    } else {
			break;
		    }
		}
	    }
	}
    } else {
	delete = call(back, fpnode_data->data);
    }

    return delete;
}


int
x1f4_lime_sldeck(void *sldeck, void *back, int (*call) (void *, const void *))
{
    int delete;
    struct fpnode_type *fpnode_data;

    fpnode_data = &sldeck(sldeck)->link_a.fpdeck.fpnode;
    if (vacant(fpnode_data)) {
	delete = 0;
    } else {
	delete = back_node(fpnode_data, back, call);
    }

    return delete;
}
