/*
 * slxx.7.c
 * Copyright (C) 2009, 2010, 2011, 2012, 2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include <slxx-defs.h>
#include <slxx-inter.h>
#include <slxx-names.h>
#include <slxx-patch.h>
#include <slxx-types.h>

#define back(miss)			((const char *) (miss))
#define fast(miss)			((const byte *) (miss))

#define sldeck(sldeck)			((struct sldeck_type *) (sldeck))

static int call_deck(void *, struct fpnode_type *, const byte *, const void *);
static int cell_deck(void *, struct fpnode_type *, struct fpnode_type *,
		     unsigned, unsigned, int);
static int case_deck(void *, struct fpnode_type *, const byte *, unsigned, int,
		     int, const byte *, const void *);
static int ever_deck(void *, struct fpnode_type *, const byte *, const void *);
static int fail_deck(void *, struct fpnode_type *, struct fpnode_type *,
		     const byte *, const void *, int, unsigned);
static int fare_deck(void *, struct fpnode_type *, struct fpnode_type *,
		     unsigned, unsigned, const byte *, const void *);
static int fast_deck(void *, struct fpnode_type *, const void *);
static int land_deck(struct fpnode_type *, unsigned, unsigned, unsigned *);
static int line_deck(void *, struct fpnode_type *, struct fpnode_type *,
		     unsigned, unsigned, const byte *, const void *);
static int long_deck(void *, struct fpnode_type *, const byte *, const void *,
		     unsigned, struct fpnode_type *);
static int mind_deck(void *, struct fpnode_type *, const byte *, const void *);
static int near_deck(void *, struct fpnode_type *, unsigned, unsigned, int,
		     const byte *, const void *);
static int news_deck(void *, struct fpnode_type *, const void *);
static int none_deck(void *, struct fpnode_type *, byte *);
static int pass_deck(void *, struct fpnode_type *, const void *);
static int post_deck(void *, struct fpnode_type *, struct fpnode_type *,
		     const byte *, const void *);
static int post_line(void *, const byte *, const void *);
static int post_name(void *, const byte *, const void *);
static int post_pair(void *, const byte *, const void *);
static int post_slip(void *, const byte *, const void *);
static int side_deck(void *, struct fpnode_type *, struct fpnode_type *,
		     unsigned, unsigned, unsigned, int, const byte *,
		     const void *);
static int trap_deck(void *, struct fpnode_type *, struct fpnode_type *,
		     const byte *, const void *, int, unsigned);
static int zero_deck(void *, struct fpnode_type *);

static int
call_deck(void *sldeck, struct fpnode_type *fpnode_data, const byte *name,
	  const void *data)
{
    int status;

    do {
	unsigned length;

	length = strlen(back(name));
	if (length < sizeof(byte *) + 2) {
	    length--;

	    fpnode_data->bits = dset(0, length);
	    fpnode_data->data = data;

	    if (length) {
		memcpy(&fpnode_data->base, name + 1, length);
	    }

	    status = 0;
	} else {
	    void *sand;

	    status = sldeck(sldeck)->link_m.link
		(sldeck(sldeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;
	    } else {
		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(0, length);
		fpnode_data->data = data;
	    }
	}
    } while (0);

    return status;
}


static int
cell_deck(void *sldeck, struct fpnode_type *fpnode_call,
	  struct fpnode_type *fpnode_data, unsigned offset, unsigned extent,
	  int zero)
{
    int status;
    void *node;

    node = fpnode_data - zero;

    status = sldeck(sldeck)->link_m.mode
	(sldeck(sldeck)->link_m.data, &node,
	 (extent + zero) * sizeof(struct fpnode_type));
    if (status) {
	status = MODE_ERROR;
    } else {
	unsigned bits, call, slip;

	fpnode_data = node;

	bits = fpnode_call->bits;

	fpnode_call->data = node;

	fpnode_call->bits = fset(bits, offset | ((extent >> 4) - 1));
	if (zero) {
	    fpnode_call->bits = zset(fpnode_call->bits);
	}

	call = Extent(bits);

	fpnode_data += zero;

	slip = Offset(bits) - offset;
	if (slip) {
	    memmove(fpnode_data + slip, fpnode_data,
		    sizeof(struct fpnode_type) * call);
	} else {
	    fpnode_data += call;
	    slip = extent - call;
	}

	for (; slip; slip--) {
	    fpnode_data->bits = lset();
	    fpnode_data++;
	}
    }

    return status;
}


static int
case_deck(void *sldeck, struct fpnode_type *fpnode_slip, const byte *slip,
	  unsigned suffix, int e, int c, const byte *name, const void *data)
{
    int status;
    void *node;

    status = sldeck(sldeck)->link_m.link
	(sldeck(sldeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	unsigned bits, prefix, length;

	prefix = slip - fpnode_slip->base;

	bits = fpnode_slip->bits;

	length = strlen(back(name));
	if (length < sizeof(byte *) + 2) {
	    if (suffix < sizeof(byte *) + 2) {
		struct fpnode_type *fpnode_data;

		fpnode_data = node;

		if (c < e + 1) {
		    if (length) {
			length--;
			memcpy(&fpnode_data->base, name + 1, length);
		    }

		    fpnode_data->bits = dset(c, length);
		    fpnode_data->data = data;

		    fpnode_data++;

		    suffix--;

		    fpnode_data->bits = pset(e, suffix, bits);
		    fpnode_data->data = fpnode_slip->data;

		    if (suffix) {
			memcpy(&fpnode_data->base, slip + 1, suffix);
		    }
		} else {
		    suffix--;

		    fpnode_data->bits = pset(e, suffix, bits);
		    fpnode_data->data = fpnode_slip->data;

		    if (suffix) {
			memcpy(&fpnode_data->base, slip + 1, suffix);
		    }

		    fpnode_data++;

		    if (length) {
			length--;
			memcpy(&fpnode_data->base, name + 1, length);
		    }

		    fpnode_data->bits = dset(c, length);
		    fpnode_data->data = data;
		}

		fpnode_slip->data = node;
	    } else {
		void *flat;

		status = sldeck(sldeck)->link_m.link
		    (sldeck(sldeck)->link_m.data, &flat, suffix);
		if (status) {
		    status = LINK_ERROR;

		    sldeck(sldeck)->link_m.free
			(sldeck(sldeck)->link_m.data, node);
		} else {
		    struct fpnode_type *fpnode_data;

		    fpnode_data = node;

		    memcpy(flat, slip + 1, suffix);

		    suffix--;

		    if (c < e + 1) {
			if (length) {
			    length--;
			    memcpy(&fpnode_data->base, name + 1, length);
			}

			fpnode_data->bits = dset(c, length);
			fpnode_data->data = data;

			fpnode_data++;

			fpnode_data->base = flat;
			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;
		    } else {
			fpnode_data->base = flat;
			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;

			fpnode_data++;

			if (length) {
			    length--;
			    memcpy(&fpnode_data->base, name + 1, length);
			}

			fpnode_data->bits = dset(c, length);
			fpnode_data->data = data;
		    }

		    fpnode_slip->data = node;
		}
	    }
	} else {
	    void *sand;

	    status = sldeck(sldeck)->link_m.link
		(sldeck(sldeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;

		sldeck(sldeck)->link_m.free(sldeck(sldeck)->link_m.data, node);
	    } else {
		if (suffix < sizeof(byte *) + 2) {
		    struct fpnode_type *fpnode_data;

		    fpnode_data = node;

		    memcpy(sand, name + 1, length);

		    length--;

		    if (c < e + 1) {
			fpnode_data->base = sand;
			fpnode_data->bits = dset(c, length);
			fpnode_data->data = data;

			fpnode_data++;

			suffix--;

			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;

			if (suffix) {
			    memcpy(&fpnode_data->base, slip + 1, suffix);
			}
		    } else {
			suffix--;

			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;

			if (suffix) {
			    memcpy(&fpnode_data->base, slip + 1, suffix);
			}

			fpnode_data++;

			fpnode_data->base = sand;
			fpnode_data->bits = dset(c, length);
			fpnode_data->data = data;
		    }

		    fpnode_slip->data = node;
		} else {
		    void *flat;

		    status = sldeck(sldeck)->link_m.link
			(sldeck(sldeck)->link_m.data, &flat, suffix);
		    if (status) {
			status = LINK_ERROR;

			sldeck(sldeck)->link_m.free
			    (sldeck(sldeck)->link_m.data, sand);
			sldeck(sldeck)->link_m.free
			    (sldeck(sldeck)->link_m.data, node);
		    } else {
			struct fpnode_type *fpnode_data;

			fpnode_data = node;

			memcpy(sand, name + 1, length);

			length--;

			memcpy(flat, slip + 1, suffix);

			suffix--;

			if (c < e + 1) {
			    fpnode_data->base = sand;
			    fpnode_data->bits = dset(c, length);
			    fpnode_data->data = data;

			    fpnode_data++;

			    fpnode_data->base = flat;
			    fpnode_data->bits = pset(e, suffix, bits);
			    fpnode_data->data = fpnode_slip->data;
			} else {
			    fpnode_data->base = flat;
			    fpnode_data->bits = pset(e, suffix, bits);
			    fpnode_data->data = fpnode_slip->data;

			    fpnode_data++;

			    fpnode_data->base = sand;
			    fpnode_data->bits = dset(c, length);
			    fpnode_data->data = data;
			}

			fpnode_slip->data = node;
		    }
		}
	    }
	}
    }

    return status;
}


static int
ever_deck(void *sldeck, struct fpnode_type *fpnode_slip, const byte *name,
	  const void *data)
{
    int status;
    void *node;

    status = sldeck(sldeck)->link_m.link
	(sldeck(sldeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	unsigned length;

	length = strlen(back(name));
	if (length < sizeof(byte *) + 2) {
	    struct fpnode_type *fpnode_data;

	    fpnode_data = node;

	    fpnode_data->bits = dset(0, 0);
	    fpnode_data->data = fpnode_slip->data;

	    fpnode_data++;

	    length--;

	    fpnode_data->bits = dset(*name, length);
	    fpnode_data->data = data;

	    if (length) {
		memcpy(&fpnode_data->base, name + 1, length);
	    }

	    fpnode_slip->bits = jset(fpnode_slip->bits, 2);
	    fpnode_slip->data = node;
	} else {
	    void *sand;

	    status = sldeck(sldeck)->link_m.link
		(sldeck(sldeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;

		sldeck(sldeck)->link_m.free(sldeck(sldeck)->link_m.data, node);
	    } else {
		struct fpnode_type *fpnode_data;

		fpnode_data = node;

		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->bits = dset(0, 0);
		fpnode_data->data = fpnode_slip->data;

		fpnode_data++;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, length);
		fpnode_data->data = data;

		fpnode_slip->bits = jset(fpnode_slip->bits, 2);
		fpnode_slip->data = node;
	    }
	}
    }

    return status;
}


static int
fail_deck(void *sldeck, struct fpnode_type *fpnode_call,
	  struct fpnode_type *fpnode_data, const byte *name, const void *data,
	  int c, unsigned mode)
{
    int status;
    unsigned count, pass, offset, zero;

    zero = Upward(mode);

    offset = Offset(mode);

    pass = offset - (c & ~15) + Extent(mode);
    fpnode_data += zero;
    if (land_deck
	(fpnode_data, Extent(mode), LINEAR_N_SCALE * pass, &count)) {
	status = cell_deck
	    (sldeck, fpnode_call, fpnode_data, c & ~15, pass, zero);
	if (status) {
	} else {
	    unsigned bits;

	    fpnode_data = (void *) fpnode_call->data;
	    fpnode_data += c & 15;

	    fpnode_data += zero;

	    bits = fpnode_data->bits;

	    status = call_deck(sldeck, fpnode_data, name, data);
	    if (status) {
	    } else {
		c &= 15;

		fpnode_data -= c;

		if (c) {
		    bits = fpnode_data->bits;
		}

		fpnode_data->bits =
		    sset(Prefix(bits) + 1, fpnode_data->bits);
	    }
	}
    } else {
	status = near_deck(sldeck, fpnode_call, mode, count, c, name, data);
    }

    return status;
}


static int
fare_deck(void *sldeck, struct fpnode_type *fpnode_call,
	  struct fpnode_type *fpnode_data, unsigned line, unsigned fast,
	  const byte *name, const void *data)
{
    int status, zero;
    unsigned c, call, fail, head, tail;

    c = *name;

    fail = fast;

    head = prefix(fpnode_data);
    if (c < head) {
	head = c;
    }

    if (head) {
	zero = 0;
    } else {
	zero = 1;
	fail--;
	if (c) {
	    head = prefix(fpnode_data + 1);
	    if (c < head) {
		head = c;
	    }
	} else {
	    head = prefix(fpnode_data);
	}
    }

    tail = prefix(fpnode_data + fast - 1);
    if (tail < c) {
	tail = c;
    }

    head &= ~15;

    call = ((tail + 16) & ~15) - head;
    if (fail * LINEAR_D_SCALE < LINEAR_N_SCALE * call) {
	status = line_deck
	    (sldeck, fpnode_call, fpnode_data, line, fast, name, data);
    } else {
	status = side_deck
	    (sldeck, fpnode_call, fpnode_data, fast, head, call, zero, name,
	     data);
    }

    return status;
}


static int
fast_deck(void *sldeck, struct fpnode_type *fpnode_call, const void *data)
{
    int status;
    struct fpnode_type *fpnode_slip;
    unsigned mode;

    fpnode_slip = (void *) fpnode_call->data;

    mode = fpnode_call->bits;
    if (Linear(mode)) {
	if (Target(mode)) {
	    status = EVER_MATCH;
	} else {
	    status = pass_deck(sldeck, fpnode_call, data);
	}
    } else {
	if (prefix(fpnode_slip)) {
	    status = news_deck(sldeck, fpnode_call, data);
	} else {
	    status = EVER_MATCH;
	}
    }

    return status;
}


static int
land_deck(struct fpnode_type *fpnode_data, unsigned extent, unsigned miss,
	  unsigned *f)
{
    unsigned left, slip = 0;

    left = LINEAR_D_SCALE;
    extent >>= 4;
    for (; extent; extent--) {
	unsigned p;

	p = prefix(fpnode_data);
	slip += p;
	left += p * LINEAR_D_SCALE;
	if (left < miss) {
	} else {
	    break;
	}

	fpnode_data += 16;
    }

    *f = slip;

    return extent;
}


static int
line_deck(void *sldeck, struct fpnode_type *fpnode_call,
	  struct fpnode_type *fpnode_data, unsigned line, unsigned fast,
	  const byte *name, const void *data)
{
    int status;

    do {
	unsigned length;

	if (fast & 3) {
	} else {
	    void *node;

	    node = fpnode_data;

	    status = sldeck(sldeck)->link_m.mode
		(sldeck(sldeck)->link_m.data, &node,
		 (fast + 4) * sizeof(struct fpnode_type));
	    if (status) {
		status = MODE_ERROR;
		if (1) {
		    break;
		}
	    } else {
		fpnode_data = node;

		fpnode_call->data = node;
	    }
	}

	length = strlen(back(name));
	if (length < sizeof(byte *) + 2) {
	    fpnode_data += line;

	    memmove(fpnode_data + 1, fpnode_data,
		    (fast - line) * sizeof(struct fpnode_type));

	    length--;

	    fpnode_data->bits = dset(*name, length);
	    fpnode_data->data = data;

	    fpnode_call->bits = iset(fpnode_call->bits, fast + 1);

	    if (length) {
		memcpy(&fpnode_data->base, name + 1, length);
	    }

	    status = 0;
	} else {
	    void *sand;

	    status = sldeck(sldeck)->link_m.link
		(sldeck(sldeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;
	    } else {
		fpnode_data += line;

		memmove(fpnode_data + 1, fpnode_data,
			(fast - line) * sizeof(struct fpnode_type));

		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, length);
		fpnode_data->data = data;

		fpnode_call->bits = iset(fpnode_call->bits, fast + 1);
	    }
	}
    } while (0);

    return status;
}


static int
long_deck(void *sldeck, struct fpnode_type *fpnode_slip, const byte *name,
	  const void *data, unsigned miss,
	  struct fpnode_type *fpnode_data)
{
    int status;
    unsigned bits;

    bits = fpnode_data->bits;

    status = call_deck(sldeck, fpnode_data, name, data);
    if (status) {
    } else {
	miss &= 15;

	fpnode_data -= miss;

	if (miss) {
	    bits = fpnode_data->bits;
	}

	fpnode_data->bits = sset(Prefix(bits) + 1, fpnode_data->bits);
    }

    return status;
}


static int
mind_deck(void *sldeck, struct fpnode_type *fpnode_data, const byte *name,
	  const void *data)
{
    return post_deck
	(sldeck, fpnode_data, (void *) fpnode_data->data, name, data);
}


static int
near_deck(void *sldeck, struct fpnode_type *fpnode_call, unsigned mode,
	  unsigned count, int c, const byte *name, const void *data)
{
    int status;

    status = _libx1f4l2_else_sldeck(sldeck, fpnode_call, count);
    if (status) {
    } else {
	struct fpnode_type *fpnode_data, *fpnode_slip;
	unsigned bits, line, mind;

	fpnode_slip = (void *) fpnode_call->data;

	fpnode_data = fpnode_slip;

	mode = fpnode_call->bits;

	mind = Spread(mode);
	Progress_node(fpnode_data, mind - 1);
	bits = fpnode_data->bits;

	line = fpnode_data - fpnode_slip;

	if (c < Prefix(bits)) {
	} else {
	    line++;
	}

	status = line_deck
	    (sldeck, fpnode_call, fpnode_slip, line, mind, name, data);
    }

    return status;
}


static int
news_deck(void *sldeck, struct fpnode_type *fpnode_call, const void *data)
{
    int status;
    struct fpnode_type *fpnode_slip;

    fpnode_slip = (void *) fpnode_call->data;

    do {
	unsigned fast;

	fast = spread(fpnode_call);

	if (fast & 3) {
	    status = 0;
	} else {
	    void *node;

	    node = fpnode_slip;

	    status = sldeck(sldeck)->link_m.mode
		(sldeck(sldeck)->link_m.data, &node,
		 (fast + 4) * sizeof(struct fpnode_type));
	    if (status) {
		status = MODE_ERROR;
		if (1) {
		    break;
		}
	    } else {
		fpnode_slip = node;

		fpnode_call->data = node;
	    }
	}

	memmove(fpnode_slip + 1, fpnode_slip,
		fast * sizeof(struct fpnode_type));

	fpnode_slip->bits = dset(0, 0);
	fpnode_slip->data = data;

	fpnode_call->bits = iset(fpnode_call->bits, fast + 1);
    } while (0);

    return status;
}


static int
none_deck(void *sldeck, struct fpnode_type *fpnode_slip, byte *slip)
{
    int status;
    unsigned bits, length;

    bits = fpnode_slip->bits;

    length = Length(bits);
    if (sizeof(byte *) < length) {
	unsigned prefix;
	void *miss;

	prefix = slip - fpnode_slip->base;

	fpnode_slip->bits = nset(prefix, bits, 2);

	miss = fpnode_slip->base;

	if (sizeof(byte *) < prefix) {
	    *slip = 0;

	    status = sldeck(sldeck)->link_m.mode
		(sldeck(sldeck)->link_m.data, &miss, prefix + 1);
	    if (status) {
		status = MODE_ERROR;
	    } else {
		byte *text;

		text = miss;

		text[prefix] = 0;

		fpnode_slip->base = miss;
	    }
	} else {
	    memcpy(&fpnode_slip->base, miss, prefix);

	    status = sldeck(sldeck)->link_m.free
		(sldeck(sldeck)->link_m.data, miss);
	    if (status) {
		status = FREE_ERROR;
		if (1) {
		    fpnode_slip->base = miss;
		}
	    }
	}
    } else {
	unsigned prefix;

	prefix = slip - (byte *) &fpnode_slip->base;

	fpnode_slip->bits = nset(prefix, bits, 2);

	status = 0;
    }

    return status;
}


static int
pass_deck(void *sldeck, struct fpnode_type *fpnode_call, const void *data)
{
    int status;

    status = zero_deck(sldeck, fpnode_call);
    if (status) {
    } else {
	struct fpnode_type *fpnode_data;

	fpnode_data = (void *) fpnode_call->data;

	fpnode_data->bits = dset(0, 0);
	fpnode_data->data = data;
    }

    return status;
}


static int
post_deck(void *sldeck, struct fpnode_type *fpnode_call,
      struct fpnode_type *fpnode_slip, const byte *name, const void *data)
{
    int c, status;

    c = *name;
    if (c) {
	do {
	    struct fpnode_type *fpnode_data;
	    unsigned bits, mind, mode, prefix;

	    mode = fpnode_call->bits;

	    mind = Spread(mode);

	    fpnode_data = fpnode_slip;

	    if (Linear(mode)) {
		unsigned offset;

		offset = Offset(mode);
		if (c < offset) {
		    status = fail_deck
			(sldeck, fpnode_call, fpnode_slip, name, data, c,
			 mode);
		    if (1) {
			break;
		    }
		} else {
		    if (c - offset < Extent(mode)) {
			fpnode_data += c - offset;
			if (Target(mode)) {
			    fpnode_data++;
			}
			if (vacant(fpnode_data)) {
			    status = long_deck
				(sldeck, fpnode_slip, name, data, c - offset,
				 fpnode_data);
			    if (1) {
				break;
			    }
			} else {
			    prefix = c;

			    bits = fpnode_data->bits;
			}
		    } else {
			status = trap_deck
			    (sldeck, fpnode_call, fpnode_slip, name, data, c,
			     mode);
			if (1) {
			    break;
			}
		    }
		}
	    } else {
		Progress_node(fpnode_data, mind - 1);

		bits = fpnode_data->bits;

		prefix = Prefix(bits);
	    }

	    if (prefix == c) {
		byte *slip;
		int e;
		unsigned length;

		length = Length(bits);

		name++;

		if (sizeof(byte *) < length) {
		    slip = fpnode_data->base;
		    if (length < Record) {
		    } else {
			length += strlen(back(slip) + Record);
		    }
		} else {
		    slip = (byte *) &fpnode_data->base;
		}

		if (1) {
		    while (1) {
			if (length) {
			    c = *name;
			    e = *slip;
			    if (c == e) {
				if (c) {
				    slip++;
				    name++;
				    length--;
				} else {
				    break;
				}
			    } else {
				break;
			    }
			} else {
			    c = *name;
			    e = 0;
			    if (1) {
				break;
			    }
			}
		    }
		}
		if (e) {
		    status = case_deck
			(sldeck, fpnode_data, slip, length, e, c, name, data);
		    if (status) {
		    } else {
			status = none_deck(sldeck, fpnode_data, slip);
		    }
		} else {
		    if (Extant(bits)) {
			status = mind_deck(sldeck, fpnode_data, name, data);
		    } else {
			if (c) {
			    status = ever_deck
				(sldeck, fpnode_data, name, data);
			} else {
			    status = EVER_MATCH;
			}
		    }
		}
	    } else {
		unsigned line;

		line = fpnode_data - fpnode_slip;

		if (c < Prefix(bits)) {
		} else {
		    line++;
		}

		status = fare_deck
		    (sldeck, fpnode_call, fpnode_slip, line, mind, name, data);
	    }
	} while (0);
    } else {
	status = fast_deck(sldeck, fpnode_call, data);
    }

    return status;
}


static int
post_line(void *sldeck, const byte *name, const void *data)
{
    int status;

    if (0) {
    } else {
	unsigned length;

	length = strlen(back(name));
	if (length < sizeof(byte *) + 2) {
	    struct fpnode_type *fpnode_data;

	    fpnode_data = &sldeck(sldeck)->link_a.fpdeck.fpnode;

	    status = 0;

	    if (length) {
		length--;
	    }

	    fpnode_data->bits = dset(*name, length);
	    fpnode_data->data = data;

	    if (length) {
		memcpy(&fpnode_data->base, name + 1, length);
	    }
	} else {
	    void *sand;

	    status = sldeck(sldeck)->link_m.link
		(sldeck(sldeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;
	    } else {
		struct fpnode_type *fpnode_data;

		fpnode_data = &sldeck(sldeck)->link_a.fpdeck.fpnode;

		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, length);
		fpnode_data->data = data;
	    }
	}
    }

    return status;
}


static int
post_name(void *sldeck, const byte *name, const void *data)
{
    struct fpnode_type *fpnode_data;

    fpnode_data = &sldeck(sldeck)->link_a.fpdeck.fpnode;

    return post_deck
	(sldeck, fpnode_data, (void *) fpnode_data->data, name, data);
}


static int
post_pair(void *sldeck, const byte *name, const void *data)
{
    int status;
    void *node;

    status = sldeck(sldeck)->link_m.link
	(sldeck(sldeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	struct fpnode_type *fpnode_data, *fpnode_slip;

	fpnode_data = &sldeck(sldeck)->link_a.fpdeck.fpnode;

	fpnode_slip = node;

	*fpnode_slip = *fpnode_data;

	fpnode_data->bits = eset(1);
	fpnode_data->data = fpnode_slip;

	status = post_deck(sldeck, fpnode_data, fpnode_slip, name, data);
    }

    return status;
}


static int
post_slip(void *sldeck, const byte *name, const void *data)
{
    struct fpnode_type fpnode;

    fpnode.bits = 0;

    return post_deck
	(sldeck, &fpnode, &sldeck(sldeck)->link_a.fpdeck.fpnode, name, data);
}


static int
side_deck(void *sldeck, struct fpnode_type *fpnode_call,
	  struct fpnode_type *fpnode_data, unsigned fast, unsigned head,
	  unsigned call, int zero, const byte *name, const void *data)
{
    int status;
    void *node;

    status = sldeck(sldeck)->link_m.link
	(sldeck(sldeck)->link_m.data, &node,
	 (call + zero) * sizeof(struct fpnode_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	struct fpnode_type *fpnode_slip;
	unsigned fail, i, slip;

	fpnode_slip = node;

	fpnode_call->data = node;

	i = call + zero;
	for (; i; i--) {
	    fpnode_slip->bits = lset();
	    fpnode_slip++;
	}

	fpnode_slip = node;

	fail = fast;

	if (zero) {
	    if (prefix(fpnode_data)) {
	    } else {
		*fpnode_slip++ = *fpnode_data++;
		fail--;
	    }
	}

	i = call >> 4;
	slip = head;
	for (; i; i--) {
	    unsigned e = 0, prefix;

	    prefix = prefix(fpnode_data) - slip;
	    while (prefix < 16) {
		fail--;
		e++;
		fpnode_slip[prefix] = *fpnode_data++;
		if (fail) {
		    prefix = prefix(fpnode_data) - slip;
		} else {
		    break;
		}
	    }

	    fpnode_slip->bits = sset(e, fpnode_slip->bits);

	    slip += 16;

	    fpnode_slip += 16;
	}

	fpnode_slip = node;

	do {
	    unsigned bits, c, length;

	    c = *name;

	    status = sldeck(sldeck)->link_m.free
		(sldeck(sldeck)->link_m.data, fpnode_data - fast);
	    if (status) {
		status = FREE_ERROR;
		if (1) {
		    break;
		}
	    }

	    if (c) {
		c -= head;
	    }

	    fpnode_slip += c;

	    if (zero) {
		fpnode_slip++;
	    }

	    bits = fpnode_slip->bits;

	    length = strlen(back(name));
	    if (length < sizeof(byte *) + 2) {
		length--;

		if (length) {
		    memcpy(&fpnode_slip->base, name + 1, length);
		}
	    } else {
		void *sand;

		status = sldeck(sldeck)->link_m.link
		    (sldeck(sldeck)->link_m.data, &sand, length);
		if (status) {
		    status = LINK_ERROR;
		    if (1) {
			break;
		    }
		} else {
		    memcpy(sand, name + 1, length);

		    length--;

		    fpnode_slip->base = sand;
		}
	    }

	    fpnode_slip->bits = dset(Prefix(bits), length);
	    fpnode_slip->data = data;

	    fpnode_slip -= c & 15;

	    bits = fpnode_slip->bits;

	    fpnode_slip->bits = sset(Prefix(bits) + 1, bits);

	    fpnode_call->bits =
		fset(fpnode_call->bits, head | ((call >> 4) - 1));
	    if (zero) {
		fpnode_call->bits = zset(fpnode_call->bits);
	    }
	} while (0);
    }

    return status;
}


static int
trap_deck(void *sldeck, struct fpnode_type *fpnode_call,
	  struct fpnode_type *fpnode_data, const byte *name, const void *data,
	  int c, unsigned mode)
{
    int status;
    unsigned count, miss, offset, zero;

    zero = Upward(mode);

    offset = Offset(mode);

    miss = c - offset;

    fpnode_data += zero;
    if (land_deck
	(fpnode_data, Extent(mode),
	 LINEAR_N_SCALE * ((miss + 16) & ~15), &count)) {
	status = cell_deck
	    (sldeck, fpnode_call, fpnode_data, offset, (miss + 16) & ~15,
	     zero);
	if (status) {
	} else {
	    unsigned bits;

	    fpnode_data = (void *) fpnode_call->data;
	    fpnode_data += miss;

	    fpnode_data += zero;

	    bits = fpnode_data->bits;

	    status = call_deck(sldeck, fpnode_data, name, data);
	    if (status) {
	    } else {
		miss &= 15;

		fpnode_data -= miss;

		if (miss) {
		    bits = fpnode_data->bits;
		}

		fpnode_data->bits =
		    sset(Prefix(bits) + 1, fpnode_data->bits);
	    }
	}
    } else {
	status = near_deck(sldeck, fpnode_call, mode, count, c, name, data);
    }

    return status;
}


static int
zero_deck(void *sldeck, struct fpnode_type *fpnode_call)
{
    int status;
    unsigned bits, extent;
    void *node;

    bits = fpnode_call->bits;

    extent = Extent(bits);

    node = (void *) fpnode_call->data;

    status = sldeck(sldeck)->link_m.mode
	(sldeck(sldeck)->link_m.data, &node,
	 (extent + 1) * sizeof(struct fpnode_type));
    if (status) {
	status = MODE_ERROR;
    } else {
	struct fpnode_type *fpnode_data;

	fpnode_data = node;

	fpnode_call->data = node;

	fpnode_call->bits = zset(bits);

	memmove(fpnode_data + 1, fpnode_data,
		sizeof(struct fpnode_type) * extent);

	if (1) {
	    fpnode_data->bits = lset();
	}
    }

    return status;
}


int
x1f4_post_sldeck(void *sldeck, const char *name, const void *data)
{
    int status;
    struct fpnode_type *fpnode_data;

    fpnode_data = &sldeck(sldeck)->link_a.fpdeck.fpnode;
    if (vacant(fpnode_data)) {
	status = post_line(sldeck, fast(name), data);
    } else {
	int prefix;

	prefix = prefix(fpnode_data);
	if (prefix) {
	    if (prefix == *name) {
		status = post_slip(sldeck, fast(name), data);
	    } else {
		status = post_pair(sldeck, fast(name), data);
	    }
	} else {
	    if (extant(fpnode_data)) {
		status = post_name(sldeck, fast(name), data);
	    } else {
		if (*name) {
		    status = post_pair(sldeck, fast(name), data);
		} else {
		    status = post_slip(sldeck, fast(name), data);
		}
	    }
	}
    }

    return status;
}
