/*
 * tcmf.8.c
 * Copyright (C) 2008-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <tcmn-bits.h>
#include <tcmn-names.h>
#include <tcmn-types.h>

#define __DATA_SELECT__			1
#define __SIGN_SELECT__			2

#define __HERE_SELECT__			__SIGN_SELECT__

#define d_count				(SIZEOF_UNSIGNED << 3)

#define d_left				(d_count - 1)

#define d_high				(1u << d_left)

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

#if __HERE_SELECT__ == __SIGN_SELECT__
static int find_d1db(void *, unsigned, unsigned, unsigned, void **);
static int find_d1dp(void *, unsigned, unsigned, unsigned, void **);
static int find_d1qb(void *, unsigned, unsigned, unsigned, void **);
static int find_d1qp(void *, unsigned, unsigned, unsigned, void **);
#endif				/* __HERE_SELECT__ == __SIGN_SELECT__ */
#if __HERE_SELECT__ == __DATA_SELECT__
static int find_list(void *, unsigned, unsigned, unsigned, void **);
#endif				/* __HERE_SELECT__ == __DATA_SELECT__ */

#if __HERE_SELECT__ == __SIGN_SELECT__
static int
find_d1db(void *tcmini, unsigned node, unsigned text, unsigned offset,
	  void **find)
{
    while (1) {
	unsigned diff, slip;
	void *extend;

	extend = M_TEXT(tcmini, node);

	slip = *(unsigned *) ((char *) extend + offset);

	diff = slip - text;
	if (!diff) {
	    *find = (char *) extend + 4;

	    break;
	} else {
	    node = L_NODE(Q_NODE(extend), (int) diff >> d_left);
	}
    }

    return 1;
}
#endif				/* __HERE_SELECT__ == __SIGN_SELECT__ */


#if __HERE_SELECT__ == __SIGN_SELECT__
static int
find_d1dp(void *tcmini, unsigned node, unsigned text, unsigned offset,
	  void **find)
{
    while (1) {
	unsigned slip;
	void *extend;

	extend = M_TEXT(tcmini, node);

	slip = *(unsigned *) ((char *) extend + offset);

	if (slip & d_high) {
	    unsigned diff;

	    diff = slip - text;

	    if (diff >> d_left) {
		node = L_NEXT(Q_NODE(extend));
		if (1) {
		    find_d1db(tcmini, node, text, offset, find);
		    if (1) {
			break;
		    }
		}
	    } else {
		if (!diff) {
		    *find = (char *) extend + 4;

		    break;
		} else {
		    node = L_LAST(Q_NODE(extend));
		}
	    }
	} else {
	    node = L_NEXT(Q_NODE(extend));
	}
    }

    return 1;
}
#endif				/* __HERE_SELECT__ == __SIGN_SELECT__ */


#if __HERE_SELECT__ == __SIGN_SELECT__
static int
find_d1qb(void *tcmini, unsigned node, unsigned text, unsigned offset,
	  void **find)
{
    while (1) {
	unsigned diff, slip;
	void *extend;

	extend = M_TEXT(tcmini, node);

	slip = *(unsigned *) ((char *) extend + offset);

	diff = slip - text;
	if (!diff) {
	    *find = (char *) extend + 4;

	    break;
	} else {
	    node = L_NODE(Q_NODE(extend), (int) diff >> d_left);
	}
    }

    return 1;
}
#endif				/* __HERE_SELECT__ == __SIGN_SELECT__ */


#if __HERE_SELECT__ == __SIGN_SELECT__
static int
find_d1qp(void *tcmini, unsigned node, unsigned text, unsigned offset,
	  void **find)
{
    while (1) {
	unsigned slip;
	void *extend;

	extend = M_TEXT(tcmini, node);

	slip = *(unsigned *) ((char *) extend + offset);

	if (slip & d_high) {
	    node = L_LAST(Q_NODE(extend));
	} else {
	    unsigned diff;

	    diff = slip - text;

	    if (diff >> d_left) {
		node = L_NEXT(Q_NODE(extend));
	    } else {
		if (!diff) {
		    *find = (char *) extend + 4;

		    break;
		} else {
		    node = L_LAST(Q_NODE(extend));
		    if (1) {
			find_d1qb(tcmini, node, text, offset, find);
			if (1) {
			    break;
			}
		    }
		}
	    }
	}
    }

    return 1;
}
#endif				/* __HERE_SELECT__ == __SIGN_SELECT__ */


#if __HERE_SELECT__ == __DATA_SELECT__
static int
find_list(void *tcmini, unsigned node, unsigned text, unsigned offset,
	  void **find)
{
    while (1) {
	unsigned copy;
	void *extend;

	extend = M_TEXT(tcmini, node);

	copy = *(unsigned *) ((char *) extend + offset);

	if (text == copy) {
	    *find = (char *) extend + 4;

	    break;
	} else {
	    if (text < copy) {
		node = L_LAST(Q_NODE(extend));
	    } else {
		node = L_NEXT(Q_NODE(extend));
	    }
	}
    }

    return 1;
}
#endif				/* __HERE_SELECT__ == __DATA_SELECT__ */


int
x1f4_look_tcfare(void *tcmini, unsigned text, unsigned slip, void **find)
{
#if __HERE_SELECT__ == __DATA_SELECT__
    find_list(tcmini, P_ROOT(Q_NODE(tcmini)), text, slip + 4, find);
#else
    if (text & d_high) {
	find_d1dp(tcmini, P_ROOT(Q_NODE(tcmini)), text, slip + 4, find);
    } else {
	find_d1qp(tcmini, P_ROOT(Q_NODE(tcmini)), text, slip + 4, find);
    }
#endif				/* __HERE_SELECT__ == __DATA_SELECT__ */

    return 1;
}
