/*
 * tcmn.f.c
 * Copyright (C) 2009-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <string.h>

#include <tcmn-bits.h>
#include <tcmn-names.h>
#include <tcmn-types.h>

#if SIZEOF_VOID_P == SIZEOF_LONG
# define integral_q			unsigned long
#else
# define integral_q			unsigned
#endif				/* SIZEOF_VOID_P == SIZEOF_LONG */

#define __UNBALANCED__			1

#define true(e)				1

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

static void copy_data(void *, unsigned, unsigned, void **);
#if SIZEOF_UNSIGNED == 4
static void copy_deck(void *, unsigned, unsigned, void **);
#endif				/* SIZEOF_UNSIGNED == 4 */
static void type_list(void *, unsigned, unsigned, unsigned);
static void type_mini(void *, unsigned, unsigned, unsigned, unsigned);

static void
copy_data(void *tcmini, unsigned node, unsigned call, void **back)
{
    do {
	unsigned slip, text;

	text = Q_NODE(M_TEXT(tcmini, node));

	slip = L_LAST(text);
	if (slip) {
#if __UNBALANCED__
	    unsigned ever, miss;
#endif				/* __UNBALANCED__ */

#if __UNBALANCED__
	    miss = Q_NODE(M_TEXT(tcmini, slip));
#endif				/* __UNBALANCED__ */

#if __UNBALANCED__
	    ever = L_LAST(miss);
	    if (ever) {
		copy_data(tcmini, ever, call, back);
	    }
#endif				/* __UNBALANCED__ */

#if __UNBALANCED__
	    if (1) {
		char *u, *v;
		unsigned i;

		v = M_TEXT(tcmini, slip + 4);
		u = *back;

		u += 4;

		i = call;
		for (; i; i--) {
		    *u++ = *v++;
		}

		*back = u;
	    }
#else
	    copy_data(tcmini, slip, call, back);
#endif				/* __UNBALANCED__ */

#if __UNBALANCED__
	    ever = L_NEXT(miss);
	    if (ever) {
		copy_data(tcmini, ever, call, back);
	    }
#endif				/* __UNBALANCED__ */
	}

	if (1) {
	    char *u, *v;
	    unsigned i;

	    v = M_TEXT(tcmini, node + 4);
	    u = *back;

	    u += 4;

	    i = call;
	    for (; i; i--) {
		*u++ = *v++;
	    }

	    *back = u;
	}

	node = L_NEXT(text);
    } while (node);
}


#if SIZEOF_UNSIGNED == 4
static void
copy_deck(void *tcmini, unsigned node, unsigned call, void **back)
{
    do {
	unsigned slip, text;

	text = Q_NODE(M_TEXT(tcmini, node));

	slip = L_LAST(text);
	if (slip) {
# if __UNBALANCED__
	    unsigned ever, miss;
# endif				/* __UNBALANCED__ */

# if __UNBALANCED__
	    miss = Q_NODE(M_TEXT(tcmini, slip));
# endif				/* __UNBALANCED__ */

# if __UNBALANCED__
	    ever = L_LAST(miss);
	    if (ever) {
		copy_deck(tcmini, ever, call, back);
	    }
# endif				/* __UNBALANCED__ */

# if __UNBALANCED__
	    if (1) {
		unsigned i, *u, *v;

		v = M_TEXT(tcmini, slip + 4);
		u = *back;

		u++;

		i = call;
		for (; i; i--) {
		    *u++ = *v++;
		}

		*back = u;
	    }
# else
	    copy_deck(tcmini, slip, call, back);
# endif				/* __UNBALANCED__ */

# if __UNBALANCED__
	    ever = L_NEXT(miss);
	    if (ever) {
		copy_deck(tcmini, ever, call, back);
	    }
# endif				/* __UNBALANCED__ */
	}

	if (1) {
	    unsigned i, *u, *v;

	    v = M_TEXT(tcmini, node + 4);
	    u = *back;

	    u++;

	    i = call;
	    for (; i; i--) {
		*u++ = *v++;
	    }

	    *back = u;
	}

	node = L_NEXT(text);
    } while (node);
}
#endif				/* SIZEOF_UNSIGNED == 4 */


static void
type_list(void *tcmini, unsigned call, unsigned lead, unsigned ever)
{
    unsigned list, miss;
    void *slip;

    list = call * lead;

    S_NODE(tcmini, P_LIST(list) | M_ROOT(Q_NODE(tcmini)));

    slip = M_TEXT(tcmini, list);

    miss = ever - lead;
    miss--;
    for (; miss; miss--) {
	list += call;
	S_NODE(slip, list);
	slip = M_TEXT(slip, call);
    }
    {
	S_NODE(slip, 0);
    }
}


static void
type_mini(void *tcmini, unsigned call, unsigned lead, unsigned mini,
	  unsigned ever)
{
    while (1) {
	unsigned last, miss = 0, next, slip;

	last = mini - lead;
	if (last) {
	    slip = (lead + mini - 1) >> 1;

	    miss |= B_LAST(call * slip);

	    if (last ^ 1) {
		type_mini(tcmini, call, lead, slip, mini);
	    } else {
		S_NODE(M_TEXT(tcmini, call * slip), B_TILT(0));
	    }
	}

	next = ever - mini - 1;
	if (next) {
	    slip = (ever + mini + 1) >> 1;

	    miss |= B_NEXT(call * slip);

	    if (next ^ 1) {
	    } else {
		S_NODE(M_TEXT(tcmini, call * slip), B_TILT(0));
		if (1) {
		    slip = 0;
		}
	    }
	} else {
	    slip = 0;
	}

	if (last ^ next) {
#if 0
	    last |= last >> 1;
	    last |= last >> 2;
	    last |= last >> 4;
	    last |= last >> 8;

	    next |= next >> 1;
	    next |= next >> 2;
	    next |= next >> 4;
	    next |= next >> 8;

	    if (last ^ next) {
		if (last < next) {
		    miss |= B_TILT(1);
		} else {
		    miss |= B_TILT(-1);
		}
	    } else {
		miss |= B_TILT(0);
	    }
#else
	    if ((last ^ next) < (last & next)) {
		miss |= B_TILT(0);
	    } else {
		if (last < next) {
		    miss |= B_TILT(1);
		} else {
		    miss |= B_TILT(-1);
		}
	    }
#endif				/* 0 */
	} else {
	    miss |= B_TILT(0);
	}

	S_NODE(M_TEXT(tcmini, call * mini), miss);

	if (slip) {
	    lead = mini + 1;
	    mini = slip;
	} else {
	    break;
	}
    }
}


/*
 * tide is assumed 2 * n + 1
 */
int
x1f4_part_tcmini(void *tcmini, void *tfmini, unsigned tide, unsigned size,
		 void **split)
{
    int delete;

    if (P_LIST(Q_NODE(tcmini))) {
	delete = 0;
    } else {
	unsigned root;
	void *slip;

	delete = 1;

	slip = M_TEXT(tfmini, 4 + size);

#if SIZEOF_UNSIGNED == 4
	if ((size & 3) | ((integral_q) tcmini & 3)) {
#endif				/* SIZEOF_UNSIGNED == 4 */
	    copy_data(tcmini, P_ROOT(Q_NODE(tcmini)), size, &slip);
#if SIZEOF_UNSIGNED == 4
	} else {
	    copy_deck(tcmini, P_ROOT(Q_NODE(tcmini)), size >> 2, &slip);
	}
#endif				/* SIZEOF_UNSIGNED == 4 */

	size += 4;

	*split = M_TEXT(tfmini, ((tide >> 1) + 1) * size + 4);

	memcpy(tcmini, tfmini, ((tide >> 1) + 1) * size);

	root = ((tide >> 1) >> 1) + 1;

	S_NODE(tcmini, B_ROOT(size * root));

	type_mini(tcmini, size, 1, root, (tide >> 1) + 1);

	type_list(tcmini, size, (tide >> 1) + 1, tide + 1);

	root = tide - ((tide >> 1) >> 1);

	S_NODE(tfmini, B_ROOT(size * root));

	type_mini(tfmini, size, (tide >> 1) + 2, root, tide + 1);

	type_list(tfmini, size, 1, (tide >> 1) + 2);
    }

    return delete;
}
