/*
 * tfqp.i.c
 * Copyright (C) 2008-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <stdlib.h>

#include <tfqp-defs.h>
#include <tfqp-inter.h>
#include <tfqp-names.h>
#include <tfqp-types.h>

#define sLEFT(___size, ___node) \
    {									      \
	unsigned ___fine;						      \
									      \
	___fine = ___node->fqnode.fine;					      \
	if (___fine & 1) {						      \
	    unsigned ___swap;						      \
									      \
	    ___fine >>= 1;						      \
									      \
	    ___swap = ___size - 1 - ___fine;				      \
	    ___size = ___fine;						      \
									      \
	    ___node->fqnode.fine = ___swap << 1;			      \
	} else {							      \
	    ___size = ___size - 1 - (___fine >> 1);			      \
	}								      \
    }

#define sLESS(___size, ___node) \
    {									      \
	unsigned ___fine;						      \
									      \
	___fine = ___node->fqnode.fine;					      \
	if (___fine & 1) {						      \
	    ___size = ___size - 1 - (___fine >> 1);			      \
	} else {							      \
	    unsigned ___swap;						      \
									      \
	    ___fine >>= 1;						      \
									      \
	    ___swap = ___size - 1 - ___fine;				      \
	    ___size = ___fine;						      \
									      \
	    ___node->fqnode.fine = ___swap << 1 | 1;			      \
	}								      \
    }

#define xLEFT(___size, ___node) \
    {									      \
	unsigned ___fine;						      \
									      \
	___fine = ___node->fqnode.fine;					      \
	if (___fine & 1) {						      \
	    ___node->fqnode.fine = (___size - 1 - (___fine >> 1)) << 1;	      \
	}								      \
    }

#define xLESS(___size, ___node) \
    {									      \
	unsigned ___fine;						      \
									      \
	___fine = ___node->fqnode.fine;					      \
	if (___fine & 1) {						      \
	} else {							      \
	    ___node->fqnode.fine = (___size - 1 - (___fine >> 1)) << 1 | 1;   \
	}								      \
    }

#define tfline(tfline)			((struct tfline_type *) (tfline))

int
_libx1f4l2_elif_tfline(void *tfline, void *text, int (*land) (void *, void *),
		       unsigned offset, unsigned flat, void **push,
		       int (*line) (void *, unsigned, void **))
{
    int status;
    struct fpnode_type **fpmiss, *fpnode_data;
    unsigned size;

    fpmiss = tfline(tfline)->link_a.fpline.fplink.fpnode_data + 0;

    size = tfline(tfline)->link_a.fpline.fplink.fqnode.fine >> 1;

    fpnode_data = fpmiss[0];

    while (1) {
	unsigned lock;

	lock = rand() % (size + 1);
	if (lock == size) {
	    struct fpnode_type *fbnode_data, **fbmiss, *fdnode_data, **fdmiss,
		*ffnode_data = NULL;
	    unsigned deck = 0;

	    fbmiss = &fbnode_data;
	    fdmiss = &fdnode_data;

	    while (1) {
		int slip;

		if (fpnode_data) {
		} else {
		    *fbmiss = NULL;
		    *fdmiss = NULL;

		    break;
		}

		slip = land(text, (char *) fpnode_data + offset);

		if (slip) {
		    if (slip < 0) {
			sLESS(size, fpnode_data);

			*fdmiss = fpnode_data;
			fdmiss = fpnode_data->fpnode_data + 0;

			fpnode_data = fdmiss[0];
		    } else {
			sLEFT(size, fpnode_data);

			*fbmiss = fpnode_data;
			fbmiss = fpnode_data->fpnode_data + 1;

			deck++;
			deck += fpnode_data->fqnode.fine >> 1;

			fpnode_data = fbmiss[0];
		    }
		} else {
		    unsigned fine;

		    ffnode_data = fpnode_data;

		    *fbmiss = fpnode_data->fpnode_data[0];
		    *fdmiss = fpnode_data->fpnode_data[1];

		    fine = fpnode_data->fqnode.fine;

		    if (fine & 1) {
			deck += size - (fine >> 1) - 1;
		    } else {
			deck += fine >> 1;
		    }

		    break;
		}
	    }

	    if (ffnode_data) {
		status = EVER_MATCH;
	    } else {
		void *tfnode;

		status = line(tfline, flat, &tfnode);
		if (status) {
		    _libx1f4l2_join_tfline(fpmiss, fbnode_data, fdnode_data);

		    break;
		} else {
		    ffnode_data = tfnode;

		    tfline(tfline)->link_a.fpline.fplink.fqnode.fine += 2;
		}
	    }

	    ffnode_data->fpnode_data[0] = fbnode_data;
	    ffnode_data->fpnode_data[1] = fdnode_data;

	    ffnode_data->fqnode.fine = deck << 1;

	    *fpmiss = ffnode_data;

	    *push = ffnode_data + 1;

	    break;
	} else {
	    int slip;

	    slip = land(text, (char *) fpnode_data + offset);

	    if (slip) {
		if (slip < 0) {
		    sLESS(size, fpnode_data);

		    *fpmiss = fpnode_data;
		    fpmiss = fpnode_data->fpnode_data + 0;

		    fpnode_data = fpmiss[0];
		} else {
		    sLEFT(size, fpnode_data);

		    *fpmiss = fpnode_data;
		    fpmiss = fpnode_data->fpnode_data + 1;

		    fpnode_data = fpmiss[0];
		}
	    } else {
		while (1) {
		    xLEFT(size, fpnode_data);

		    lock = rand() % size;
		    if (lock < fpnode_data->fqnode.fine >> 1) {
			struct fpnode_type *fnnode_data, *fqnode_data;
			unsigned fast;

			fnnode_data = fpnode_data->fpnode_data[0];
			fqnode_data = fnnode_data->fpnode_data[1];
			fpnode_data->fpnode_data[0] = fqnode_data;

			fast = fpnode_data->fqnode.fine >> 1;

			xLESS(size, fpnode_data);

			xLEFT(fast, fnnode_data);

			fast = fnnode_data->fqnode.fine >> 1;

			size = size - 1 - fast;

			*fpmiss = fnnode_data;
			fpmiss = fnnode_data->fpnode_data + 1;
		    } else {
			if (lock < size - 1) {
			    struct fpnode_type *fnnode_data, *fqnode_data;
			    unsigned fast;

			    fnnode_data = fpnode_data->fpnode_data[1];
			    fqnode_data = fnnode_data->fpnode_data[0];
			    fpnode_data->fpnode_data[1] = fqnode_data;

			    fast = size - 1 - (fpnode_data->fqnode.fine >> 1);

			    xLESS(fast, fnnode_data);

			    fast = fnnode_data->fqnode.fine >> 1;

			    size = size - 1 - fast;

			    *fpmiss = fnnode_data;
			    fpmiss = fnnode_data->fpnode_data + 0;
			} else {
			    *fpmiss = fpnode_data;

			    break;
			}
		    }
		}

		status = EVER_MATCH;

		*push = fpnode_data + 1;

		break;
	    }
	}
    }

    return status;
}
