/*
 * tsline.h
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_TSLINE_H__
#define __LIBx1f4_TSLINE_H__

#include <errors.h>

/*
 * attribute bit definitions
 */
#define X1f4_TSLINE_TRANS_MASK		(1u << 0x00)

/*
 * error definitions
 */
#define X1f4_TSLINE_DECK_ERROR		LIBx1f4L2_DECK_ERROR
#define X1f4_TSLINE_FREE_ERROR		LIBx1f4L2_FREE_ERROR
#define X1f4_TSLINE_LINK_ERROR		LIBx1f4L2_LINK_ERROR
#define X1f4_TSLINE_MODE_ERROR		LIBx1f4L2_MODE_ERROR

/*
 * error definitions (continued)
 */
#define X1f4_TSLINE_EVER_MATCH		LIBx1f4L2_EVER_MATCH
#define X1f4_TSLINE_MISS_CLASS		LIBx1f4L2_MISS_CLASS

typedef struct x1f4_tsline_type {
    void *trans;
} x1f4_tsline_type;

#if !defined __x1f4_trans_type_defined__
typedef struct x1f4_trans_type {
    int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	(*mode) (void *, void **, unsigned);
    void *context;
} x1f4_trans_type;
#endif				/* !__x1f4_trans_type_defined__ */

#if !defined __x1f4_trans_type_defined__
# define __x1f4_trans_type_defined__
#endif				/* !__x1f4_trans_type_defined__ */

extern int x1f4_base_tsline(unsigned *);
extern int x1f4_call_tsline(unsigned *);
extern int x1f4_case_tsline(void *);
extern int x1f4_cast_tsline
    (void *, void *, int (*) (void *, void *), unsigned);
extern int x1f4_ever_tsline(void *, void **);
extern int x1f4_fast_tsline(void *, unsigned, struct x1f4_tsline_type *);
extern int x1f4_fgxd_tsline
    (void *, void *, unsigned, int (*) (void *, void *), void **);
extern int x1f4_find_tsline
    (void *, void *, unsigned, int (*) (void *, void *), void **);
extern int x1f4_fini_tsline(void **);
extern int x1f4_flat_tsline(void *);
extern int x1f4_fmxd_tsline
    (void *, void *, unsigned, int (*) (void *, void *), void **);
extern int x1f4_fxxd_tsline
    (void *, void *, unsigned, int (*) (void *, void *), void **);
extern int x1f4_high_tsline(void *);
extern int x1f4_init_tsline(void **, unsigned, struct x1f4_tsline_type *);
extern int x1f4_land_tsline
    (void *, void *, int (*) (void *, void *), unsigned, unsigned, void **);
extern int x1f4_last_tsline(void **, void *);
extern int x1f4_lead_tsline(void *, void **);
extern int x1f4_lime_tsline(void *, void *, int (*) (void *, void *));
extern int x1f4_llxd_tsline
    (void *, void *, unsigned, int (*) (void *, void *), void **);
extern int x1f4_lmxd_tsline
    (void *, void *, unsigned, int (*) (void *, void *), void **);
extern int x1f4_lxxd_tsline
    (void *, void *, unsigned, int (*) (void *, void *), void **);
extern int x1f4_near_tsline
    (void *, void *, int (*) (void *, void *), unsigned, void *,
     int (*) (void *, void *));
extern int x1f4_next_tsline(void **, void *);
extern int x1f4_pass_tsline
    (void *, void *, int (*) (void *, void *), unsigned);
extern int x1f4_post_tsline
    (void *, void *, int (*) (void *, void *), unsigned, unsigned, void **);
extern int x1f4_scan_tsline
    (void *, void *, int (*) (void *, void *, unsigned));
extern int x1f4_shut_tsline
    (void *, void *, int (*) (void *, void *), unsigned, void *,
     int (*) (void *, void *));
extern int x1f4_side_tsline(void *, void *);
extern int x1f4_skew_tsline(void *, void *);

#endif				/* __LIBx1f4_TSLINE_H__ */
