/*
 * tvxx.c.c
 * Copyright (C) 2008-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <tvxx-defs.h>
#include <tvxx-inter.h>
#include <tvxx-names.h>
#include <tvxx-types.h>
#include <trans.h>

static int fini_list(void **);
static int init_list(void **, struct tvlist_type *);
static int text_list(struct tvlist_type *, struct tvlist_type *,
		     int (*) (void *, unsigned *, const void *),
		     int (*) (void *, void *, const void *),
		     int (*) (void *, void *), void *);

static int
fini_list(void **text)
{
    return _libx1f4l2_fini_tvlist(text);
}


static int
init_list(void **text, struct tvlist_type *tvlist_data)
{
    int status;
    struct screen_type screen;
    struct trans_type trans;

    screen.trans_data = &trans;

    trans.data = tvlist_data->link_m.data;
    trans.free = tvlist_data->link_m.free;
    trans.link = tvlist_data->link_m.link;
    trans.mode = tvlist_data->link_m.mode;

    status = _libx1f4l2_init_tvlist(text, TRANS_MASK, &screen);

    return status;
}


static int
text_list(struct tvlist_type *tvlist_data, struct tvlist_type *tvlist_text,
	  int (*size) (void *, unsigned *, const void *),
	  int (*copy) (void *, void *, const void *),
	  int (*free) (void *, void *), void *context)
{
    int status;

    if (tvlist_text->link_a.fplist.fplink.fqnode.size) {
	status = _libx1f4l2_rail_tvlist
	    (tvlist_data, tvlist_text, size, copy, free, context);
    } else {
	status = 0;
    }

    return status;
}


int
x1f4_text_tvlist(void **tvlist, void *tctext,
		 int (*size) (void *, unsigned *, const void *),
		 int (*copy) (void *, void *, const void *),
		 int (*free) (void *, void *), void *context)
{
    int status;
    void *text;

    status = init_list(&text, tctext);
    if (status) {
    } else {
	status = text_list(text, tctext, size, copy, free, context);
	if (status) {
	    fini_list(&text);
	} else {
	    *tvlist = text;
	}
    }

    return status;
}
