/*
 * tvxx.o.c
 * Copyright (C) 2008-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <tvxx-defs.h>
#include <tvxx-names.h>
#include <tvxx-types.h>

#define tvlist(tvlist)			((struct tvlist_type *) (tvlist))

static int copy_miss(struct tvlist_type *, struct fpnode_type *,
		     struct fpnode_type *,
		     int (*) (void *, unsigned *, const void *),
		     int (*) (void *, void *, const void *),
		     int (*) (void *, void *), void *);
static int copy_node(struct tvlist_type *, struct fpnode_type **,
		     struct fpnode_type *,
		     int (*) (void *, unsigned *, const void *),
		     int (*) (void *, void *, const void *),
		     int (*) (void *, void *), void *);
static int free_node(struct tvlist_type *, struct fpnode_type *,
		     int (*) (void *, void *), void *);

static int
copy_miss(struct tvlist_type *tvlist_data, struct fpnode_type *fpnode_copy,
	  struct fpnode_type *fpnode_data,
	  int (*size) (void *, unsigned *, const void *),
	  int (*copy) (void *, void *, const void *),
	  int (*free) (void *, void *), void *context)
{
    int status;

    do {
	struct fpnode_type *fpnode_text;

	fpnode_text = fpnode_data->fpnode_data[0];
	if (fpnode_text) {
	    status = copy_node
		(tvlist_data, fpnode_copy->fpnode_data + 0, fpnode_text, size,
		 copy, free, context);
	    if (status) {
		break;
	    } else {
		fpnode_copy->fpnode_data[0]->uplink_data = fpnode_copy;
	    }
	} else {
	    fpnode_copy->fpnode_data[0] = NULL;
	}

	fpnode_text = fpnode_data->fpnode_data[1];
	if (fpnode_text) {
	    status = copy_node
		(tvlist_data, fpnode_copy->fpnode_data + 1, fpnode_text, size,
		 copy, free, context);
	    if (status) {
		fpnode_text = fpnode_copy->fpnode_data[0];
		if (fpnode_text) {
		    free_node(tvlist_data, fpnode_text, free, context);
		}
	    } else {
		fpnode_copy->fpnode_data[1]->uplink_data = fpnode_copy;
	    }
	} else {
	    status = 0;

	    fpnode_copy->fpnode_data[1] = NULL;
	}
    } while (0);

    return status;
}


static int
copy_node(struct tvlist_type *tvlist_data, struct fpnode_type **fpnode,
	  struct fpnode_type *fpnode_data,
	  int (*size) (void *, unsigned *, const void *),
	  int (*copy) (void *, void *, const void *),
	  int (*free) (void *, void *), void *context)
{
    int status;
    unsigned down;
    void *tvnode;

    size(context, &down, fpnode_data + 1);

    status = tvlist_data->link_m.link
	(tvlist_data->link_m.data, &tvnode, sizeof(struct fpnode_type) + down);
    if (status) {
	status = LINK_ERROR;
    } else {
	struct fpnode_type *fpnode_copy;

	fpnode_copy = tvnode;

	status = copy(context, fpnode_copy + 1, fpnode_data + 1);
	if (status) {
	    status = CALL_ERROR;

	    tvlist_data->link_m.free(tvlist_data->link_m.data, tvnode);
	} else {
	    status = copy_miss
		(tvlist_data, fpnode_copy, fpnode_data, size, copy, free,
		 context);
	    if (status) {
		free(context, fpnode_copy + 1);

		tvlist_data->link_m.free(tvlist_data->link_m.data, tvnode);
	    } else {
		fpnode_copy->fqnode = fpnode_data->fqnode;

		*fpnode = tvnode;
	    }
	}
    }

    return status;
}


static int
free_node(struct tvlist_type *tvlist_data, struct fpnode_type *fpnode_data,
	  int (*free) (void *, void *), void *context)
{
    struct fpnode_type *fpnode_text;

    fpnode_text = fpnode_data->fpnode_data[0];
    if (fpnode_text) {
	free_node(tvlist_data, fpnode_text, free, context);
    }

    fpnode_text = fpnode_data->fpnode_data[1];
    if (fpnode_text) {
	free_node(tvlist_data, fpnode_text, free, context);
    }

    free(context, fpnode_data + 1);

    tvlist_data->link_m.free(tvlist_data->link_m.data, fpnode_data);

    return -1;
}


int
_libx1f4l2_rail_tvlist(void *tcdata, void *tctext,
		       int (*size) (void *, unsigned *, const void *),
		       int (*copy) (void *, void *, const void *),
		       int (*free) (void *, void *), void *context)
{
    int status;

    status = copy_node
	(tcdata, tvlist(tcdata)->link_a.fplist.fplink.fpnode_data + 0,
	 tvlist(tctext)->link_a.fplist.fplink.fpnode_data[0], size, copy, free,
	 context);
    if (status) {
    } else {
	tvlist(tcdata)->link_a.fplist.fplink.fpnode_data[0]->uplink_data =
	    NULL;
	tvlist(tcdata)->link_a.fplist.fplink.fqnode =
	    tvlist(tctext)->link_a.fplist.fplink.fqnode;
    }

    return status;
}
