/*
 * bqfa.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <errno.h>
#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <count2.h>
#include <inter.h>
#include <libx1f4l2.h>

#if SIZEOF_VOID_P == SIZEOF_LONG
# define integral_q			unsigned long
#else
# define integral_q			unsigned
#endif				/* SIZEOF_VOID_P == SIZEOF_LONG */

#if SIZEOF_VOID_P == 8
# define HALF				040
#else
# define HALF				020
#endif				/* SIZEOF_VOID_P == 8 */

#define LX				16

#define line(pick) \
    ((struct line_type *) (pick))

typedef struct line_type {
    const char *data;
    unsigned size;
} line_type;

typedef struct context_type {
    const char *self;
    int do_profile, do_reverse;
    unsigned size;
} context_type;

static int deck_miss(void *, void *);
static int free_list(void *);
static int lain_look(void *, void *, unsigned *);
static int lain_fare(void *, void *);
static int lain_move(void *, void *, unsigned);
static int lain_pick(void *, void *, unsigned *);
static int land_list(struct context_type *, void *);
static int lime_list(struct context_type *, void *);
static int post_line(struct context_type *, void *, char *, unsigned);
static int read_list(struct context_type *, void *);
static int slip_data(void *, void *);
static int slip_flat(void *, void *);

static void usage(void);

static int
deck_miss(void *deck, void *miss)
{
    int delete;
    unsigned char *s, *v;

    s = deck;
    v = miss;

    while (1) {
	unsigned char c;

	c = *s;
	if (c ^ *v) {
	    delete = (int) c - (int) *v;

	    break;
	} else {
	    if (c) {
		s++;
		v++;
	    } else {
		delete = 0;

		break;
	    }
	}
    }

    return delete;
}


static int
free_list(void *bqfx)
{
    return x1f4_lime_bqfset(bqfx, NULL, slip_flat);
}


static int
lain_fare(void *text, void *node)
{
    int status;
    unsigned lead;

    lead = *(unsigned char *) node;
    if (lead == LX - 1) {
	unsigned rule;

	rule = line(text)->size;
	if (rule < LX - 1 - sizeof(char *)) {
	    status = memcmp(line(text)->data, (char *) node + 1, rule + 1);
	} else {
	    status = memcmp
		(line(text)->data, (char *) node + 1, LX - 1 - sizeof(char *));
	    if (status) {
	    } else {
		status = deck_miss
		    ((char *) line(text)->data + LX - 1 - sizeof(char *),
		     *(char **) ((char *) node + LX - sizeof(char *)));
	    }
	}
    } else {
	unsigned rule;

	rule = line(text)->size;
	if (lead < rule) {
	    rule = lead;
	}

	status = memcmp(line(text)->data, (char *) node + 1, rule + 1);
    }

    return status;
}


static int
lain_look(void *text, void *node, unsigned *pick)
{
    int status;
    unsigned size;

    size = *(integral_q *) node;
    if (size) {
	unsigned look = 1;
	void *call;

	call = (char *) node + LX;

	size--;
	while (size) {
	    unsigned half;
	    void *slip;

	    half = (size + 1) >> 1;
	    slip = (char *) call + LX * half;
	    if (lain_fare(text, slip) < 0) {
		size = half - 1;
	    } else {
		look += half;
		call = slip;
		size -= half;
	    }
	}

	status = lain_fare(text, call);
	if (status) {
	    if (status < 0) {
		look--;
	    } else {
	    }

	    status = 0;

	    *pick = look;
	} else {
	    status = 1;

	    *pick = look;
	}
    } else {
	status = 0;

	*pick = 0;
    }

    return status;
}


static int
lain_move(void *target, void *source, unsigned count)
{
    memmove(target, source, count * LX);

    return 0;
}


static int
lain_pick(void *text, void *node, unsigned *pick)
{
    int record = 1, status;
    unsigned half = HALF, slip = 0;

    do {
	int effect;

	effect = lain_fare(text, (char *) node + LX * (slip + half));
	if (effect < 0) {
	} else {
	    slip += half;
	    record = effect;
	}

	half >>= 1;
    } while (half ^ 1);

    if (*(integral_q *) node & (integral_q) 2 << slip) {
	int effect;

	effect = lain_fare(text, (char *) node + LX * (slip + 1));
	if (effect < 0) {
	    *pick = slip + 0;
	    status = !record;
	} else {
	    *pick = slip + 1;
	    status = !effect;
	}
    } else {
	*pick = slip + 1;
	status = !record;
    }

    return status;
}


static int
land_list(struct context_type *context_data, void *list)
{
    count2_decls;
    int status;

    if (context_data->do_profile) {
	start_count2();
    }

    status = read_list(context_data, list);

    if (status) {
    } else {
	if (context_data->do_profile) {
	    finish_count2_l("delete", 55);
	}
    }

    return status;
}


static int
lime_list(struct context_type *context_data, void *bqfx)
{
    int status;

    status = x1f4_lime_bqfset(bqfx, NULL, slip_data);
    if (status) {
	fprintf(stderr, "%s: cannot traverse data\n", context_data->self);
    } else {
    }

    return status;
}


static int
post_line(struct context_type *context_data, void *bqfx, char *data,
	  unsigned size)
{
    int status;
    struct line_type line;
    void *post;

    line.data = data;
    line.size = size;

    status = x1f4_post_bqfset(bqfx, &line, &post);
    if (status) {
	fprintf(stderr, "%s: cannot insert data\n", context_data->self);
	if (status == X1f4_BQFSET_EVER_MATCH) {
	    fprintf(stderr, "%s: `%s' must be already in set\n",
		    context_data->self, data);
	} else {
	    fprintf(stderr, "%s: %s\n", context_data->self, strerror(errno));
	}
    } else {
	unsigned char *text;

	text = post;
	if (size < LX - 1) {
	    *text = size;
	    memcpy(text + 1, data, size + 1);
	} else {
	    unsigned side;

	    side = size + sizeof(char *) + 2 - LX;
	    post = malloc(side);
	    if (post) {
		void **tail;

		*text = LX - 1;
		memcpy(text + 1, data, LX - 1 - sizeof(char *));
		text += LX - sizeof(char *);
		tail = (void *) text;
		*tail = post;
		memcpy(post, data + LX - 1 - sizeof(char *), side);
	    } else {
		*text = 1;
		status = 1;
		fprintf(stderr, "%s: %s\n", context_data->self,
			strerror(errno));
	    }
	}
    }

    return status;
}


static int
read_list(struct context_type *context_data, void *list)
{
    int status;

    while (1) {
	char *data;
	int done = 0;
	unsigned size = 0;

	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (!done || size) {
		status = post_line(context_data, list, data, size);
	    }

	    if (1) {
		free(data);
	    }

	    if (status) {
		break;
	    }

	    if (done) {
		break;
	    }
	}
    }

    return status;
}


static int
slip_data(void *slip, void *data)
{
    unsigned lead;

    lead = *(unsigned char *) data;

    if (lead ^ (LX - 1)) {
	fputs((char *) data + 1, stdout);
    } else {
	fwrite((char *) data + 1, LX - 1 - sizeof(char *), 1, stdout);
	fputs(*(char **) ((char *) data + LX - sizeof(char *)), stdout);
    }

    fputc('\n', stdout);

    return 0;
}


static int
slip_flat(void *slip, void *data)
{
    if (*(unsigned char *) data ^ (LX - 1)) {
    } else {
	free(*(char **) ((char *) data + LX - sizeof(char *)));
    }

    return 0;
}


static void
usage(void)
{
    puts("Usage: bqfa [OPTIONS]\n\
Test the BQ generic set, output sorted input.\n\
\n\
Options:\n\
  -p, --profile			time operations, suppress output\n\
  -r, --reverse			remove set items in insertion reverted order\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status;
    struct context_type context;

    context.do_profile = 0;
    context.do_reverse = 0;

    unsetenv("POSIXLY_CORRECT");

    {
	const char *size_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "profile",	    0x00,   NULL,   0x70    },
		{   "reverse",	    0x00,   NULL,   0x72    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "pr", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'N':
		size_specs = optarg;
		break;
	    case 'p':
		context.do_profile = 1;
		break;
	    case 'r':
		context.do_reverse = 1;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	struct x1f4_bqfset_type bqfset;
	void *bqfx;

	if (1) {
	    context.size = LX;
	} else {
	}

	if (1) {
	    bqfset.size = LX;
	} else {
	}

	if (1) {
	    bqfset.fare = lain_fare;
	    bqfset.look = lain_look;
	    bqfset.move = lain_move;
	    bqfset.pick = lain_pick;
	} else {
	}

	status = x1f4_init_bqfset(&bqfx, X1f4_BQFSET_SIZE_FRAME, &bqfset);
	if (status) {
	    perror(argv[0]);
	} else {
	    int excess;

	    context.self = argv[0];

	    status = land_list(&context, bqfx);
	    if (status) {
	    } else {
		if (context.do_profile) {
		} else {
		    status = lime_list(&context, bqfx);
		}
	    }

	    free_list(bqfx);

	    excess = x1f4_fini_bqfset(&bqfx);
	    if (excess) {
		perror(argv[0]);

		if (status) {
		} else {
		    status = excess;
		}
	    }
	}
    }

    return status;
}
