/*
 * bqfm.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/time.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <count2.h>
#include <libx1f4l2.h>

#define __UNROLL_PICK__			0

#define __SIMPLE_PICK__			0

#if SIZEOF_VOID_P == SIZEOF_LONG
# define integral_q			unsigned long
#else
# define integral_q			unsigned
#endif				/* SIZEOF_VOID_P == SIZEOF_LONG */

#if SIZEOF_VOID_P == 8
# define HALF				040
#else
# define HALF				020
#endif				/* SIZEOF_VOID_P == 8 */

#define LX				10

#define VX				65536

typedef struct context_type {
    const char *self;
    int do_profile, do_regular, forward;
    unsigned count, seed, size;
} context_type;

static int cast_list(struct context_type *, void *, void *);
static int free_list(void *);
static int lain_look(void *, void *, unsigned *);
static int lain_fare(void *, void *);
static int lain_move(void *, void *, unsigned);
static int lain_pick(void *, void *, unsigned *);
static int link_list(struct context_type *, void **);
static int miss_list(struct context_type *, void *, void *);
static int pass_list(struct context_type *, void *, void *);
static int post_list(struct context_type *, void *, void *, void *);
static int sort_days(const void *, const void *);
static int sort_fare(const void *, const void *);
static int time_list(struct context_type *);
static int trap_list(struct context_type *, void *);
static int type_list(struct context_type *);

static void usage(void);

static const char digits[] = "0123456789";

static int
cast_list(struct context_type *context_data, void *bqfx, void *deck)
{
    int shift, status;
    unsigned class, i;
    char *data;

    data = malloc(1);
    if (!data) {
	perror(context_data->self);
	status = -1;
    } else {
	shift = context_data->size;

	class = 1;

	i = context_data->count;
	while (i) {
	    const void *pick;
	    unsigned j;

	    i--;
	    status = x1f4_lift_scdeck(deck, &pick, &data, &j, &class);
	    if (status) {
	    } else {
		if (pick) {
		    status = 1;
		} else {
		    if (j != shift) {
			status = 1;
		    }
		}
	    }
	    if (status) {
		fprintf(stderr, "%s: cannot retrieve reference\n",
			context_data->self);
		if (1) {
		    break;
		}
	    } else {
		void *miss;

		status = x1f4_lead_bqfset(bqfx, &miss);
		if (status) {
		    fprintf(stderr, "%s: cannot find data\n",
			    context_data->self);
		    if (1) {
			break;
		    }
		} else {
		    if (memcmp(miss, data, shift)) {
			fprintf(stderr, "%s: cannot match data\n",
				context_data->self);

			status = 1;

			if (1) {
			    break;
			}
		    } else {
			status = x1f4_mind_bqfset(bqfx);
			if (status) {
			    fprintf(stderr, "%s: cannot delete data\n",
				    context_data->self);
			    if (1) {
				break;
			    }
			} else {
			    status = x1f4_mind_scdeck(deck);
			    if (status) {
				fprintf(stderr, "%s: cannot delete reference"
					"\n", context_data->self);
				if (1) {
				    break;
				}
			    }
			}
		    }
		}
	    }
	}

	if (i) {
	    status = 1;
	} else {
	    if (x1f4_case_bqfset(bqfx)) {
		fprintf(stderr, "%s: set not void\n", context_data->self);

		status = 1;
	    } else {
		status = 0;
	    }
	}

	free(data);
    }

    return status;
}


static int
free_list(void *list)
{
    free(list);

    return 0;
}


static int
lain_fare(void *text, void *node)
{
    return memcmp(text, node, LX);
}


static int
lain_look(void *text, void *node, unsigned *pick)
{
#if 0
    int status;
    unsigned size;

    size = *(integral_q *) node;
    if (size) {
	unsigned look = 1;
	void *call;

	call = (char *) node + LX;

	size--;
	while (size) {
	    unsigned half;
	    void *slip;

	    half = (size + 1) >> 1;
	    slip = (char *) call + LX * half;
	    if (memcmp(text, slip, LX) < 0) {
		size = half - 1;
	    } else {
		look += half;
		call = slip;
		size -= half;
	    }
	}

	if (memcmp(call, text, LX)) {
	    status = 0;

	    if (memcmp(text, call, LX) < 0) {
		look--;
	    } else {
	    }

	    *pick = look;
	} else {
	    status = 1;

	    *pick = look;
	}
    } else {
	status = 0;

	*pick = 0;
    }

    return status;
#else
    int effect, status;
    unsigned size;
    unsigned const char *aime, *line;

    size = *(integral_q *) node;
    if (size) {
	unsigned i, look = 1;
	void *call;

	call = (char *) node + LX;

	size--;
	while (size) {
	    unsigned half;
	    void *slip;

	    half = (size + 1) >> 1;
	    slip = (char *) call + LX * half;
	    aime = text;
	    line = slip;
	    i = LX;
	    do {
		effect = (int) *aime - (int) *line;
		if (effect) {
		    break;
		}

		i--;
		aime++;
		line++;
	    } while (i);
	    if (effect < 0) {
		size = half - 1;
	    } else {
		look += half;
		call = slip;
		size -= half;
	    }
	}

	aime = text;
	line = call;
	i = LX;
	do {
	    effect = (int) *aime - (int) *line;
	    if (effect) {
		break;
	    }

	    i--;
	    aime++;
	    line++;
	} while (i);
	if (effect) {
	    status = 0;

	    if (effect < 0) {
		look--;
	    } else {
	    }

	    *pick = look;
	} else {
	    status = 1;

	    *pick = look;
	}
    } else {
	status = 0;

	*pick = 0;
    }

    return status;
#endif				/* 0 */
}


static int
lain_move(void *target, void *source, unsigned count)
{
    memmove(target, source, count * LX);

    return 0;
}


static int
lain_pick(void *text, void *node, unsigned *pick)
{
#if __SIMPLE_PICK__
    int status;
    unsigned half = HALF, slip = 0;

    do {
	if (memcmp(text, (char *) node + LX * (slip + half), LX) < 0) {
	} else {
	    slip += half;
	}

	half >>= 1;
    } while (half ^ 1);

    if (*(integral_q *) node & (integral_q) 2 << slip) {
	if (memcmp(text, (char *) node + LX * (slip + 1), LX) < 0) {
	    *pick = slip + 0;
	    if (slip) {
		if (memcmp(text, (char *) node + LX * (slip + 0), LX)) {
		    status = 0;
		} else {
		    status = 1;
		}
	    } else {
		status = 0;
	    }
	} else {
	    *pick = slip + 1;
	    if (memcmp(text, (char *) node + LX * (slip + 1), LX)) {
		status = 0;
	    } else {
		status = 1;
	    }
	}
    } else {
	*pick = slip + 1;
	if (slip) {
	    if (memcmp(text, (char *) node + LX * (slip + 0), LX)) {
		status = 0;
	    } else {
		status = 1;
	    }
	} else {
	    if (1) {
		status = 0;
	    }
	}
    }

    return status;
#elif 0
    int record = 1, status;
    unsigned half = HALF, slip = 0;

    do {
	int effect;

	effect = memcmp(text, (char *) node + LX * (slip + half), LX);
	if (effect < 0) {
	} else {
	    slip += half;
	    record = effect;
	}

	half >>= 1;
    } while (half ^ 1);

    if (*(integral_q *) node & (integral_q) 2 << slip) {
	int effect;

	effect = memcmp(text, (char *) node + LX * (slip + 1), LX);
	if (effect < 0) {
	    *pick = slip + 0;
	    status = !record;
	} else {
	    *pick = slip + 1;
	    status = !effect;
	}
    } else {
	*pick = slip + 1;
	status = !record;
    }

    return status;
#else
    int record = 1, status;
    unsigned half = HALF, slip = 0;
    unsigned const char *aime, *line;

    do {
	int effect;
	unsigned i;

	aime = text;
	line = (unsigned char *) node + LX * (slip + half);
	i = LX;
	do {
	    effect = (int) *aime - (int) *line;
	    if (effect) {
		break;
	    }

	    i--;
	    aime++;
	    line++;
	} while (i);
	if (effect < 0) {
	} else {
	    slip += half;
	    record = effect;
	}

	half >>= 1;
    } while (half ^ 1);

    if (*(integral_q *) node & (integral_q) 2 << slip) {
	int effect;
	unsigned i;

	aime = text;
	line = (unsigned char *) node + LX * (slip + 1);
	i = LX;
	do {
	    effect = (int) *aime - (int) *line;
	    if (effect) {
		break;
	    }

	    i--;
	    aime++;
	    line++;
	} while (i);
	if (effect < 0) {
	    *pick = slip + 0;
	    status = !record;
	} else {
	    *pick = slip + 1;
	    status = !effect;
	}
    } else {
	*pick = slip + 1;
	status = !record;
    }

    return status;
#endif				/* __SIMPLE_PICK__ */
}


static int
link_list(struct context_type *context_data, void **link)
{
    int status;
    void *list;

    list = malloc(LX * context_data->count);
    if (list) {
	unsigned i, v;

	v = context_data->seed;

	status = 0;

	*link = list;

	i = context_data->count;
	while (i) {
	    char *o;
	    unsigned e;

	    e = i * 523 + 342;

	    v *= 1033;

	    i--;

	    e ^= v;

#if SIZEOF_UNSIGNED != 4
	    e &= 0xffffffffu;
#endif				/* SIZEOF_UNSIGNED != 4 */

	    e = e >> 16 | e << 16;

	    o = (char *) list + LX;

	    do {
		--o;
		*o = digits[e % 10];
		e /= 10;
	    } while (e);
	    while (o != list) {
		--o;
		*o = 32;
	    }

	    list = (char *) list + LX;
	}
    } else {
	status = -1;
    }

    return status;
}


static int
miss_list(struct context_type *context_data, void *bqfx, void *list)
{
    int shift, status;
    unsigned i;

    shift = context_data->size;

    i = context_data->count;
    while (i) {
	i--;
	status = x1f4_cast_bqfset(bqfx, list);
	if (status) {
	    fprintf(stderr, "%s: cannot delete data\n",
		    context_data->self);
	    if (1) {
		break;
	    }
	} else {
	    list = (char *) list + shift;
	}
    }

    if (i) {
	status = 1;
    } else {
	if (x1f4_case_bqfset(bqfx)) {
	    fprintf(stderr, "%s: set not void\n", context_data->self);

	    status = 1;
	} else {
	    status = 0;
	}
    }

    return status;
}


static int
pass_list(struct context_type *context_data, void *bqfx, void *deck)
{
    int shift, status;
    unsigned class, i;
    char *data;

    data = malloc(1);
    if (!data) {
	perror(context_data->self);
	status = -1;
    } else {
	shift = context_data->size;

	class = 1;

	i = context_data->count;
	while (i) {
	    const void *pick;
	    unsigned j;

	    i--;
	    status = x1f4_plot_scdeck(deck, &pick, &data, &j, &class);
	    if (status) {
	    } else {
		if (pick) {
		    status = 1;
		} else {
		    if (j != shift) {
			status = 1;
		    }
		}
	    }
	    if (status) {
		fprintf(stderr, "%s: cannot retrieve reference\n",
			context_data->self);
		if (1) {
		    break;
		}
	    } else {
		void *miss;

		status = x1f4_ever_bqfset(bqfx, &miss);
		if (status) {
		    fprintf(stderr, "%s: cannot find data\n",
			    context_data->self);
		    if (1) {
			break;
		    }
		} else {
		    if (memcmp(miss, data, shift)) {
			fprintf(stderr, "%s: cannot match data (expected `%."
				"*s', found `%.*s')\n", context_data->self, LX,
				data, LX, (char *) miss);

			status = 1;

			if (1) {
			    break;
			}
		    } else {
			status = x1f4_slip_bqfset(bqfx);
			if (status) {
			    fprintf(stderr, "%s: cannot delete data\n",
				    context_data->self);
			    if (1) {
				break;
			    }
			} else {
			    status = x1f4_slip_scdeck(deck);
			    if (status) {
				fprintf(stderr, "%s: cannot delete reference"
					"\n", context_data->self);
				if (1) {
				    break;
				}
			    }
			}
		    }
		}
	    }
	}

	if (i) {
	    status = 1;
	} else {
	    if (x1f4_case_bqfset(bqfx)) {
		fprintf(stderr, "%s: set not void\n", context_data->self);

		status = 1;
	    } else {
		status = 0;
	    }
	}

	free(data);
    }

    return status;
}


static int
post_list(struct context_type *context_data, void *bqfx, void *deck,
	  void *list)
{
    int shift, status;
    unsigned i;

    shift = context_data->size;

    i = context_data->count;
    while (i) {
	void *post;

	i--;
	status = x1f4_post_bqfset(bqfx, list, &post);
	if (status) {
	    fprintf(stderr, "%s: cannot insert data\n", context_data->self);
	    if (1) {
		break;
	    }
	} else {
	    memcpy(post, list, shift);

	    if (deck) {
		status = x1f4_post_scpath(deck, list, shift, NULL);
		if (status) {
		    fprintf(stderr, "%s: cannot insert reference\n",
			    context_data->self);
		    if (1) {
			break;
		    }
		}
	    }

	    list = (char *) list + shift;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;
    }

    return status;
}


static int
sort_days(const void *text, const void *node)
{
    return -memcmp(text, node, LX);
}


static int
sort_fare(const void *text, const void *node)
{
    return memcmp(text, node, LX);
}


static int
time_list(struct context_type *context_data)
{
    int status;
    struct x1f4_bqfset_type bqfset;
    void *bqfx;

    bqfset.size = context_data->size;

    bqfset.fare = lain_fare;
    bqfset.look = lain_look;
    bqfset.move = lain_move;
    bqfset.pick = lain_pick;

    status = x1f4_init_bqfset(&bqfx, X1f4_BQFSET_SIZE_FRAME, &bqfset);
    if (status) {
	perror(context_data->self);
    } else {
	int excess;

	if (0) {
	} else {
	    void *list;

	    status = link_list(context_data, &list);
	    if (status) {
		perror(context_data->self);
	    } else {
		status = post_list(context_data, bqfx, NULL, list);
		if (status) {
		} else {
		    count2_decls;

		    if (context_data->do_regular) {
			if (context_data->forward) {
			    qsort(list, context_data->count, LX, sort_fare);
			} else {
			    qsort(list, context_data->count, LX, sort_days);
			}
		    } else {
		    }

		    start_count2();

		    if (context_data->do_regular) {
			status = miss_list(context_data, bqfx, list);
		    } else {
			status = trap_list(context_data, bqfx);
		    }
		    if (status) {
			fprintf(stderr, "%s: the sequence seed was %u\n",
				context_data->self, context_data->seed);
		    } else {
			if (context_data->forward) {
			    finish_count2_l("forward delete", 55);
			} else {
			    finish_count2_l("backward delete", 55);
			}
		    }
		}

		free_list(list);
	    }
	}

	excess = x1f4_fini_bqfset(&bqfx);
	if (excess) {
	    perror(context_data->self);

	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
trap_list(struct context_type *context_data, void *bqfx)
{
    int (*cast) (void *), status;
    unsigned i;

    if (context_data->forward) {
	cast = x1f4_mind_bqfset;
    } else {
	cast = x1f4_slip_bqfset;
    }

    i = context_data->count;
    while (i) {
	i--;
	status = cast(bqfx);
	if (status) {
	    fprintf(stderr, "%s: cannot delete data\n",
		    context_data->self);
	    if (1) {
		break;
	    }
	}
    }

    if (i) {
	status = 1;
    } else {
	if (x1f4_case_bqfset(bqfx)) {
	    fprintf(stderr, "%s: set not void\n", context_data->self);

	    status = 1;
	} else {
	    status = 0;
	}
    }

    return status;
}


static int
type_list(struct context_type *context_data)
{
    int status;
    struct x1f4_bqfset_type bqfset;
    void *bqfx;

    bqfset.size = context_data->size;

    bqfset.fare = lain_fare;
    bqfset.look = lain_look;
    bqfset.move = lain_move;
    bqfset.pick = lain_pick;

    status = x1f4_init_bqfset(&bqfx, X1f4_BQFSET_SIZE_FRAME, &bqfset);
    if (status) {
	perror(context_data->self);
    } else {
	int excess;
	void *deck;

	status = x1f4_init_scdeck(&deck, 0, NULL);
	if (status) {
	    perror(context_data->self);
	} else {
	    void *list;

	    status = link_list(context_data, &list);
	    if (status) {
		perror(context_data->self);
	    } else {
		status = post_list(context_data, bqfx, deck, list);
		if (status) {
		} else {
		    if (context_data->forward) {
			status = cast_list(context_data, bqfx, deck);
		    } else {
			status = pass_list(context_data, bqfx, deck);
		    }
		    if (status) {
			fprintf(stderr, "%s: the sequence seed was %u\n",
				context_data->self, context_data->seed);
		    } else {
		    }
		}

		free_list(list);
	    }


	    excess = x1f4_fini_scdeck(&deck);
	    if (excess) {
		perror(context_data->self);

		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = x1f4_fini_bqfset(&bqfx);
	if (excess) {
	    perror(context_data->self);

	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: bqfm [OPTIONS]\n\
Test the BQ generic set end delete methods.\n\
\n\
Options:\n\
  -N, --size SIZE		test SIZE wide sets\n\
  -b, --backward		start with the last\n\
  -f, --forward			start with the first (default)\n\
  -p, --profile			time operations, suppress output\n\
  -r, --regular			test the regular, non end special delete\n\
  -t, --time			seed random sequence with current time\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    struct context_type context;

    context.do_profile = 0;
    context.do_regular = 0;

    context.forward = 1;

    context.count = VX;

    context.seed = 1;

    unsetenv("POSIXLY_CORRECT");

    {
	const char *size_specs = NULL;
	int fast = ~0, time_seed = 0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "backward",	    0x00,   NULL,   0x62    },
		{   "forward",	    0x00,   NULL,   0x66    },
		{   "profile",	    0x00,   NULL,   0x70    },
		{   "regular",	    0x00,   NULL,   0x72    },
		{   "size",	    0x01,   NULL,   0x4e    },
		{   "time",	    0x00,   NULL,   0x74    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "N:bfprt", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'N':
		size_specs = optarg;
		break;
	    case 'b':
		context.forward = 0;
		break;
	    case 'f':
		context.forward = 1;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'p':
		context.do_profile = 1;
		break;
	    case 'r':
		context.do_regular = 1;
		break;
	    case 't':
		time_seed = 1;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (size_specs) {
	    if (x1f4_parse_wxcardinal(&context.count, size_specs, NULL, 0)) {
		fprintf(stderr, "%s: cannot parse test set width specificati"
			"on: `%s'\n", argv[0], size_specs);

		return 1;
	    }
	}

	if (time_seed) {
	    struct timeval now;

	    gettimeofday(&now, NULL);
	    context.seed = now.tv_sec ^ (now.tv_usec * 1000001);
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	context.size = LX;

	if (context.do_profile) {
	    time_list(&context);
	} else {
	    type_list(&context);
	}
    }

    return 0;
}
