/*
 * fs12.c
 * Copyright (C) 2002-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <inter.h>
#include <libx1f4l2.h>

typedef struct context_type {
    const char *self;
    unsigned line;
    void *tc12;
} context_type;

typedef struct line_type {
    void *line;
} line_type;

static int case_away(struct context_type *);
static int dump_list(struct context_type *, void *);
static int fast_away(struct context_type *);
static int fini_list(struct context_type *, void **);
static int head_line(struct context_type *, void *, char *, unsigned);
static int init_list(struct context_type *, void **);
static int link_node(void *, const char *, unsigned, void *);
static int read_list(struct context_type *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = read_list(context_data, list);
	if (status) {
	} else {
	    status = dump_list(context_data, list);
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
dump_list(struct context_type *context_data, void *list)
{
    int status;
    char *line;

    line = (char *) malloc(context_data->line + 1);
    if (!line) {
	status = -1;
    } else {
	unsigned size;
	void *tc12;

	status = 0;

	tc12 = context_data->tc12;

	x1f4_size_tclist(tc12, &size);
	while (size) {
	    struct line_type *line_data;
	    void *peek, *slip;

	    size--;

	    x1f4_peek_tclist(tc12, size, &peek);

	    line_data = peek;
	    slip = line;
	    x1f4_type_fspost(line_data->line, &slip);
	    strcpy(slip, (char *) (line_data + 1));

	    puts(line);
	}

	free(line);
    }

    return status;
}


static int
fast_away(struct context_type *context_data)
{
    int status;

    status = x1f4_init_tclist(&context_data->tc12, 0, NULL);
    if (status) {
    } else {
	int excess;

	status = case_away(context_data);

	excess = x1f4_fini_tclist(&context_data->tc12);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_fspost(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
head_line(struct context_type *context_data, void *list, char *data,
	  unsigned size)
{
    int status;

    status = x1f4_push_fspost(list, data, size);
    if (status) {
	fprintf(stderr, "%s: cannot insert item\n", context_data->self);
    } else {
	if (context_data->line < size) {
	    context_data->line = size;
	}
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;
    struct x1f4_fspost_type fspost;

    fspost.link = link_node;
    fspost.text = context_data->tc12;

    status = x1f4_init_fspost(list, 0, &fspost);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
	context_data->line = 0;
    }

    return status;
}


static int
link_node(void *text, const char *data, unsigned size, void *node)
{
    int status;
    void *slip;

    status = x1f4_push_tclist
	(text, 0, size + sizeof(struct line_type) + 1, &slip);
    if (status) {
    } else {
	char *line;
	struct line_type *line_data;

	line_data = slip;

	line_data->line = node;

	line = (char *) (line_data + 1);
	memcpy(line, data, size);
	line[size] = 0;
    }

    return status;
}


static int
read_list(struct context_type *context_data, void *list)
{
    int status;

    do {
	char *data;
	int done = 0;
	unsigned size = 0;

	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (!done || size) {
		status = head_line(context_data, list, data, size);
	    }

	    if (1) {
		free(data);
	    }

	    if (status) {
		break;
	    }

	    if (done) {
		break;
	    } else {
		while (1) {
		    status = libx1f4l2_read_line
			(context_data->self, &done, &data, &size);
		    if (status) {
			break;
		    }

		    if (!done || size) {
			status = head_line(context_data, list, data, size);
		    }

		    if (1) {
			free(data);
		    }

		    if (status) {
			break;
		    }

		    if (done) {
			break;
		    }
		}
	    }
	}
    } while (0);

    return status;
}


static void
usage(void)
{
    puts("Usage: fs12 [OPTIONS]\n\
Test fs_insert(), output input.\n\
\n\
Options:\n\
  --help			display this help and exit\n\
  --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	struct context_type context;

	context.self = argv[0];

	status = fast_away(&context);
    }

    return status;
}
