/*
 * ls18.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <inter.h>
#include <libx1f4l2.h>
#include <lsas.h>

#define __BINARY_FRINGE__		0

#if __BINARY_DECK__
#else
# undef __BINARY_FRINGE__
# define __BINARY_FRINGE__		0
#endif				/* __BINARY_DECK__ */

#if SIZEOF_VOID_P == SIZEOF_LONG
# define integral_q			unsigned long
#else
# define integral_q			unsigned
#endif				/* SIZEOF_VOID_P == SIZEOF_LONG */

#if __BINARY_FRINGE__
# define pick_line			land_line
#else
# define pick_line			fast_line
#endif				/* __BINARY_FRINGE__ */

typedef struct context_type {
    const char *self;
    int ignore_duplicates, probe;
    unsigned class, frame;
} context_type;

typedef struct nail_type {
    struct context_type *context_data;
    void *list;
} nail_type;

static int case_away(struct context_type *);
static int fail_list(struct context_type *, void *, void *);
static int fail_node(void *, void *);
static int fast_line(struct context_type *, void *, void *, unsigned, char *,
		     unsigned);
static int fini_list(struct context_type *, void **);
static int fini_rate(struct context_type *, void **);
static int free_node(void *, void *);
static int init_list(struct context_type *, void **);
static int init_rate(struct context_type *, void **);
#if __BINARY_FRINGE__
static int land_line(struct context_type *, void *, void *, unsigned, char *,
		     unsigned);
#endif				/* __BINARY_FRINGE__ */
static int nail_list(struct context_type *, void *, void *);
static int nail_node(void *, void *);
static int read_list(struct context_type *, void *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *rate;

	status = init_rate(context_data, &rate);
	if (status) {
	} else {
	    status = read_list(context_data, list, rate);
	    if (status) {
	    } else {
		if (context_data->probe) {
		    status = fail_list(context_data, list, rate);
		} else {
		    status = nail_list(context_data, list, rate);
		}
	    }

	    excess = fini_rate(context_data, &rate);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
fail_list(struct context_type *context_data, void *list, void *rate)
{
    struct nail_type nail;

    nail.context_data = context_data;
    nail.list = list;

    return x1f4_lime_qsrate(rate, &nail, fail_node);
}


static int
fail_node(void *miss, void *node)
{
    int status;
    void **call;
    struct nail_type *nail_data;
    void *deck = NULL;

    nail_data = miss;

    call = node;

#if __BILL_FOR_NOTE__
    status = x1f4_post_lsbill
	(nail_data->list, *call, (integral_q) call[2], 0, &deck);
#else
    status = x1f4_post_lsnote(nail_data->list, *call, 0, &deck);
#endif				/* __BILL_FOR_NOTE__ */
    if (status == X1f4_LSNOTE_EVER_MATCH) {
	if (deck) {
	    if (*(const unsigned *) deck != (integral_q) call[1]) {
		fprintf(stderr, "%s: cannot match `%s' (expected %u, found %"
			"u)\n", nail_data->context_data->self, (char *) *call,
			*(const unsigned *) deck,
			 (unsigned) (integral_q) call[1]);
	    } else {
		status = 0;
	    }
	} else {
	    fprintf(stderr, "%s: buggy find `%s'\n",
		    nail_data->context_data->self, (char *) *call);
	}
    } else {
	status = -1;
	fprintf(stderr, "%s: cannot find `%s'\n",
		nail_data->context_data->self, (char *) *call);
    }

    return status;
}


static int
fast_line(struct context_type *context_data, void *list, void *rate,
	  unsigned index, char *data, unsigned size)
{
    int status;
    void *fail;

#if __BILL_FOR_NOTE__
    status = x1f4_post_lsbill(list, data, size, sizeof(unsigned), &fail);
#else
    status = x1f4_post_lsnote(list, data, sizeof(unsigned), &fail);
#endif				/* __BILL_FOR_NOTE__ */
    if (status == X1f4_S3DECK_EVER_MATCH) {
	if (context_data->ignore_duplicates) {
	    status = 0;
	} else {
	    fprintf(stderr, "%s: duplicate `%s' line\n", context_data->self,
		    data);
	}
    } else {
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);
	} else {
	    void *pick;

	    status = x1f4_tail_qsrate(rate, &pick);
	    if (status) {
		fprintf(stderr, "%s: cannot insert item\n",
			context_data->self);
	    } else {
		unsigned *bill;
		void **call;

		call = pick;
		*call = data;
		call[1] = (void *) (integral_q) index;
		call[2] = (void *) (integral_q) size;
		bill = fail;
		*bill = index;
	    }
	}
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_lsnote(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy tree\n", context_data->self);
    }

    return status;
}


static int
fini_rate(struct context_type *context_data, void **rate)
{
    int status;

    x1f4_lime_qsrate(*rate, NULL, free_node);

    status = x1f4_fini_qsrate(rate);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
free_node(void *miss, void *node)
{
    free(*(void **) node);

    return 0;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;
    struct x1f4_lsnote_type lsnote;

#if __FIX_INTERNAL__
    lsnote.inner = context_data->class;
#endif				/* __FIX_INTERNAL__ */
#if __FIX_EXTERNAL__
    lsnote.outer = context_data->frame;
#endif				/* __FIX_EXTERNAL__ */

    status = x1f4_init_lsnote
	(list, X1f4_LSNOTE_INNER_MASK | X1f4_LSNOTE_OUTER_MASK, &lsnote);
    if (status) {
	fprintf(stderr, "%s: cannot create tree\n", context_data->self);
    } else {
    }

    return status;
}


static int
init_rate(struct context_type *context_data, void **rate)
{
    int status;
    struct x1f4_qsrate_type qsrate;

    qsrate.size = sizeof(void *) * 3;

    status = x1f4_init_qsrate(rate, 0, &qsrate);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
    }

    return status;
}


#if __BINARY_FRINGE__
static int
land_line(struct context_type *context_data, void *list, void *rate,
	  unsigned index, char *data, unsigned size)
{
    char *fast;
    int status;
    unsigned m;

    fast = malloc(size * 3 + 1);
    if (fast) {
	char *s;

	m = size;
	s = fast;

	for (; m; m--) {
	    *s++ = 0;
	    *s++ = *data++;
	    *s++ = 0;
	}

	*s = 0;

	status = fast_line(context_data, list, rate, index, fast, 3 * size);
	if (status) {
	    free(fast);
	} else {
	    free(data - size);
	}
    } else {
	status = 1;
    }

    return status;
}
#endif				/* __BINARY_FRINGE__ */


static int
nail_list(struct context_type *context_data, void *list, void *rate)
{
    struct nail_type nail;

    nail.context_data = context_data;
    nail.list = list;

    return x1f4_lime_qsrate(rate, &nail, nail_node);
}


static int
nail_node(void *miss, void *node)
{
    int status;
    void **call;
    struct nail_type *nail_data;
    void *deck;

    nail_data = miss;

    call = node;

#if __BILL_FOR_NOTE__
    status = x1f4_find_lsbill
	(nail_data->list, *call, (integral_q) call[2], &deck);
#else
    status = x1f4_find_lsnote(nail_data->list, *call, &deck);
#endif				/* __BILL_FOR_NOTE__ */
    if (status) {
	if (*(const unsigned *) deck != (integral_q) call[1]) {
	    fprintf(stderr, "%s: cannot match `%s' (expected %u, found %u)\n",
		    nail_data->context_data->self, (char *) *call,
		    *(const unsigned *) deck, (unsigned) (integral_q) call[1]);
	} else {
	    status = 0;
	}
    } else {
	status = -1;
	fprintf(stderr, "%s: cannot find `%s'\n",
		nail_data->context_data->self, (char *) *call);
    }

    return status;
}


static int
read_list(struct context_type *context_data, void *list, void *rate)
{
    char *data;
    int done = 0, status;
    unsigned i = 0, size = 0;

    while (1) {
	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    int slip = 1;

	    if (done) {
	    } else {
#if __MISS_VOID_LINES__
		if (size) {
#endif				/* __MISS_VOID_LINES__ */
		    status = pick_line
			(context_data, list, rate, i, data, size);
		    if (status) {
		    } else {
			slip = 0;
		    }
#if __MISS_VOID_LINES__
		}
#endif				/* __MISS_VOID_LINES__ */
	    }

	    if (slip) {
		free(data);
	    }

	    if (status) {
		break;
	    }

	    i++;

	    if (done) {
		break;
	    }
	}
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: ls18 [OPTIONS]\n\
Test " lsname "_find().\n\
\n\
Options:\n\
  -d, --ignore-duplicates	ignore duplicate input lines\n\
  -e, --external L		set external nodes size as L bytes\n\
  -i, --internal L		set internal nodes size as L bytes\n\
  -p, --probe			find data by duplicate key insertion\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.ignore_duplicates = 0;

    context.class = 256;
    context.frame = 256;

    context.probe = 0;

    {
	const char *class_specs = NULL, *frame_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "external",	    0x01,   NULL,   0x65    },
		{   "ignore-duplicates",
				    0x00,   NULL,   0x64    },
		{   "internal",	    0x01,   NULL,   0x69    },
		{   "probe",	    0x00,   NULL,   0x70    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "de:i:p", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'd':
		context.ignore_duplicates = 1;
		break;
	    case 'e':
		frame_specs = optarg;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'i':
		class_specs = optarg;
		break;
	    case 'p':
		context.probe = 1;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (class_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.class, class_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse internal node width specif"
			"ication: `%s'\n", argv[0], class_specs);

		return 1;
	    } else {
		if (context.class < 4) {
		    fprintf(stderr, "%s: will not use internal node widths l"
			    "ess than 4 (like the indicated 1 << %u)\n",
			    argv[0], context.class);

		    return 1;
		}
	    }
	}

	if (frame_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.frame, frame_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse external node width specif"
			"ication: `%s'\n", argv[0], frame_specs);

		return 1;
	    } else {
		if (context.frame < 1) {
		    fprintf(stderr, "%s: will not use external node widths l"
			    "ess than 4 (like the indicated %u << 2)\n",
			    argv[0], context.frame);

		    return 1;
		}
	    }
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	status = case_away(&context);
    }

    return status;
}
