/*
 * mn11.c
 * Copyright (C) 2002-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <errno.h>
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <inter.h>
#include <libx1f4l2.h>
#include <tcmn.h>

typedef struct context_type {
    const char *self;
    unsigned mini;
} context_type;

typedef struct node_type {
    union {
	char node[4];
	void *data;
    } data[2];
} node_type;

static int case_away(struct context_type *);
static int deck_fast(void *, void *);
static int deck_miss(void *, void *);
static int dump_list(struct context_type *, void *);
static int dump_node(void *, void *);
static int fast_line(struct context_type *, void *, void *, unsigned, char *,
		     unsigned);
static int fini_list(struct context_type *, void **);
static int fini_mini(struct context_type *, void **);
static int init_list(struct context_type *, void **);
static int init_mini(struct context_type *, void **);
static int read_list(struct context_type *, void *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list, *mini;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = init_mini(context_data, &mini);
	if (status) {
	} else {
	    status = read_list(context_data, list, mini);
	    if (status) {
	    } else {
		status = dump_list(context_data, mini);
	    }

	    excess = fini_mini(context_data, &mini);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
deck_fast(void *deck, void *miss)
{
    return deck_miss
	(deck, ((struct node_type *) ((char *) miss - 4))->data[1].data);
}


static int
deck_miss(void *deck, void *miss)
{
    int delete;
    unsigned char *s, *v;

    s = deck;
    v = miss;

    while (1) {
	unsigned char c;

	c = *s;
	if (c ^ *v) {
	    delete = (int) c - (int) *v;

	    break;
	} else {
	    if (c) {
		s++;
		v++;
	    } else {
		delete = 0;

		break;
	    }
	}
    }

    return delete;
}


static int
dump_list(struct context_type *context_data, void *list)
{
    x1f4_lime_tcmini(list, NULL, dump_node);

    return 0;
}


static int
dump_node(void *text, void *node)
{
    puts(((struct node_type *) ((char *) node - 4))->data[1].data);

    return 0;
}


static int
fast_line(struct context_type *context_data, void *list, void *mini,
	  unsigned index, char *data, unsigned size)
{
    int status;
    void *slip;

    status = x1f4_post_tcline(list, data, deck_miss, 0, size + 1, &slip);
    if (status == X1f4_TCLINE_EVER_MATCH) {
	status = x1f4_port_tcmini(mini, data, deck_fast, 0, &slip);
	if (status == X1f4_TCMINI_EVER_MATCH) {
	    status = 0;
	} else {
	    status = -1;

	    fprintf(stderr, "%s: broken mini BST already\n",
		    context_data->self);
	}
    } else {
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);
	} else {
	    void *hook;

	    memcpy(slip, data, size + 1);

	    status = x1f4_port_tcmini(mini, data, deck_fast, 0, &hook);
	    if (status) {
		fprintf(stderr, "%s: cannot insert item (mini size is %u)\n",
			context_data->self, context_data->mini);
	    } else {
		struct node_type *node_data;

		node_data = (struct node_type *) ((char *) hook - 4);
		node_data->data[1].data = slip;

		context_data->mini++;
	    }
	}
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_tcline(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
fini_mini(struct context_type *context_data, void **mini)
{
    free(*mini);

    return 0;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_tcline(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
    }

    return status;
}


static int
init_mini(struct context_type *context_data, void **mini)
{
    int status;
    void *list;

    list = malloc(X1f4_TCMINI_SIZE_BOUND);
    if (list) {
	x1f4_type_tcmini
	    (list, sizeof(struct node_type) - 4, X1f4_TCMINI_SIZE_BOUND);

	*mini = list;

	status = 0;
    } else {
	status = -1;

	fprintf(stderr, "%s: %s\n", context_data->self, strerror(errno));
    }

    return status;
}


static int
read_list(struct context_type *context_data, void *list, void *mini)
{
    char *data;
    int done = 0, status;
    unsigned i = 0, size = 0;

    context_data->mini = 0;

    while (1) {
	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (!done || size) {
		status = fast_line(context_data, list, mini, i, data, size);
	    }

	    if (1) {
		free(data);
	    }

	    if (status) {
		break;
	    } else {
		i++;
	    }

	    if (done) {
		break;
	    }
	}
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: mn11 [OPTIONS]\n\
Test " tcname "_insert(), output sorted, uniq-ed input.\n\
\n\
Options:\n\
  --help			display this help and exit\n\
  --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	status = case_away(&context);
    }

    return status;
}
