/*
 * qs13.c
 * Copyright (C) 2002-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <inter.h>
#include <libx1f4l2.h>
#include <qxcc.h>

#define QS11				1000000

typedef struct context_type {
    const char *self;
    unsigned count;
} context_type;

static int case_away(struct context_type *);
static int fini_rate(struct context_type *, void **);
static int flip_rate(struct context_type *, void *);
static int init_rate(struct context_type *, void **);
static int peek_rate(struct context_type *, void *, void *);
static int push_rate(struct context_type *, void *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_rate(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *rate;

	status = init_rate(context_data, &rate);
	if (status) {
	} else {
	    status = push_rate(context_data, list, rate);
	    if (status) {
	    } else {
		status = flip_rate(context_data, rate);
		if (status) {
		} else {
		    status = peek_rate(context_data, list, rate);
		}
	    }

	    excess = fini_rate(context_data, &rate);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_rate(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
fini_rate(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_qxrate(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
flip_rate(struct context_type *context_data, void *list)
{
    int status;

    status = x1f4_flip_qxrate(list);
    if (status) {
	fprintf(stderr, "%s: cannot reverse list\n", context_data->self);
    }

    return status;
}


static int
init_rate(struct context_type *context_data, void **list)
{
    int status;
    struct x1f4_qxrate_type qxrate;

#if __MISS_FOUR_RATE__
#else
    qxrate.quarter = 1 << 4;
#endif				/* __MISS_FOUR_RATE__ */
    qxrate.rate = 5;
    qxrate.size = sizeof(unsigned);

    status = x1f4_init_qxrate
	(list, X1f4_QXRATE_FOUR_FRAME | X1f4_QXRATE_RATE_FRAME, &qxrate);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    }

    return status;
}


static int
peek_rate(struct context_type *context_data, void *list, void *rate)
{
    int status = 0;
    unsigned i;

    i = context_data->count;
    while (i) {
	void *plus;

	i--;

	status = x1f4_peek_qxrate(list, i, &plus);
	if (status) {
	    fprintf
		(stderr, "%s: cannot retrieve item\n", context_data->self);

	    break;
	} else {
	    void *data;

	    status = x1f4_peek_qxrate(rate, i, &data);
	    if (status) {
		fprintf(stderr, "%s: cannot retrieve item\n",
			context_data->self);

		break;
	    } else {
		if (*(unsigned *) data ^ *(unsigned *) plus) {
		    fprintf(stderr, "%s: cannot match lists\n",
			    context_data->self);
		    fprintf(stderr, "%s: index was %u, tclist was %u, " qxname
			    "rate was %u\n", context_data->self, i,
			    *(unsigned *) plus, *(unsigned *) data);

		    status = 1;

		    break;
		}
	    }
	}
    }

    return status;
}


static int
push_rate(struct context_type *context_data, void *list, void *rate)
{
    int status = 0;
    unsigned i;

    srand(0);

    i = 0;
    while (i < context_data->count) {
	unsigned e;
	void *data;

	e = rand();

	status = x1f4_push_qxrate(list, i, &data);
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);

	    break;
	} else {
	    unsigned *slip;

	    i++;

	    slip = data;
	    *slip = i;

	    status = x1f4_push_qxrate(rate, 0, &data);
	    if (status) {
		fprintf
		    (stderr, "%s: cannot insert item\n", context_data->self);

		break;
	    } else {
		slip = data;
		*slip = i;
	    }
	}
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: qs13 [OPTIONS]\n\
Test " qxname "_reverse().\n\
\n\
Options:\n\
  -N, --size SIZE		test SIZE wide sets\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.count = QS11;

    {
	const char *size_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "size",	    0x01,   NULL,   0x4e    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "N:", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'N':
		size_specs = optarg;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (size_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.count, size_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse list size specification: `"
			"%s'\n", argv[0], size_specs);

		return 1;
	    }
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	status = case_away(&context);
    }

    return status;
}
