/*
 * rate.c
 * Copyright (C) 2008-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>

#define __DUMB_SEEK__			0
#define __HOLE_SEEK__			1

#define __HERE_SEEK__			__HOLE_SEEK__

#define EVER_MAX			(1 << 013)
#define LINE_MAX			(1 << 011)

#if SIZEOF_VOID_P == SIZEOF_LONG
# define integral_q			unsigned long
#else
# define integral_q			unsigned
#endif				/* SIZEOF_VOID_P == SIZEOF_LONG */

#if SIZEOF_UNSIGNED_LONG == 4
# define LINE_HOLE			0x0a0a0a0aul
# define MISS_HOLE			0x7efefefful
#else
# define LINE_HOLE			0x0a0a0a0a0a0a0a0aul
# define MISS_HOLE			0x7efefefefefefefful
#endif				/* SIZEOF_UNSIGNED_LONG == 4 */

static int copy_data(const char *, unsigned, unsigned, unsigned, void **,
		     unsigned *);
static int find_line(void *, unsigned, unsigned *);

static unsigned here_ever, here_size = 0;
static unsigned char here_data[EVER_MAX];

static int
copy_data(const char *self, unsigned miss, unsigned ever, unsigned fast,
	  void **line, unsigned *size)
{
    int status;
    void *data;

    data = malloc(fast + 1);
    if (!data) {
	status = -1;

	fprintf(stderr, "%s: %s\n", self, strerror(errno));
    } else {
	char *text;

	status = 0;

	memcpy(data, here_data + ever, fast);
	text = data;
	text[fast] = 0;

	*size = fast;

	fast++;

	here_ever = ever + fast;
	here_size = miss - fast;

	*line = data;
    }

    return status;
}


static int
find_line(void *data, unsigned size, unsigned *fast)
{
    char *line;
#if __HERE_SEEK__ == __HOLE_SEEK__
    unsigned slip;
#endif				/* __HERE_SEEK__ == __HOLE_SEEK__ */

    line = data;

#if __HERE_SEEK__ == __DUMB_SEEK__
    do {
	if (*line == 10) {
	    break;
	} else {
	    line++;
	    size--;
	}
    } while (size);
#else
    slip =
	(SIZEOF_UNSIGNED_LONG
	 - ((integral_q) data & (SIZEOF_UNSIGNED_LONG - 1)))
	& (SIZEOF_UNSIGNED_LONG - 1);
    if (slip) {
	if (size < slip) {
	    slip = size;
	}

	size -= slip;

	do {
	    if (*line == 10) {
		size = 0;

		break;
	    } else {
		line++;
		slip--;
	    }
	} while (slip);
    }
#endif				/* __HERE_SEEK__ == __HOLE_SEEK__ */

#if __HERE_SEEK__ == __HOLE_SEEK__
    if (size) {
	slip = size / SIZEOF_UNSIGNED_LONG;
	if (slip) {
	    unsigned long *deck;

	    deck = (unsigned long *) line;

	    do {
		unsigned long this;

		this = *deck ^ LINE_HOLE;
		if (((this + MISS_HOLE) ^ ~this) & ~MISS_HOLE) {
		    line = (char *) deck;

		    if (*line == 10) {
			size = 0;
			if (1) {
			    break;
			}
		    }

		    line++;

		    if (*line == 10) {
			size = 0;
			if (1) {
			    break;
			}
		    }

		    line++;

		    if (*line == 10) {
			size = 0;
			if (1) {
			    break;
			}
		    }

		    line++;

		    if (*line == 10) {
			size = 0;
			if (1) {
			    break;
			}
		    }

# if SIZEOF_UNSIGNED_LONG == 8
		    line++;

		    if (*line == 10) {
			size = 0;
			if (1) {
			    break;
			}
		    }

		    line++;

		    if (*line == 10) {
			size = 0;
			if (1) {
			    break;
			}
		    }

		    line++;

		    if (*line == 10) {
			size = 0;
			if (1) {
			    break;
			}
		    }

		    line++;

		    if (*line == 10) {
			size = 0;
			if (1) {
			    break;
			}
		    }
# endif				/* SIZEOF_UNSIGNED_LONG == 8 */

		    deck++;

		    slip--;
		} else {
		    slip--;
		    if (slip) {
			deck++;
		    } else {
			size = size & (SIZEOF_UNSIGNED_LONG - 1);

			line = (char *) (deck + 1);

			break;
		    }
		}
	    } while (1);
	}

	if (size) {
	    do {
		if (*line == 10) {
		    break;
		} else {
		    line++;
		    size--;
		}
	    } while (size);
	}
    }
#endif				/* __HERE_SEEK__ == __HOLE_SEEK__ */

    *fast = line - (char *) data;

    return 0;
}


int
libx1f4l2_rate_line(const char *self, const char *path, int file, int *done,
		    char **line, unsigned *slip)
{
    int status;

    do {
	unsigned ever, fast, miss;

	miss = here_size;
	if (miss) {
	    ever = here_ever;

	    find_line(here_data + ever, miss, &fast);
	    if (fast < miss) {
		status = copy_data
		    (self, miss, ever, fast, (void *) line, slip);

		break;
	    } else {
		if (ever) {
		    memmove(here_data, here_data + ever, miss);
		}
	    }
	} else {
	    ever = 0;
	}

	if (miss < EVER_MAX) {
	    ever = miss;

	    while (1) {
		status = read(file, here_data + miss, EVER_MAX - miss);
		if (status == -1) {
		    if (errno == EINTR) {
		    } else {
			status = -1;

			fprintf(stderr, "%s: cannot read `%s': %s\n", self,
				path, strerror(errno));

			if (1) {
			    break;
			}
		    }
		} else {
		    if (status) {
			miss = miss + status;
			if (miss < EVER_MAX) {
			} else {
			    status = 0;
			    if (1) {
				break;
			    }
			}
		    } else {
			if (ever == miss) {
			    *done = 1;

			    status = copy_data
				(self, 1, 0, 0, (void *) line, slip);

			    break;
			} else {
			    status = 0;
			    if (1) {
				break;
			    }
			}
		    }
		}
	    }

	    if (status) {
		break;
	    }

	    find_line(here_data, miss, &fast);
	    if (fast < miss) {
		status = copy_data(self, miss, 0, fast, (void *) line, slip);

		break;
	    } else {
		if (miss == EVER_MAX) {
		} else {
		    status = copy_data
			(self, miss + 1, 0, miss, (void *) line, slip);

		    break;
		}
	    }
	}

	if (1) {
	    void *data;

	    data = malloc((EVER_MAX << 1) + 1);
	    if (!data) {
		status = -1;

		fprintf(stderr, "%s: %s\n", self, strerror(errno));
	    } else {
		unsigned less;

		less = miss + EVER_MAX;

		while (1) {
		    status = read(file, (char *) data + miss, less - miss);
		    if (status == -1) {
			if (errno == EINTR) {
			} else {
			    status = -1;

			    fprintf(stderr, "%s: cannot read `%s': %s\n",
				    self, path, strerror(errno));

			    if (1) {
				break;
			    }
			}
		    } else {
			if (status) {
			    ever = status;

			    find_line((char *) data + miss, ever, &fast);
			    if (fast < ever) {
				char *text;

				memcpy(data, here_data, EVER_MAX);

				miss += fast;

				*line = data;
				*slip = miss;

				text = (char *) data + miss;
				*text = 0;

				ever -= fast + 1;

				here_ever = 0;
				here_size = ever;

				memcpy(here_data, text + 1, ever);

				status = 0;

				break;
			    } else {
				miss = miss + ever;
			    }
			} else {
			    char *text;

			    memcpy(data, here_data, EVER_MAX);

			    *line = data;
			    *slip = miss;

			    text = data;
			    text[miss] = 0;

			    here_size = 0;

			    break;
			}
		    }

		    if (miss == less) {
			void *swap;

			less += EVER_MAX;

			swap = realloc(data, less + 1);
			if (swap) {
			    data = swap;
			} else {
			    free(data);

			    status = -1;

			    fprintf(stderr, "%s: %s\n", self, strerror(errno));

			    break;
			}
		    }
		}
	    }
	}
    } while (0);

    return status;
}
