/*
 * sx31.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <inter.h>
#include <libx1f4l2.h>
#include <s3xx.h>

#define __BINARY_FRINGE__		1

#define __COPY_TEST__			0
#define __TEXT_TEST__			0

#if __BINARY_DECK__
#else
# undef __BINARY_FRINGE__
# define __BINARY_FRINGE__		0
#endif				/* __BINARY_DECK__ */

#if __BINARY_FRINGE__
# define pick_line			land_line
#else
# define pick_line			fast_line
#endif				/* __BINARY_FRINGE__ */

typedef struct context_type {
    const char *self;
    int ignore_duplicates;
} context_type;

static int case_away(struct context_type *);
#if __COPY_TEST__
static int copy_list(struct context_type *, void **);
#endif				/* __COPY_TEST__ */
#if __TEXT_TEST__
static int copy_node(void *, const void **, const void *);
#endif				/* __TEXT_TEST__ */
static int dump_list(struct context_type *, void *, unsigned);
static int dump_node(void *, const char *, unsigned);
static int fast_line(struct context_type *, void *, unsigned, char *,
		     unsigned);
static int fini_list(struct context_type *, void **);
#if __TEXT_TEST__
static int free_node(void *, const void *);
#endif				/* __TEXT_TEST__ */
static int init_list(struct context_type *, void **);
#if __BINARY_FRINGE__
static int land_line(struct context_type *, void *, unsigned, char *,
		     unsigned);
#endif				/* __BINARY_FRINGE__ */
static int read_list(struct context_type *, void *, unsigned *);
#if __TEXT_TEST__
static int text_list(struct context_type *, void **);
#endif				/* __TEXT_TEST__ */

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	unsigned copy;

	status = read_list(context_data, list, &copy);
	if (status) {
	} else {
#if __COPY_TEST__
	    status = copy_list(context_data, &list);
	    if (status) {
	    } else {
#endif				/* __COPY_TEST__ */
#if __TEXT_TEST__
		status = text_list(context_data, &list);
		if (status) {
		} else {
#endif				/* __TEXT_TEST__ */
		    status = dump_list(context_data, list, copy);
#if __TEXT_TEST__
		}
#endif				/* __TEXT_TEST__ */
#if __COPY_TEST__
	    }
#endif				/* __COPY_TEST__ */
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


#if __COPY_TEST__
static int
copy_list(struct context_type *context_data, void **fare)
{
    int status;
    void *copy, *list;

    list = *fare;

    status = x1f4_copy_s3deck(&copy, list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy tree\n", context_data->self);
    } else {
	*fare = copy;

	status = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __COPY_TEST__ */


#if __TEXT_TEST__
static int
copy_node(void *none, const void **link, const void *data)
{
    *link = data;

    return 0;
}
#endif				/* __TEXT_TEST__ */


static int
dump_list(struct context_type *context_data, void *list, unsigned copy)
{
    int status;
#if __HEAD_LIST_DECK__
    void *miss;
#endif				/* __HEAD_LIST_DECK__ */

#if __HEAD_LIST_DECK__
    miss = malloc(copy + 1);
    if (miss) {
	status = x1f4_list_s3deck(list, NULL, dump_node, miss);
#endif				/* __HEAD_LIST_DECK__ */
#if __HEAD_NOTE_DECK__
	status = x1f4_note_s3deck(list, NULL, dump_node);
#endif				/* __HEAD_NOTE_DECK__ */
#if __HEAD_LIST_DECK__

	free(miss);
    } else {
	status = -1;
    }
#endif				/* __HEAD_LIST_DECK__ */

    return status;
}


static int
dump_node(void *text, const char *node, unsigned size)
{
#if __BINARY_FRINGE__
    size /= 3;
    node++;
    for (; size; size--) {
	putchar(*node);
	node += 3;
    }
    putchar('\n');
#else
    puts(node);
#endif				/* __BINARY_FRINGE__ */

    return 0;
}


static int
fast_line(struct context_type *context_data, void *list, unsigned index,
	  char *data, unsigned size)
{
    int status;

#if __PATH_FOR_DECK__
    status = x1f4_post_s3path(list, data, size, NULL);
#else
    status = x1f4_post_s3deck(list, data, NULL);
#endif				/* __PATH_FOR_DECK__ */
    if (status == X1f4_S3DECK_EVER_MATCH) {
	if (context_data->ignore_duplicates) {
	    status = 0;
	} else {
	    fprintf(stderr, "%s: duplicate `%s' line\n", context_data->self,
		    data);
	}
    } else {
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);
	} else {
	}
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_s3deck(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy tree\n", context_data->self);
    }

    return status;
}


#if __TEXT_TEST__
static int
free_node(void *none, const void *data)
{
    return 0;
}
#endif				/* __TEXT_TEST__ */


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_s3deck(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create tree\n", context_data->self);
    } else {
    }

    return status;
}


#if __BINARY_FRINGE__
static int
land_line(struct context_type *context_data, void *list, unsigned index,
	  char *data, unsigned size)
{
    char *fast;
    int status;
    unsigned m;

    fast = malloc(size * 3 + 1);
    if (fast) {
	char *s;

	m = size;
	s = fast;

	for (; m; m--) {
	    *s++ = 0;
	    *s++ = *data++;
	    *s++ = 0;
	}

	*s = 0;

	status = fast_line(context_data, list, index, fast, 3 * size);

	free(fast);
    } else {
	status = 1;
    }

    return status;
}
#endif				/* __BINARY_FRINGE__ */


static int
read_list(struct context_type *context_data, void *list, unsigned *copy)
{
    char *data;
    int done = 0, status;
    unsigned i = 0, miss = 0, size = 0;

    while (1) {
	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (done) {
	    } else {
#if __MISS_VOID_LINES__
		if (size) {
#endif				/* __MISS_VOID_LINES__ */
		    status = pick_line(context_data, list, i, data, size);
		    if (miss < size) {
			miss = size;
		    }
#if __MISS_VOID_LINES__
		}
#endif				/* __MISS_VOID_LINES__ */
	    }

	    if (1) {
		free(data);
	    }

	    if (status) {
		break;
	    } else {
		i++;
	    }

	    if (done) {
		break;
	    }
	}
    }

    if (status) {
    } else {
	*copy = miss;
    }

    return status;
}


#if __TEXT_TEST__
static int
text_list(struct context_type *context_data, void **fare)
{
    int status;
    void *copy, *list;

    list = *fare;

    status = x1f4_text_s3deck(&copy, list, copy_node, free_node, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot destroy tree\n", context_data->self);
    } else {
	*fare = copy;

	status = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __TEXT_TEST__ */


static void
usage(void)
{
    puts("Usage: sx31 [OPTIONS]\n\
Test " sxname "_insert(), output sorted input.\n\
\n\
Options:\n\
  -d, --ignore-duplicates	ignore duplicate input lines\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.ignore_duplicates = 0;

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "ignore-duplicates",
				    0x00,   NULL,   0x64    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "d", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'd':
		context.ignore_duplicates = 1;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
#if __POINTER_DECK__
	fprintf(stderr, "%s: cannot work for data structures not storing the"
		" string keys such as " sxname "\n", argv[0]);

	status = 1;
#else
	context.self = argv[0];

	status = case_away(&context);
#endif				/* __POINTER_DECK__ */
    }

    return status;
}
