/*
 * sx35.c
 * Copyright (C) 2002-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <inter.h>
#include <libx1f4l2.h>
#include <s3xx.h>

typedef struct context_type {
    const char *self;
} context_type;

static int case_away(struct context_type *);
static int dump_list(struct context_type *, void *, unsigned);
static int fast_line(struct context_type *, void *, unsigned, char *,
		     unsigned);
static int fini_list(struct context_type *, void **);
static int init_list(struct context_type *, void **);
static int line_node(void *, const char *, unsigned);
static int read_list(struct context_type *, void *, unsigned *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	unsigned copy;

	status = read_list(context_data, list, &copy);
	if (status) {
	} else {
	    status = dump_list(context_data, list, copy);
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
dump_list(struct context_type *context_data, void *list, unsigned copy)
{
    return x1f4_bank_s3deck(list, NULL, line_node);
}


static int
fast_line(struct context_type *context_data, void *list, unsigned index,
	  char *data, unsigned size)
{
    int status;

#if __PATH_FOR_DECK__
    status = x1f4_post_s3path(list, data, size, NULL);
#else
    status = x1f4_post_s3deck(list, data, NULL);
#endif				/* __PATH_FOR_DECK__ */
    if (status == X1f4_S3DECK_EVER_MATCH) {
	fprintf(stderr, "%s: duplicate `%s' line\n", context_data->self, data);
    } else {
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);
	} else {
	}
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_s3deck(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy tree\n", context_data->self);
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_s3deck(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create tree\n", context_data->self);
    } else {
    }

    return status;
}


static int
line_node(void *text, const char *node, unsigned size)
{
    if (size) {
	fwrite(node, size, 1, stdout);
    } else {
    }

    fputc('\n', stdout);

    return 0;
}


static int
read_list(struct context_type *context_data, void *list, unsigned *copy)
{
    char *data;
    int done = 0, status;
    unsigned i = 0, miss = 0, size = 0;

    while (1) {
	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (done) {
	    } else {
		if (size) {
		    status = fast_line(context_data, list, i, data, size);
		    if (miss < size) {
			miss = size;
		    }
		}
	    }

	    if (1) {
		free(data);
	    }

	    if (status) {
		break;
	    } else {
		i++;
	    }

	    if (done) {
		break;
	    }
	}
    }

    if (status) {
    } else {
	*copy = miss;
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: sx35 [OPTIONS]\n\
Dump " sxname "tree suffixes.\n\
\n\
Options:\n\
  --help			display this help and exit\n\
  --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
#if __POINTER_DECK__
	fprintf(stderr, "%s: cannot work for data structures not storing the"
		" string keys such as " sxname "\n", argv[0]);

	status = 1;
#else
	context.self = argv[0];

	status = case_away(&context);
#endif				/* __POINTER_DECK__ */
    }

    return status;
}
