/*
 * sx39.c
 * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, Ciprian Niculescu
 * Copyright (C) 2009, 2010, 2011, 2012, 2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#if defined HAVE_LIBx1f4i0
# include <libx1f4i0.h>
#endif				/* HAVE_LIBx1f4i0 */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <count2.h>
#include <inter.h>
#include <libx1f4l2.h>
#include <s3xx.h>
#if !defined HAVE_LIBx1f4i0
# include <vkeyhook.h>
# include <vkeytree.h>
#endif				/* !HAVE_LIBx1f4i0 */

#define sans__________________		NULL

#define _______MODE			0

#define SANS_LOGIQUE_ARGS_0 \
    struct context_type *, const struct test_type *
#define SANS_LOGIQUE_ARGS_1 \
    struct context_type *context_data, const struct test_type *test_data
#define SANS_LOGIQUE_ARGS_2 \
    context_data, test_data

#define EXCISE_MODE \
    "random delete all in 1,000,000 strings list"

#define EXSECT_MODE \
    "delete all dictionary words"

#define HXSECT_MODE \
    "delete every other dictionary word"

#define LOOKUP_MODE \
    "retrieve all dictionary words"

#define RENDER_MODE \
    "callback traverse the dictionary words"

#define SELECT_MODE \
    "random retrieve all from 1,000,000 strings list"

#define STREAM_TEST \
    "all the tests"

#define STRING_MODE \
    "insert 1,000,000 strings"

#define STRIPE_MODE \
    "insert the dictionary words"

typedef struct context_type {
    const char *dictionary, *self;
} context_type;

typedef struct test_type {
    const char *description;
    int (*test) (SANS_LOGIQUE_ARGS_0);
} test_type;

static int deck_miss(void *, void *);
static int fini_list(struct context_type *, void **);
static int fini_path(struct context_type *, void **);
static int fini_rate(struct context_type *, void **);
static int init_list(struct context_type *, void **);
static int init_path(struct context_type *, void **);
static int init_rate(struct context_type *, void **);
static int list_link(void *, const void *, const char *, unsigned);
static int pick_rate(struct context_type *, void **, unsigned *);
static int miss_node(void *, const char *, unsigned);
static int sans_logique_100806_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100806_1a(SANS_LOGIQUE_ARGS_0, void *, void *);
static int sans_logique_100806_2_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100806_2a(SANS_LOGIQUE_ARGS_0, void *, void *);
static int sans_logique_100810_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100810_0a(SANS_LOGIQUE_ARGS_0, void *, void *);
static int sans_logique_100812_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100812_0a(SANS_LOGIQUE_ARGS_0, void *, void *);
static int sans_logique_100812_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100812_1a(SANS_LOGIQUE_ARGS_0, void *, void *);
static int sans_logique_100812_2_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100812_2a(SANS_LOGIQUE_ARGS_0, void *, void *);
static int sans_logique_120922_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_120922_0a(SANS_LOGIQUE_ARGS_0, void *, unsigned);
static int sans_logique_130221_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_130221_0a(SANS_LOGIQUE_ARGS_0, void *, void *);
static int sans_logique__________(SANS_LOGIQUE_ARGS_0);

static void class(void);
static void usage(void);

static const char digits[] = "0123456789abcdef";
static const struct test_type test_3_9[] = {
/* *INDENT-OFF* */
    {	STREAM_TEST,		sans_logique__________			},
    {	STRING_MODE,		sans_logique_100806_1_			},
    {	SELECT_MODE,		sans_logique_100806_2_			},
    {	EXCISE_MODE,		sans_logique_100810_0_			},
    {	STRIPE_MODE,		sans_logique_100812_0_			},
    {	LOOKUP_MODE,		sans_logique_100812_1_			},
    {	EXSECT_MODE,		sans_logique_100812_2_			},
    {   RENDER_MODE,		sans_logique_120922_0_			},
    {	HXSECT_MODE,		sans_logique_130221_0_			},
    {   _______MODE,		sans__________________			}
/* *INDENT-ON* */
};

static const struct x1f4_vkeytree_type link___e[] = {
/* *INDENT-OFF* */
    {	"rase",		{	test_3_9 + 6	},	NULL		},
    {	"xcise",	{	test_3_9 + 3	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_3_9[] = {
/* *INDENT-OFF* */
    {	"all",		{	test_3_9 + 0	},	NULL		},
    {	"e",		{	NULL		},	link___e	},
    {	"find",		{	test_3_9 + 2	},	NULL		},
    {	"half",		{	test_3_9 + 8	},	NULL		},
    {	"lookup",	{	test_3_9 + 5	},	NULL		},
    {	"render",	{	test_3_9 + 7	},	NULL		},
    {	"string",	{	test_3_9 + 1	},	NULL		},
    {	"words",	{	test_3_9 + 4	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link____[] = {
/* *INDENT-OFF* */
    {	"",		{	NULL		},	link_3_9	},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
};

static int
deck_miss(void *deck, void *miss)
{
    int delete;
    unsigned char *s, *v;

    s = deck;
    v = miss;

    while (1) {
	unsigned char c;

	c = *s;
	if (c ^ *v) {
	    delete = (int) c - (int) *v;

	    break;
	} else {
	    if (c) {
		s++;
		v++;
	    } else {
		delete = 0;

		break;
	    }
	}
    }

    return delete;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_s3deck(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy tree\n", context_data->self);
    }

    return status;
}


static int
fini_path(struct context_type *context_data, void **path)
{
    free(*path);

    return 0;
}


static int
fini_rate(struct context_type *context_data, void **rate)
{
    int status;

    status = x1f4_fini_l2list(rate);
    if (status) {
	fprintf(stderr, "%s: cannot destroy dictionary\n", context_data->self);
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_s3deck(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create tree\n", context_data->self);
    } else {
    }

    return status;
}


static int
init_path(struct context_type *context_data, void **path)
{
#define __DUMP_PROFILE__		0
#define __LAME_MATCHES__		0

    int status;
    void *line;

    status = x1f4_init_tvlist(&line, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create tree\n", context_data->self);
    } else {
	unsigned copy = 0;
#if __LAME_MATCHES__
	unsigned matches = 0;
#endif				/* __LAME_MATCHES__ */
#if __DUMP_PROFILE__
	unsigned profile[24];
#endif				/* __DUMP_PROFILE__ */
	unsigned size = 0;

	srand(0);

#if __DUMP_PROFILE__
	memset(profile, 0, 24 * sizeof(unsigned));
#endif				/* __DUMP_PROFILE__ */

	while (size < 1000000) {
	    char buffer[24], *string;
	    unsigned extend, length;
	    void *data;

	    length = rand() % (256 - 112) + 112;
	    length = length * length * length;
	    length = (1 << 25) / length;

	    extend = length + 1;

	    string = buffer;
	    for (; length; length--) {
		*string++ = digits[rand() & 15];
	    }

	    *string = 0;

	    status = x1f4_post_tvlist
		(line, buffer, deck_miss, 0, extend, &data);
	    if (status == X1f4_TVLINE_EVER_MATCH) {
#if __LAME_MATCHES__
		matches++;
#endif				/* __LAME_MATCHES__ */
	    } else {
		if (status) {
		    break;
		} else {
#if __DUMP_PROFILE__
		    profile[extend - 1]++;
#endif				/* __DUMP_PROFILE__ */
		    memcpy(data, buffer, extend);
		    if (1) {
			size++;
			if (1) {
			    copy += extend;
			}
		    }
		}
	    }
	}

	if (status) {
	} else {
	    x1f4_rand_tvlist(line);

#if __DUMP_PROFILE__
	    if (1) {
		unsigned i = 3, *p;

		p = profile;

		for (; i; i--) {
		    unsigned j = 8;

		    for (; j; j--) {
			fprintf(stderr, " %7u", *p++);
		    }

		    fputc('\n', stderr);
		}
	    }
#endif				/* __DUMP_PROFILE__ */
	}

	if (status) {
#if __LAME_MATCHES__
	} else {
	    fprintf(stderr, "matched for %u times while building dictionary\n",
		    matches);
#endif				/* __LAME_MATCHES__ */
	}

	if (status) {
	} else {
	    void *text;

	    text = malloc(copy + 1000000);
	    if (!text) {
		status = -1;
	    } else {
		unsigned char *slip;
		void *data;

		*path = text;

		slip = text;

		x1f4_lead_tvlist(line, &data);

		do {
		    unsigned extend;

		    extend = strlen(data) + 1;

		    *slip++ = extend;

		    memcpy(slip, data, extend);

		    slip += extend;
		} while (x1f4_next_tvlist(&data, data));
	    }
	}

	x1f4_fini_tvlist(&line);
    }

    return status;

#undef __LAME_MATCHES__
#undef __DUMP_PROFILE__
}


static int
init_rate(struct context_type *context_data, void **rate)
{
    unsigned e;

    return pick_rate(context_data, rate, &e);
}


static int
list_link(void *file, const void *node, const char *data, unsigned size)
{
    fwrite("  ", 2, 1, file);
    fwrite(data, size, 1, file);
    fwrite("                ", 16 - size, 1, file);
    fwrite(" - ", 3, 1, file);
    fputs(((const struct test_type *) node)->description, file);
    fputc('\n', file);

    return 0;
}


static int
miss_node(void *text, const char *name, unsigned size)
{
    return 0;
}


static int
pick_rate(struct context_type *context_data, void **rate, unsigned *fail)
{
    int status;

    status = x1f4_init_l2list(rate, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create dictionary\n", context_data->self);
    } else {
	if (!context_data->dictionary) {
	    fprintf(stderr, "%s: no dictionary defined\n", context_data->self);

	    status = -1;
	} else {
	    int file;

	    file = open(context_data->dictionary, O_RDONLY);
	    if (file == -1) {
		status = 1;

		fprintf(stderr, "%s: cannot open `%s': %s\n",
			context_data->self, context_data->dictionary,
			strerror(errno));
	    } else {
		int done = 0, excess;
		unsigned slip = 0;
		void *list;

		list = *rate;

		while (1) {
		    char *data;
		    unsigned size;

		    status = libx1f4l2_rate_line
			(context_data->self, context_data->dictionary, file,
			 &done, &data, &size);
		    if (status) {
			break;
		    } else {
#if __MISS_VOID_LINES__
			if (size) {
#elif 0
			}
#else
			if (size || !done) {
#endif				/* __MISS_VOID_LINES__ */
			    void *line;

			    status = x1f4_tail_l2list
				(list, sizeof(unsigned) + size + 1, &line);
			    if (status) {
				break;
			    } else {
				unsigned *copy;

				copy = line;

				*copy = size;

				memcpy(copy + 1, data, size + 1);

				if (slip < size) {
				    slip = size;
				}
			    }
			}

			free(data);

			if (done) {
			    break;
			}
		    }
		}

		excess = close(file);
		if (excess) {
		    if (status) {
		    } else {
			status = excess;
		    }

		    fprintf(stderr, "%s: cannot close `%s': %s\n",
			    context_data->self, context_data->dictionary,
			    strerror(errno));
		}

		if (status) {
		} else {
		    *fail = slip;
		}
	    }
	}

	if (status) {
	    x1f4_fini_l2list(rate);
	}
    }

    return status;
}


static int
sans_logique_100806_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_path(context_data, &line);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100806_1a(SANS_LOGIQUE_ARGS_2, line, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }

	    excess = fini_path(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
sans_logique_100806_1a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    unsigned i = 1000000;
    unsigned char *slip;

    slip = line;

    while (i) {
	unsigned extend;

	i--;

	extend = *slip++;

#if __PATH_FOR_DECK__
	status = x1f4_post_s3path(list, (char *) slip, extend, NULL);
#else
	status = x1f4_post_s3deck(list, (char *) slip, NULL);
#endif				/* __PATH_FOR_DECK__ */
	if (status) {
	    break;
	} else {
	    if (1) {
		slip += extend;
	    }
	}
    }

    return status;
}


static int
sans_logique_100806_2_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_path(context_data, &line);
	if (status) {
	} else {
	    status = sans_logique_100806_1a(SANS_LOGIQUE_ARGS_2, line, list);
	    if (status) {
	    } else {
		count2_decls;

		start_count2();

		status = sans_logique_100806_2a
		    (SANS_LOGIQUE_ARGS_2, line, list);

		if (status) {
		} else {
		    finish_count2_l(test_data->description, 55);
		}
	    }

	    excess = fini_path(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
sans_logique_100806_2a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    unsigned i = 1000000;
    unsigned char *slip;

    slip = line;

    while (1) {
	const void *text;

	i--;

#if __PATH_FOR_DECK__
	status = x1f4_find_s3path(list, (char *) (slip + 1), *slip, &text);
#else
	status = x1f4_find_s3deck(list, (char *) (slip + 1), &text);
#endif				/* __PATH_FOR_DECK__ */
	if (status) {
	    if (i) {
		slip += *slip + 1;
	    } else {
		status = 0;
		if (1) {
		    break;
		}
	    }
	} else {
	    fprintf(stderr, "%s: cannot find `%s'\n", context_data->self,
		    (char *) (slip + 1));
	    status = 1;
	    if (1) {
		break;
	    }
	}
    }

    return status;
}


static int
sans_logique_100810_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_path(context_data, &line);
	if (status) {
	} else {
	    status = sans_logique_100806_1a(SANS_LOGIQUE_ARGS_2, line, list);
	    if (status) {
	    } else {
		count2_decls;

		start_count2();

		status = sans_logique_100810_0a
		    (SANS_LOGIQUE_ARGS_2, line, list);

		if (status) {
		} else {
		    finish_count2_l(test_data->description, 55);
		}
	    }

	    excess = fini_path(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
sans_logique_100810_0a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    unsigned i = 1000000;
    unsigned char *slip;

    slip = line;

    while (i) {
	i--;

#if __PATH_FOR_DECK__
# if __HEAD_CAST_DECK__
	status = x1f4_cast_s3path(list, (char *) (slip + 1), *slip);
# else
	status = x1f4_pass_s3path(list, (char *) (slip + 1), *slip);
# endif				/* __HEAD_CAST_DECK__ */
#else
# if __HEAD_CAST_DECK__
	status = x1f4_cast_s3deck(list, (char *) (slip + 1));
# else
	status = x1f4_pass_s3deck(list, (char *) (slip + 1));
# endif				/* __HEAD_CAST_DECK__ */
#endif				/* __PATH_FOR_DECK__ */
	if (status) {
	    fprintf(stderr, "%s: cannot delete %u long `%s'\n",
		    context_data->self, *slip, (char *) (slip + 1));
	    break;
	} else {
	    if (1) {
		slip += *slip + 1;
	    }
	}
    }

    return status;
}


static int
sans_logique_100812_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_rate(context_data, &line);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100812_0a(SANS_LOGIQUE_ARGS_2, line, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }

	    excess = fini_rate(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
sans_logique_100812_0a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    void *data;

    status = x1f4_lead_l2list(line, &data);
    if (status == X1f4_L2LIST_DECK_ERROR) {
	status = 0;
    } else {
	do {
	    unsigned *call;

	    call = data;
#if __PATH_FOR_DECK__
	    status = x1f4_post_s3path(list, (void *) (call + 1), *call, NULL);
#else
	    status = x1f4_post_s3deck(list, (void *) (call + 1), NULL);
#endif				/* __PATH_FOR_DECK__ */
	    if (status) {
		if (status == X1f4_L2CLIP_EVER_MATCH) {
		    status = 0;
		} else {
		    break;
		}
	    }
	} while (x1f4_next_l2list(&data, data));
    }

    return status;
}


static int
sans_logique_100812_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_rate(context_data, &line);
	if (status) {
	} else {
	    status = sans_logique_100812_0a(SANS_LOGIQUE_ARGS_2, line, list);
	    if (status) {
	    } else {
		count2_decls;

		start_count2();

		status = sans_logique_100812_1a
		    (SANS_LOGIQUE_ARGS_2, line, list);

		if (status) {
		} else {
		    finish_count2_l(test_data->description, 55);
		}
	    }

	    excess = fini_rate(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
sans_logique_100812_1a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    void *data;

    status = x1f4_lead_l2list(line, &data);
    if (status == X1f4_L2LIST_DECK_ERROR) {
	status = 0;
    } else {
	do {
	    unsigned *call;
	    const void *text;

	    call = data;
#if __PATH_FOR_DECK__
	    status = x1f4_find_s3path(list, (void *) (call + 1), *call, &text);
#else
	    status = x1f4_find_s3deck(list, (void *) (call + 1), &text);
#endif				/* __PATH_FOR_DECK__ */
	    if (status) {
		status = 0;
	    } else {
		fprintf(stderr, "%s: cannot find %u long `%s'\n",
			context_data->self, *call, (char *) (call + 1));
		status = 1;
		if (1) {
		    break;
		}
	    }
	} while (x1f4_next_l2list(&data, data));
    }

    return status;
}


static int
sans_logique_100812_2_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_rate(context_data, &line);
	if (status) {
	} else {
	    status = sans_logique_100812_0a(SANS_LOGIQUE_ARGS_2, line, list);
	    if (status) {
	    } else {
		count2_decls;

		start_count2();

		status = sans_logique_100812_2a
		    (SANS_LOGIQUE_ARGS_2, line, list);

		if (status) {
		} else {
		    finish_count2_l(test_data->description, 55);
		}
	    }

	    excess = fini_rate(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
sans_logique_100812_2a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    void *data;

    status = x1f4_lead_l2list(line, &data);
    if (status == X1f4_L2LIST_DECK_ERROR) {
	status = 0;
    } else {
	do {
	    unsigned *call;

	    call = data;
#if __PATH_FOR_DECK__
# if __HEAD_CAST_DECK__
	    status = x1f4_cast_s3path(list, (void *) (call + 1), *call);
# else
	    status = x1f4_pass_s3path(list, (void *) (call + 1), *call);
# endif				/* __HEAD_CAST_DECK__ */
#else
# if __HEAD_CAST_DECK__
	    status = x1f4_cast_s3deck(list, (void *) (call + 1));
# else
	    status = x1f4_pass_s3deck(list, (void *) (call + 1));
# endif				/* __HEAD_CAST_DECK__ */
#endif				/* __PATH_FOR_DECK__ */
	    if (status) {
		if (status == X1f4_L2CLIP_MISS_CLASS) {
		    if (1) {
			fprintf(stderr, "%s: cannot delete %u long `%s'\n",
				context_data->self, *call,
				(char *) (call + 1));
			break;
		    } else {
			/*
			 * have an option to allow this one?
			 */
			status = 0;
		    }
		} else {
		    fprintf(stderr, "%s: cannot delete %u long `%s'\n",
			    context_data->self, *call, (char *) (call + 1));
		    break;
		}
	    }
	} while (x1f4_next_l2list(&data, data));
    }

    return status;
}


static int
sans_logique_120922_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	unsigned copy;
	void *line = NULL;

	status = pick_rate(context_data, &line, &copy);
	if (status) {
	} else {
	    status = sans_logique_100812_0a(SANS_LOGIQUE_ARGS_2, line, list);
	    if (status) {
	    } else {
		count2_decls;

		start_count2();

		status = sans_logique_120922_0a
		    (SANS_LOGIQUE_ARGS_2, list, copy);

		if (status) {
		} else {
		    finish_count2_l(test_data->description, 55);
		}
	    }

	    excess = fini_rate(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
sans_logique_120922_0a(SANS_LOGIQUE_ARGS_1, void *list, unsigned copy)
{
    int status;
#if __HEAD_LIST_DECK__
    void *miss;
#endif				/* __HEAD_LIST_DECK__ */

#if __HEAD_LIST_DECK__
    miss = malloc(copy + 1);
    if (miss) {
	status = x1f4_list_s3deck(list, NULL, miss_node, miss);
#endif				/* __HEAD_LIST_DECK__ */
#if __HEAD_NOTE_DECK__
	status = x1f4_note_s3deck(list, NULL, miss_node);
#endif				/* __HEAD_NOTE_DECK__ */
#if __HEAD_LIST_DECK__

	free(miss);
    } else {
	status = -1;
    }
#endif				/* __HEAD_LIST_DECK__ */

    return status;
}


static int
sans_logique_130221_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_rate(context_data, &line);
	if (status) {
	} else {
	    status = sans_logique_100812_0a(SANS_LOGIQUE_ARGS_2, line, list);
	    if (status) {
	    } else {
		count2_decls;

		start_count2();

		status = sans_logique_130221_0a
		    (SANS_LOGIQUE_ARGS_2, line, list);

		if (status) {
		} else {
		    finish_count2_l(test_data->description, 55);
		}
	    }

	    excess = fini_rate(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
sans_logique_130221_0a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    void *data;

    status = x1f4_lead_l2list(line, &data);
    if (status == X1f4_L2LIST_DECK_ERROR) {
	status = 0;
    } else {
	do {
	    unsigned *call;

	    call = data;
#if __PATH_FOR_DECK__
# if __HEAD_CAST_DECK__
	    status = x1f4_cast_s3path(list, (void *) (call + 1), *call);
# else
	    status = x1f4_pass_s3path(list, (void *) (call + 1), *call);
# endif				/* __HEAD_CAST_DECK__ */
#else
# if __HEAD_CAST_DECK__
	    status = x1f4_cast_s3deck(list, (void *) (call + 1));
# else
	    status = x1f4_pass_s3deck(list, (void *) (call + 1));
# endif				/* __HEAD_CAST_DECK__ */
#endif				/* __PATH_FOR_DECK__ */
	    if (status) {
		if (status == X1f4_L2CLIP_MISS_CLASS) {
		    if (1) {
			fprintf(stderr, "%s: cannot delete %u long `%s'\n",
				context_data->self, *call,
				(char *) (call + 1));
			break;
		    } else {
			/*
			 * have an option to allow this one?
			 */
			status = 0;
		    }
		} else {
		    fprintf(stderr, "%s: cannot delete %u long `%s'\n",
			    context_data->self, *call, (char *) (call + 1));
		    break;
		}
	    }

	    x1f4_next_l2list(&data, data);
	} while (x1f4_next_l2list(&data, data));
    }

    return status;
}


static int
sans_logique__________(SANS_LOGIQUE_ARGS_1)
{
    int status;

    test_data = test_3_9 + 1;

    do {
	status = test_data->test(SANS_LOGIQUE_ARGS_2);
	if (status) {
	    break;
	} else {
	    test_data++;
	}
    } while (test_data->test);

    return status;
}


static void
class(void)
{
    x1f4_hook_vtkey(stdout, list_link, link____);
}


static void
usage(void)
{
    puts("Usage: sx39 [OPTIONS] TEST\n\
Profile the " sxname "tree associative array.\n\
\n\
Options:\n\
  -d, --dictionary PATH		set dictionary as PATH\n\
      --list			lists available tests and exit\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    context.dictionary = NULL;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "dictionary",   0x01,   NULL,   0x64    },
		{   "list",	    0x00,   NULL,   0x6c    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "d:", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'd':
		context.dictionary = optarg;
		break;
	    case 'h':
	    case 'l':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'l':
		class();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (argc - optind != 1) {
	fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' for "
		"more information.\n", argv[0], argv[0]);

	return 1;
    } else {
	const void *test;

	if (x1f4_find_vtkey(argv[optind], link____, &test)) {
	    status = 1;

	    fprintf(stderr, "%s: cannot parse test specification: `%s'\n",
		    argv[0], argv[optind]);
	} else {
	    context.self = argv[0];

	    status = ((struct test_type *) test)->test(&context, test);
	}
    }

    return status;
}
