/*
 * tc12.c
 * Copyright (C) 2002-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <inter.h>
#include <libx1f4l2.h>
#include <tcxx.h>

#define __SKIP_COPY_LIST__		1
#define __SKIP_TILE_LIST__		1

#define __LINK_COPY_LIST__		!__SKIP_COPY_LIST__
#define __LINK_TILE_LIST__		!__SKIP_TILE_LIST__

typedef struct context_type {
    const char *self;
} context_type;

static int case_away(struct context_type *);
#if __LINK_COPY_LIST__
static int copy_list(struct context_type *, void **);
#endif				/* __LINK_COPY_LIST__ */
#if __LINK_COPY_LIST__
static int copy_node(void *, void *, const void *);
#endif				/* __LINK_COPY_LIST__ */
static int dump_list(struct context_type *, void *);
static int fast_line(struct context_type *, void *, char *, unsigned,
		     unsigned);
static int fini_list(struct context_type *, void **);
static int flat_list(struct context_type *, void *);
#if __LINK_COPY_LIST__
static int free_node(void *, void *);
#endif				/* __LINK_COPY_LIST__ */
static int head_line(struct context_type *, void *, char *, unsigned);
static int init_list(struct context_type *, void **);
static int lock_line(struct context_type *, void *, char *, unsigned,
		     unsigned *);
#if __LINK_TILE_LIST__
static int miss_list(struct context_type *, void *, unsigned);
#endif				/* __LINK_TILE_LIST__ */
static int post_line(struct context_type *, void *, char *, unsigned);
static int read_list(struct context_type *, void *);
static int seek_miss(void *, void *);
#if __LINK_COPY_LIST__
static int size_node(void *, unsigned *, const void *);
#endif				/* __LINK_COPY_LIST__ */

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = read_list(context_data, list);
	if (status) {
	} else {
#if __LINK_COPY_LIST__
	    status = copy_list(context_data, &list);
	    if (status) {
	    } else {
#endif				/* __LINK_COPY_LIST__ */
		status = dump_list(context_data, list);
		if (status) {
		} else {
		    status = flat_list(context_data, list);
		}
#if __LINK_COPY_LIST__
	    }
#endif				/* __LINK_COPY_LIST__ */
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


#if __LINK_COPY_LIST__
static int
copy_list(struct context_type *context_data, void **away)
{
    int status;
    void *copy, *list;

    list = *away;

    status = x1f4_text_tclist
	(&copy, list, size_node, copy_node, free_node, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot copy list\n", context_data->self);
    } else {
	*away = copy;

	x1f4_fini_tclist(&list);
    }

    return status;
}
#endif				/* __LINK_COPY_LIST__ */



#if __LINK_COPY_LIST__
static int
copy_node(void *text, void *copy, const void *data)
{
    strcpy(copy, data);

    return 0;
}
#endif				/* __LINK_COPY_LIST__ */


static int
dump_list(struct context_type *context_data, void *list)
{
    int status = 0;
    unsigned i = 0, size;

    x1f4_size_tclist(list, &size);
    while (i < size) {
	void *peek;

	status = x1f4_peek_tclist(list, i, &peek);
	if (status) {
	    break;
	} else {
	    puts(peek);
	    i++;
	}
    }

    return status;
}


static int
fast_line(struct context_type *context_data, void *list, char *data,
	  unsigned size, unsigned post)
{
    int status;
    void *slip;

    status = x1f4_push_tclist(list, post, size + 1, &slip);
    if (status) {
	fprintf(stderr, "%s: cannot insert item\n", context_data->self);
    } else {
	if (0) {
	    status = x1f4_flow_tclist(list);
	}
	if (status) {
	    fprintf(stderr, "%s: corrupt list\n", context_data->self);
	} else {
	    memcpy(slip, data, size + 1);
	}
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_tclist(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
flat_list(struct context_type *context_data, void *list)
{
    int status = 0;
    unsigned miss, size;

    srand(time(NULL));

    x1f4_size_tclist(list, &size);

    miss = sqrt(size);

    miss |= miss >> 001;
    miss |= miss >> 002;
    miss |= miss >> 004;
    miss |= miss >> 010;
    miss |= miss >> 020;

    miss >>= 1;

    while (size) {
#if __SKIP_TILE_LIST__
	status = x1f4_miss_tclist(list, rand() % size);
#else
	status = miss_list(context_data, list, rand() % size);
#endif				/* __SKIP_TILE_LIST__ */
	if (status) {
	    break;
	} else {
	    if (1) {
		size--;
	    }
	    if (!(size & miss)) {
		status = x1f4_flow_tclist(list);
		if (status) {
		    fprintf(stderr, "%s: corrupt list\n", context_data->self);

		    break;
		}
	    }
	}
    }

    return status;
}


#if __LINK_COPY_LIST__
static int
free_node(void *text, void *data)
{
    return 0;
}
#endif				/* __LINK_COPY_LIST__ */


static int
head_line(struct context_type *context_data, void *list, char *data,
	  unsigned size)
{
    int status;
    void *slip;

    status = x1f4_push_tclist(list, 0, size + 1, &slip);
    if (status) {
	fprintf(stderr, "%s: cannot insert item\n", context_data->self);
    } else {
	status = x1f4_flow_tclist(list);
	if (status) {
	    fprintf(stderr, "%s: corrupt list\n", context_data->self);
	} else {
	    memcpy(slip, data, size + 1);
	}
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_tclist(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    }

    return status;
}


static int
lock_line(struct context_type *context_data, void *list, char *data,
	  unsigned size, unsigned *lock)
{
    if (x1f4_fgxp_tclist(list, data, 0, seek_miss, lock)) {
    } else {
	x1f4_size_tclist(list, lock);
    }

    return 0;
}


#if __LINK_TILE_LIST__
static int
miss_list(struct context_type *context_data, void *list, unsigned tile)
{
    int status;
    void *data;

    status = x1f4_peek_tclist(list, tile, &data);
    if (status) {
    } else {
	unsigned back;

	status = x1f4_tile_tclist(&back, data);
	if (status) {
	} else {
	    if (back != tile) {
		fprintf(stderr, "%s: failed index test (got %u, was looking "
			"for %u)\n", context_data->self, back, tile);

		status = 1;
	    } else {
		status = x1f4_miss_tclist(list, tile);
	    }
	}
    }

    return status;
}
#endif				/* __LINK_TILE_LIST__ */


static int
post_line(struct context_type *context_data, void *list, char *data,
	  unsigned size)
{
    int status;
    unsigned post;

    status = lock_line(context_data, list, data, size, &post);
    if (status) {
    } else {
	status = fast_line(context_data, list, data, size, post);
    }

    return status;
}


static int
read_list(struct context_type *context_data, void *list)
{
    int status;

    do {
	char *data;
	int done = 0;
	unsigned size = 0;

	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (!done || size) {
		status = head_line(context_data, list, data, size);
	    }

	    if (1) {
		free(data);
	    }

	    if (status) {
		break;
	    }

	    if (done) {
		break;
	    } else {
		while (1) {
		    status = libx1f4l2_read_line
			(context_data->self, &done, &data, &size);
		    if (status) {
			break;
		    }

		    if (!done || size) {
			status = post_line(context_data, list, data, size);
		    }

		    if (1) {
			free(data);
		    }

		    if (status) {
			break;
		    }

		    if (done) {
			break;
		    }
		}
	    }
	}
    } while (0);

    return status;
}


static int
seek_miss(void *deck, void *miss)
{
    int delete;
    unsigned char *s, *v;

    s = deck;
    v = miss;

    while (1) {
	unsigned char c;

	c = *s;
	if (c ^ *v) {
	    delete = (int) c - (int) *v;

	    break;
	} else {
	    if (c) {
		s++;
		v++;
	    } else {
		delete = 0;

		break;
	    }
	}
    }

    return delete;
}


#if __LINK_COPY_LIST__
static int
size_node(void *text, unsigned *size, const void *data)
{
    *size = strlen(data) + 1;

    return 0;
}
#endif				/* __LINK_COPY_LIST__ */


static void
usage(void)
{
    puts("Usage: tc12 [OPTIONS]\n\
Test " tcname "_delete(), output sorted input.\n\
\n\
Options:\n\
  --help			display this help and exit\n\
  --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	struct context_type context;

	context.self = argv[0];

	status = case_away(&context);
    }

    return status;
}
