/*
 * tc33.c
 * Copyright (C) 2002-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <inter.h>
#include <libx1f4l2.h>
#include <tcqp.h>

typedef struct context_type {
    const char *self;
    int second;
    void *slip;
} context_type;

static int case_away(struct context_type *);
static int deck_miss(void *, void *);
static int dump_list(struct context_type *, void *);
#if !__HEAD_NEXT_LINE__
static int dump_node(void *, void *);
#endif				/* !__HEAD_NEXT_LINE__ */
static int fast_away(struct context_type *);
static int fast_line(struct context_type *, void *, unsigned, char *,
		     unsigned);
static int fini_list(struct context_type *, void **);
static int fini_slip(struct context_type *);
static int init_list(struct context_type *, void **);
static int init_slip(struct context_type *);
static int miss_half(struct context_type *, void *);
static int read_list(struct context_type *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = read_list(context_data, list);
	if (status) {
	} else {
	    status = miss_half(context_data, list);
	    if (status) {
	    } else {
		status = dump_list(context_data, list);
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
deck_miss(void *deck, void *miss)
{
    int delete;
    unsigned char *s, *v;

    s = deck;
    v = miss;

    while (1) {
	unsigned char c;

	c = *s;
	if (c ^ *v) {
	    delete = (int) c - (int) *v;

	    break;
	} else {
	    if (c) {
		s++;
		v++;
	    } else {
		delete = 0;

		break;
	    }
	}
    }

    return delete;
}


static int
dump_list(struct context_type *context_data, void *list)
{
#if !__HEAD_NEXT_LINE__
    x1f4_lime_tcline(list, NULL, dump_node);
#else
    if (x1f4_case_tcline(list)) {
	void *node;

	if (x1f4_lead_tcline(list, &node)) {
	} else {
	    do {
		puts(node);
	    } while (x1f4_next_tcline(&node, node));
	}
    }
#endif				/* !__HEAD_NEXT_LINE__ */

    return 0;
}


#if !__HEAD_NEXT_LINE__
static int
dump_node(void *text, void *node)
{
    puts(node);

    return 0;
}
#endif				/* !__HEAD_NEXT_LINE__ */


static int
fast_away(struct context_type *context_data)
{
    int status;

    status = init_slip(context_data);
    if (status) {
    } else {
	int excess;

	status = case_away(context_data);

	excess = fini_slip(context_data);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
fast_line(struct context_type *context_data, void *list, unsigned index,
	  char *data, unsigned size)
{
    int status;
    void *slip;

    status = x1f4_post_tcline(list, data, deck_miss, 0, size + 1, &slip);
    if (status == X1f4_TCLINE_EVER_MATCH) {
	status = 0;
    } else {
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);
	} else {
	    int second;

	    memcpy(slip, data, size + 1);

	    second = context_data->second;
	    if (second) {
		context_data->second = 0;
	    } else {
		void *miss;

		context_data->second = 1;

		second = x1f4_tail_tslist
		    (context_data->slip, sizeof(void *), &miss);
		if (second) {
		} else {
		    void **record;

		    record = miss;
		    *record = slip;
		}
	    }
	}
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_tcline(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
fini_slip(struct context_type *context_data)
{
    int status;

    status = x1f4_fini_tslist(&context_data->slip);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_tcline(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
    }

    return status;
}


static int
init_slip(struct context_type *context_data)
{
    int status;

    status = x1f4_init_tslist(&context_data->slip, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
    }

    return status;
}


static int
miss_half(struct context_type *context_data, void *list)
{
    int status;
    unsigned size;
    void *ever;

    ever = context_data->slip;

    x1f4_size_tslist(ever, &size);
    if (size) {
	void *peek;

	x1f4_peek_tslist(ever, 0, &peek);
	do {
	    status = x1f4_pass_tcline(list, *(void **) peek, deck_miss, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot delete item\n",
			context_data->self);
		if (1) {
		    break;
		}
	    }
	} while (x1f4_next_tslist(&peek, peek));
    } else {
	status = 0;
    }

    return status;
}


static int
read_list(struct context_type *context_data, void *list)
{
    char *data;
    int done = 0, status;
    unsigned i = 0, size = 0;

    while (1) {
	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (!done || size) {
		status = fast_line(context_data, list, i, data, size);
	    }

	    if (1) {
		free(data);
	    }

	    if (status) {
		break;
	    } else {
		i++;
	    }

	    if (done) {
		break;
	    }
	}
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: tc33 [OPTIONS]\n\
Test " tcname "_delete(), output every second input line.\n\
\n\
Options:\n\
  -1, --first			start with first line\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.second = 0;

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "first",	    0x00,   NULL,   0x31    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "1", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case '1':
		context.second = 1;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	status = fast_away(&context);
    }

    return status;
}
