/*
 * ts13.c
 * Copyright (C) 2002-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libx1f4l2.h>
#include <tcxx.h>

typedef struct context_type {
    char **argv;
    const char *self;
    int argc, do_reverse;
} context_type;

typedef struct pair_type {
    void *deck, *miss;
} pair_type;

static int case_away(struct context_type *);
static int deck_miss(void *, void *, void *);
static int dump_list(struct context_type *, void *);
static int fast_line(struct context_type *, void *, char *, char *);
static int fini_list(struct context_type *, void **);
static int init_list(struct context_type *, void **);
static int read_list(struct context_type *, void *);
static int sort_list(struct context_type *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = read_list(context_data, list);
	if (status) {
	} else {
	    status = sort_list(context_data, list);
	    if (status) {
	    } else {
		status = dump_list(context_data, list);
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
deck_miss(void *text, void *deck, void *miss)
{
    int delete;
    unsigned char *s, *v;

    s = ((struct pair_type *) deck)->deck;
    v = ((struct pair_type *) miss)->deck;

    while (1) {
	unsigned char c;

	c = *s;
	if (c ^ *v) {
	    delete = (int) c - (int) *v;

	    break;
	} else {
	    if (c) {
		s++;
		v++;
	    } else {
		delete = 0;

		break;
	    }
	}
    }

    return delete;
}


static int
dump_list(struct context_type *context_data, void *list)
{
    unsigned size;

    x1f4_size_tclist(list, &size);
    if (size) {
#if !__HEAD_NEXT_LIST__
	unsigned i = 1;
#endif				/* !__HEAD_NEXT_LIST__ */
	void *pair;

	x1f4_peek_tclist(list, 0, &pair);

	fputs(((struct pair_type *) pair)->deck, stdout);
	fputs(" ", stdout);
	fputs(((struct pair_type *) pair)->miss, stdout);

	size--;

	while (size) {
#if !__HEAD_NEXT_LIST__
	    x1f4_peek_tclist(list, i, &pair);
#else
	    x1f4_next_tclist(&pair, pair);
#endif				/* !__HEAD_NEXT_LIST__ */

	    fputs("   ", stdout);
	    fputs(((struct pair_type *) pair)->deck, stdout);
	    fputs(" ", stdout);
	    fputs(((struct pair_type *) pair)->miss, stdout);

#if !__HEAD_NEXT_LIST__
	    i++;
#endif				/* !__HEAD_NEXT_LIST__ */

	    size--;
	}

	fputs("\n", stdout);
    }

    return 0;
}


static int
fast_line(struct context_type *context_data, void *list, char *deck,
	  char *miss)
{
    int status;
    void *slip;

    status = x1f4_tail_tclist(list, sizeof(struct pair_type), &slip);
    if (status) {
	fprintf(stderr, "%s: cannot insert item\n", context_data->self);
    } else {
	struct pair_type *pair_data;

	pair_data = slip;

	pair_data->deck = deck;
	pair_data->miss = miss;
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_tclist(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_tclist(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
    }

    return status;
}


static int
read_list(struct context_type *context_data, void *list)
{
    int status;
    unsigned size;

    size = context_data->argc >> 1;
    if (size) {
	char **argv;

	argv = context_data->argv;

	while (1) {
	    status = fast_line(context_data, list, *argv, *(argv + 1));
	    if (status) {
		break;
	    } else {
		size--;
		if (size) {
		    argv += 2;
		} else {
		    break;
		}
	    }
	}
    } else {
	status = 0;
    }

    return status;
}


static int
sort_list(struct context_type *context_data, void *list)
{
    return x1f4_sort_tclist(list, NULL, 0, deck_miss);
}


static void
usage(void)
{
    puts("Usage: ts13 [OPTIONS] [MAJOR MINOR]...\n\
Test " tcname "_sort(), output (by first) sorted command line string pairs.\n\
\n\
Options:\n\
  -r, --reverse			does nothing\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.do_reverse = 0;

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "reverse",	    0x00,   NULL,   0x72    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "r", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'r':
		context.do_reverse = 1;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if ((argc - optind) & 1) {
	fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' for "
		"more information.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	context.argc = argc - optind;
	context.argv = argv + optind;

	status = case_away(&context);
    }

    return status;
}
