/*
 * ts20.c
 * Copyright (C) 2002-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libx1f4l2.h>
#include <tcxx.h>

typedef struct context_type {
    char **argv, *seek;
    const char *self;
    int argc, do_reverse;
} context_type;

static int case_away(struct context_type *);
static int deck_miss(void *, void *, void *);
static int dump_list(struct context_type *, void *);
static int ever_seek(const char *, const char *, unsigned);
static int fast_line(struct context_type *, void *, char *);
static int fini_list(struct context_type *, void **);
static int init_list(struct context_type *, void **);
static int peek_list(struct context_type *, void *);
static int read_list(struct context_type *, void *);
static int seek_list(struct context_type *, void *);
static int seek_miss(void *, void *);
static int sort_list(struct context_type *, void *);
static int tick_list(struct context_type *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = read_list(context_data, list);
	if (status) {
	} else {
	    status = sort_list(context_data, list);
	    if (status) {
	    } else {
		status = dump_list(context_data, list);
		if (status) {
		} else {
		    status = tick_list(context_data, list);
		    if (status) {
		    } else {
			status = dump_list(context_data, list);
			if (status) {
			} else {
			    status = seek_list(context_data, list);
			    if (status) {
			    } else {
				status = dump_list(context_data, list);
				if (status) {
				} else {
				    status = peek_list(context_data, list);
				    if (status) {
				    } else {
					status = dump_list(context_data, list);
					if (status) {
					} else {
					}
				    }
				}
			    }
			}
		    }
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
deck_miss(void *text, void *deck, void *miss)
{
    int delete;
    unsigned char *s, *v;

    s = (unsigned char *) *(char **) deck;
    v = (unsigned char *) *(char **) miss;

    while (1) {
	unsigned char c;

	c = *s;
	if (c ^ *v) {
	    delete = (int) c - (int) *v;

	    break;
	} else {
	    if (c) {
		s++;
		v++;
	    } else {
		delete = 0;

		break;
	    }
	}
    }

    return delete;
}


static int
dump_list(struct context_type *context_data, void *list)
{
    unsigned size;

    x1f4_size_tclist(list, &size);
    if (size) {
	void *miss;

	fputs("   ", stdout);

	x1f4_peek_tclist(list, 0, &miss);

	fputs(*(char **) miss, stdout);

	size--;

	while (size) {
	    x1f4_next_tclist(&miss, miss);

	    fputs("   ", stdout);
	    fputs(*(char **) miss, stdout);

	    size--;
	}

	fputs("\n", stdout);
    }

    return 0;
}


static int
ever_seek(const char *ever, const char *slip, unsigned seek)
{
    fputs(ever, stdout);

    for (; seek; seek--) {
	fputs("    ", stdout);
    }

    fputs(" ", stdout);
    fputs(slip, stdout);
    fputs("\n", stdout);

    return 0;
}


static int
fast_line(struct context_type *context_data, void *list, char *deck)
{
    int status;
    void *slip;

    status = x1f4_tail_tclist(list, sizeof(char *), &slip);
    if (status) {
	fprintf(stderr, "%s: cannot insert item\n", context_data->self);
    } else {
	*(char **) slip = deck;
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_tclist(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_tclist(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
    }

    return status;
}


static int
peek_list(struct context_type *context_data, void *list)
{
    unsigned index;

    if (x1f4_seek_tclist(list, context_data->seek, 0, seek_miss, &index)) {
	ever_seek("*", " ^ ", index);
    } else {
	fputs("*\n", stdout);
    }

    return 0;
}


static int
read_list(struct context_type *context_data, void *list)
{
    int status;
    unsigned size;

    size = context_data->argc;
    if (size) {
	char **argv;

	argv = context_data->argv;

	while (1) {
	    status = fast_line(context_data, list, *argv);
	    if (status) {
		break;
	    } else {
		size--;
		if (size) {
		    argv++;
		} else {
		    break;
		}
	    }
	}
    } else {
	status = 0;
    }

    return status;
}


static int
seek_list(struct context_type *context_data, void *list)
{
    unsigned index;

    if (x1f4_llxp_tclist(list, context_data->seek, 0, seek_miss, &index)) {
	ever_seek("L", " ^>", index);
    } else {
	fputs("L\n", stdout);
    }
    if (x1f4_lxxp_tclist(list, context_data->seek, 0, seek_miss, &index)) {
	ever_seek("l", " ^>", index);
    } else {
	fputs("l\n", stdout);
    }
    if (x1f4_fxxp_tclist(list, context_data->seek, 0, seek_miss, &index)) {
	ever_seek("g", "<^ ", index);
    } else {
	fputs("g\n", stdout);
    }
    if (x1f4_fgxp_tclist(list, context_data->seek, 0, seek_miss, &index)) {
	ever_seek("G", "<^ ", index);
    } else {
	fputs("G\n", stdout);
    }

    return 0;
}


static int
seek_miss(void *deck, void *miss)
{
    int delete;
    unsigned char *s, *v;

    s = deck;
    v = (unsigned char *) *(char **) miss;

    while (1) {
	unsigned char c;

	c = *s;
	if (c ^ *v) {
	    delete = (int) c - (int) *v;

	    break;
	} else {
	    if (c) {
		s++;
		v++;
	    } else {
		delete = 0;

		break;
	    }
	}
    }

    return delete;
}


static int
sort_list(struct context_type *context_data, void *list)
{
    return x1f4_sort_tclist(list, NULL, 0, deck_miss);
}


static int
tick_list(struct context_type *context_data, void *list)
{
    unsigned index;

    if (x1f4_fmxp_tclist(list, context_data->seek, 0, seek_miss, &index)) {
	ever_seek("F", "<^ ", index);
    } else {
	fputs("F\n", stdout);
    }
    if (x1f4_lmxp_tclist(list, context_data->seek, 0, seek_miss, &index)) {
	ever_seek("L", " ^>", index);
    } else {
	fputs("L\n", stdout);
    }

    return 0;
}


static void
usage(void)
{
    puts("Usage: ts20 [OPTIONS] SEEK [STRING]...\n\
Test " tcname "_find()s, look for SEEK in STRING(s).\n\
\n\
Options:\n\
  -r, --reverse			does nothing\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.do_reverse = 0;

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "reverse",	    0x00,   NULL,   0x72    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "r", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'r':
		context.do_reverse = 1;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if ((argc - optind) < 1) {
	fprintf(stderr, "%s: too few arguments\nTry `%s --help' for more inf"
		"ormation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	context.seek = argv[optind];

	optind++;

	context.argc = argc - optind;
	context.argv = argv + optind;

	status = case_away(&context);
    }

    return status;
}
