/*
 * tv59.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBx1f4i0
# include <libx1f4i0.h>
#endif				/* HAVE_LIBx1f4i0 */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libx1f4l2.h>
#if !defined HAVE_LIBx1f4i0
# include <vkeyhook.h>
# include <vkeytree.h>
#endif				/* !HAVE_LIBx1f4i0 */

#if SIZEOF_VOID_P == SIZEOF_LONG
# define integral_q			unsigned long
#else
# define integral_q			unsigned
#endif				/* SIZEOF_VOID_P == SIZEOF_LONG */

#define pointer(e) \
    ((void *) (integral_q) (e))

#define here_trans() \
    ((struct x1f4_trans_type *) &trans)

#define SANS_LOGIQUE_ARGS_0 \
    struct context_type *, const struct test_type *
#define SANS_LOGIQUE_ARGS_1 \
    struct context_type *context_data, const struct test_type *test_data
#define SANS_LOGIQUE_ARGS_2 \
    context_data, test_data

#define SANS_LOGIQUE_args_0 \
    struct context_type *, const struct test_type *, void **, const void **
#define SANS_LOGIQUE_args_1 \
    struct context_type *context_data, const struct test_type *test_data,     \
    void **pset, const void **tile
#define SANS_LOGIQUE_args_2 \
    context_data, test_data, pset, tile

#define FOUR87_MODE \
    "insert and delete 4 pointers"

#define CHANGE_MODE \
    "insert 1,000,000, insert and delete 1,000,000 unmatched"

#define INSERT_MODE \
    "insert and delete 1,000,000 distinct pointers"

#define STREAM_TEST \
    "all the tests"

typedef struct context_type {
    const char *self;
} context_type;

typedef struct pair_type {
    const char *line, *miss;
} pair_type;

typedef struct test_type {
    const char *description;
    int (*test) (SANS_LOGIQUE_ARGS_0);
} test_type;

static int deck_slip(void *, void *);
static int fini_list(struct context_type *, void **, const void **);
static int free_data(void *, void *);
static int init_list(struct context_type *, void **, const void **);
static int link_data(void *, void **, unsigned);
static int list_link(void *, const void *, const char *, unsigned);
static int lose_node(void *, void *);
static int miss_deck(void *, void *, void *);
static int miss_line(void *, void *);
static int miss_node(void *, void *);
static int mode_data(void *, void **, unsigned);
static int sans_logique_091002_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_091002_0a(SANS_LOGIQUE_args_0, void *);
static int sans_logique_091002_0b(SANS_LOGIQUE_args_0, void *);
static int sans_logique_091002_0c(SANS_LOGIQUE_args_0, void *);
static int sans_logique_091002_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_091002_1a(SANS_LOGIQUE_args_0, const void *);
static int sans_logique_091002_2_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_091002_2a(SANS_LOGIQUE_args_0, void *);
static int sans_logique_091002_2b(SANS_LOGIQUE_args_0, void *);
static int sans_logique_091002_2c(SANS_LOGIQUE_args_0, void *);
static int sans_logique__________(SANS_LOGIQUE_ARGS_0);

static void class(void);
static void usage(void);

static const struct pair_type test_5_9[] = {
/* *INDENT-OFF* */
    {	"11",	"1"	},
    {	"111",	"1"	},
    {	"1111",	"1"	},
    {	"1112",	"12"	},
    {	"112",	"12"	},
    {	"1121",	"12"	},
    {	"1122",	"12"	},
    {	"1123",	"123"	},
    {	"1132",	"123"	},
    {	"12",	"12"	},
    {	"121",	"12"	},
    {	"1211",	"12"	},
    {	"1212",	"12"	},
    {	"1213",	"123"	},
    {	"122",	"12"	},
    {	"1221",	"12"	},
    {	"1222",	"12"	},
    {	"1223",	"123"	},
    {	"123",	"123"	},
    {	"1231",	"123"	},
    {	"1232",	"123"	},
    {	"1233",	"123"	},
    {	"1234",	"1234"	},
    {	"1243",	"1234"	},
    {	"1312",	"123"	},
    {	"132",	"123"	},
    {	"1321",	"123"	},
    {	"1322",	"123"	},
    {	"1323",	"123"	},
    {	"1324",	"1234"	},
    {	"1332",	"123"	},
    {	"1342",	"1234"	},
    {	"1423",	"1234"	},
    {	"1432",	"1234"	},
    {	"21",	"12"	},
    {	"211",	"12"	},
    {	"2111",	"12"	},
    {	"2112",	"12"	},
    {	"2113",	"123"	},
    {	"212",	"12"	},
    {	"2121",	"12"	},
    {	"2122",	"12"	},
    {	"2123",	"123"	},
    {	"213",	"123"	},
    {	"2131",	"123"	},
    {	"2132",	"123"	},
    {	"2133",	"123"	},
    {	"2134",	"1234"	},
    {	"2143",	"1234"	},
    {	"221",	"12"	},
    {	"2211",	"12"	},
    {	"2212",	"12"	},
    {	"2213",	"123"	},
    {	"2221",	"12"	},
    {	"2231",	"123"	},
    {	"231",	"123"	},
    {	"2311",	"123"	},
    {	"2312",	"123"	},
    {	"2313",	"123"	},
    {	"2314",	"1234"	},
    {	"2321",	"123"	},
    {	"2331",	"123"	},
    {	"2341",	"1234"	},
    {	"2413",	"1234"	},
    {	"2431",	"1234"	},
    {	"3112",	"123"	},
    {	"312",	"123"	},
    {	"3121",	"123"	},
    {	"3122",	"123"	},
    {	"3123",	"123"	},
    {	"3124",	"1234"	},
    {	"3132",	"123"	},
    {	"3142",	"1234"	},
    {	"321",	"123"	},
    {	"3211",	"123"	},
    {	"3212",	"123"	},
    {	"3213",	"123"	},
    {	"3214",	"1234"	},
    {	"3221",	"123"	},
    {	"3231",	"123"	},
    {	"3241",	"1234"	},
    {	"3312",	"123"	},
    {	"3321",	"123"	},
    {	"3412",	"1234"	},
    {	"3421",	"1234"	},
    {	"4123",	"1234"	},
    {	"4132",	"1234"	},
    {	"4213",	"1234"	},
    {	"4231",	"1234"	},
    {	"4312",	"1234"	},
    {	"4321",	"1234"	},
    {	NULL,	NULL	}
/* *INDENT-ON* */
};
static const struct test_type test_1_9[] = {
/* *INDENT-OFF* */
    {	INSERT_MODE,		sans_logique_091002_0_			},
    {	FOUR87_MODE,		sans_logique_091002_1_			},
    {	CHANGE_MODE,		sans_logique_091002_2_			},
    {	STREAM_TEST,		sans_logique__________			}
/* *INDENT-ON* */
};
static const struct x1f4_trans_type trans = {
/* *INDENT-OFF* */
    free_data,
    link_data,
    mode_data,
    NULL
/* *INDENT-ON* */
};
static const struct x1f4_vkeytree_type link_1_a[] = {
/* *INDENT-OFF* */
    {	"ll",		{	test_1_9 + 3	},	NULL		},
    {	"a",		{	test_1_9 + 0	},	NULL		},
    {	"b",		{	test_1_9 + 1	},	NULL		},
    {	"c",		{	test_1_9 + 2	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_9[] = {
/* *INDENT-OFF* */
    {	"a",		{	NULL		},	link_1_a	},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link____[] = {
/* *INDENT-OFF* */
    {	"",		{	NULL		},	link_1_9	},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
};

static int
deck_slip(void *deck, void *miss)
{
    int true;

    miss = *(void **) miss;

    if ((char *) deck < (char *) miss) {
	true = -1;
    } else {
	if ((char *) miss < (char *) deck) {
	    true = 1;
	} else {
	    true = 0;
	}
    }

    return true;
}


static int
fini_list(struct context_type *context_data, void **pset, const void **tile)
{
    int status;

    status = x1f4_fini_tcvset(pset, tile, here_trans());
    if (status) {
	fprintf(stderr, "%s: cannot destroy set\n", context_data->self);
    }

    return status;
}


static int
free_data(void *context, void *data)
{
    free(data);

    return 0;
}


static int
init_list(struct context_type *context_data, void **pset, const void **tile)
{
    int status;

    status = x1f4_init_tcvset(pset, tile);
    if (status) {
	fprintf(stderr, "%s: cannot create set\n", context_data->self);
    } else {
    }

    return status;
}


static int
link_data(void *context, void **link, unsigned size)
{
    int status = 0;
    void *data;

    data = malloc(size);
    if (!data) {
	status = -1;
    } else {
	*link = data;
    }

    return status;
}


static int
list_link(void *file, const void *node, const char *data, unsigned size)
{
    fwrite("  ", 2, 1, file);
    fwrite(data, size, 1, file);
    fwrite("           ", 13 - size, 1, file);
    fwrite(" - ", 3, 1, file);
    fputs(((const struct test_type *) node)->description, file);
    fputc('\n', file);

    return 0;
}


static int
lose_node(void *miss, void *node)
{
    int status;
    void *data;

    status = x1f4_lead_tclist(miss, &data);
    if (status) {
    } else {
	if (*(void **) data == node) {
	    do {
		status = x1f4_mind_tclist(miss);
		if (status) {
		    break;
		} else {
		    status = x1f4_lead_tclist(miss, &data);
		    if (status) {
			unsigned size;

			x1f4_size_tclist(miss, &size);
			if (size) {
			} else {
			    status = 0;
			}

			break;
		    }
		}
	    } while (*(void **) data == node);
	} else {
	    status = 1;
	}
    }

    return status;
}


static int
miss_deck(void *node, void *deck, void *miss)
{
    int true;

    deck = *(void **) deck;
    miss = *(void **) miss;

    if ((char *) deck < (char *) miss) {
	true = -1;
    } else {
	if ((char *) miss < (char *) deck) {
	    true = 1;
	} else {
	    true = 0;
	}
    }

    return true;
}


static int
miss_line(void *miss, void *node)
{
    const char *line;
    int status;

    line = *(const char **) miss;
    if (pointer((int) *line) == node) {
	status = 0;

	*(const char **) miss = line + 1;
    } else {
	status = 1;
    }

    return status;
}


static int
miss_node(void *miss, void *node)
{
    int status;

    if (*(void **) miss) {
	if (*(void **) *(void **) miss == node) {
	    status = x1f4_next_tsline(miss, *(void **) miss);
	    if (status) {
		status = 0;
	    } else {
		*(void **) miss = NULL;
	    }
	} else {
	    status = 1;
	}
    } else {
	status = 1;
    }

    return status;
}


static int
mode_data(void *context, void **link, unsigned size)
{
    int status = 0;
    void *data;

    data = realloc(*link, size);
    if (!data) {
	status = -1;
    } else {
	*link = data;
    }

    return status;
}


static int
sans_logique_091002_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    const struct x1f4_tccase_type *tccase_data;
    void *pset;

    status = init_list(context_data, &pset, (void *) &tccase_data);
    if (status) {
    } else {
	int excess;
	void *line;

	status = x1f4_init_tsline(&line, 0, NULL);
	if (status) {
	} else {
	    status = sans_logique_091002_0a
		(SANS_LOGIQUE_ARGS_2, &pset, (void *) &tccase_data, line);
	    if (status) {
	    } else {
		status = sans_logique_091002_0b
		    (SANS_LOGIQUE_ARGS_2, &pset, (void *) &tccase_data, line);
		if (status) {
		} else {
		    status = sans_logique_091002_0c
			(SANS_LOGIQUE_ARGS_2, &pset, (void *) &tccase_data,
			 line);
		}
	    }

	    excess = x1f4_fini_tsline(&line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	if (status) {
	    fprintf(stdout, "%55s: failed\n", test_data->description);
	} else {
	    fprintf(stdout, "%55s: passed\n", test_data->description);
	}

	excess = fini_list(context_data, &pset, (void *) &tccase_data);
    }

    return status;
}


static int
sans_logique_091002_0a(SANS_LOGIQUE_args_1, void *line)
{
#define E			64

    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	unsigned e;

	e = i * 523 + 342;

	e ^= e << 003;
	e += e >> 005;
	e ^= e << 002;
	e += e >> 017;
	e ^= e << 012;

	if (e < E) {
	} else {
	    e = e & (E - 1);
	}

	status = (*(const struct x1f4_tccase_type **) tile)->push
	    (pset, tile, pointer(((i << 6) - e) << 4), here_trans());
	if (status) {
	    break;
	} else {
	    void *data;

	    status = x1f4_post_tsline
		(line, pointer(((i << 6) - e) << 4), deck_slip, 0,
		 sizeof(void *), &data);
	    if (status) {
		break;
	    } else {
		*((void **) data) = pointer(((i << 6) - e) << 4);

		i--;
	    }
	}
    }

    return status;

#undef E
}


static int
sans_logique_091002_0b(SANS_LOGIQUE_args_1, void *line)
{
    int status;
    void *data;

    x1f4_lead_tsline(line, &data);

    status = (*(const struct x1f4_tccase_type **) tile)->lime
	(*pset, &data, miss_node);

    return status;
}


static int
sans_logique_091002_0c(SANS_LOGIQUE_args_1, void *line)
{
    int status;
    void *data;


    x1f4_lead_tsline(line, &data);
    do {
	status = (*(const struct x1f4_tccase_type **) tile)->miss
	    (pset, tile, *((void **) data), here_trans());
    } while (x1f4_next_tsline(&data, data));

    if (status) {
    } else {
	if (*pset) {
	    status = 1;
	}
    }

    return status;
}


static int
sans_logique_091002_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    const struct x1f4_tccase_type *tccase_data;
    void *pset;

    status = init_list(context_data, &pset, (void *) &tccase_data);
    if (status) {
    } else {
	int excess;
	const struct pair_type *pair_data;

	pair_data = test_5_9;

	while (pair_data->line) {
	    status = sans_logique_091002_1a
		(SANS_LOGIQUE_ARGS_2, &pset, (void *) &tccase_data, pair_data);
	    if (status) {
		break;
	    } else {
		pair_data++;
	    }
	}

	if (status) {
	    fprintf(stdout, "%55s: failed\n", test_data->description);
	} else {
	    fprintf(stdout, "%55s: passed\n", test_data->description);
	}

	excess = fini_list(context_data, &pset, (void *) &tccase_data);
    }

    return status;
}


static int
sans_logique_091002_1a(SANS_LOGIQUE_args_1, const void *pair)
{
    const char *line;
    int c, status;

    line = ((const struct pair_type *) pair)->line;
    c = *line;
    do {
	status = (*(const struct x1f4_tccase_type **) tile)->push
	    (pset, tile, pointer(c), here_trans());
	if (status) {
	    break;
	} else {
	    line++;
	    c = *line;
	}
    } while (c);

    if (status) {
    } else {
	line = ((const struct pair_type *) pair)->miss;

	status = (*(const struct x1f4_tccase_type **) tile)->lime
	    (*pset, &line, miss_line);
	if (status) {
	} else {
	    status = (*(const struct x1f4_tccase_type **) tile)->high
		(pset, tile, here_trans());
	}
    }

    return status;
}


static int
sans_logique_091002_2_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    const struct x1f4_tccase_type *tccase_data;
    void *pset;

    status = init_list(context_data, &pset, (void *) &tccase_data);
    if (status) {
    } else {
	int excess;
	void *list;

	status = x1f4_init_tclist(&list, 0, NULL);
	if (status) {
	} else {
	    status = sans_logique_091002_2a
		(SANS_LOGIQUE_ARGS_2, &pset, (void *) &tccase_data, list);
	    if (status) {
	    } else {
		status = sans_logique_091002_2b
		    (SANS_LOGIQUE_ARGS_2, &pset, (void *) &tccase_data, list);
		if (status) {
		} else {
		    status = sans_logique_091002_2c
			(SANS_LOGIQUE_ARGS_2, &pset, (void *) &tccase_data,
			 list);
		    if (status) {
		    } else {
			status = tccase_data->high
			    (&pset, (void *) &tccase_data, here_trans());
			if (status) {
			} else {
			    if (pset) {
				status = 1;
			    }
			}
		    }
		}
	    }

	    excess = x1f4_fini_tclist(&list);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	if (status) {
	    fprintf(stdout, "%55s: failed\n", test_data->description);
	} else {
	    fprintf(stdout, "%55s: passed\n", test_data->description);
	}

	excess = fini_list(context_data, &pset, (void *) &tccase_data);
    }

    return status;
}


static int
sans_logique_091002_2a(SANS_LOGIQUE_args_1, void *line)
{
    int status;
    unsigned i;

    i = 0;
    while (i < 1000000) {
	unsigned e;

	e = (rand() & 1023) + 1;

	status = (*(const struct x1f4_tccase_type **) tile)->push
	    (pset, tile, pointer(e), here_trans());
	if (status) {
	    break;
	} else {
	    void *data;

	    i++;

	    status = x1f4_push_tclist(line, rand() % i, sizeof(void *), &data);
	    if (status) {
		break;
	    } else {
		*((void **) data) = pointer(e);
	    }
	}
    }

    return status;
}


static int
sans_logique_091002_2b(SANS_LOGIQUE_args_1, void *list)
{
    int status;
    unsigned i;

    i = 0;
    while (i < 1000000) {
	unsigned e;

	e = (rand() & 1023) + 1;

	status = (*(const struct x1f4_tccase_type **) tile)->push
	    (pset, tile, pointer(e), here_trans());
	if (status) {
	    break;
	} else {
	    void *data;

	    i++;

	    status = x1f4_push_tclist
		(list, rand() % 1000000, sizeof(void *), &data);
	    if (status) {
		break;
	    } else {
		*((void **) data) = pointer(e);

		status = x1f4_lead_tclist(list, &data);
		if (status) {
		    break;
		} else {
		    status = (*(const struct x1f4_tccase_type **) tile)->miss
			(pset, tile, *((void **) data), here_trans());
		    if (status) {
			break;
		    } else {
			status = x1f4_mind_tclist(list);
			if (status) {
			    break;
			}
		    }
		}
	    }
	}
    }

    if (status) {
    } else {
	unsigned size;

	x1f4_size_tclist(list, &size);
	if (size != 1000000) {
	    status = 1;
	}
    }

    return status;
}


static int
sans_logique_091002_2c(SANS_LOGIQUE_args_1, void *list)
{
    int status;

    status = x1f4_sort_tclist(list, NULL, 0, miss_deck);
    if (status) {
    } else {
	status = (*(const struct x1f4_tccase_type **) tile)->lime
	    (*pset, list, lose_node);
    }

    return status;
}


static int
sans_logique__________(SANS_LOGIQUE_ARGS_1)
{
    int status;

    test_data = test_1_9;

    do {
	status = test_data->test(SANS_LOGIQUE_ARGS_2);
	if (status) {
	    break;
	} else {
	    test_data++;
	}
    } while (test_data->test != sans_logique__________);

    return status;
}


static void
class(void)
{
    x1f4_hook_vtkey(stdout, list_link, link____);
}


static void
usage(void)
{
    puts("Usage: tv59 [OPTIONS] TEST\n\
Test the pointer set.\n\
\n\
Options:\n\
  --list			lists available tests and exit\n\
  --help			display this help and exit\n\
  --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "list",	    0x00,   NULL,   0x6c    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'l':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'l':
		class();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (argc - optind != 1) {
	fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' for "
		"more information.\n", argv[0], argv[0]);

	return 1;
    } else {
	const void *test;

	if (x1f4_find_vtkey(argv[optind], link____, &test)) {
	    status = 1;

	    fprintf(stderr, "%s: cannot parse test specification: `%s'\n",
		    argv[0], argv[optind]);
	} else {
	    struct context_type context;

	    context.self = argv[0];

	    status = ((struct test_type *) test)->test(&context, test);
	}
    }

    return status;
}

